/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/
/*! \file TeApplicationUtils.h
    This contains some utilitary functions to deal with TerraLib databases
*/
#ifndef  __TERRALIB_INTERNAL_APPLICATIONUTILS_H
#define  __TERRALIB_INTERNAL_APPLICATIONUTILS_H

#include "TeThemeApplication.h"
#include "TeVisual.h"
#include "TeDatabase.h"
#include "TeGeneralizedProxMatrix.h"

#include <string>
#include <vector>
#include <set>
using namespace std;

struct TeKernelParams;


struct ObjectInfo
{
	int	status_;						// collection: c_object_status
	map<string, int> uniqueMap_;	// uniqueId to grid status (collection: grid_status)
};

/** @defgroup ExtTheme Extended Theme Functions
    @ingroup  DatabaseUtils
	A set of functions to manipulate the extension of a Theme as proposed by TerraView application. 
	It can be used as a model for other TerraLib based applications.
	\sa TeThemeApplication.h
@{
*/
/**	Creates a table to store extra information about the extended Theme
	\param db			a pointer to a TerraLib database
	\param integrity	a flag to indicate that the a integrity relation should be created (with Theme)
	\param tableName	name of the table that persists the extended theme
*/
bool createThemeApplicationTable(TeDatabase* db, bool integrity=false, const string& tableName="te_theme_application");

/** Populates the extended theme table with default parameters
	\param db			a pointer to a TerraLib database
	\param tableName	name of the table that persists the extended theme
*/
bool populateThemeApplicationTable(TeDatabase* db, const string& tableName="te_theme_application");


/** Delete an extended theme

	\param db			a pointer to a TerraLib database
	\param themeId		identifier of the extended theme
	\param tableName	name of the table that persists the extended theme
*/
bool deleteThemeApplication (TeDatabase* db, int themeId, const string& tableName="te_theme_application");

/** Loads a set of views that belongs to a user (considering that it has extended themes)
	\param db		a pointer to a TerraLib database
	\param user		user name
*/
bool loadViewSet(TeDatabase* db, const string& user);

/** Updates the parameters of a view (considering that it has extended themes)
	\param db	a pointer to a TerraLib database
	\param vier	a pointer to a view
*/
bool updateView(TeDatabase* db, TeView *view);

/** Loads the view tree that contains a view (considering that it has extended themes)
	\param db	a pointer to a TerraLib database
	\param view	a pointer to a view
	\return a pointer to a view tree structure
*/
TeViewTree* loadViewTree(TeDatabase* db, TeView* view, int id);

/** Updates a  view tree structure in a database (considering that it has extended themes)
	\param db	a pointer to a TerraLib database
	\param tree	a pointer to a view tree
	\return a pointer to a view tree structure
*/
bool updateViewTree (TeDatabase* db, TeViewTree *tree);
/** @} */ 

/** @defgroup Media Media Persistence
   @ingroup  DatabaseUtils
   A set of functions to manipulate the media data persistence model proposed by TerraLib  
@{
*/
/** Creates a table to store multimedia documents 
	\param db		pointer do a TerraLib database
	\param tname	name of the table to be created 
*/
bool createMediaTable(TeDatabase* db, const string& tname);

/** Creates a table to store the association between objects of a layer and its multimedia attributes
	\param db		pointer do a TerraLib database
	\param layerId	identifier of the layer that will contain the media
	\param tname	name of the table to be created 
*/
bool createLayerMediaTable(TeDatabase* db, int layerId, const string& tname);

/** Inserts a new media file in a database
	\param db			pointer do a TerraLib database
	\param tableName	identifier of the layer that will contain the media
	\param fileName		name of the file that contains the media 
	\param description	description of the media
	\param type			type of the media
	\param blob			flag to indicate that the media should be turned to a blob in the databse
*/
bool insertMedia(TeDatabase* db, const string& tableName, string& fileName, const string& description, const string& type, bool blob=true);

/** Associates a media data to an object of a layer
	\param db			pointer do a TerraLib database
	\param layer		layer that contains the object
	\param objectId		identifier of the object 
	\param mediaName	name of the media
	\param mediaTable	name of the table tha contains the media
*/
bool insertMediaObjLayer (TeDatabase* db, TeLayer* layer, const string& objectId, const string& mediaName, const string& mediaTable); 

/** Sets a media as a default media of an object of a layer
	\param db			pointer do a TerraLib database
	\param layer		layer that contains the object
	\param objectId		identifier of the object 
	\param mediaName	name of the media
	\param mediaTable	name of the table tha contains the media
*/
bool setDefaultMedia(TeDatabase* db, TeLayer* layer, const string& objectId, const string& mediaName, const string& mediaTable);

/** Updates a media description in a database
	\param db			pointer do a TerraLib database
	\param mediaName	name of the media
	\param mediaTable	name of the table tha contains the media
	\param desc			description of the media
*/
bool updateMediaDescription(TeDatabase* db, const string& mediaName, const string& mediaTable, const string& desc);

//! Inserts a new general information of a layer or database
/*
	\param db			pointer do a TerraLib database
	\param layerId		if < 0 it is database information, else it is layer information
	\param tableName	table name
	\param fileName		name of the file that contains the media information
	\param type			file type. It can be ".TXT", ".DOC", ".MP3", etc. If URL it should be ".URL"
*/
bool insertInfo2Blob(TeDatabase* db, int layerId, const string& tableName,  const string& fileName,  const string& type);

//! read blob and copy to temporary file
/*
	\param db			pointer do a TerraLib database
	\param layerId		if < 0 it is database information, else it is layer information
	\param tableName	table name
	\param type			file type. It can be ".TXT", ".DOC", ".MP3", etc. If URL it should be ".URL"
	\return the temporary file that contains the blob
*/
string getInfoBlob2TemporaryFile(TeDatabase* db, int layerId, const string& tableName, string& type);
/** @} */ 

/** @defgroup PBCharts Pie/Bar Charts
  @ingroup  DatabaseUtils
  A set of functions to manipulate the charts (pie and/or bar) persistence model used by TerraView application.
  It can be used as a model for other TerraLib based applications.
  @{
 */
/** Creates a table to store the parameters of bar/pie charts 
	\param db			pointer do a TerraLib database
	\param integrity	flag to indicate a integrity relation (with the theme) should be created 
*/
bool createChartParamsTable(TeDatabase* db, bool integrity=false);	

/** Updates the color associated to a selection mode for a theme
	\param db		pointer do a TerraLib database
	\param themeId	theme extended identifier
	\param sel		selection mode
	\param color	color definition
*/
bool updateGridColor(TeDatabase* db, int themeId, TeObjectSelectionMode sel, TeColor& color);
/** @} */ 

/** @defgroup RampColor Color Manipulation Functions
   @ingroup  DatabaseUtils
   A set of functions to manipulate a ramp color persistence model used by TerraView application.
   It can be used as a model for other TerraLib based applications.
 @{
 */
/** Creates a standard color table to store color ramps 
 	\param db		pointer to an opened database connection
	\param tname	name of the table that will store the color ramps
 	\note The table created will have the following scheme:
 
   \par  --------------------------------------------------
   \par | ramp_name  |  color_order  | red  | green | blue |
   \par ----------------------------------------------------
   \par | red        |      1        | 255  |  256  | 256  |
   \par | red        |      2        | 196  |  195  | 256  |
   \par | red        |      3        | 123  |  124  | 256  |
   \par | green      |      1        | 0    |  256  | 123  |
   \par | ...        |     ...       | ...  |  ...  | ...  |
   \par ---------------------------------------------------
*/
bool createColorSchemeTable(TeDatabase* db, const string& tname="te_color_scheme");

/** Generates a graduated color scale following a set of color ramps, based on a color scheme table stored in a TerraLib database
	\param db		pointer do a TerraLib database
	\param ramps	vector with the sequence color ramps used to build the scale 
	\param nc		desired number of colors on the scale
	\param asc		flag to indicate that within each ramp goes from light to dark colors
	\param colors	a vector with resulting color sequence
	\param tname	name of the table from where the colors were retrived
	\returns true if color scale was successfully generated and false otherwise
*/
bool getColors(TeDatabase* db, std::vector<std::string>& ramps, int nc, bool asc, vector<TeColor>& colors, const string& tname="te_color_scheme");

/** Populates a color scheme table with 30 ramp colors, each ramp with 13 shades of colors
	\param db		pointer do a TerraLib database
	\param tname	name of the table where the colors will be stored
*/
bool populateColorSchemeTable(TeDatabase* db, const string& tname="te_color_scheme");

/** Insert a ramp color entry in the database
	\param db		pointer do a TerraLib database
	\param name		name of the color
	\param color	color definition
	\param tname	name of the table where the colors will be stored
*/
bool insertRampColor(TeDatabase* db, const string& name, int order, TeColor& color, const string& tableName = "te_color_scheme");
/** @} */ 

/** @defgroup Kernel Map Kernel Metadata Functions
    @ingroup  DatabaseUtils
 *  A set of functions to persist in a TerraLib database the metadata associated to Kernel Map algorithm.
 *	\sa TeKernelParams.h
 *  @{
 */
/** Creates a table to store information about the results of the Kernel map algorithm
	\param db pointer do a TerraLib database
*/
bool createKernelMetadataTable(TeDatabase* db);

/** Saves the parameters associated to the execution of a Kernel Map algorithm in a TerraLib database
	\param db			pointer do a TerraLib database
	\param kernelPar	reference to a structure of parameters associated to a Kernel Map 
*/
bool insertKernelMetadata(TeDatabase* db, TeKernelParams& kernelPar);

/** Updates in the database the parameters associated to the execution of a Kernel Map  -- inserts if it does not exist
	\param db			pointer do a TerraLib database
	\param kernelPar	reference to a structure of parameters associated to a Kernel Map 
*/
bool updateKernelMetadata(TeDatabase* db, TeKernelParams& kernelPar);

/** Loads some stored Kernel parameters associated to a theme used as the support to the Kernel Map
	\param db			pointer do a TerraLib database
	\param supThemeId	identifier of the theme that has the support to the Kernel Map
	\param supTableName	name of the attribute table has stores the Kernel result
	\param supAttrName	name of the column table has stores the Kernel result
	\param KernelPar	reference to a structure of parameters associated to a Kernel Map	
*/
bool loadKernelParams (TeDatabase* db, int supThemeId, const string& supTableName, const string& supAttrName, TeKernelParams& KernelPar);

/** Loads some stored Kernel parameters associated to a layer 
	\param db			pointer do a TerraLib database
	\param LayerId		identifier layer
	\param KernelPar	reference to a structure of parameters associated to a Kernel Map	
*/
bool loadKernelParams (TeDatabase* db, int LayerId, TeKernelParams& KernelPar);
/** @} */ 

/** @defgroup Metadata Layer and Attribute table metadata
    @ingroup  DatabaseUtils
    A set of functions to persist application dependent metadata about layers and attribute tables
  @{
 */
/** Creates a table to store some information about a layer layers 
	The table is called te_layer_metadata and has the attributes: [layer_id, name, author, source, description, date_, hour_, transf]
	\param db	pointer do a TerraLib database
*/
int createTeLayerMetadata(TeDatabase* db);

/** Creates a table to store metadata about attribute tables 
	The table is called te_table_metadata and has the attributes: [table_id, name, author, source, quality, description, date_, hour_, transf]
	\param db	pointer do a TerraLib database
*/
int createTeTableMetadata(TeDatabase* db);
/** @} */ 


/** Removes tables that contains the visual of texts associated to a layer
	\param layer	pointer do a TerraLib layer
*/
bool deleteTextVisual(TeLayer* layer);


/** Delete the objects of a layer that are pointed in a theme
	\param theme	pointer to the extended theme were the objects are pointed
	\param box		to return the bounding box that contains the objects that were deleted
	\return the error message of the database when the function fails
*/
string deletePointedObjects(TeTheme* theme, TeBox& box);

/** Inserts a new object into a theme
	\param theme	pointer to a theme (the object is already in the layer)
	\param p		labels position location
	\param newId	the identifier of the object
	\param newTId	the identifier of the instance of the object (for temporal objects)
*/
string insertObjectIntoCollection(TeTheme* theme, TeCoord2D p, string newId, string newTId = "");
/** @} */ 

/** @defgroup GPM Generalized Proximity Matrix
    @ingroup  DatabaseUtils
    Functions to persist a generalized proximity matrix
	\sa TeGeneralizedProxMatrix
 *  @{
*/
/** Creates a table to store information about the generalized proximity matrix
	\param db			pointer do a TerraLib database
	\param integrity	flag to indicate a integrity relation (with the theme) should be created 
*/
bool createGPMMetadataTable(TeDatabase* db, const bool& integrity=true);

/** Creates a table to store information about the neighbourhood of a the objects selected in a theme
	\param db			pointer do a TerraLib database
	\param tableName	name of the table 
*/
bool createNeighbourhoodTable(TeDatabase* db, const string& tableName);

/** Persists a generalized proximity matrix in a TerraLib database
	\param db	pointer do a TerraLib database
	\param gpm	a proximity matrix
	\param objs	vector with the identifier of the objects in the matrix
*/
template<typename set> 
bool insertGPM(TeDatabase* db, TeGeneralizedProxMatrix<set>* gpm, vector<string>& objs);

/** Loads from the database the dafault GPM associated to a theme
	\param db		pointer do a TerraLib database
	\param themeId	identification of the theme
	\param gpm		a structure of a GPM in memory
*/
template<typename set> 
bool loadDefaultGPM(TeDatabase* db, const int& themeId, TeGeneralizedProxMatrix<set>* &gpm);

/** Delete all generalized proximity matrix of a theme (neighbourhoodTable and metadata) or only the gpm with a specific id (gpmId)
	\param db		pointer do a TerraLib database
	\param themeId	identification of the theme
	\param gpmId	GPM identifier
*/
bool deleteGPMTable(TeDatabase* db, const int& themeId, const int& gpmId=-1);
/** @} */ 

/** @} */ // end of group  DatabaseUtils


/** Loads the parameters of an extended theme
	\param db			a pointer to a TerraLib database
	\param theme		pointer to an instance of an extended theme
	\param tableName	name of the table that persists the extended theme
*/
bool loadThemeApplication (TeDatabase* db, TeThemeApplication *theme, const string& tableName="te_theme_application");

/** Saves the parameters of an extended theme 
	\param db			a pointer to a TerraLib database
	\param theme		pointer to an instance of an extended theme
	\param tableName	name of the table that persists the extended theme 
*/
bool insertThemeApplication (TeDatabase* db, TeThemeApplication *theme, const string& tableName="te_theme_application");

/** Updates the parameters of an extended theme
	\param db			a pointer to a TerraLib database
	\param theme		pointer to an instance of an extended theme
	\param tableName	name of the table that persists the extended theme 
*/
bool updateThemeApplication (TeDatabase* db, TeThemeApplication *theme, const string& tableName="te_theme_application");

/** Saves in the database the color parameters associated to a theme
	\param db			pointer do a TerraLib database
	\param theme	pointer to a theme as defined by TerraView
*/
bool insertChartParams(TeDatabase* db, TeThemeApplication *theme);

/** Deletes the chart parameters of a theme in the database
	\param db			pointer do a TerraLib database
	\param theme	pointer to a theme as defined by TerraView
*/
bool deleteChartParams (TeDatabase* db, TeThemeApplication *theme);

/** Verifies if there is a pie/bar char around a given location, for a given theme and select it
	\param db		pointer do a TerraLib database
	\param theme	pointer to a theme as defined by TerraView
	\param pt		a location 
	\param delta	a tolerance around the location
*/
bool locatePieBar (TeDatabase* db, TeThemeApplication* theme, TeCoord2D pt, double delta);

/** Loads the charts parameters associated to a theme
	\param db		pointer do a TerraLib database
	\param theme	pointer to a theme as defined by TerraView
*/
bool loadChartParams (TeDatabase* db, TeThemeApplication *theme);

/** Updates the location of the selected pie/bar chart of a theme
	\param db		pointer do a TerraLib database
	\param theme	pointer to a theme as defined by TerraView
	\param pt		the new location for the chart
*/
bool updatePieBarLocation (TeDatabase* db, TeThemeApplication* theme, TeCoord2D pt);

/** Updates the location of selected object of an extended theme with TePoint geometry
	\param db		pointer do a TerraLib database
	\param theme	pointer to an extended theme
	\param pt		new location of the point
*/
bool updatePointLocation (TeDatabase* db, TeThemeApplication* theme, TeCoord2D pt); 


//! generate Count legends 
bool generateCountLegends(TeThemeApplication* theme, string table, string field);

void changeObjectStatus(ObjectInfo& info, string& uniqueId, string mode, int pointed = 1);

bool saveObjectStatus(TeThemeApplication *theme, set<string>& objectSet, map<string, ObjectInfo>& objectMap);

TeStatisticType getStatisticFromString(const string& s);


void getVisual(TeThemeApplication* theme, TeGeomRep geomRep,
			   map<string, TeVisual>& objVisualMap, set<string>& oidSet, bool allObjects);

vector<string> getUidVec(const string& oid, TeThemeApplication* theme);

template <typename Iterator>
vector<string> getUidVec(Iterator& begin, Iterator& end, TeThemeApplication* theme);

// Implementation of the template functions
template<typename set>
bool insertGPM(TeDatabase* db, TeGeneralizedProxMatrix<set>* gpm, vector<string>& objs)
{
	if(!db)
		return false; 

	TeProxMatrixConstructionParams* cParams = gpm->getConstructionParams();
	TeProxMatrixSlicingParams*	sParams = gpm->getSlicingParams();
	TeProxMatrixWeightsParams*   wParams = gpm->getWeightsParams();

	if((!cParams) || (!sParams) || (!wParams))
		return false;
		
	//update todas as gpms desse tema para default igual a zero
	string up =  " UPDATE te_gpm SET gpm_default = 0 WHERE theme_id1 = "+ Te2String(cParams->theme_id1_);
	db->execute (up);

	// ----- progress bar
	int step = 0;
	if(TeProgress::instance())
		TeProgress::instance()->setTotalSteps(objs.size());
	// -----
	
	//insere na tabela de metadados 
	string ins = " INSERT INTO te_gpm (gpm_default, theme_id1, geom_rep1, theme_id2, ";
	ins += " geom_rep2, neighbours_table, construction_date, construction_strategy, ";
	ins += " const_max_distance, const_num_neighbours, const_max_net_distance, ";
	ins += " const_max_con_distance, slicing_strategy, slicing_zone_dist, ";
	ins += " slicing_zone_local, weight_strategy, weight_norm, weight_a, ";
	ins += " weight_b, weight_c, weight_factor, weight_dist_raio, weigth_max_local_dist ) ";
	ins += " VALUES ( ";
	ins += " 1 ";
	ins += ", "+ Te2String(cParams->theme_id1_);
	ins += ", "+ Te2String((int)cParams->geom_rep1_);
	
	if(cParams->theme_id2_<0)
		ins += ", NULL ";
	else
        ins += ", "+ Te2String(cParams->theme_id2_);
	
	ins += ", "+ Te2String((int)cParams->geom_rep2_);
	ins += ", '"+ gpm->neighbourhood_table_ +"'";
	ins += ", NULL ";
	ins += ", "+ Te2String((int)cParams->strategy_); 
	ins += ", "+ Te2String(cParams->max_distance_, 8);
	ins += ", "+ Te2String(cParams->num_neighbours_);
	ins += ", "+ Te2String(cParams->max_net_distance_, 8);
	ins += ", "+ Te2String(cParams->max_connection_distance_, 8);
	ins += ", "+ Te2String((int)sParams->strategy_); 
	ins += ", "+ Te2String(sParams->zone_dist_, 8);
	ins += ", "+ Te2String((int)sParams->zone_local_);
	ins += ", "+ Te2String((int)wParams->strategy_);
	ins += ", "+ Te2String((int)wParams->norm_);
	ins += ", "+ Te2String(wParams->a_, 8);
	ins += ", "+ Te2String(wParams->b_, 8);
	ins += ", "+ Te2String(wParams->c_, 8);
	ins += ", "+ Te2String(wParams->factor_, 8);
	ins += ", "+ Te2String(wParams->dist_ratio_, 8);
	ins += ", "+ Te2String(wParams->max_local_dist_, 8);
	ins += " )";

	if(!db->execute (ins))
	{
		if (TeProgress::instance())
			TeProgress::instance()->reset();
		return false;
	}

	//insere na tabela de vizinhos
	ins = " INSERT INTO "+ gpm->neighbourhood_table_;
	ins += " (object_id1, object_id2, centroid_distance, weight, ";
	ins += " slice, order_neig, borders_length, net_objects_distance, ";
	ins += " net_minimum_path ) VALUES ( ";

	vector<string>::iterator it = objs.begin(); 
	while(it != objs.end())
	{
		string objId1 = (*it);

		//get the neighbours
		TeNeighboursMap neigs = gpm->getMapNeighbours(objId1);
		TeNeighboursMap::iterator itNeig = neigs.begin();
		while(itNeig != neigs.end())
		{
			string objId2 = itNeig->first;
			TeProxMatrixAttributes attributes = itNeig->second;
			//sql
			string sql = "'"+ objId1 +"'";
			sql += ", '"+ objId2 +"'";
			sql += ", "+ Te2String(attributes.CentroidDistance(), 8);
			sql += ", "+ Te2String(attributes.Weight(), 8);
			sql += ", "+ Te2String(attributes.Slice());
			sql += ", "+ Te2String(attributes.Order());
			sql += ", "+ Te2String(attributes.BorderLength(), 8);
			sql += ", "+ Te2String(attributes.NetworkObjectsDistance(), 8);
			sql += ", "+ Te2String(attributes.NetworkMinimumPath(), 8);
			sql += " )";

			if(!db->execute(ins + sql))
			{
				if (TeProgress::instance())
					TeProgress::instance()->reset();
				return false;
			}
			++itNeig;
		}

		//progress bar
		if(TeProgress::instance())
		{
			if (TeProgress::instance()->wasCancelled())
			{
				TeProgress::instance()->reset();
				return false;
			}
			else
				TeProgress::instance()->setProgress(step);
		}	
		++step;
		++it;
	}
	if (TeProgress::instance())
		TeProgress::instance()->reset();
	return true;
}

template<typename set>
bool loadDefaultGPM(TeDatabase* db, const int& themeId, TeGeneralizedProxMatrix<set>* &gpm)
{
	if(!db)
		return false;

	TeDatabasePortal* portal = db->getPortal();
	if(!portal)
		return false;

	//load the table name
	string sql = " SELECT gpm_id, neighbours_table, weight_strategy, weight_norm, ";
	sql += " weight_a, weight_b, weight_c, weight_factor, weight_dist_raio, ";
	sql += " weigth_max_local_dist FROM te_gpm ";
	sql += " WHERE theme_id1 = "+ Te2String(themeId) +" AND gpm_default = 1";
	if(!portal->query(sql) || !portal->fetchRow())
	{
		delete portal;
		return false;
	}

	string table = portal->getData(1);
	int gpmId = portal->getInt(0);
	int weightStrategy = portal->getInt(2); 

	//load weight strategy
	TeProxMatrixWeightsStrategy* ws = 0;
	if((TeGPMWeightsStrategy)weightStrategy==TeNoWeightsStrategy)
		ws = new TeProxMatrixNoWeightsStrategy((bool)portal->getInt(3));

	else if((TeGPMWeightsStrategy)weightStrategy==TeInverseDistanceStrategy)
		ws = new TeProxMatrixInverseDistanceStrategy (portal->getDouble(4), portal->getDouble(5), 
				portal->getDouble(6), portal->getDouble(7), (bool)portal->getInt(3));

	else if((TeGPMWeightsStrategy)weightStrategy==TeSquaredInverseDistStrategy)
		ws = new TeProxMatrixSquaredInverseDistanceStrategy (portal->getDouble(4), portal->getDouble(5), 
				portal->getDouble(6), portal->getDouble(7), (bool)portal->getInt(3));

	else if((TeGPMWeightsStrategy)weightStrategy==TeConnectionStrenghtStrategy)
		ws = new TeProxMatrixConnectionStrenghtStrategy(portal->getDouble(8), portal->getDouble(9), 
				portal->getDouble(7),(bool)portal->getInt(3));

	portal->freeResult();

	sql= " SELECT object_id1, object_id2, centroid_distance, weight, ";
	sql += " slice, order_neig, borders_length, net_objects_distance, ";
	sql += " net_minimum_path FROM "+ table;
	if(!portal->query(sql))
	{
		delete portal;
		return false;
	}

	TeProxMatrixImplementation* impl = TeProxMatrixAbstractFactory::MakeConcreteImplementation();
	while(portal->fetchRow())
	{
		TeProxMatrixAttributes attr(portal->getDouble(3), portal->getInt(4), 
			portal->getInt(5),  portal->getDouble(2), portal->getDouble(6), 
			portal->getDouble(7), portal->getDouble(8));
		
		impl->connectObjects (string(portal->getData(0)), string(portal->getData(1)), attr);
	}

	gpm = new TeGeneralizedProxMatrix<set>(gpmId, table, impl, ws);

	delete portal;
	return true;
}


template <typename Iterator>
vector<string> getUidVec(Iterator& begin, Iterator& end, TeThemeApplication* theme)
{
	Iterator it;
	string oid;
	vector<string> uidVec;

	vector<TeTable> tableVec = theme->attrTables();
	if (tableVec.size() == 1)
	{
		for (it = begin; it != end; ++it)
			uidVec.push_back(*it);
		return uidVec;
	}

	unsigned int i;
	string uid;
	TeDatabase* db = theme->layer()->database();
	TeDatabasePortal* portal = db->getPortal();

	string w = " WHERE " + tableVec[0].uniqueName();
	string q, whereClause;
	for (it = begin; it != end; ++it)
	{
		oid = *it;
		whereClause = w + " = '" + oid + "'";
		q = theme->sqlJoin() + whereClause;
		if (!portal->query(q))
		{
			delete portal;
			return uidVec;
		}

		while(portal->fetchRow())
		{
			for (i = 1; i < tableVec.size(); ++i)
			{
				uid = oid;
				uid += portal->getData(tableVec[i].uniqueName());
			}
			uidVec.push_back(uid);
		}
		portal->freeResult();
	}

	delete portal;
	return uidVec;
}

#endif



