/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

#include "TeOCIOracle.h"

#include <sys/stat.h>

bool	
TeOCIOracle::createSequence(const string &tableName)
{
	string nameSeq = getNameSequence(tableName);
	
	string seq = " CREATE SEQUENCE " + nameSeq;
	seq += " START WITH 1 INCREMENT BY 1 ORDER ";
	if (!execute(seq))
	{	
		errorMessage_ = "Error creating sequence to table " + tableName + " !";
		return false;
	}

	return true;
}

bool
TeOCIOracle::createAutoIncrementTrigger(const string &tableName, const string &fieldName)
{
	string nameTri = getNameTrigger(tableName);
	string nameSeq = getNameSequence(tableName);

	string tri;
	tri = "CREATE TRIGGER " + nameTri; 
	tri += " BEFORE INSERT ON "+tableName; 
	tri += " for each row";
	tri += " begin";
	tri += " select "+nameSeq+".NEXTVAL";
	tri += " into :new."+fieldName;
	tri += " from dual;";
	tri += " end;";

	if(!execute(tri))
	{	
		errorMessage_ = "Error creating trigger to table " + tableName + " !";
		return false;
	}
	return true;
}

string
TeOCIOracle::getNameSequence(const string &tableName)
{
	string name;
	if(tableName.size()>21)
		name = tableName.substr(0,20) + "_seq";
	else
		name = tableName + "_seq";

	return name;
}

string
TeOCIOracle::getNameTrigger(const string &tableName)
{
	string name;
	if(tableName.size()>21)
		name = tableName.substr(0,20) + "_tri";
	else
		name = tableName + "_tri";

	return name;
}

string  
TeOCIOracle::escapeSequence (const string& from)
{
	int  fa = 0;
    string to = from;
    to.insert(0, " ");
    string::iterator it = to.begin();
    while(it != to.end())
    {
            int f = to.find("'", fa);
            if(f > fa)
            {
                    to.insert(f, "'");
                    fa = f + 2;
            }
            else
                    break;
    }
    to = to.substr(1, to.size() - 1);
    return to;
}

TeOCIOracle::TeOCIOracle() 
{
	connection_ = new TeOCIConnection();  
	dbmsName_ = "OracleOCI";
	sequenceCont_ = -1;
	sequenceName_ = "";
}


TeOCIOracle::~TeOCIOracle() 
{
	if (connection_)
		delete (connection_); //disconnect
	connection_ = NULL;
}

bool 
TeOCIOracle::newDatabase(const string& database, const string& user, const string& password, const string& host, const int& port, bool terralibModel) 
{ 
	if (!connect(host,user,password,database,port))
       return false;
    if (terralibModel)
    {
        //create conceptual model
        if(!this->createConceptualModel())  
			return false;
    }
    return true;
}

bool
TeOCIOracle::connect(const string& host, const string& user, const string& password, const string& database, int port)
{
	if (connection_->isConnected())
	{ 
		delete (connection_);
		connection_ = new TeOCIConnection();
	}

	isConnected_ = false;
	if (connection_->connect(host.c_str(),user.c_str(),password.c_str()))
	{
		isConnected_ = true;
		host_ = host;
		user_ = user;
		password_ = password;
		database_ = database;
		portNumber_ = port;
		return true;
	}
	else
	{
		isConnected_ = false;
		errorMessage_ = "Error connecting to database server!";
		delete (connection_); //disconect
		connection_ = NULL;
		return false;
	}
}

void
TeOCIOracle::close() 
{
	clear();
	connection_->disconnect();
	isConnected_ = false;
}

bool 
TeOCIOracle::listTables(vector<string>& tableList)
{
	tableList.clear();
	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	string exist = " SELECT table_name FROM all_tables WHERE ";
	exist += " OWNER = '" + TeConvertToUpperCase(user_) + "'";

	if(!ocip->query(exist))
	{
		delete ocip;
		return false;
	}

	while(ocip->fetchRow())
		tableList.push_back (ocip->getData(0));
			
	delete ocip;
	return true;
}


bool 
TeOCIOracle::tableExist(const string& table)
{
	bool	status;
	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	string exist = " SELECT table_name FROM all_tables WHERE";
	exist += " TABLE_NAME = '" + TeConvertToUpperCase(table) + "'";
	exist += " AND OWNER = '" + TeConvertToUpperCase(user_) + "'";

	if(!ocip->query(exist))
	{
		delete ocip;
		return false;
	}

	if(ocip->fetchRow())
		status = true;
	else
		status = false;

	delete ocip;
	return (status);
}

bool 
TeOCIOracle::columnExist(const string& table, const string& column, TeAttribute& attr)
{
	bool	status = false;
	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	string exist = " SELECT DATA_TYPE, DATA_LENGTH, DATA_SCALE FROM ALL_TAB_COLUMNS WHERE";
	exist += " TABLE_NAME = '" + TeConvertToUpperCase(table) + "'";
	exist += " AND COLUMN_NAME = '" + TeConvertToUpperCase(column) + "'";
	exist += " AND OWNER = '" + TeConvertToUpperCase(user_) + "'";

	if(!ocip->query(exist))
	{
		delete ocip;
		return false;
	}
	
	if(ocip->fetchRow())
	{	
		attr.rep_.name_ = column;
		
		string	dataType = string(ocip->getData(0)); 
		int		dataLength = atoi(ocip->getData(1)); 
		int		dataScale = atoi(ocip->getData(2));
		bool    number = false;
		
		if(dataType=="VARCHAR2")
		{
			attr.rep_.type_ = TeSTRING;
			attr.rep_.numChar_ = dataLength;
		}
		else if (dataType=="BLOB")
		{
			attr.rep_.type_ = TeBLOB;
			attr.rep_.numChar_ = dataLength;
		}
		else if (dataType=="NUMBER")
		{
			number = true;
		}
		else if (dataType=="SDO_GEOMETRY")
		{
			attr.rep_.type_ = TeOBJECT;
			attr.rep_.numChar_ = dataLength;
		}
		else if (dataType== "CHAR")
		{	
			attr.rep_.type_ = TeCHARACTER;
			attr.rep_.numChar_ = dataLength;
		}
		else if (dataType=="DATE")
		{
			attr.rep_.type_ = TeDATETIME;
		}
		else
		{
			attr.rep_.type_ = TeSTRING;
			attr.rep_.numChar_ = dataLength;
		}

		if(number)
		{
			if(dataScale > 0)
				attr.rep_.type_ = TeREAL;
			else
				attr.rep_.type_ = TeINT;
		}
		status = true;
	}
	
	delete ocip;
	return (status);
}

bool
TeOCIOracle::createTable(const string& table, TeAttributeList &attr)
{
	short	cont=0;
	string pkeys ="";
	bool	hasAutoNumber=false;
	string	fieldName="";

	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	TeAttributeList::iterator it = attr.begin();
	string tablec;
	tablec = "CREATE TABLE " + table +" (";
	
	while ( it != attr.end())
	{
		if (cont)
			tablec += ", ";
			
		switch ((*it).rep_.type_)
		{
			case TeSTRING:
				tablec += (*it).rep_.name_ + " VARCHAR2(" + Te2String((*it).rep_.numChar_) + ")";
			break;
			
			case TeREAL:
				if((*it).rep_.decimals_>0)
					tablec += (*it).rep_.name_ +" NUMBER(*,"+ Te2String((*it).rep_.decimals_) +") ";
				else
					tablec += (*it).rep_.name_ +" NUMBER(*,38) ";
			break;
			
			case TeINT:
			case TeUNSIGNEDINT:
				// auto number
        		tablec += (*it).rep_.name_ + " NUMBER(32) ";
				if((*it).rep_.isAutoNumber_)
				{
					hasAutoNumber=true;
					fieldName=(*it).rep_.name_;
				}
			break;

			case TeDATETIME:
				tablec += (*it).rep_.name_ + " DATE ";
			break;

			case TeCHARACTER:
				tablec += (*it).rep_.name_ + " CHAR ";
			break;

			case TeBLOB:
				tablec += (*it).rep_.name_ + " BLOB ";
			break;

			default:
				tablec += (*it).rep_.name_ + " VARCHAR2(255) ";
			break;
		}
		//default values
		if(!((*it).rep_.defaultValue_.empty()))
			tablec += " DEFAULT '" + (*it).rep_.defaultValue_ + "' ";

		//not null values
		if(!((*it).rep_.null_))
			tablec += " NOT NULL ";

		// check if column is part of primary key
		if ((*it).rep_.isPrimaryKey_ && (*it).rep_.type_ != TeBLOB )
		{
			if (!pkeys.empty())
				pkeys += ", ";
			pkeys += (*it).rep_.name_;
			tablec += " NOT NULL";
		}

		++it;
		cont++;
	}

	if(!pkeys.empty())
		tablec += ", PRIMARY KEY (" + pkeys + ") ";

	tablec += ")";

	if(!execute(tablec))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table;

		return false;
	}

	if(hasAutoNumber)
	{
		string dropSql = " DROP TRIGGER "+ getNameTrigger(table); 
		execute(dropSql);
		dropSql = " DROP SEQUENCE "+ getNameSequence(table); 
		execute(dropSql);

		if(!createSequence(table))
		{
			deleteTable(table);
			return false;
		}
		
		if(!createAutoIncrementTrigger(table,fieldName))
		{
			deleteTable(table);
			string sql= "DROP SEQUENCE "+ getNameSequence(table);
			execute(sql); 
			return false;
		}
	}
	return true;
}

bool 
TeOCIOracle::addColumn (const string& table, TeAttributeRep &rep)
{
	if(!tableExist(table))
		return false;

	string field = TeGetExtension(rep.name_.c_str());
	if(field.empty())
		field = rep.name_;

	string tab;
	tab = " ALTER TABLE " + table + " ADD ( ";
	tab += field + "  ";
	
	switch (rep.type_)
	{
		case TeSTRING:
			tab += "VARCHAR2(" + Te2String(rep.numChar_) + ") ";
			break;
			
		case TeREAL:
			tab += "NUMBER(*,38)";	
			break;
			
		case TeINT:
			tab += "NUMBER(32)";
			break;

		case TeDATETIME:
			tab += "DATE";

		case TeCHARACTER:
			tab += "CHAR";
		
		case TeBLOB:
			tab += "BLOB";
		
		default:
			tab += "VARCHAR2(" + Te2String(rep.numChar_) + ") ";
			break;
	}

	tab += " ) ";

	if(!connection_->execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error inserting a column to table " + table + " !";
		return false;
	}

	alterTableInfoInMemory(table);
	return true;
}

bool
TeOCIOracle::deleteTable (const string& table)
{
	int f =	table.find ("te_collection");

	if( table=="te_theme" ||
		table=="te_layer" ||
		table=="te_representation" ||
		table=="te_tables_relation" ||
		table=="te_layer_table" ||
		table=="te_color_scheme" ||
		table=="te_raster_metadata" ||
		table=="te_projection" ||
		table=="te_view" ||
		table=="te_chart_params" ||
		table=="te_legend" ||
		table=="te_visual" ||
		f == 0)
	{
		errorMessage_ = "No  possvel deletar tabelas do modelo!";
		return false;
	}

	string del = "DROP TABLE " + table;
	if(!execute(del))
		return false;

	string seq = "DROP SEQUENCE " + getNameSequence(table);
	if(!execute(seq))
		return false;

	return true;
}

bool 
TeOCIOracle::execute (const string &q)
{
	bool result = connection_->execute(q);
	
	if(!result)
		errorMessage_ = connection_->getErrorMessage();

	return result;
}

TeDatabasePortal*  
TeOCIOracle::getPortal ()
{
	TeOCIOraclePortal* ocip = new TeOCIOraclePortal (this);
	return ocip;
}

bool 
TeOCIOracle::createViewTable ()
{
	string table;
	table = "CREATE TABLE te_view ( ";
	table += "view_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", projection_id ";
	table += "NUMBER(32) NOT NULL";
	table += " CONSTRAINT view_proj UNIQUE ";
	table += ", name ";
	table += "VARCHAR2(255) NOT NULL";
	table += ", user_name ";
	table += "VARCHAR2(255) ";
	table += ", visibility ";
	table += "NUMBER(32) ";

	table += ", PRIMARY KEY(view_id))";
	
	if(!execute(table))
	{	
		errorMessage_ = "Error creating table te_view!";
		return false;
	}
	
	//creates a sequence to table 
	if(!createSequence("te_view"))
	{
		errorMessage_ = "Error creating sequence to table te_view!";
		return false;
	}

	//creates index
	if(!createIndex ("te_view", "view_index", "name, user_name"))
	{
		errorMessage_ = "Error creating index to table te_view!";
		return false;
	}
	
	return true;
}

bool 
TeOCIOracle::createThemeTable ()
{
	string table;
	table = "CREATE TABLE te_theme (";
	table += "theme_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", layer_id ";
	table += "NUMBER(32) NULL";
	table += ", view_id ";
	table += "NUMBER(32) NULL";
	table += ", name ";
	table += "VARCHAR2(255) NOT NULL";
	table += ", parent_id ";
	table += "NUMBER(32) ";
	table += ", priority ";
	table += "NUMBER(32) ";	
	table += ", node_type ";
	table += "NUMBER(32) NOT NULL";	
		
	table += ", min_scale ";
	table += "NUMBER(*,38)";
	table += ", max_scale ";
	table += "NUMBER(*,38)";

	table += ", generate_attribute_where ";
	table += "VARCHAR2(500)";
	table += ", generate_spatial_where ";
	table += "VARCHAR2(500)";
	table += ", generate_temporal_where ";
	table += "VARCHAR2(500)";
	
	table += ", collection_table ";
	table += "VARCHAR2(255)";

	table += ", visible_rep ";
	table += "NUMBER(32) NULL";
	table += ", enable_visibility ";
	table += "NUMBER(32) NULL";

	table += ", PRIMARY KEY (theme_id))";
	
	if(!execute(table))
	{	
		errorMessage_ = "Error creating table te_theme!";
		return false;
	}

	//creates a sequence to the tabela
	if(!createSequence("te_theme"))
	{
		errorMessage_ = "Error creating sequence to table te_theme!";
		return false;
	}

	//creates index
	if(!createIndex ("te_theme", "theme_view_index", "view_id"))
	{
		errorMessage_ = "Error creating index to table te_theme!";
		return false;
	}

	//creates index
	if(!createIndex ("te_theme", "theme_name_index", "name"))
	{
		errorMessage_ = "Error creating index to table te_theme!";
		return false;
	}
	return true;
}

bool
TeOCIOracle::createGroupingTable()
{
	string table;
	table = "CREATE TABLE te_grouping (";
	table += "theme_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", grouping_number ";
	table += "NUMBER(32)";
	table += ", grouping_attr ";
	table += "VARCHAR2(255)";
	table += ", grouping_attr_type ";
	table += "NUMBER(32)";
	table += ", grouping_mode ";
	table += "NUMBER(32)";
	table += ", grouping_norm_attr ";
	table += "VARCHAR2(255)";
	table += ", grouping_std_dev ";
	table += "NUMBER(*,38) DEFAULT 0.0";
	table += ", grouping_precision ";
	table += "NUMBER(32)";
	table += ", grouping_function ";
	table += "VARCHAR2(20)";
	table += ", grouping_chronon ";
	table += "NUMBER(32)";
	table += ", PRIMARY KEY (theme_id))";
	
	if(!execute(table))
	{	
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table te_grouping!";
		return false;
	}
	return true;
}


bool 
TeOCIOracle::createThemeTablesTable()
{
	string table = "CREATE TABLE te_theme_table  (";
	table += " theme_table_id ";
	table += " NUMBER(32) NOT NULL";
	table += ", theme_id ";
	table += " NUMBER(32) NOT NULL ";
	table += ", table_id ";
	table += " NUMBER(32) NOT NULL ";
	table += ", relation_id ";
	table += " NUMBER(32)";
	table += ", table_order ";
	table += " NUMBER(32) ";
	table += ", PRIMARY KEY (theme_table_id) ";
	table += ")";
	
	if(!execute(table))
	{	
		errorMessage_ = "Error creating table te_theme_table!";
		return false;
	}

	//creates a sequence to the tabela
	if(!createSequence("te_theme_table"))
	{
		errorMessage_ = "Error creating sequence to table te_theme_table!";
		return false;
	}

	//creates index
	if(!createIndex("te_theme_table", "theme_table_index", "theme_id"))
	{
		errorMessage_ = "Error creating index to table te_theme_table!";
		return false;
	}

	//creates index
	if(!createIndex("te_theme_table", "theme_table_table_id", "table_id"))
	{
		errorMessage_ = "Error creating index to table te_theme_table!";
		return false;
	}

	//creates index
	if(!createIndex("te_theme_table", "theme_table_relation_id", "relation_id"))
	{
		errorMessage_ = "Error creating index to table te_theme_table!";
		return false;
	}
	return true;
}


bool 
TeOCIOracle::createLayerTable()
{
	string table;
	table = "CREATE TABLE te_layer (";
	table += "layer_id ";
	table += "NUMBER(32) NOT NULL ";
	table += ", projection_id ";
	table += "NUMBER(32) NOT NULL ";
	table += "CONSTRAINT layer_proj UNIQUE ";
	table += ", name ";
	table += "VARCHAR2(255) NOT NULL ";
	table += ", lower_x ";
	table += "NUMBER(*,38)";
	table += ", lower_y ";
	table += "NUMBER(*,38)";
	table += ", upper_x ";
	table += "NUMBER(*,38) ";
	table += ", upper_y ";
	table += "NUMBER(*,38) ";
	table += ", initial_time ";
	table += "DATE ";
	table += ", final_time ";
	table += "DATE ";

	table += ", PRIMARY KEY (layer_id))";

	if(!execute(table))
	{
		errorMessage_ = "Error creating table te_layer!";
		return false;
	}

	//creates a sequence to table 
	if(!createSequence("te_layer"))
	{
		errorMessage_ = "Error creating sequence to table te_layer!";
		return false;
	}

	//creates index
	if(!createIndex("te_layer", "layer_name_id", "name"))
	{
		errorMessage_ = "Error creating index to table te_layer!";
		return false;
	}

	return true;
}


bool 
TeOCIOracle::createLayerTableTable()
{
	string table;
	table = "CREATE TABLE te_layer_table (";
	table += "table_id ";
	table += " NUMBER(32) NOT NULL ";
	table += ", layer_id ";
	table += "NUMBER(32) ";
	table += ", attr_table ";
	table += "VARCHAR2(255) NOT NULL ";
	table += ", unique_id ";
	table += "VARCHAR2(255) ";
	table += ", attr_link ";
	table += "VARCHAR2(255) ";
	table += ", attr_initial_time ";
	table += "VARCHAR2(255) ";
	table += ", attr_final_time ";
	table += "VARCHAR2(255) ";
	table += ", attr_time_unit ";
	table += "NUMBER(32) ";
	table += ", attr_table_type ";
	table += "NUMBER(32) ";
	table += ", user_name ";
	table += "VARCHAR2(255) ";
	table += ", initial_time ";
	table += "DATE ";
	table += ", final_time ";
	table += "DATE ";

	table += ", PRIMARY KEY (table_id))";

	if(!execute(table))
	{
		errorMessage_ = "Error creating table te_layer_table!";
		return false;
	}

	//creates a sequence to table 
	if(!createSequence("te_layer_table"))
	{
		errorMessage_ = "Error creating sequence to table te_layer_table!";
		return false;
	}

	//creates index
	if(!createIndex("te_layer_table", "layer_table_id", "layer_id"))
	{
		errorMessage_ = "Error creating index to table te_layer!";
		return false;
	}

	//creates index
	if(!createIndex("te_layer_table", "layer_attr_table_id", "attr_table"))
	{
		errorMessage_ = "Error creating index to table te_layer!";
		return false;
	}

	return true;
}


bool 
TeOCIOracle::createTablesRelationTable()
{
	string table;
	table = "CREATE TABLE te_tables_relation (";
	table += "relation_id ";
	table += "NUMBER(32) NOT NULL ";
	table += ", related_table_id ";
	table += "NUMBER(32) NOT NULL ";
	table += ", related_attr ";
	table += "VARCHAR2(255) NOT NULL";
	table += ", external_table_name ";
	table += "VARCHAR2(255) NOT NULL";
	table += ", external_attr ";
	table += "VARCHAR2(255) NOT NULL ";

	table += ", PRIMARY KEY (relation_id))";

	if(!execute(table))
	{
		errorMessage_ = "Error creating table te_tables_relation!";
		return false;
	}

	//creates a sequence to table 
	if(!createSequence("te_tables_relation"))
	{
		errorMessage_ = "Error creating sequence to table te_tables_relation!";
		return false;
	}

	return true;
}


bool 
TeOCIOracle::createRepresentationTable ()
{
	string table;
	table = "CREATE TABLE te_representation (";
	table += " repres_id ";
	table += " NUMBER(32) NOT NULL "; 
	table += ", layer_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", geom_type ";
	table += "NUMBER(32) NOT NULL";
	table += ", geom_table ";
	table += "VARCHAR2(255) ";
	table += ", description "; 
	table += "VARCHAR2(255) ";
	table += ", lower_x ";
	table += "NUMBER(*,38) ";
	table += ", lower_y ";
	table += "NUMBER(*,38) ";
	table += ", upper_x ";
	table += "NUMBER(*,38) ";
	table += ", upper_y ";
	table += "NUMBER(*,38) ";
	table += ", res_x ";
	table += "NUMBER(*,38) ";
	table += ", res_y ";
	table += "NUMBER(*,38) ";
	table += ", num_cols ";
	table += "NUMBER(32)";
	table += ", num_rows ";
	table += "NUMBER(32)";
	table += ", initial_time ";
	table += "DATE";
	table += ", final_time ";
	table += "DATE";
	
	table += ", PRIMARY KEY (repres_id))";

	if(!execute(table))
	{	
		errorMessage_ = "Error creating table te_representation!";
		return false;
	}
	//creates a sequence to table 
	if(!createSequence("te_representation"))
	{
		errorMessage_ = "Error creating sequence to table te_representation!";
		return false;
	}

	//creates index
	if(!createIndex("te_representation", "representation_index", "layer_id, geom_type"))
	{
		errorMessage_ = "Error creating index to table te_representation!";
		return false;
	}
	return true;
}


bool 
TeOCIOracle::createLegendTable ()
{
	string table;
	table = "CREATE TABLE te_legend (";
	table += "  legend_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", theme_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", group_id ";
	table += "NUMBER(32)";
	table += ", num_objs ";
	table += "NUMBER(32)";
	table += ", lower_value ";
	table += "VARCHAR2(255)";
	table += ", upper_value ";
	table += "VARCHAR2(255)";
	table += ", label ";
	table += "VARCHAR2(255)";

	table += ", PRIMARY KEY (legend_id))";

	if(!execute(table))
	{
		errorMessage_ = "Error creating table te_legend!";
		return false;
	}
	
	//creates a sequence to the table 
	if(!createSequence("te_legend"))
	{
		errorMessage_ = "Error creating sequence to table te_legend!";
		return false;
	}

	//creates index
	if(!createIndex("te_legend", "legend_index", "theme_id"))
	{
		errorMessage_ = "Error creating index to table te_legend!";
		return false;
	}
	return true;
}



bool 
TeOCIOracle::createVisualTable()
{
	string table;
	table = "CREATE TABLE te_visual (";
	table += "  legend_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", geom_type ";
	table += "NUMBER(32) NOT NULL";

	table += ", lib_name ";
	table += "VARCHAR2(255) ";
	table += ", symb_id ";
	table += "NUMBER(32) ";
	table += ", red ";
	table += "NUMBER(32) ";
	table += ", green ";
	table += "NUMBER(32) ";
	table += ", blue ";
	table += "NUMBER(32) ";
	table += ", transparency ";
	table += "NUMBER(32) ";
	table += ", width ";
	table += "NUMBER(32) ";
	
	table += ", contour_lib_name ";
	table += "VARCHAR2(255) ";
	table += ", contour_symb_id ";
	table += "NUMBER(32) ";
	table += ", contour_red ";
	table += "NUMBER(32) ";
	table += ", contour_green ";
	table += "NUMBER(32) ";
	table += ", contour_blue ";
	table += "NUMBER(32) ";
	table += ", contour_transp ";
	table += "NUMBER(32) ";
	table += ", contour_width ";
	table += "NUMBER(32) ";

	table += ", size_value ";
	table += "NUMBER(32) ";
	table += ", pt_angle ";
	table += "NUMBER(32) ";
	
	table += ", family ";
	table += "VARCHAR2(100)";
	table += ", bold ";
	table += "NUMBER(32) ";
	table += ", italic ";
	table += "NUMBER(32) ";
	table += ", alignment_vert ";
	table += "NUMBER(*,38)";
	table += ", alignment_horiz ";
	table += "NUMBER(*,38)";
	table += ", tab_size ";
	table += "NUMBER(32) ";
	table += ", line_space ";
	table += "NUMBER(32) ";
	table += ", fixed_size ";
	table += "NUMBER(32) ";

	table += ", PRIMARY KEY (legend_id, geom_type))";

	if(!execute(table))
	{
		errorMessage_ = "Error creating table te_visual!";
		return false;
	}
	
	return true;
}

bool 
TeOCIOracle::createVisualRasterTable()
{
	string create = "CREATE TABLE te_visual_raster (";
	create += " theme_id		NUMBER(32) NOT NULL,";
	create += " band_in			NUMBER(32) NOT NULL,";
	create += " band_out		NUMBER(32), ";
	create += " transf_type		NUMBER(32), ";
	create += " param1			NUMBER(*,38),";
	create += " param2			NUMBER(*,38),";
	create += " lut_table		VARCHAR2(255), ";
	create += " PRIMARY KEY (theme_id, band_in))";
	
	if(!execute(create))
	{
		errorMessage_ = "Error creating table te_visual_raster!";
		return false;
	}
	
	return true;
}

bool 
TeOCIOracle::createDatabaseTable()
{
	string create = "CREATE TABLE te_database (";
	create += "db_version		VARCHAR(50) NOT NULL,";
	create += "db_creation		DATE,";
	create += "PRIMARY KEY (db_version))";

	if(!execute(create))
	{
		errorMessage_ = "Error creating table te_projection!";
		return false;
	}
	return true;
}

bool 
TeOCIOracle::createProjectionTable ()
{
	string table;
	table = "CREATE TABLE te_projection (";
	table += "projection_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", name ";
	table += "VARCHAR2(255) NOT NULL";
	table += ", long0 ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", lat0 ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", offx ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", offy ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", stlat1 ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", stlat2 ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", unit ";
	table += "VARCHAR2(100) ";
	table += ", scale ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", hemis ";
	table += "NUMBER(32) ";
	table += ", datum ";
	table += "VARCHAR2(100) ";
	table += ", radius ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", flattening ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", dx ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", dy ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	table += ", dz ";
	table += "NUMBER(*,38) DEFAULT 0.0000 NOT NULL ";
	
	table += ", PRIMARY KEY (projection_id))";

	if(!execute(table))
	{
		errorMessage_ = "Error creating table te_projection!";
		return false;
	}
	
	//creates a sequence to table 
	if(!createSequence("te_projection"))
	{
		errorMessage_ = "Error creating sequence to table te_projection!";
		return false;
	}

	return true;
}

bool
TeOCIOracle::createRasterMetadataTable(const string& tableName)
{
	if(tableExist(tableName))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string table;
	table = "CREATE TABLE "+ tableName +" (";
	table += " geom_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", band_id ";
	table += "NUMBER(32) NOT NULL";
	table += ", min_value ";
	table += "NUMBER(*,38) DEFAULT 0.0 ";
	table += ", max_value ";
	table += "NUMBER(*,38) DEFAULT 0.0 ";
	table += ", num_bits ";
	table += "NUMBER(32)";
	table += ", data_type ";
	table += "NUMBER(32)";
	table += ", photometric_type ";
	table += "NUMBER(32)";
	table += ", compression_type ";
	table += "NUMBER(32)";
	table += ", dummy ";
	table += "NUMBER(*,38) ";
	
	table += ", PRIMARY KEY (geom_id, band_id))";

	if(!execute(table))
	{	
		errorMessage_ = "Error creating table "+ tableName +"!";
		return false;
	}
	
	return true;
}


bool 
TeOCIOracle::createLUTTable(const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "index_id ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", r_val ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", g_val ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", b_val ";
	tab += "NUMBER(32) NOT NULL ";
	
	tab += ", PRIMARY KEY (index_id))";

	if(!execute(tab))
	{
		errorMessage_ = "Error creating table " + table ;
		return false;
	}

	return true;
}


bool
TeOCIOracle::createCollectionTable(const string& tableName)
{
	
	if(tableExist(tableName))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string table = "CREATE TABLE "+ tableName +" (";
	table += " c_object_id ";
	table += "VARCHAR2(255) NOT NULL";
	table += ", c_legend_id  ";
	table += "NUMBER(32) ";
	table += ", label_x ";
	table += "NUMBER(*,38) ";
	table += ", label_y ";
	table += "NUMBER(*,38) ";
	table += ", c_legend_own ";
	table += "NUMBER(32) ";
	table += ", c_object_status ";
	table += "NUMBER(32) ";
	
	table += ", PRIMARY KEY (c_object_id))";

	if(!execute(table))
	{	
		errorMessage_ = "Error creating table " + tableName + "!";
		return false;
	}

	if(!createIndex(tableName, tableName+"_index_1", "c_legend_id"))
	{
		errorMessage_ = "Error creating index to table "+ tableName +" !";
		return false;
	}
	
	if(!createIndex(tableName, tableName+"_index_2", "c_legend_own"))
	{
		errorMessage_ = "Error creating index to table "+ tableName +" !";
		return false;
	}
	return true;
}

bool 
TeOCIOracle::createCellGeometry (const string& table)
{
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table + "(";
	tab += "geom_id ";
	tab += "NUMBER(32) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR2(255) NULL";
	tab += ", col_number ";
	tab += "NUMBER(32) NOT NULL";
	tab += ", row_number ";
	tab += "NUMBER(32) NOT NULL";
	tab += ", lower_x ";
	tab += "NUMBER(*,38)";
	tab += ", lower_y ";
	tab += "NUMBER(*,38)";
	tab += ", upper_x ";
	tab += "NUMBER(*,38)";
	tab += ", upper_y ";
	tab += "NUMBER(*,38)";
	tab += ", PRIMARY KEY (geom_id))";

	if (!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createIndex(table, table + "box_idx", "lower_x, lower_y, upper_x, upper_y"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createIndex(table, table + "rc_idx", "row_number, col_number"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}

	//creates a sequence to table 
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}

	return true;
}

bool 
TeOCIOracle::createTextGeometry (const string& table)
{
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table +"(";
	tab += "geom_id ";
	tab += "NUMBER(32) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR2(255) ";
	tab += ", x ";
	tab += "NUMBER(*,38) NULL";
	tab += ", y ";
	tab += "NUMBER(*,38) NULL";
	tab += ", text_value ";
	tab += "VARCHAR2(255) NULL";
	tab += ", angle ";
	tab += "NUMBER(*,38) ";
	tab += ", height ";
	tab += "NUMBER(*,38) ";
	tab += ", alignment_vert ";
	tab += "NUMBER(*,38) ";
	tab += ", alignment_horiz ";
	tab += "NUMBER(*,38) ";
	
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table  " + table + " !";
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createIndex(table, table + "xy_idx", "x,y"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
		
	//creates a sequence to table 
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}

	return true;
}

bool 
TeOCIOracle::createArcGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string tab;
	tab = "CREATE TABLE " + table +"(";
	tab += "geom_id ";
	tab += "NUMBER(32) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR2(255) NULL";
	tab += ", from_node ";
	tab += "NUMBER(32) NOT NULL";
	tab += ", to_node ";
	tab += "NUMBER(32) NOT NULL";
	
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		errorMessage_ = "Error creating table " + table ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createIndex(table, table + "fn_idx", "from_node"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createIndex(table, table + "tn_idx", "to_node"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
		
	//creates a sequence to table 
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}
	return true;
}


bool 
TeOCIOracle::createNodeGeometry (const string& table)
{
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string tab;
	tab = "CREATE TABLE " + table + " (";
	tab += "geom_id ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", object_id ";
	tab += "VARCHAR2(255) NULL";
	tab += ", x ";
	tab += "NUMBER(*,38) NOT NULL";
	tab += ", y ";
	tab += "NUMBER(*,38) NOT NULL";
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table  " + table + " !";
		return false;
	}
	
	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createIndex(table, table + "xy_idx", "x,y"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
		
	//creates a sequence to table 
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}
	return true;
}

bool 
TeOCIOracle::createRasterGeometry (const string& table)
{
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string create = "CREATE TABLE " + table + " (";
	create += "geom_id		NUMBER(32) NOT NULL,";
	create += "object_id	VARCHAR2(255) NOT NULL,";
	create += "raster_table VARCHAR2(255) NOT NULL,";
	create += "lut_table	VARCHAR2(255) ,";
	create += "res_x		NUMBER(*,38) DEFAULT 0.0,";
	create += "res_y		NUMBER(*,38) DEFAULT 0.0,";
	create += "num_bands	NUMBER(32),";
	create += "num_cols		NUMBER(32),";
	create += "num_rows		NUMBER(32),";
	create += "block_height NUMBER(32),";
	create += "block_width	NUMBER(32),";
	create += "lower_x		NUMBER(*,38) DEFAULT 0.0,";
	create += "lower_y		NUMBER(*,38) DEFAULT 0.0,";
	create += "upper_x		NUMBER(*,38) DEFAULT 0.0,";
	create += "upper_y		NUMBER(*,38) DEFAULT 0.0,";
	create += "tiling_type  NUMBER(32),";
	create += "PRIMARY KEY (geom_id))";
	
	if(!execute (create))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}
	
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}

	//raster geometry
	if(!createAutoIncrementTrigger(table,"geom_id"))
	{
		deleteTable(table);
		string sql= "DROP SEQUENCE "+ getNameSequence(table);
		execute(sql); 
		return false;
	}
	
	return true;
}

bool 
TeOCIOracle::createRasterTable (const string& tableName)
{
	string tab;
	tab = "CREATE TABLE "+ tableName + " ( ";
	tab += "block_id ";
	tab += " VARCHAR2(50) NOT NULL";
	tab += ", band_id ";
	tab += " NUMBER(32) NOT NULL";
	tab += ", resolution_factor ";
	tab += " NUMBER(32) ";
	tab += ", subband ";
	tab += " NUMBER(32) ";
	tab += ", lower_x ";
	tab += " NUMBER(*,38) DEFAULT 0.0 ";
	tab += ", lower_y ";
	tab += " NUMBER(*,38) DEFAULT 0.0 ";
	tab += ", upper_x ";
	tab += " NUMBER(*,38) DEFAULT 0.0 ";
	tab += ", upper_y ";
	tab += " NUMBER(*,38) DEFAULT 0.0 ";
	tab += ", block_size ";
	tab += " NUMBER(32) ";
	tab += ", spatial_data ";
	tab += " BLOB ";
	
	tab += ", PRIMARY KEY (block_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + tableName ;
		return false;
	}
	
	if(!createIndex(tableName, tableName + "box_idx", "lower_x, lower_y, upper_x, upper_y"))
	{
		errorMessage_ = "Error creating index to table "+ tableName +" !";
		return false;
	}
	
	if(!createIndex(tableName, tableName + "sbnd_idx", "subband"))
	{
		errorMessage_ = "Error creating index to table "+ tableName +" !";
		return false;
	}
	
	if(!createIndex(tableName, tableName + "rf_idx", "resolution_factor"))
	{
		errorMessage_ = "Error creating index to table "+ tableName +" !";
		return false;
	}
	return true;
}

bool 
TeOCIOracle::createPolygonGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}
	
	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "geom_id ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", object_id ";
	tab += "VARCHAR2(255) NULL";
	tab += ", num_coords ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", num_holes ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", parent_id ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", lower_x ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", lower_y ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", upper_x ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", upper_y ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", ext_max ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", spatial_data ";
	tab += "BLOB ";	
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
	
	if(!createIndex(table, table + "parent_idx", "parent_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}

	if(!createIndex(table, table + "extm_idx", "ext_max"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}

	if(!createIndex(table, table + "box_idx", "lower_x, lower_y, upper_x, upper_y"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}

	//creates a sequence to table 
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}
	return true;
}


bool 
TeOCIOracle::createLineGeometry (const string& table)
{
	
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "geom_id ";
	tab += "NUMBER(32) NOT NULL ";
	tab += ", object_id ";
	tab += "VARCHAR2(255) NULL";
	tab += ", num_coords ";	
	tab += "NUMBER(32) NOT NULL ";
	tab += ", lower_x ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", lower_y ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", upper_x ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", upper_y ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", ext_max ";
	tab += "NUMBER(*,38) NOT NULL ";
	tab += ", spatial_data ";
	tab += "BLOB ";	
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}

	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}

	if(!createIndex(table, table + "box_idx", "lower_x, lower_y, upper_x, upper_y"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
			
	//creates a sequence to table 
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}
	return true;
}


bool 
TeOCIOracle::createPointGeometry (const string& table)
{
	if(tableExist(table))
	{
		errorMessage_= "Table already exist!";
		return false;
	}

	string tab;
	tab = "CREATE TABLE " + table + " ( ";
	tab += "geom_id ";
	tab += "NUMBER(32) NOT NULL";
	tab += ", object_id ";
	tab += "VARCHAR2(255) NULL";
	tab += ", x ";
	tab += "NUMBER(*,38) ";
	tab += ", y ";	
	tab += "NUMBER(*,38) ";
	tab += ", PRIMARY KEY (geom_id))";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error creating table " + table + " !" ;
		return false;
	}	
		
	//creates indexes
	if(!createIndex(table, table + "obj_idx", "object_id"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}

	if(!createIndex(table, table + "xy_idx", "x,y"))
	{
		errorMessage_ = "Error creating index to table "+ table +" !";
		return false;
	}
		
	//creates a sequence to table 
	if(!createSequence(table))
	{
		deleteTable(table);
		return false;
	}
	return true;
}

bool 
TeOCIOracle::insertRelationInfo(const int tableId, const string& tField,
						    const string& eTable, const string& eField, int& relId)
{
	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	relId = -1;
	string sel = "SELECT relation_id FROM te_tables_relation WHERE";
	sel += " related_table_id = " + Te2String(tableId);
	sel += " AND related_attr = '" + tField + "'";
	sel += " AND external_table_name = '" + eTable + "'";
	sel += " AND external_attr = '" + eField + "'";
	if (!ocip->query(sel))
	{
		delete ocip;
		return false;
	}

	if (ocip->fetchRow())
	{
		relId = atoi(ocip->getData(0));
		delete ocip;
		return true;
	}

	string seq = getNameSequence("te_tables_relation");	
	string sql = " INSERT INTO te_tables_relation( ";
	sql += " relation_id, related_table_id, related_attr, ";
	sql += " external_table_name, external_attr) ";
	sql += " VALUES( ";
	sql += seq +".NEXTVAL "; 
	sql += ", "+ Te2String(tableId);
	sql += ",'" + escapeSequence(tField) + "'";
	sql += ",'" + escapeSequence(eTable) + "'";
	sql += ",'" + escapeSequence(eField) + "')";

	if(!execute(sql))
	{
		errorMessage_ = "Error inserting tables information!";   
		delete ocip;
		return false;
	}

	ocip->freeResult();
	sql = "SELECT "+ seq +".CURRVAL FROM DUAL";
	if (!ocip->query(sql))
	{
		errorMessage_ = "Error in the sequence te_table_relation_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	relId = atoi((const char*)ocip->getData(0));
	delete ocip;
	return true;
}

bool 
TeOCIOracle::insertTableInfo (int layerId, TeTable &table, const string& user)
{
	string ins, seq;

	seq = getNameSequence("te_layer_table");
	ins = "INSERT INTO te_layer_table ( "; 
	ins += " table_id, "; 
	ins += " layer_id, attr_table, unique_id, attr_link, ";
	ins += " attr_initial_time, attr_final_time, attr_time_unit, ";
	ins += " attr_table_type, user_name) VALUES ( ";
	ins += seq + ".NEXTVAL "; 
	
	if(layerId>0)
		ins += ","+ Te2String(layerId);
	else
		ins += ", NULL ";
	
	ins += ", '" + escapeSequence(table.name()) + "'";
	ins += ", '" + escapeSequence(table.uniqueName()) + "'";
	ins += ", '" + escapeSequence(table.linkName()) + "'";
	ins += ", '" + escapeSequence(table.attInitialTime()) + "'";
	ins += ", '" + escapeSequence(table.attFinalTime()) + "'";
	ins += ",  " + Te2String(table.attTimeUnit());
	ins += ",  " + Te2String(table.tableType());
	ins += ", '" +  escapeSequence(user) + "'";
	ins += ")";

	if(!execute(ins))
	{
		errorMessage_ = "Error inserting tables information!";   
		return false;
	}

	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	ins = "SELECT "+ seq +".CURRVAL FROM DUAL";
	if (!ocip->query(ins))
	{
		errorMessage_ = "Error in the sequence te_layer_table_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	int index = atoi((const char*)ocip->getData(0));
	table.setId(index);
	delete ocip;
	return true;
}


TeDBRelationType 
TeOCIOracle::existRelation(const string& tableName, const string& relName)
{
	TeOCIOraclePortal  *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return TeNoRelation;
	
	string integ = " SELECT DELETE_RULE  ";
	integ += " FROM   all_constraints ";
	integ += " WHERE  CONSTRAINT_TYPE = 'R'"; 
	integ += " AND    OWNER = '" + TeConvertToUpperCase(user_) + "'";
	integ += " AND    CONSTRAINT_NAME = '" + TeConvertToUpperCase(relName) + "'";
	
	string temp = " AND    TABLE_NAME = '" + TeConvertToUpperCase(tableName) + "' ";   
	integ += temp;
	
	if(!ocip->query(integ))
	{
		delete ocip;
		return TeNoRelation;
	}

	string cascade;
	if(ocip->fetchRow())
	{
		cascade = string(ocip->getData(0));
		if(cascade== "CASCADE")
		{
			delete ocip;
			return TeRICascadeDeletion;
		}
		
		delete ocip;
		return TeRINoCascadeDeletion;
	}

	delete ocip;
	return TeNoRelation;	
}


bool 
TeOCIOracle::createRelation (const string& name, const string& table, const string& fieldName, const string& relatedTable, const string& relatedField, bool cascadeDeletion)
{
	string relation = "ALTER TABLE " + table + " ADD ";
	relation += " CONSTRAINT " + name;
	relation += " FOREIGN KEY ( " + fieldName + ") "; 
	relation += " REFERENCES " + relatedTable + "(" + relatedField + ")";
		
	if (cascadeDeletion)
		relation += " ON DELETE CASCADE ";

	if(!execute(relation))
		return false;

	return true;
}

string
TeOCIOracle::getSQLStatistics (TeGroupingAttr& attrs)
{
	string sql = "";
	string virg = "";

	TeGroupingAttr::iterator it = attrs.begin();
	int count = 0;
	while(it != attrs.end())
	{
		if(count>0)
			virg = ",";

		switch ((*it).second)
		{
			case TeSUM:
				sql += virg +" SUM( "+ (*it).first.name_ +") AS SUM_"+ Te2String(count);
				(*it).second = TeNOSTATISTIC;
				++count;
				break;
			case TeMAXVALUE:
				sql += virg +" MAX( "+ (*it).first.name_ +") AS MAX_"+ Te2String(count);
				(*it).second = TeNOSTATISTIC;
				++count;
				break;
			case TeMINVALUE:
				sql += virg +" MIN( "+ (*it).first.name_ +") AS MIN_"+ Te2String(count);
				(*it).second = TeNOSTATISTIC;
				++count;
				break;
			case TeCOUNT:
				sql += virg +" COUNT( "+ (*it).first.name_ +") AS COUNT_"+ Te2String(count);
				(*it).second = TeNOSTATISTIC;
				++count;
				break;
			case TeMEAN:
				sql += virg +" AVG( "+ (*it).first.name_ +") AS AVG_"+ Te2String(count);
				(*it).second = TeNOSTATISTIC;
				++count;
				break;
			case TeSTANDARDDEVIATION:
				sql += virg +" STDDEV( "+ (*it).first.name_ +") AS STDDEV_"+ Te2String(count);
				(*it).second = TeNOSTATISTIC;
				++count;
				break;
			case TeVARIANCE:
				sql += virg +" VARIANCE( "+ (*it).first.name_ +") AS VAR_"+ Te2String(count);
				(*it).second = TeNOSTATISTIC;
				++count;
				break;
			default:
				break;
		}
		++it;
	}
	return sql;
}


string 
TeOCIOracle::getSQLAutoNumber(const string& table)
{
	return (getNameSequence(table) +".NEXTVAL");
}


bool 
TeOCIOracle::insertRasterBlock(const string& table, const string& blockId, const TeCoord2D& ll, const TeCoord2D& ur, 
								   unsigned char *buf,unsigned long size, int band, unsigned int res, unsigned int subband)
{
	if (blockId.empty()) // no block identifier provided
	{
		errorMessage_ = "bloco sem identificador";
		return false;
	}

	TeOCIOraclePortal* portal = (TeOCIOraclePortal*) this->getPortal();
	if (!portal)
		return false;

	bool update = false;
	string q =" SELECT * FROM " + table; 
	q += " WHERE block_id='" + blockId + "'";

	if (!portal->query(q))
	{
		delete portal;
		return false;
	}
	// check if this block is alread in the database
	if (portal->fetchRow())
		update = true;

	delete portal;

	try
	{
		if (!update)
		{
			q = "INSERT INTO "+ table +" (block_id, band_id, subband, ";
			q += " resolution_factor, lower_x, lower_y, upper_x, upper_y, ";
			q += " block_size, spatial_data) VALUES ( ";
			q += "'" + blockId + "'";
			q += ", " + Te2String(band);
			q += ", " + Te2String(subband);
			q += ", " + Te2String(res);
			q += ", " + Te2String(ll.x(), 10);
			q += ", " + Te2String(ll.y(), 10);
			q += ", " + Te2String(ur.x(), 10);
			q += ", " + Te2String(ur.y(), 10);
			q += ", " + Te2String(size);
			q += ", :blobValue";
			q += ")";
			if (!connection_->executeBLOBSTM(q, buf, size, ":blobValue"))
				return false;
		}
		else
		{
			q = " UPDATE "+ table +" SET spatial_data=:blobValue ";
			q += " WHERE block_id='" + blockId + "'";
			if (!connection_->executeBLOBSTM(q, buf, size, ":blobValue"))
				return false;
		}
		sizeToAllocate_[table]=size;
	}
	catch(...)
	{
		errorMessage_ = "Error inserting raster block!";
		return false;
	}
	return true;
}

bool 
TeOCIOracle::insertBlob (const string& tableName, const string& columnBlob, TeAttributeRep& columnId, const string& valueId, unsigned char* data, int size)
{
	string whereClause = columnId.name_ +" = ";
	switch (columnId.type_ )
	{
		case TeSTRING:
			whereClause += "'"+ valueId + "'";
			break;
		default:
			whereClause += valueId;
			break;
	}
	return (insertBlob (tableName, columnBlob, whereClause, data, size));
}
	

bool 
TeOCIOracle::insertBlob (const string& tableName, const string& columnBlob, TeAttributeRep& columnId, const string& valueId, const string& fileName)
{
	string whereClause = columnId.name_ +" = ";
	switch (columnId.type_ )
	{
		case TeSTRING:
			whereClause += "'"+ valueId + "'";
			break;
		default:
			whereClause += valueId;
			break;
	}
	return (insertBlob(tableName, columnBlob, whereClause, fileName));
}

bool 
TeOCIOracle::insertBlob (const string& tableName, const string& columnBlob, const string& whereClause, unsigned char* data, int size)
{
	TeOCIOraclePortal* portal = (TeOCIOraclePortal*) getPortal();
	if (!portal)
		return false;

	try
	{
		string q = "SELECT * FROM "+ tableName +" WHERE "+ whereClause;
		if((!portal->query(q)) || (!portal->fetchRow()))
		{
			delete portal;
			return false;
		}
		delete portal;

		q = " UPDATE "+ tableName +" SET "+ columnBlob +" = :blobValue";
		q += " WHERE "+ whereClause;
		if (!connection_->executeBLOBSTM(q, data, size, ":blobValue"))
			return false;
		
		sizeToAllocate_[tableName]=size;
	}
	catch(...)
	{
		errorMessage_ = "Error!";
		return false;
	}
	return true;
}

bool 
TeOCIOracle::insertBlob (const string& tableName, const string& columnBlob, const string& whereClause, const string& fileName)
{
	unsigned char	*cdata = 0;
	int		size;
	FILE	*fp = 0;
	
	struct	stat buf;
	int		result;
	
	result = stat(fileName.c_str(), &buf);
	
	if( result != 0 )
		return false;
	
	size = buf.st_size;

	cdata = new unsigned char[size];
	fp = fopen(fileName.c_str(), "rb");
	fread(cdata, sizeof(unsigned char), size, fp); 

	bool status = insertBlob (tableName, columnBlob, whereClause, cdata, size);

	if(fp)
		fclose(fp);

	if (cdata)
		delete []cdata;

	return status;
}
	
bool 
TeOCIOracle::insertTable(TeTable &table)
{
	string tableName = table.name();
	int size = table.size();
	TeAttributeList att = table.attributeList();
	TeAttributeList::iterator it = att.begin();
	TeAttributeList::iterator itEnd = att.end();

	TeTableRow row;
	int i;
	unsigned int j;
	for ( i = 0; i < size; i++  )
	{
		row = table[i];
		it = att.begin();
		string q = 	"INSERT INTO "+tableName+" VALUES(";
		j = 1;
		int jj = 0;
		while ( it != itEnd )
		{
			if (j != 1)
				q += ", ";

			if(row[jj].empty())
			{
				q += " NULL ";
				++it;
				j++;
				jj++;
				continue;
			}

  			string oracleFormat="";
			string dateTime="";
			if(((*it).rep_.type_==TeDATETIME) && (!row[jj].empty()))
			{
				TeTime t  (row[jj], (*it).dateChronon_, (*it).dateTimeFormat_, (*it).dateSeparator_,
						   (*it).timeSeparator_, (*it).indicatorPM_);
				
				dateTime = t.getDateTime("DDsMMsYYYYsHHsmmsSS");
				oracleFormat = " TO_DATE('" + dateTime + "', 'DD/MM/YYYY HH24:MI:SS')";
			}
			
			switch ((*it).rep_.type_)
  			{
  				case TeSTRING:
					q += "'"+ escapeSequence(row[jj]) +"'";
  				break;
  				case TeREAL:
					q += row[jj]; 
  				break;
  				case TeINT:
					q += row[jj];
  				break;
				case TeDATETIME:
					q += oracleFormat;
				break;
				case TeCHARACTER:
					q += "'" + escapeSequence(row[jj]) + "'";
  				break;
				case TeBLOB:
					q += " EMPTY_BLOB() ";
				break;
  				default:
					q += "'"+ escapeSequence(row[jj]) +"'";
  				break;
  			}
			++it;
			j++;
			jj++;
		}
		q += ")";
		
		if (!execute(q))
			continue;
	}
	return true;
}
	
bool 
TeOCIOracle::alterTable (const string& table, TeAttributeRep &rep, const string& oldColName)
{
	if(!tableExist(table))
		return false;

	if(!oldColName.empty() && oldColName != rep.name_ ) //rename column
	{
		string add = " ALTER TABLE "+ table +" RENAME COLUMN "+ oldColName +" TO "+ rep.name_;
		if(!execute(add))
		{
			if(errorMessage_.empty())
				errorMessage_ = "Error alter table " + table + " !";
			return false;
		}
	}

	string tab = " ALTER TABLE "+ table +" MODIFY ( ";
	tab += rep.name_ + "  ";
	switch (rep.type_)
	{
		case TeSTRING:
			if ( rep.numChar_ < 256 )
				tab  += "VARCHAR2(" + Te2String(rep.numChar_) + ") ";
			else
				tab += "VARCHAR(2000) "; 
			break;

		case TeREAL:
			if(rep.decimals_>0)
				tab += " NUMBER(*,"+ Te2String(rep.decimals_) +") ";
			else
				tab += " NUMBER(*,38) ";
		break;
		
		case TeINT:
			tab += " NUMBER(32) ";
			break;

		case TeDATETIME:
			tab += " DATE ";
			break;

		case TeCHARACTER:
			tab += " CHAR ";
			break;
		
		default:
			tab += " VARCHAR2(" + Te2String(rep.numChar_) + ") ";
			break;
	}

	tab += " )";

	if(!execute(tab))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error alter table " + table + " !";
		return false;
	}
	
	string tableId;
	TeDatabasePortal* portal = getPortal();
	string sql = "SELECT table_id FROM te_layer_table WHERE attr_table = '" + table + "'";
	if(portal->query(sql) && portal->fetchRow())
		tableId = portal->getData(0);

	delete portal;

	if(tableId.empty() == false)
	{
		if(oldColName.empty() == false) // column name changed
		{
			 // update relation
			sql = "UPDATE te_tables_relation SET related_attr = '" + rep.name_ + "'";
			sql += " WHERE related_table_id = " + tableId;
			sql += " AND related_attr = '" + oldColName + "'";
			if(execute(sql) == false)
				return false;

			sql = "UPDATE te_tables_relation SET external_attr = '" + rep.name_ + "'";
			sql += " WHERE external_table_name = '" + table + "'";
			sql += " AND external_attr = '" + oldColName + "'";
			if(execute(sql) == false)
				return false;

			 // update grouping
			sql = "UPDATE te_grouping SET grouping_attr = '" + table + "." + rep.name_ + "'";
			sql += " WHERE grouping_attr = '" + table + "." + oldColName + "'";
			if(execute(sql) == false)
				return false;
		}
		else // column type changed
		{
			// delete relation
			sql = "DELETE FROM te_tables_relation WHERE (related_table_id = " + tableId;
			sql += " AND related_attr = '" + rep.name_ + "')";
			sql += " OR (external_table_name = '" + table + "'";
			sql += " AND external_attr = '" + rep.name_ + "')";
			if(execute(sql) == false)
				return false;

			// delete grouping
			TeDatabasePortal* portal = getPortal();
			sql = "SELECT theme_id FROM te_grouping WHERE grouping_attr = '" + table + "." + oldColName + "'";
			if(portal->query(sql) && portal->fetchRow())
			{
				string themeId = portal->getData(0);

				sql = "DELETE FROM te_legend WHERE theme_id = " + themeId + " AND group_id >= 0";
				if(execute(sql) == false)
				{
					delete portal;
					return false;
				}
			}
			delete portal;

			sql = "DELETE FROM te_grouping";
			sql += " WHERE grouping_attr = '" + table + "." + oldColName + "'";
			if(execute(sql) == false)
				return false;
		}
	}
	alterTableInfoInMemory(table);
	return true;
}

bool 
TeOCIOracle::alterTable(const string& oldTableName, const string& newTableName)
{
	string sql = " ALTER TABLE "+ oldTableName +" RENAME TO "+ newTableName;
	if(!this->execute(sql))
		return false;

	//update te_layer_table
	sql = " UPDATE te_layer_table ";
	sql += " SET attr_table = '"+ newTableName +"'";
	sql += " WHERE attr_table = '"+ oldTableName +"'";
	execute(sql);

	//update te_tables_relation
	sql = " UPDATE te_tables_relation ";
	sql += " SET external_table_name = '"+ newTableName +"'";
	sql += " WHERE external_table_name = '"+ oldTableName +"'";
	execute(sql);
	
	alterTableInfoInMemory(newTableName, oldTableName);
	return true;
}


bool
TeOCIOracle::updateTable (TeTable &table)
{
	string tableName = table.name();
	TeAttributeList att = table.attributeList();
	TeAttributeList::iterator it = att.begin();

	TeTableRow row;
	unsigned int i;
	unsigned int j;
	string uniqueName = table.uniqueName();
	string uniqueVal;
	for ( i = 0; i < table.size(); i++  )
	{
		row = table[i];
		it = att.begin();
		string q = 	"UPDATE "+tableName+" SET ";
		j = 1;
		int jj = 0;
		while ( it != att.end() )
		{
			string oracleFormat;
			string dateTime; 
			if((*it).rep_.type_==TeDATETIME)
			{
				TeTime t  (row[jj], (*it).dateChronon_, (*it).dateTimeFormat_, (*it).dateSeparator_,
						   (*it).timeSeparator_, (*it).indicatorPM_);
				
				dateTime = t.getDateTime();
				oracleFormat = "DD/MM/YYYY HH24:MI:SS"; 
			}
			
			if (uniqueName != (*it).rep_.name_)
			{
				q += (*it).rep_.name_ + "=";
  				switch ((*it).rep_.type_)
  				{
  					case TeSTRING:
						q += "'"+escapeSequence(row[jj])+"'";
  					break;
  					case TeREAL:
						q += row[jj];
  					break;
  					case TeINT:
						q += row[jj];
  					break;
					case TeDATETIME:
						q += " TO_DATE('" + dateTime + "', '"+ oracleFormat +"')";
  					break;
					case TeCHARACTER:
						q += "'" + escapeSequence(row[jj]) + "'";
  					break;
  					default:
						q += "'"+escapeSequence(row[jj])+"'";
  					break;
  				}
				if (j<att.size())
					q+= ",";
			}
			else
				uniqueVal = row[jj];

			++it;
			j++;
			jj++;
		}
		q += " WHERE " + uniqueName + " = " + uniqueVal;
		if (!execute(q))
			continue;
	}
	return true;
}
	

bool 
TeOCIOracle::insertProjection (TeProjection *proj)
{
	string insert = "INSERT INTO te_projection (";
	insert += " projection_id, "; 
	insert += " name, long0, lat0,";
	insert += " offx, offy, stlat1, stlat2, unit, scale, hemis, datum, ";
	insert += " radius, flattening, dx , dy, dz ) VALUES ( ";
	insert += " te_projection_seq.NEXTVAL ";
	insert += ", '" + escapeSequence(proj->name()) + "'";
	insert += ", " + Te2String(proj->params().lon0*TeCRD,10);
	insert += ", " + Te2String(proj->params().lat0*TeCRD,10);
	insert += ", " + Te2String(proj->params().offx,10);
	insert += ", " + Te2String(proj->params().offy,10);
	insert += ", " + Te2String(proj->params().stlat1*TeCRD,10);
	insert += ", " + Te2String(proj->params().stlat2*TeCRD,10);
	insert += ", '" + escapeSequence(proj->params().units) + "'";
	insert += ", " + Te2String(proj->params().scale,10);
	insert += ", " + Te2String(proj->params().hemisphere, 10);
	insert += ", '" + escapeSequence(proj->params().datum.name()) + "'";
	insert += ", " + Te2String(proj->params().datum.radius(),10);
	insert += ", " + Te2String(proj->params().datum.flattening(),10);
	insert += ", " + Te2String(proj->params().datum.xShift(),10);
	insert += ", " + Te2String(proj->params().datum.yShift(),10);
	insert += ", " + Te2String(proj->params().datum.zShift(),10);
	insert += ")";

	if(!execute(insert))
	{
		errorMessage_ = "Error inserting in the table te_projection!";   
		return false;
	}

	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*) getPortal();
	if(!ocip)
		return false;

	string seq = "SELECT te_projection_seq.CURRVAL FROM DUAL";
	if(!ocip->query(seq))
	{
		errorMessage_ = "Error in the sequence te_projection_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	proj->id(atoi((const char*)ocip->getData(0))); 
	delete ocip;
	return true;
}


bool
TeOCIOracle::insertRepresentation (int layerId, TeRepresentation& rep)
{
	if (layerId <= 0)
		return false;
	
	string ins;
	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*) getPortal();
	if(!ocip)
		return false;

	ins = " INSERT INTO te_representation ( ";
	ins += " repres_id, ";
	ins += " layer_id, geom_type, geom_table, ";
	ins += " description, lower_x, lower_y, upper_x, upper_y, res_x, res_y, num_cols, ";
	ins += " num_rows) VALUES (";
	ins += " te_representation_seq.NEXTVAL ";
	ins += ", " + Te2String(layerId);
	ins += ", " + Te2String(static_cast<int>(rep.geomRep_));
	ins += ", '" + escapeSequence(rep.tableName_) + "'";
	ins += ", '" + escapeSequence(rep.description_) + "'";
	ins += ", " + Te2String(rep.box_.x1(),10);
	ins += ", " + Te2String(rep.box_.y1(),10);
	ins += ", " + Te2String(rep.box_.x2(),10);
	ins += ", " + Te2String(rep.box_.y2(),10);
	ins += ", " + Te2String(rep.resX_,10);
	ins += ", " + Te2String(rep.resY_,10);
	ins += ", " + Te2String(rep.nCols_);
	ins += ", " + Te2String(rep.nLins_);
	ins += ")";

	if(!execute(ins))
	{
		errorMessage_ = "Error inserting in the table te_representation!";  
		delete ocip;
		return false;
	}
	
	string seq = "SELECT te_representation_seq.CURRVAL FROM DUAL";
	if(!ocip->query(seq))
	{
		errorMessage_ = "Error in the sequence te_representation_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	int id = atoi((const char*)ocip->getData(0)); 
	rep.id_ = id;
	delete ocip;
	return true;
}


bool
TeOCIOracle::insertLegend (TeLegendEntry* leg)
{
	TeOCIOraclePortal  *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	string ins = "INSERT INTO te_legend ( ";
	ins += " legend_id, ";
	ins += " theme_id, group_id, ";
	ins += " num_objs, lower_value, upper_value, label) VALUES ( ";
	ins += "te_legend_seq.NEXTVAL";
	ins += ", " + Te2String(leg->theme());
	ins += ", " + Te2String(leg->group());
	ins += ", " + Te2String(leg->count());
	ins += ", '" + escapeSequence(leg->from()) + "'";
	ins += ", '" + escapeSequence(leg->to()) + "'";
	ins += ", '" + escapeSequence(leg->label()) + "'";
	ins += ")";
		
	if (!execute(ins))
	{
		errorMessage_ = "Error inserting in the table te_legend!"; 
		delete ocip;
		return false;
	}

	string seq = "SELECT te_legend_seq.CURRVAL FROM DUAL";
	if(!ocip->query(seq))
	{
		errorMessage_ = "Error in the sequence te_theme_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";;
		delete ocip;
		return false;
	}

	int index = atoi((const char*)ocip->getData(0)); 
	leg->id(index);
	
	delete ocip;
	legendMap_[leg->id()] = leg;
	return insertVisual(leg);
}

bool
TeOCIOracle::insertPolygonSet(const string& table, TePolygonSet &ps)
{
	if(!getValueSequence(table))
		return false;

	for (unsigned int i = 0; i < ps.size(); i++ )
	{
		TePolygon& poly = ps [i];
		if (!insertPolygon (table,poly))		
			return false;
	}
	return true;
}

bool 
TeOCIOracle::insertPolygon (const string& table, TePolygon &poly)
{
	double	extmax;
	unsigned int i, k, ni, size;
	double	*points	= 0;
	try
	{
		long parentId = 0;

		for ( k = 0; k < poly.size(); k++ )
		{
			TeLinearRing ring (poly[k]);
			TeBox b = ring.box();
			size = ring.size(); 
			ni = 0;
			
			// max extent 
			extmax = MAX(b.width(),b.height());
		
			// number of holes
			if (k==0)
				ni = poly.size()-1;
			else
				ni = 0;
			
			++sequenceCont_;
			if(k==0)
				parentId=sequenceCont_;

			// insert data
			string ins = "INSERT INTO " + table + " ( ";
			ins += " geom_id, ";
			ins += " object_id, num_coords, num_holes, ";
			ins += " parent_id, lower_x, lower_y, upper_x, upper_y, "; 
			ins += " ext_max, spatial_data) VALUES ( ";
			ins +=  getNameSequence(table) +".NEXTVAL";
			ins += ", '" + escapeSequence(poly.objectId()) + "'";
			ins += ","+ Te2String(size);
			ins += ","+ Te2String(ni);
			ins += ","+ Te2String((long)parentId);
			ins += ","+ Te2String(b.lowerLeft().x(), 10);
			ins += ","+ Te2String(b.lowerLeft().y(), 10);
			ins += ","+ Te2String(b.upperRight().x(), 10);
			ins += ","+ Te2String(b.upperRight().y(), 10);
			ins += ","+ Te2String(extmax);
			ins += ", :blobValue ";
			ins += ")";

			//get coord
			points = new double[2*ring.size()];  //2*sizeof(double)*size
			int iac = 0;
			for (i=0;i<ring.size();i++)
			{
				points[iac++]=ring[i].x();
				points[iac++]=ring[i].y();
			}
			unsigned long bLen = 2*(ring.size())*sizeof(double)*sizeof(unsigned char);

			if (!connection_->executeBLOBSTM(ins, (unsigned char*)points, bLen, ":blobValue"))
			{
				errorMessage_ = "Error inserting in the table " + table + "!";  
				delete[] points;
				return false;
			}

			poly.geomId(sequenceCont_);
			if(points)
				delete[] points;
			points=NULL;

			if(bLen > sizeToAllocate_[table])
				sizeToAllocate_[table] = bLen;
		}
	}
	catch(...)
	{
		errorMessage_ = "Error inserting polygon!";
		if(points)
			delete[] points;
		return false;
	}

	return true;
}

bool 
TeOCIOracle::updatePolygon (const string& table, TePolygon &poly)
{
	double	extmax;
	unsigned int k, ni, size, i;
	double	*points	= NULL;

	try
	{
		TeBox b = poly[0].box();
		for ( k = 0; k < poly.size(); k++ )
		{
			//Gets each ring
			TeLinearRing ring ( poly[k] );
			size = ring.size();
			ni = 0;
			if (k==0)
			{
				extmax = MAX(b.width(),b.height());
				ni = poly.size()-1;
			}
			else
				ni = 0;

			//update information about each ring
			string sql;
			sql =  " UPDATE " + table + " SET ";
			sql += ", object_id = '" + poly.objectId() + "'";
			sql += ", num_coords = " + Te2String(size);
			sql += ", num_holes = " + Te2String(ni);
			sql += ", parent_id = " + ((long)poly[0].geomId());
			sql += ", lower_x = " + Te2String(b.lowerLeft().x(), 10);
			sql += ", lower_y = " + Te2String(b.lowerLeft().y(), 10);
			sql += ", upper_x = " + Te2String(b.upperRight().x(), 10);
			sql += ", upper_y = " + Te2String(b.upperRight().y(), 10);
			sql += ", ext_max = " + Te2String(extmax);
			sql += ", spatial_data = :blobValue ";
			sql += " WHERE geom_id = "+ Te2String(ring.geomId());

			//get coord
			points = new double[2*ring.size()];  //2*sizeof(double)*size
			int iac = 0;
			for (i=0;i<ring.size();i++)
			{
				points[iac++]=ring[i].x();
				points[iac++]=ring[i].y();
			}

			int bLen = 2*(ring.size())*sizeof(double)*sizeof(unsigned char);

			if (!connection_->executeBLOBSTM(sql, (unsigned char*)points, bLen, ":blobValue"))
			{
				errorMessage_ = "Error inserting in the table " + table + "!";  
				delete[] points;
				return false;
			}

			if(points)
				delete[] points;
			points=NULL;
		}
	}
	catch(...)
	{
		errorMessage_ = "Error updating polygons!";
		if(points)
			delete[] points;
		return false;
	}
	return true;
}

bool
TeOCIOracle::getValueSequence(const string& table)
{
	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	if (getNameSequence(table) != sequenceName_)
	{
		string seq = "SELECT "+ getNameSequence(table) +".NEXTVAL FROM DUAL";
		
		if(!ocip->query(seq) || !ocip->fetchRow())
		{
			errorMessage_ = "Error in the sequence " + table + "_seq!"; 
			sequenceName_ = "";
			delete ocip;
			return false;
		}

		sequenceName_ = getNameSequence(table);
		sequenceCont_ = atoi(ocip->getData(0));
	}

	delete ocip;
	return true;
}

bool 
TeOCIOracle::insertLineSet(const string& table, TeLineSet &ls)
{
	if(!getValueSequence(table))
		return false;

	for (unsigned int i = 0; i < ls.size(); i++ )
	{
		TeLine2D& line = ls [i];
		if (!insertLine (table,line))
			return false;
	}
	return true;
}

bool
TeOCIOracle::insertLine (const string& table, TeLine2D &line)
{
	double	extmax;
	unsigned int i, size;
	double	*points	= 0;
	
	try
	{
		TeBox b = line.box();
		size = line.size(); 
		// max extent 
		extmax = MAX(b.width(),b.height());
		
		// insert data
		string ins = "INSERT INTO " + table + " ( ";
		ins += " geom_id, ";
		ins += " object_id, num_coords, ";
		ins += " lower_x, lower_y, upper_x, upper_y, "; 
		ins += " ext_max, spatial_data) VALUES ( ";
		ins +=  getNameSequence(table) +".NEXTVAL";
		ins += ", '" + escapeSequence(line.objectId()) + "'";
		ins += ","+ Te2String(size);
		ins += ","+ Te2String(b.lowerLeft().x(), 10);
		ins += ","+ Te2String(b.lowerLeft().y(), 10);
		ins += ","+ Te2String(b.upperRight().x(), 10);
		ins += ","+ Te2String(b.upperRight().y(), 10);
		ins += ","+ Te2String(extmax);
		ins += ", :blobValue";
		ins += ")";

		//get coord
		points = new double[2*line.size()];  //2*sizeof(double)*size
		int iac = 0;
		for (i=0;i<line.size();i++)
		{
			points[iac++]=line[i].x();
			points[iac++]=line[i].y();
		}
		unsigned long bLen = 2*(line.size())*sizeof(double)*sizeof(unsigned char);

		if (!connection_->executeBLOBSTM(ins, (unsigned char*)points, bLen, ":blobValue"))
		{
			errorMessage_ = "Error inserting in the table " + table + "!";  
			delete[] points;
			return false;
		}
		++sequenceCont_;
		line.geomId(sequenceCont_);
		if(points)
			delete[] points;
		points=NULL;
		
		if(bLen > sizeToAllocate_[table])
			sizeToAllocate_[table] = bLen;
	}
	catch(...)
	{
		errorMessage_ = "Error inserting line!";
		if(points)
			delete[] points;
		return false;
	}

	return true;
}


bool 
TeOCIOracle::updateLine(const string& table, TeLine2D &line)
{
	double	extmax;
	unsigned int size, i;
	double	*points	= NULL;

	try
	{
		TeBox b = line.box();
		size = line.size();
		extmax = MAX(b.width(),b.height());
		
		//update information about each line
		string sql =  "UPDATE " + table + " SET ";
		sql += "  object_id= '" + line.objectId() + "'";
		sql += ", num_coords= " + ((long)size);
		sql += ", lower_x= " + Te2String(b.lowerLeft().x(), 10);
		sql += ", lower_y= " + Te2String(b.lowerLeft().y(), 10);
		sql += ", upper_x= " + Te2String(b.upperRight().x(), 10);
		sql += ", upper_y= " + Te2String(b.upperRight().y(), 10);
		sql += ", ext_max= " + Te2String(extmax);
		sql += ", spatial_data = :blobValue ";
		sql += " WHERE geom_id = "+ Te2String(line.geomId());

		//get coord
		points = new double[2*line.size()];  //2*sizeof(double)*size
		int iac = 0;
		for (i=0;i<line.size();i++)
		{
			points[iac++]=line[i].x();
			points[iac++]=line[i].y();
		}
		int bLen = 2*(line.size())*sizeof(double)*sizeof(unsigned char);

		if (!connection_->executeBLOBSTM(sql, (unsigned char*)points, bLen, ":blobValue"))
		{
			errorMessage_ = "Error inserting in the table " + table + "!";  
			delete[] points;
			return false;
		}
		
		if(points)
			delete[] points;
		points = NULL;
	}
	catch(...)
	{
		errorMessage_ = "Error updating lines!";
		if(points)
			delete[] points;
		return false;
	}
	return true;
}

bool 
TeOCIOracle::insertPoint(const string& table, TePoint &point)
{
	string ins = "INSERT INTO " + table + " ( ";
	ins += " geom_id, ";
	ins += " object_id, x, y) ";
	ins += " VALUES ( ";
	ins += getNameSequence(table) +".NEXTVAL";
	ins += ", '" + escapeSequence(point.objectId()) + "'";
	ins += ",  " + Te2String(point.location().x(), 10);
	ins += ",  " + Te2String(point.location().y(), 10);
	ins += " ) ";

	if(!execute(ins))
	{
		errorMessage_ = "Error inserting points!"; 
		return false;
	}
	return true;
}


bool 
TeOCIOracle::insertText	(const string& table, TeText &text)
{
	string ins = "INSERT INTO " + table + " ( ";
	ins += " geom_id, ";
	ins += " object_id, x, y, text_value, angle, height, alignment_vert, ";
	ins += " alignment_horiz) VALUES ( ";
	ins += getNameSequence(table) +".NEXTVAL";
	ins += ", '" + escapeSequence(text.objectId()) + "'";
	ins += ",  " + Te2String(text.location().x(),10);
	ins += ",  " + Te2String(text.location().y(),10);
	ins += ", '" + escapeSequence(text.textValue()) + "'";
	ins += ",  " + Te2String(text.angle(),10);
	ins += ",  " + Te2String(text.height(),10);
	ins += ",  " + Te2String(text.alignmentVert(),10);
	ins += ",  " + Te2String(text.alignmentHoriz(),10);
	ins += " )";
	
	if(!execute(ins))
	{
		errorMessage_ = "Error inserting texts!"; 
		return false;
	}
	return true;
}

bool 
TeOCIOracle::insertArc (const string& table, TeArc &arc)
{
	string ins = "INSERT INTO " + table + " (";
	ins += " geom_id, ";
	ins += " object_id, from_node, to_node ) ";
	ins += " VALUES ( ";
	ins += getNameSequence(table) +".NEXTVAL";
	ins += ", '" + escapeSequence(arc.objectId()) + "'";
	ins += ",  " + Te2String(arc.fromNode().geomId());
	ins += ",  " + Te2String(arc.toNode().geomId());
	ins += " )";
	
	if(!execute(ins))
	{
		errorMessage_ = "Error inserting arc!"; 
		return false;
	}
	return true;
}

bool 
TeOCIOracle::insertNode (const string& table, TeNode &node)
{	
	string ins = "INSERT INTO " + table + " ( ";
	ins += " geom_id, "; 
	ins += " object_id, x, y) ";
	ins += " VALUES ( ";
	ins += getNameSequence(table) +".NEXTVAL";
	ins += ", '" + escapeSequence(node.objectId()) + "'";
	ins += ",  " + Te2String(node.location().x(), 10);
	ins += ",  " + Te2String(node.location().y(), 10);
	ins += " ) ";

	if(!execute(ins))
	{
		errorMessage_ = "Error inserting nodes!"; 
		return false;
	}
	return true;
}
	
bool 
TeOCIOracle::insertCell (const string& table, TeCell &cell )
{
	TeBox b = cell.box();
	
	string ins = "INSERT INTO " + table + " ( ";
	ins += " geom_id, ";
	ins += " object_id, col_number, row_number, lower_x, ";
	ins += " lower_y, upper_x, upper_y ) ";
	ins += " VALUES ( ";
	ins += getNameSequence(table) +".NEXTVAL";
	ins += ", '" + escapeSequence(cell.objectId ()) + "'";
	ins += ",  " + Te2String(cell.column());
	ins += ",  " + Te2String(cell.line());
	ins += ",  " + Te2String(b.lowerLeft().x(), 10);
	ins += ",  " + Te2String(b.lowerLeft().y(), 10);
	ins += ",  " + Te2String(b.upperRight().x(), 10);
	ins += ",  " + Te2String(b.upperRight().y(), 10);
	ins += " )";
		
	if(!execute(ins))
	{
		errorMessage_ = "Error inserting cells!"; 
		return false;
	}
	return true;
}

bool 
TeOCIOracle::insertLayer(TeLayer* layer)
{	
	int index;
	TeProjection* proj = layer->projection();
	if (!proj || !insertProjection(proj))
	{
		errorMessage_ = "No  possvel inserir layer sem projeo";
		return false;
	}
	string ins = "INSERT INTO te_layer (";
	ins += " layer_id, ";
	ins += " projection_id, name ";
	ins += ", lower_x, lower_y, upper_x, upper_y) ";
	ins += " VALUES ( ";
	ins += "te_layer_seq.NEXTVAL";
	ins += ", "+ Te2String(proj->id());
	ins += ", '" + escapeSequence(layer->name()) + "'";
	ins += ", " + Te2String(layer->box().x1(),10);
	ins += ", " + Te2String(layer->box().y1(),10);
	ins += ", " + Te2String(layer->box().x2(),10);
	ins += ", " + Te2String(layer->box().y2(),10);
	ins += ")";

	if(!execute(ins))
	{
		errorMessage_ = "Error inserting in the table te_layer!";   
		return false;
	}

	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	ins = "SELECT te_layer_seq.CURRVAL FROM DUAL";
	if (!ocip->query(ins))
	{
		errorMessage_ = "Error in the sequence te_layer_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	index = atoi((const char*)ocip->getData(0));
	layer->id(index);
	delete ocip;
	layerMap_[layer->id()] = layer;
	return true;
}


bool
TeOCIOracle::deleteLayer(int layerId)
{
	//Delete attributes tables
	if(!deleteLayerTable(layerId))
		return false;

	TeOCIOraclePortal* portal = (TeOCIOraclePortal*) this->getPortal();
	if (!portal)
		return false;

	string geomTable;
	string sql = "SELECT projection_id FROM te_layer WHERE layer_id = ";
	sql += Te2String(layerId);

	if (!portal->query(sql))
	{	
		delete portal;
		return false;
	}

	if (!portal->fetchRow())
	{
		delete portal;
		return false;
	}
	string projId = portal->getData("projection_id");
	portal->freeResult();

	// Get all representations that are associated to this layer
	sql = "SELECT * FROM te_representation WHERE layer_id = "+ Te2String(layerId);
	if (!portal->query (sql))
	{
		delete portal;
		return false;
	}

	while (portal->fetchRow())
	{	// Delete the geometry tables
		geomTable = portal->getData("geom_table");
	
		// Delete lut table
		TeGeomRep rep = TeGeomRep(atoi(portal->getData("geom_type")));
		if (rep == TeRASTER)
		{
			TeOCIOraclePortal* portal2 = (TeOCIOraclePortal*)this->getPortal();
			sql = "SELECT lut_table, raster_table FROM " + geomTable;
			string tabName;
			if (!portal2->query (sql))
			{
				delete portal2;
				continue;
			}

			while (portal2->fetchRow())
			{
				// remove lut table
				tabName = portal2->getData(0);
				if (!tabName.empty() && this->tableExist(tabName))
				{
					sql = "DROP TABLE " + tabName;
					this->execute(sql);

					sql= "DROP SEQUENCE "+ getNameSequence(tabName);
					this->execute(sql); 
				}
				// remove raster table
				tabName = portal2->getData(1);
				if (!tabName.empty() && this->tableExist(tabName))
				{
					sql = "DROP TABLE " + tabName;
					this->execute(sql);

					sql= "DROP SEQUENCE "+ getNameSequence(tabName);
					this->execute(sql); 
				}
			}
			delete portal2;
			// remove raster metadata table
			tabName = geomTable + "_metadata";
			if (!tabName.empty() && this->tableExist(tabName))
			{
				sql = "DROP TABLE " + tabName;
				this->execute(sql);

				sql= "DROP SEQUENCE "+ getNameSequence(tabName);
				this->execute(sql); 

			}

		}
		if (this->tableExist(geomTable))
		{
			sql = "DROP TABLE " + geomTable;
			if (!this->execute(sql) )
			{
				delete portal;
				return false;
			}
			
			sql= "DROP SEQUENCE "+ getNameSequence(geomTable);
			this->execute(sql);
		}
	}

	portal->freeResult();
	if (existRelation("te_representation","fk_rep_layer_id") != TeRICascadeDeletion)
	{
		// Delete entries into representations table
		sql = "DELETE FROM te_representation WHERE layer_id = " +Te2String(layerId);
		if (!this->execute(sql) )
		{
			delete portal;
			return false;
		}
	}

	// delete layer themes
	sql = "SELECT theme_id FROM te_theme WHERE layer_id=" + Te2String(layerId);
	if (!portal->query (sql))
	{
		delete portal;
		return false;
	}
	
	int themeId;
	while (portal->fetchRow())
	{	
		themeId = atoi(portal->getData("theme_id"));
		this->deleteTheme(themeId);
	}
	
	sql = "DELETE FROM te_layer WHERE layer_id=" + Te2String(layerId);
	if (!this->execute(sql))
	{
		delete portal;
		return false;
	}

	// delete layer projection
	sql = "DELETE FROM te_projection WHERE projection_id = "+ projId;
	if (!this->execute(sql))
	{	
		delete portal;
		return false;
	}

	// remove all the items themes associated to the layer to be removed
	TeThemeMap::iterator it;
	for (it = themeMap_.begin(); it != themeMap_.end(); ++it)
	{
		TeTheme *theme = it->second;
		if (theme->layer()->id() == layerId)
		{
			themeMap_.erase(theme->id());
			delete theme;
		}
	}

	// Delete layer and its entry in the layer map
	TeLayer* layer = layerMap_[layerId];
	layerMap_.erase(layerId);
	delete layer;

	delete portal;
	return true;
}

bool
TeOCIOracle::insertTheme (TeTheme *theme)
{
	double maxScale = theme->maxScale ();
	if(maxScale==TeMAXFLOAT)
		maxScale = 0.;

	string ins = "INSERT INTO te_theme (";
	ins += " theme_id, ";
	ins += " layer_id, view_id, name, ";
	ins += " parent_id, priority, node_type, min_scale, max_scale, ";
	ins += " generate_attribute_where, generate_spatial_where, generate_temporal_where, ";
	ins += " collection_table, visible_rep, enable_visibility) VALUES (";
	ins += "te_theme_seq.NEXTVAL";
	ins += ", " + Te2String(theme->layerId());
	ins += ", " + Te2String(theme->view());
	ins += ", '" + escapeSequence(theme->name()) + "'";
	ins += ", " + Te2String(theme->parentId ());
	ins += ", " + Te2String(theme->priority());
	ins += ", " + Te2String(theme->type ());
	ins += ", " + Te2String (theme->minScale(),10);
	ins += ", " + Te2String (maxScale,10);
	ins += ", '" + escapeSequence(theme->attributeRest()) + "'";
	ins += ", '" + escapeSequence(theme->spatialRest()) + "'";
	ins += ", '" + escapeSequence(theme->temporalRest()) + "'";
	ins += ", '" + escapeSequence(theme->collectionTable()) + "'";
	ins += ", " + Te2String(theme->visibleRep());
	ins += ", " + Te2String(theme->visibility());
	ins += ")";

	if(!execute(ins))
	{
		errorMessage_ = "Error inserting in the table te_theme!";   
		return false;
	}

	TeOCIOraclePortal  *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	string seq = "SELECT te_theme_seq.CURRVAL FROM DUAL";
	if(!ocip->query(seq))
	{
		errorMessage_ = "Error in the sequence te_theme_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	int index = atoi((const char*)ocip->getData(0)); 
	theme->id(index);
	delete ocip;
	string colName = theme->collectionTable(); 

	//insert collection name
	if(colName.empty())
	{
		colName = "te_collection_" + Te2String(index);
		theme->collectionTable(colName);	
		
		ins = "UPDATE te_theme SET ";
		ins += " collection_table = '" + escapeSequence(colName) + "'";
		ins += " WHERE theme_id = " + Te2String (index);
		
		if(!execute(ins))
		{
			errorMessage_ = "Error inserting in the table te_theme!";   
			return false;
		}
	}
	
	bool status;

	// insert grouping
	int numSlices = 0;
	if(theme->grouping().groupMode_ != TeNoGrouping)
	{
		if(!insertGrouping (theme->id(), theme->grouping()))
			return false;
		numSlices = theme->grouping().groupNumSlices_;
	}

	// insert legend
	theme->outOfCollectionLegend().group(-1); 
	theme->outOfCollectionLegend().theme(theme->id()); 
	status = insertLegend (&(theme->outOfCollectionLegend())); 
	if (!status)
		return status;

	theme->withoutDataConnectionLegend().group(-2); 
	theme->withoutDataConnectionLegend().theme(theme->id()); 
	status = insertLegend (&(theme->withoutDataConnectionLegend())); 
	if (!status)
		return status;

	theme->defaultLegend().group(-3); 
	theme->defaultLegend().theme(theme->id()); 
	status = insertLegend (&(theme->defaultLegend())); 
	if (!status)
		return status;

	theme->pointingLegend().group(-4); 
	theme->pointingLegend().theme(theme->id()); 
	status = insertLegend (&(theme->pointingLegend())); 
	if (!status)
		return status;

	theme->queryLegend().group(-5); 
	theme->queryLegend().theme(theme->id()); 
	status = insertLegend (&(theme->queryLegend())); 
	if (!status)
		return status;

	theme->queryAndPointingLegend().group(-6); 
	theme->queryAndPointingLegend().theme(theme->id()); 
	status = insertLegend (&(theme->queryAndPointingLegend())); 
	if (!status)
		return status;

	for (int i = 0; i < numSlices; i++)
	{
		theme->legend()[i].group(i);
		theme->legend()[i].theme(theme->id());
		status = insertLegend (&(theme->legend()[i]));
		if (!status)
			return status;
	}
	if (!status)
		return status;

	themeMap_[theme->id()] = theme;
	return updateThemeTable(theme);
}


bool 
TeOCIOracle::insertThemeTable (int themeId, int tableId, int relationId, int tableOrder)
{
	string ins = "INSERT INTO te_theme_table ( ";
	ins += " theme_table_id, ";
	ins += " theme_id, table_id, relation_id, table_order)";
	ins += " VALUES ( ";
	ins += getNameSequence("te_theme_table") +".NEXTVAL ";
	ins += ", "+ Te2String(themeId);
	ins += ", "+ Te2String(tableId);
	
	if(relationId>0)
		ins += ", "+ Te2String(relationId);
	else
		ins += ", null ";

	ins += ", "+ Te2String(tableOrder);
	ins += ")";

	if(!execute(ins))
	{
		errorMessage_ = "Error inserting in the table te_theme_table!";   
		return false;
	}

	return true;
}

bool 
TeOCIOracle::insertThemeGroup(TeViewTree* tree)
{
	string ins = "INSERT INTO te_theme (theme_id, view_id, name, ";
	ins += " parent_id, priority, node_type ) VALUES (";
	ins += "te_theme_seq.NEXTVAL";
	ins += ", " + Te2String(tree->view());
	ins += ", '" + escapeSequence(tree->name()) + "'";
	ins += ", " + Te2String(tree->parentId ());
	ins += ", " + Te2String(tree->priority());
	ins += ", " + Te2String(1);
	ins += ")";

	TeOCIOraclePortal  *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;
	
	try
	{
		if(!execute(ins))
		{
			errorMessage_ = "Error inserting in the table te_theme!";   
			delete ocip;
			return false;
		}

		string seq = "SELECT te_theme_seq.CURRVAL FROM DUAL";
		if(!ocip->query(seq))
		{
			errorMessage_ = "Error in the sequence te_theme_seq!";  
			delete ocip;
			return false;
		}

		if(!ocip->fetchRow())
		{
			errorMessage_ = "Sequence value not found!";
			delete ocip;
			return false;
		}

		int index = atoi((const char*)ocip->getData(0)); 
		tree->id(index);
		
	}
	catch(...)
	{
		errorMessage_ = "Error inserting in the table te_theme!";
		if(ocip)
			delete ocip;
		return false;
	}

	delete ocip;
	return true;
}

bool 
TeOCIOracle::generateLabelPositions (TeTheme *theme)
{
	string	geomTable, upd;
	string	collTable = theme->collectionTable();
	
	if((collTable.empty()) || (!tableExist(collTable)))
		return false;

	if( theme->layer()->hasGeometry(TeCELLS)    || 
		theme->layer()->hasGeometry(TePOLYGONS) ||
		theme->layer()->hasGeometry(TeLINES) )
	{
		geomTable = theme->layer()->tableName(TeCELLS);
		
		if(geomTable.empty())
		{
			geomTable = theme->layer()->tableName(TePOLYGONS);
			if(geomTable.empty())
				geomTable = theme->layer()->tableName(TeLINES);
		}

		upd= "UPDATE " + collTable + " SET ";
		upd += " label_x = (SELECT MAX(lower_x + (upper_x - lower_x)/2) ";
		upd += "FROM " + geomTable + " WHERE object_id = c_object_id), ";
		
		upd += " label_y = (SELECT MAX(lower_y + (upper_y - lower_y)/2) ";
		upd += "FROM " + geomTable + " WHERE object_id = c_object_id) ";
		upd += " WHERE label_x IS NULL OR label_y IS NULL";
	}
	else if(theme->layer()->hasGeometry(TePOINTS))
	{
		geomTable = theme->layer()->tableName(TePOINTS);
		
		upd= " UPDATE " + collTable + " SET ";
		upd += " label_x = (SELECT MAX(x) ";
		upd += " FROM " + geomTable + " p WHERE object_id = c_object_id), ";
		
		upd += " label_y = (SELECT MAX(y) ";
		upd += " FROM " + geomTable + " p WHERE object_id = c_object_id) ";
		upd += " WHERE label_x IS NULL OR label_y IS NULL";
	}

	if(!execute(upd))
		return false;

	return true;
}

bool 
TeOCIOracle::loadLayer(TeLayer* layer)
{
	if(!TeDatabase::loadLayer(layer))
		return false;
    
	string tableName = "";
	if(layer->hasGeometry(TePOLYGONS))
		tableName = layer->tableName(TePOLYGONS);
	else if(layer->hasGeometry(TeLINES))
		tableName = layer->tableName(TeLINES);
	else if (layer->hasGeometry(TeRASTER))
		tableName = layer->raster()->params().fileName_;
	else
		return true;

	map<string, unsigned long>::iterator it = sizeToAllocate_.find (tableName);
	if(it!=sizeToAllocate_.end())
		return true;

	string sql;
	unsigned long maxSize = 0;
	TeDatabasePortal* portal = this->getPortal();
	if(!portal)
		return false;
    
	if(layer->hasGeometry(TePOLYGONS) || layer->hasGeometry(TeLINES))
	{
		sql = " SELECT MAX(NUM_COORDS) FROM "+ tableName;
		if(!portal->query(sql) || !portal->fetchRow())
		{
			delete portal;
			return false;
		}
		maxSize = (portal->getInt(0))*2*sizeof(double)*sizeof(unsigned char);
	}
	else if (layer->hasGeometry(TeRASTER))
	{
		sql = " SELECT MAX(LENGTH(SPATIAL_DATA)) FROM "+ tableName;
		if(!portal->query(sql) || !portal->fetchRow())
		{
			delete portal;
			return false;
		}
		maxSize = portal->getInt(0);
	}
	sizeToAllocate_[tableName] = maxSize;
	delete portal;
	return true;
}

bool 
TeOCIOracle::insertView(TeView *view)
{
	// save its projection
	TeProjection* proj = view->projection();
	if ( !proj || !insertProjection(proj))
	{
		errorMessage_ = "No  possvel inserir vista sem projeo";
		return false;
	}

	string ins = "INSERT INTO te_view (view_id, projection_id, name, user_name, visibility)";
	ins += " VALUES (";
	ins += "te_view_seq.NEXTVAL";
	ins += ", " + Te2String(proj->id());
	ins += ", '" + escapeSequence(view->name ()) + "'";
	ins += ", '" + escapeSequence(view->user ()) + "'";
	ins += ", " + Te2String((int)view->isVisible());
	ins += " )";
	
	if(!execute (ins))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error inserting in the table te_view!";   
		return false;
	}

	TeOCIOraclePortal *ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	ins = "SELECT te_view_seq.CURRVAL FROM DUAL";
	if (!ocip->query(ins))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error in the sequence te_view_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	int index = atoi((const char*)ocip->getData(0));
	view->id(index);
	delete ocip;

	int size = view->size();

	for (int th=0; th<size; th++)
	{
		TeViewNode* node = view->get(th);
		if (node->type() == TeTHEME)
		{
			TeTheme *theme = (TeTheme*) node;
			insertTheme (theme);
		}
		else
		{
			TeViewTree* tree = (TeViewTree*)node;
			insertViewTree (tree);
		}
	}

	// Insert view in the view map
	viewMap_[view->id()] = view;
	return true;
}

bool
TeOCIOracle::insertViewTree (TeViewTree *tree)
{
	
	string ins = "INSERT INTO te_theme (theme_id, view_id, name, ";
	ins += " parent_id, node_type, priority) VALUES (";
	ins += " te_theme_seq.NEXTVAL";
	ins += ", " + Te2String(tree->view());
	ins += ", '" + escapeSequence(tree->name()) + "'";
	ins += ", " + Te2String(tree->parentId());
	ins += ", " + Te2String(tree->type());
	ins += ", " + Te2String(tree->priority());
	ins += ")";
	
	if(!execute(ins))
	{
		if(errorMessage_.empty())
			errorMessage_ = "Error inserting in the table te_theme!";
		return false;
	}

	TeOCIOraclePortal	*ocip = (TeOCIOraclePortal*)getPortal();
	if(!ocip)
		return false;

	string seq = "SELECT te_theme_seq.CURRVAL FROM DUAL";
	if(!ocip->query(seq))
	{
		errorMessage_ = "Error in the sequence te_theme_seq!";  
		delete ocip;
		return false;
	}

	if(!ocip->fetchRow())
	{
		errorMessage_ = "Sequence value not found!";
		delete ocip;
		return false;
	}

	int index = atoi((const char*)ocip->getData(0)); 
	tree->id(index);
	delete ocip;
	
	return true;
}

string TeOCIOracle::concatValues(vector<string>& values, const string& unionString)
{
	string concat = "";
	
	for(unsigned int i = 0; i < values.size(); ++i)
	{
		if(i != 0)
		{
			concat += " || ";

			if(!unionString.empty())
			{
				concat += "'";
				concat += unionString;
				concat += "'";
				concat += " || ";
			}
		}

		concat += values[i];
	}

	return concat;
}

bool TeOCIOracle::getAttributeList(const string& tableName,TeAttributeList& attList)
{
	string sql = " SELECT COLUMN_NAME, DATA_TYPE, DATA_LENGTH, DATA_SCALE ";
	sql += " FROM USER_TAB_COLUMNS WHERE TABLE_NAME = '";
	sql += TeConvertToUpperCase(tableName) +"'";
	
	attList.clear();
	TeDatabasePortal* portal = this->getPortal();
	if(!portal)
		return false;
	if(!portal->query(sql))
	{
		delete portal;
		return false;
	}
	
	while(portal->fetchRow())
	{
		TeAttribute at;
		at.rep_.name_ = portal->getData(0);
		string type = TeConvertToUpperCase(portal->getData(1));
		int scale = portal->getInt(3);
		if((type=="NUMBER" && scale>0) || (type=="FLOAT"))
	        at.rep_.type_ = TeREAL;
		else if(type=="NUMBER")
			at.rep_.type_ = TeINT;
		else if(type=="VARCHAR2")
		{
			at.rep_.type_ = TeSTRING;
			at.rep_.numChar_=portal->getInt(2);
		}
		else if(type=="BLOB")
			at.rep_.type_ = TeBLOB;
		else if(type=="CHAR")
			at.rep_.type_ = TeCHARACTER;
		else if(type=="DATE")
			at.rep_.type_ = TeDATETIME;
		else 
		{
			at.rep_.type_ = TeSTRING;
			at.rep_.numChar_=portal->getInt(2);
		}
		attList.push_back(at);
	}
	delete portal;
	return true;
}

string TeOCIOracle::toUpper(const string& value)
{
	string result  = "upper(";
	       result += value;
		   result += ")";

	return result;
}

string  
TeOCIOracle::getSQLTime(TeTime& time)
{
	string dateTime = time.getDateTime();	
	string result;
	result = "TO_DATE('";
	result += dateTime;
	result += "  ";
	result += "','DD/MM/YYYY HH24:MI:SS')"; 
	return result;
}

//---------------------------------- Oracle portal

TeOCIOraclePortal::TeOCIOraclePortal(TeOCIOracle *pDatabase) 
{
	unsigned long size = 0;
	map<string, unsigned long>::iterator it = pDatabase->sizeToAllocate_.begin();
	while(it!=pDatabase->sizeToAllocate_.end())
	{
		if(size<it->second)
			size = it->second;
		++it;
	}
	cursor_ = new TeOCICursor(pDatabase->connection_, size);
	db_ = pDatabase;
	numRows_ = 0;
	numFields_ = 0;
	curRow_ = 0;
}


TeOCIOraclePortal::~TeOCIOraclePortal () 
{
	if(cursor_)
		delete cursor_;
	cursor_ = NULL;
}


bool TeOCIOraclePortal::isConnected()
{
	return (cursor_->conn()->isConnected());
}


bool TeOCIOraclePortal::isEOF() 
{
	if (!isConnected())
		return true;

	return false;
}

bool
TeOCIOraclePortal::getLinearRing (TeLine2D& line)
{

	double	*points	= NULL;
	try
	{
		int index = atoi (getData ("geom_id"));
		TeBox b (getDouble("lower_x"),getDouble("lower_y"),getDouble("upper_x"),getDouble("upper_y"));
		line.objectId (string (getData ("object_id")));
		line.geomId (index);
		line.setBox (b);

		unsigned int ncoords = getInt("num_coords");
		unsigned int nords = ncoords*2;
			
		if (!cursor_->readBlob(&points))
			return false;
		
		line.clear();

		for (unsigned int i=0;i<nords;++i) 
		{
			double x = points[i];
			++i;
			double y = points[i];
			
			TeCoord2D c = TeCoord2D(x,y);
			line.add(c);
		}
	}
	catch(...)
	{
		errorMessage_ = "Error getting media!";
		return false;
	}
	return true;
}


bool
TeOCIOraclePortal::moveFirst() 
{
	if (cursor_->moveFirst())
		return true;
	else
		return false;
}


bool 
TeOCIOraclePortal::moveNext() 
{
	if(cursor_->moveNext())
		return true;
	return false;
}

bool TeOCIOraclePortal::query (const string &q,TeCursorLocation /* l */ , TeCursorType /* t */, TeCursorEditType /* e */, TeCursorDataType /* dt */)  //Ok
{
	errorMessage_.clear ();

	if (!cursor_->isOpen())
	{
		if(!cursor_->open())
		{
			numRows_ = 0;
			return false;
		}
	}

	if (!cursor_->query(q))
	{
		this->errorMessage_ = cursor_->getErrorMessage();
		numRows_ = 0;
		return false;
	}
	
	numFields_= this->cursor_->numCol();
	
	attList_.clear ();
	int i;
	for(i = 1; i <= numFields_ ; i++)
	{
		TeAttribute attribute;

		switch (cursor_->colType(i))
		{
			case 3: //INTEGER
			attribute.rep_.type_ = TeINT;
			break;

			case 2:  //NUMBER
			case 4: //FLOAT DOUBLE
			attribute.rep_.type_ = TeREAL;
			break;

			case 12: //Date
			attribute.rep_.type_ = TeDATETIME;
			attribute.dateChronon_ = TeSECOND;
			attribute.dateTimeFormat_ = "DD/MM/YYYY HH24:MI:SS";
			break;

			case 113: //Blob
			attribute.rep_.type_ = TeBLOB;
			break;

			case 96: //CHAR
			case 9: //VARCHAR:
			case 1: //VARCHAR2:
			attribute.rep_.type_ = TeSTRING;
			break;

			case 108: //OBJECT: // SDO_GEOMETRY
			attribute.rep_.type_ = TeOBJECT;
			break;
			default :
			attribute.rep_.type_ = TeUNKNOWN;
			break;
		} 
		
		//attribute.rep_.name_ = TeConvertToLowerCase(cursor_->ColName(i)); 
		attribute.rep_.name_ = cursor_->colName(i); 
		attribute.rep_.numChar_ = cursor_->colSize(i);
		attList_.push_back ( attribute );
	}
	curRow_=-1;
	return true;
}


void TeOCIOraclePortal::freeResult () 
{
	cursor_->freeResult();
}


bool TeOCIOraclePortal::fetchRow () 
{
	try
	{	
		if( !isConnected() )
			return false;
		if( numFields_ == 0/* || numRows_ < 1*/)
			return false;
		
		if (curRow_ == -1)
		{
			if(moveFirst())
				curRow_++;
			else
				return false;
		}
		else
		
		if (moveNext())
		{
			curRow_++;
			return true;
		}
		else
			return false;
	}
	
	catch(...) //_com_error &e)
	{
		return false;
	}

	return true;
}

bool TeOCIOraclePortal ::fetchRow (int i)
{
	try
	{	
		if( !isConnected() || (numFields_ == 0))
			return false;
		
		if (curRow_ == -1)
		{
			if(moveFirst())
				curRow_++;
			else
				return false;
		}

		if(curRow_ == i)
			return true;

		if (cursor_->moveTo(i))
		{
			curRow_ = cursor_->currentRow();
			return true;
		}
		return false;
	}
	catch(...)
	{
		return false;
	}
}

char* TeOCIOraclePortal::getData (int i) 
{
	char* result;
	if (i > numFields_ || i < 0)
		return "";
	
	try
	{
		result = cursor_->getFieldValue(i+1);
	}
	catch(...)
	{
		errorMessage_ = "Error!";
		return "";
	}
	return result;
	
}

char* TeOCIOraclePortal::getData (const string &s)
{
	char* result;

	string fieldName;
	size_t pos = s.find(".", 0, 1);
	if (pos != string::npos)
		fieldName = s.substr(pos+1);
	else
		fieldName = s;

	int index = getColumnIndex (fieldName);
	if(index == -1)
		return "";
	else
		result = getData(index);

	return result;
}


int TeOCIOraclePortal::getInt (int i) 
{
	int value = atoi(getData(i));
	return value;
}

int  
TeOCIOraclePortal::getInt (const string& s)
{
	int value = atoi(getData(s));
	return value;
}

double TeOCIOraclePortal ::getDouble (int i) 
{
	string strValue=getData(i);
	size_t find;
	find = strValue.find(",");
	if(find<strValue.size())
		strValue.replace(find, 1, ".");
		
	return atof(strValue.c_str());
}

double  
TeOCIOraclePortal::getDouble (const string& s)
{
	string strValue=getData(s);
	size_t find;
	find = strValue.find(",");
	if(find<strValue.size())
		strValue.replace(find, 1, ".");
	
	return atof(strValue.c_str());
}


bool
TeOCIOraclePortal::getBool(int i)
{
	char* fieldChar=0;
	
	fieldChar = cursor_->getFieldValue(i+1);
	if(fieldChar == 0)
		return false;
	
	return true;
}


bool    
TeOCIOraclePortal::getBool (const string& s)
{
	string fieldName;
	size_t pos = s.find(".", 0, 1);
	if (pos != string::npos)
		fieldName = s.substr(pos+1);
	else
		fieldName = s;

	int index = getColumnIndex (fieldName);
	if(index == -1)
		return false;
	return (getBool(index));
}

bool		
TeOCIOraclePortal::getBlob(const string& s, unsigned char* &data, long& size)
{
	try
	{
		unsigned int len = cursor_->getBlobSize(s);
		data = new unsigned char[len];
		if(!cursor_->readBlob(data, len, s))
			return false;
		size = len;


	}
	catch(...)
	{
		errorMessage_ = "Error getting media!";
		return false;
	}
	return false;
}


TeTime 
TeOCIOraclePortal::getDate (int i) 
{ 
	TeTime temp;
	string result;

	if (i > numFields_ || i < 0)
		return temp;
	
	try
	{
		result = cursor_->getFieldValue(i+1);
	}
	catch(...)
	{
		errorMessage_ = "Error!";
		return temp;
	}
		
	TeTime t(result, TeSECOND, "DDsMMsYYYYsHHsmmsSS");
	return t;
}
	

TeTime 
TeOCIOraclePortal::getDate (const string& s) 
{ 
	string  result;
	TeTime temp;

	string fieldName;
	size_t pos = s.find(".", 0, 1);
	if (pos != string::npos)
		fieldName = s.substr(pos+1);
	else
		fieldName = s;

	int index = getColumnIndex (fieldName);
	if(index == -1)
		return temp;
	else
		result = getData(index);

	TeTime t(result, TeSECOND, "DDsMMsYYYYsHHsmmsSS") ;
	return t;
}

string
TeOCIOraclePortal::getDateAsString(int i)
{
	TeTime t = this->getDate(i);
	string date = t.getDateTime ();

	if (!date.empty())
	{		string tval = " TO_DATE ('"+ date +"','DDsMMsYYYYsHHsmmsSS') ";
		return tval;
	}
	else
		return "";
}

string 
TeOCIOraclePortal::getDateAsString(const string& s)
{
	TeTime t = this->getDate(s);
	string date = t.getDateTime ();

	if (!date.empty())
	{		string tval = " TO_DATE ('"+ date +"','DDsMMsYYYYsHHsmmsSS') ";
		return tval;
	}
	else
		return "";
}


bool
TeOCIOraclePortal::fetchGeometry (TePolygon& poly)
{
	try 
	{
		TeLine2D line;
		int numHoles = atoi(this->getData("num_holes"));
		bool flag = this->fetchGeometry(line);

		TeLinearRing ring(line);
		poly.objectId ( ring.objectId());
		poly.geomId ( ring.geomId() );
		poly.add ( ring );
		int parentId = poly.geomId();

		while (flag && numHoles>0 && (atoi(this->getData("parent_id")) == parentId)) // get holes
		{
			TeLine2D aux;
			this->getLinearRing(aux);
			TeLinearRing hole(aux);
			poly.add (hole);
			flag = this->fetchRow();
		}
		return flag;
	}
	catch(...)
	{
		errorMessage_ = "Error fetchGeometry!";
		return false;
	}
}

bool 
TeOCIOraclePortal::fetchGeometry (TeLine2D& line)
{	
	try
	{
		getLinearRing(line);
		return (this->fetchRow());
	}
	catch(...)
	{
		errorMessage_ = cursor_->getErrorMessage();
		return false;
	}
}

bool 
TeOCIOraclePortal::fetchGeometry(TePoint& p)
{

	try 
	{
		p.geomId( atoi(getData("geom_id")));
		p.objectId( string(getData("object_id")));
		double x = getDouble("x");
		double y = getDouble("y");
		TeCoord2D c(x, y);
		p.add(c);
		return(this->fetchRow());
	} 
	catch(...)
	{
		errorMessage_ = "Error fetchGeometry!";
		return false;
	}
}

bool 
TeOCIOraclePortal::fetchGeometry(TeNode& n)
{
	try 
	{
		TeCoord2D c(getDouble("x"), getDouble("y"));
		n.geomId( atol(getData("geom_id")));
		n.objectId( string(getData("object_id")));
		n.add(c);
		return(this->fetchRow());
	} 
	catch(...)
	{
		errorMessage_ = "Error fetchGeometry!";
		return false;
	}
	return true;
}

bool
TeOCIOraclePortal::getRasterBlock(unsigned long& size, unsigned char* ptData)
{
	// gets the spatial data
	// expects that the data size will be a block of double maximum

	try
	{
		unsigned int len = 0;
		if(!getCursor()->readBlob(ptData, len, "spatial_data"))
			return false;
		size = len;


	}
	catch(...)
	{
		size = 0;
		errorMessage_ = "ERRO!";
		return false;
	}
	return true;
}


