/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeImportExport.h
    This file contains functions to deal with the importing/exporting
	of geographical data to and from  a TerraLib database in usual
	formats
*/
#ifndef __TERRALIB_INTERNAL_IMPORTEXPORT_H
#define __TERRALIB_INTERNAL_IMPORTEXPORT_H

#include "TeTable.h"
#include "TeBox.h"
#include "TeDataTypes.h"

#include <string>
using namespace std;

class TeLayer;
class TeSPRFile;
class TeAsciiFile;
class TeProjection;
class TeDatabase;

/** @defgroup ImpExp Importing and exporting 
 *  TerraLib importing and exporting functions
 *  @{
 */

/** @defgroup MIDMIF MID/MIF format
 *  @ingroup ImpExp
 *  Functions related to MID MIF format
 *  @{
 */
 /** Imports a geo data in MID/MIF format to a TerraLib database
	\param mifFileName	MIF file name
	\param db			pointer to a TerraLib database
	\param layerName	name of the layer that will contain the data (optional)
	\return a pointer to created layer if the data was imported successfully and 0 otherwise
*/
TeLayer* TeImportMIF(const string& mifFileName, TeDatabase* db, const string& layerName="");

/** Imports a MID/MIF file to a specific layer TerraLib
	\param layer			pointer to a layer
	\param mifFileName		MIF file name
	\param attrTableName	the name that the attribute table will be saved
	\param objectIdAttr		name of the attribute that has the identification of objects
	\param chunkSize		the number of objects in memory before save them in the layer
	\return TRUE if the data was imported successfully and FALSE otherwise
*/
bool TeImportMIF(TeLayer* layer, const string& mifFileName, string attrTableName="",  
				 string objectIdAttr="", int unsigned chunkSize = 60);

/** Reads the attributes definition of a data in MID/MIF format
	\param mifFileName		MIF file name
	\param attList			to return the list of attributes defined in the file
*/
void TeReadMIFAttributeList(const string& mifFileName, TeAttributeList& attList);

/** Reads the geographical projection of a data in MID/MIF format
	\param mifFileName		MIF file name
	\return a pointer to a TeProjection
*/
TeProjection* TeReadMIFProjection(const string& mifFileName);

/** Exports a layer in a TerraLib database to a file in MID/MIF format
   \param layer			pointer to the layer
   \param mifFileName	name of the output MID/MIF file
   \param tableName		name of the attribute table to be exported
   \return true if the data was successfully exported and FALSE otherwise
*/
bool TeExportMIF(TeLayer* layer, const string& mifFileName, const string& tableName);

/** @} */ // End of subgroup MID/MIF format

/** @defgroup SPR SPRING GEO/TAB format
   @ingroup ImpExp
   Functions related to SPRING format
   @{
 */ 
/** Imports a geo data in SPRING geo/tab format to a TerraLib database
	\param geoFileName	GEO/TAB file name
	\param db			pointer to a TerraLib database
	\param layerName	name of the layer that will contain the data (optional)
	\return a pointer to created layer if the data was imported successfully and 0 otherwise
*/
TeLayer* TeImportGEO(const string& geoFileName, TeDatabase* db, const string& layerName="");

/** Imports a GEO/TAB data to a layer TerraLib
    \param layer			pointer to a layer
	\param geoFileName		GEO/TAB file name
	\param attrTableName	the name that the attribute table will be saved
	\param chunkSize the number of objects in memory before save them in the layer
*/
bool TeImportGEO(TeLayer* layer, const string& geoFileName, string attrTableName="", unsigned int chunkSize=60);

/** Reads the list of attributes  of GEO/TAB file
	\param geoFileName	DBF input file name
	\param attList		to return the DBF list of attributes
	\param labelName	to return the name of the column that is the link with geometries
	\return true if the list of attributes were was successfully read and false otherwise
*/
void TeReadGeoAttributeList(const string& geoFileName, TeAttributeList& attList, string& labelName);

/** Exports a layer in a TerraLib database to a file in ASCII SPRING format
    \param layer		pointer to the layer
    \param sprFileName	name of the output SPRING file
    \param tableName	name of the attribute table to be exported
    \param cat			Spring's category
	\param name			name of the column that has the object identifier or class indication
    \return TRUE if the data was successfully exported and FALSE otherwise
    \note Parameter 'name' is the Object name if exporting to categories Cadastral or Network .
   Is the name of the column that has the quote value if Numeric category or the name
   of the column that has the class information if category Thematic.
*/
bool TeExportSPR(TeLayer* layer, const string& sprFileName, const string& tableName, 
			 TeSpringModels cat, const string& name);
/** @} */ // End of subgroup SPRING format

/** @defgroup DBF DBF tables
 *  @ingroup ImpExp
 *  Functions related to DBF table format
 *  @{
 */

/** Reads the list of attributes specified in DBF file 
   \param dbfFileName	DBF input file name
   \param attList		returns the DBF list of attributes
   \return TRUE if the list of attributes were was successfully read and FALSE otherwise
*/
bool TeReadDBFAttributeList(const string& dbfFileName, TeAttributeList& attList);

/** Reads a set of records from a shapelib to a structure in memory
	\param dbfFileName	DBF file name
	\param table		to return the set of records in memory
	\param nrecords		number of records to be read
	\param rinitial		number of the initial record
	\param createAuto	flag used to indicate if an auto number columns should be added to each record
*/
bool TeReadDBFAttributes (const string& dbfFileName, TeTable& table, 
						  int nrecords=-1, int rinitial=0, bool createAuto=false);
/** Imports a DBF table to a TerraLib database
	This function imports a DBF to a TerraLib database. An attribute table can be associated to a
	particular layer, or as an external table that will be linked to any layer when genereated a 
	theme. A column called object_id_ will be added with unique values automatically generated.
	\param dbfFileName	DBF file name
	\param db			pointer to a TerraLib database
	\param layer		pointer to a layer to which the table will associated
	\param tableType	type of the attribute table
*/
bool TeImportDBFTable(const string& dbfFileName, TeDatabase* db, TeLayer* layer=0,TeAttrTableType tableType=TeAttrExternal, string objectIdAtt = "object_id_");

/** @} */ // End of subgroup DBF format

/** @defgroup CSV format
 *  @ingroup ImpExp
 *  Functions related to CSV format
 *  @{
 */
/** Creates an attribute table in a TerraLib database and imports a CSV file to this table 
	\param csvFileName			CSV file name
	\param attrTable			attribute table specifications
	\param db					pointer to a TerraLib database
	\param layer				pointer to a layer to which the table will associated
	\param separator			character used to separate the data in the CSV file 
	\param firstLineIsHeader	indicates if the first line is a header
	\param chunkSize			number of objects in memory before save them in the database
	\param createAutoNum		indicates if the table first attribute is an autonumber which must be generated  
	\return TRUE if the data was imported successfully and FALSE otherwise
*/
bool TeImportCSVFile(const string& csvFileName, TeTable& attrTable, TeDatabase* db, TeLayer* layer=0, 
					 const char& separator=',', const bool& firstLineIsHeader=false,   
					 const unsigned int& chunkSize=100, const bool& createAutoNum = false);

/** Imports a CSV file to a layer in a TerraLib database, its geometries (points) and attributes
	\param csvFileName			CSV file name pointer to a layer
	\param attributeX			attribute index which provides the coordinate X
	\param attributeY			attribute index which provides the coordinate Y
	\param attrTable			attribute table specifications
	\param db					pointer to a TerraLib database
	\param layer				pointer to an empty layer which will be filled from CSV file
	\param separator			character used to separate the data in the CSV file 
	\param firstLineIsHeader	indicates if the first line is a header
	\param chunkSize			number of objects in memory before save them in the database
	\param createAutoNum		indicates if the table first attribute is an autonumber which must be generated  
	\return TRUE if the data was imported successfully and FALSE otherwise
*/
bool TeImportCSVFile(const string& csvFileName, const int& attributeX, const int& attributeY, 
					 TeTable& attrTable, TeDatabase* db, TeLayer* layer, 
					 const char& separator=',', const bool& firstLineIsHeader=false,   
					 const unsigned int& chunkSize=100, const bool& createAutoNum = false);

/** @} */ // End of subgroup CSV format


/** @defgroup SHP Shapefile format
 *  @ingroup ImpExp
 *  Functions related to Shapefile format. Uses some functions related to DBF attribute table format.
 *  @{
 */ 
/** Imports a geo data in a shapefile format to a TerraLib database
	\param shpFileName	Shapefile name
	\param db			pointer to a TerraLib database
	\param layerName	name of the layer that will contain the data (optional)
	\return a pointer to created layer if the data was imported successfully and 0 otherwise
*/
TeLayer* TeImportShape(const string& shpFileName, TeDatabase* db, const string& layerName="");

/** Imports a shapefile to a layer TerraLib
	\param layer			pointer to a layer
	\param shpFileName		shapefile file name
	\param attrTableName	the name that the attribute table will be saved
	\param objectIdAttr		name of the attribute that has the identification of objects
	\param chunkSize		the number of objects in memory before save them in the layer
*/
bool TeImportShape (TeLayer* layer, const string& shpFileName, string attrTableName="",  
					string objectIdAttr="", unsigned int  chunkSize = 60);

/** Exports a layer in a TerraLib database to a file in MID/MIF format
   \param layer			pointer to the layer
   \param shpFileName	name of the output shapefile
   \param tableName		name of the attribute table to be exported
   \return TRUE if the data was successfully exported and FALSE otherwise
*/
bool TeExportShapefile(TeLayer* layer, const string& shpFileName, const string& tableName);
/** @} */ // End of subgroup SHAPEFILE format

/** @defgroup BNA BNA format
 *  @ingroup ImpExp
 *  Functions related to BNA format (Atlas GIS BNA)
 *  @{
 */
 
/** Imports a geo data in BNA format to a TerraLib database
	\param bnaFileName	BNA file name
	\param db			pointer to a TerraLib database
	\param layerName	name of the layer that will contain the data (optional)
	\return a pointer to created layer if the data was imported successfully and 0 otherwise
*/
TeLayer* TeImportBNA(const string& bnaFileName, TeDatabase* db, const string& layerName="");

/** Imports a BNA file to a specific layer TerraLib
	\param layer			pointer to a layer
	\param bnaFileName		BNA file name
	\param attrTableName	the name that the attribute table will be saved
	\param attList			to return the list of attributes defined in the bna file, used only to name the fields
	\param chunkSize		the number of objects in memory before save them to the layer
	\param linkCol			name of the column that will contain the link to geometries
	\return TRUE if the data was imported successfully and FALSE otherwise
*/
bool TeImportBNA(TeLayer* layer, const string& bnaFileName, string attrTableName,  
				 TeAttributeList& attList, int unsigned chunkSize, const string& linkCol = "");

/** Reads the attributes definition of a data in BNA format 
	\param bnaFileName	BNA file name
	\param attList		to return the attribute list
	\return TRUE if the attribute list is successfully read and FALSE otherwise
*/
bool TeReadBNAAttributeList(const string& bnaFileName, TeAttributeList& attList);

/** @} */ // End of subgroup BNA format

/** \example importMIDMIF.cpp
 Shows how to import geographical data in MID/MIF format
 */

/** \example importShape.cpp
 Shows how to import geographical data in shapefile format
 */

/** \example importDBF.cpp
 Shows how to import a DBF table
 */

/** \example importGeoTab.cpp
 Shows how to import geographical data in SPRING GEO/TAB format
 */

/** \example importCSV.cpp
 Shows how to import a MID/MIF data file into a new layer and a CSV file as a second attribute table of the layer
*/

/** @} */ // End of group ImpExp format

#endif

