/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

/**
 * @file TePDIAgnostic.hpp
 * @brief This file contains a set of macros, routines and classes to deal with
 * system checking and logging facility.
 * They should NOT be used by anyone because the support and interfaces 
 * can be changed in future. THIS IS FOR INTERNAL USE ONLY.
 * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
 *
 * @note The following macros will be used:
 * 
 * @param PDIAGN_DEBUG_MODE If defined, debug code macros will be compiled and
 * used.
 * @param PDIAGN_ENABLE_STDOUT_LOG If defined, messages also will be logged
 * to STDOUT ( otherwise only TeErrorLog will log messages ).
 */

#ifndef TEPDIAGNOSTIC_HPP
  #define TEPDIAGNOSTIC_HPP

  #include <stdlib.h>
  #include <stdio.h>
  #include <sstream>
  #include <string>
  #include <iostream>
  #include <exception>
  
  #include <TeException.h>
  #include <TeErrorLog.h>   

  /**
   * TePDIAgnostic debug mode selection based on NDEBUG define.
   */
  #ifndef PDIAGN_DEBUG_MODE
    #ifndef NDEBUG
      #define PDIAGN_DEBUG_MODE
    #endif
  #endif   

  /**
   * Logs a message to stdout
   *
   * @param message Message to be logged.
   */
  #ifdef PDIAGN_ENABLE_STDOUT_LOG
    #define PDIAGN_LOGMSG_STDOUT( message ) \
      std::cout << std::endl << "Message : " \
        << __FILE__ \
        << ":" << __LINE__ \
        << " - " << TePDIAgnostic::to_string( message ) \
        << std::endl;
  #else
    #define PDIAGN_LOGMSG_STDOUT( message ) {};
  #endif

    /**
    * Logs a error message to stderr
    *
    * @param message Message to be logged.
    */
  #ifdef PDIAGN_ENABLE_STDOUT_LOG
    #define PDIAGN_LOGERR_STDOUT( message ) \
      std::cerr << std::endl << "Error : " \
        << __FILE__ \
        << ":" << __LINE__ \
        << " - " << TePDIAgnostic::to_string( message ) \
        << std::endl;
  #else
    #define PDIAGN_LOGERR_STDOUT( message ) {};
  #endif

    /**
    * Logs a warning message to stdout
    *
    * @param message Message to be logged.
    */
  #ifdef PDIAGN_ENABLE_STDOUT_LOG
    #define PDIAGN_LOGWARN_STDOUT( message ) \
      std::cout << std::endl << "Warning : " \
        << __FILE__ \
        << ":" << __LINE__ \
        << " - " << TePDIAgnostic::to_string( message ) \
        << std::endl;
  #else
    #define PDIAGN_LOGWARN_STDOUT( message ) {};
  #endif
        

  /**
   * Logs a message.
   *
   * @param message Message to be logged.
   */
  #define PDIAGN_LOGMSG( message ) \
  { \
    TeErrorLog::instance().insert( LOG_MESSAGE, \
      TePDIAgnostic::to_string( message ) ); \
    PDIAGN_LOGMSG_STDOUT( message ); \
  };

  /**
   * Logs a message.
   *
   * @param message Message to be logged.
   */
  #define PDIAGN_LOGERR( message ) \
  { \
    TeErrorLog::instance().insert( UNKNOWN_ERROR_TYPE, \
      TePDIAgnostic::to_string( message ) ); \
    PDIAGN_LOGERR_STDOUT( message ); \
  };
      
  /**
   * Logs a warning message.
   *
   * @param message Message to be logged.
   */
  #define PDIAGN_LOGWARN( message ) \
  { \
    TeErrorLog::instance().insert( LOG_MESSAGE, \
     TePDIAgnostic::to_string( message ) ); \
    PDIAGN_LOGWARN_STDOUT( message ); \
  };
      
  /**
   * Logs a message to stderr and throws.
   *
   * @param message Message to be logged.
   */
  #define PDIAGN_LOG_AND_THROW( message ) \
  { \
    PDIAGN_LOGERR_STDOUT( message ); \
    throw TeException( UNKNOWN_ERROR_TYPE, \
      TePDIAgnostic::to_string( message ), false ); \
  };
      
  /**
   *  Checks if value is true and throws an exception if not.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_TRUE_OR_THROW( value , message ) \
    if( value == 0 ) { \
      PDIAGN_LOGERR_STDOUT( message ); \
      throw TeException( UNKNOWN_ERROR_TYPE, \
        TePDIAgnostic::to_string( message ), false ); \
    };      

  /**
   * Variable watching.
   *
   * @param variable Variable to be logged.
   */
  #define PDIAGN_WATCH( variable ) \
    { \
      PDIAGN_LOGMSG( "WATCH - " + TePDIAgnostic::to_string( #variable ) + \
      "=" + TePDIAgnostic::to_string( variable ) ); \
    };

  /**
   *  Checks if value is true and logs an warning message if not.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_TRUE_OR_LOG( value , message ) \
    if( value == 0 ) { \
      PDIAGN_LOGWARN( message ); \
    };

  /**
   * Checks if value is true. For false values a warning message will be logged 
   * and a return of context with false value will be done.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_TRUE_OR_RETURN( value , message ) \
    if( value == 0 ) { \
      PDIAGN_LOGWARN( message ); \
      return false; \
    };

  /**
   * Logs a warning message will and return false.
   *
   * @param message Message to be logged.
   */
  #define PDIAGN_LOG_AND_RETURN( message ) \
    { \
      PDIAGN_LOGWARN( message ); \
      return false; \
    };

  /**
   *  Checks if value is false and logs an warning message if not.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_FALSE_OR_LOG( value , message ) \
    if( value != 0 ) { \
      PDIAGN_LOGWARN( message ); \
    };

  /**
   *  Checks if two values are equal and throws an exception if not.
   *
   * @param value1 Value to be checked.
   * @param value2 Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_CHECK_EQUAL( value1 , value2 , message ) \
    PDIAGN_TRUE_OR_THROW( ( ((double)value1) == ((double)value2) ), \
    std::string( "Values must be equal [" ) + \
    TePDIAgnostic::to_string( value1 ) + "!=" + \
    TePDIAgnostic::to_string( value2 ) + "] - " + \
    TePDIAgnostic::to_string( message ) );

  /**
   *  Checks if two values are diferent and throws an exception if not.
   *
   * @param value1 Value to be checked.
   * @param value2 Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_CHECK_NOTEQUAL( value1 , value2 , message ) \
    PDIAGN_TRUE_OR_THROW( ( ((double)value1) != ((double)value2) ), \
    std::string( "Values can't be equal [" ) + \
    TePDIAgnostic::to_string( #value1 ) + std::string( "==" ) + \
    TePDIAgnostic::to_string( #value2 ) + std::string( "==" ) + \
    TePDIAgnostic::to_string( value1 ) + std::string( "]" ) );

  /**
   *  Checks if two values are equal ( within an EPS ) and
   *  throws an exception if not.
   *
   * @param value1 Value to be checked.
   * @param value2 Value to be checked.
   * @param eps EPS ( threshold )
   * @param message Message to be logged.
   */
  #define PDIAGN_CHECK_EPS( value1 , value2 , eps , message ) \
    { \
      PDIAGN_TRUE_OR_THROW( ( eps >= 0), "Invalid eps" ); \
      double PDIAGN_CHECK_EPS_double_diff = 0; \
      double PDIAGN_CHECK_EPS_double_value1 = (double)value1; \
      double PDIAGN_CHECK_EPS_double_value2 = (double)value2; \
      double PDIAGN_CHECK_EPS_double_eps = (double)eps; \
      if( PDIAGN_CHECK_EPS_double_value1 < PDIAGN_CHECK_EPS_double_value2 ) { \
        PDIAGN_CHECK_EPS_double_diff = ( PDIAGN_CHECK_EPS_double_value2 - \
          PDIAGN_CHECK_EPS_double_value1 ); \
      } else { \
        PDIAGN_CHECK_EPS_double_diff = ( PDIAGN_CHECK_EPS_double_value1 - \
          PDIAGN_CHECK_EPS_double_value2 ); \
      }; \
      PDIAGN_TRUE_OR_THROW( \
        ( PDIAGN_CHECK_EPS_double_diff <= PDIAGN_CHECK_EPS_double_eps ), \
        std::string( "Values are not equal: " ) + \
        TePDIAgnostic::to_string( #value1 ) + \
        std::string( "=[") + \
        TePDIAgnostic::to_string( PDIAGN_CHECK_EPS_double_value1 ) + \
        std::string( "] " ) + \
        TePDIAgnostic::to_string( #value2 ) + \
        std::string( "=[") + \
        TePDIAgnostic::to_string( PDIAGN_CHECK_EPS_double_value2 ) + \
        std::string( "] eps=[") + \
        TePDIAgnostic::to_string( PDIAGN_CHECK_EPS_double_eps ) + \
        std::string( "] diff=[") + \
        TePDIAgnostic::to_string( PDIAGN_CHECK_EPS_double_diff ) + \
        std::string( "] - " ) + \
        TePDIAgnostic::to_string( message ) \
        ); \
    };

  /**
   * Throws an exception for not implemented source.
   */
  #define PDIAGN_NOT_IMPLEMENTED \
    PDIAGN_LOG_AND_THROW( "Not Implemented." );

  /**
   *  Checks if Debug mode is enabled and throws an exception if not.
   */
  #define PDIAGN_DEBUG_MODE_CHECK \
    PDIAGN_TRUE_OR_THROW( TePDIAgnostic::debugModeCheck() , \
    "PDIAGN_DEBUG_MODE not enabled at TePDI compilation time." );
    
  /**
   * Checks if value is true and throws an exception if not.
   *
   * @note This macro will be disabled for non PDI debug mode.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #ifdef PDIAGN_DEBUG_MODE
    #define PDIAGN_DEBUG_CONDITION( value , message ) \
      PDIAGN_TRUE_OR_THROW( value , message );
  #else
    #define PDIAGN_DEBUG_CONDITION( value , message ) {};
  #endif
  
  /**
   * Checks if value is true. For false values a warning message will be logged 
   * and a return of context with false value will be done.
   *
   * @note This macro will be disabled for non PDI debug mode.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #ifdef PDIAGN_DEBUG_MODE
    #define PDIAGN_DEBUG_RETURN( value , message ) \
      PDIAGN_TRUE_OR_RETURN( value , message );
  #else
    #define PDIAGN_DEBUG_RETURN( value , message ) {};
  #endif  

  /**
   * Checks if value is true and throws an exception if not.
   *
   * @note This macro will be disabled for non PDI debug mode.
   * @note DEPRECATED - WILL BE REMOVED IN THE FUTURE.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_PRE_CONDITION( value , message ) \
    PDIAGN_DEBUG_CONDITION( value , message );

  /**
   * Checks if value is true and throws an exception if not.
   *
   * @note This macro will be disabled for non debug mode.
   * @note DEPRECATED - WILL BE REMOVED IN THE FUTURE.
   *
   * @param value Value to be checked.
   * @param message Message to be logged.
   */
  #define PDIAGN_POS_CONDITION( value , message ) \
    PDIAGN_DEBUG_CONDITION( value , message );

  /**
   *  This namespace contains a set of routines and classes to deal with
   *  system checking and logging facility.
   * @ingroup PDIModule
   */
  namespace TePDIAgnostic{

    /**
     * Data conversion to string.
     *
     * @param data Data to be converted.
     * @return The converted string.
     */
     template< class T >
     std::string to_string( const T& data )
     {
        std::stringstream temp_ss;
        temp_ss.setf(std::ios_base::fixed);
        temp_ss << data;
        return temp_ss.str();
     }

    /**
     * Checks if the code was compiled with debug mode.
     *
     * @return true if debug mode was used at compilation time. false if not.
     */
     bool debugModeCheck();

  };

#endif //TEPDIAGNOSTIC_HPP
