#include "TePDIBlending.hpp"

#include <TePDIBlendStratFactory.hpp>
#include <TePDIStatistic.hpp>
#include <TePDIContrast.hpp>
#include <TePDIUtils.hpp>
#include <TePDIAgnostic.hpp>

#include <TeOverlay.h>
#include <TeGeometryAlgorithms.h>


TePDIBlending::TePDIBlending()
{
}


TePDIBlending::~TePDIBlending()
{
}


void TePDIBlending::ResetState( const TePDIParameters& )
{
}


bool TePDIBlending::CheckParameters( const TePDIParameters& parameters ) const
{
  /* Checking input_raster1 */
  
  TePDITypes::TePDIRasterPtrType input_raster1;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "input_raster1", 
    input_raster1 ),
    "Missing parameter: input_raster1" );
  PDIAGN_TRUE_OR_RETURN( input_raster1.isActive(),
    "Invalid parameter: input_raster1 inactive" );
  PDIAGN_TRUE_OR_RETURN( input_raster1->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: input_raster1 not ready" );    
    
  /* Checking input_raster2 */
  
  TePDITypes::TePDIRasterPtrType input_raster2;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "input_raster2", 
    input_raster2 ),
    "Missing parameter: input_raster2" );
  PDIAGN_TRUE_OR_RETURN( input_raster2.isActive(),
    "Invalid parameter: input_raster2 inactive" );
  PDIAGN_TRUE_OR_RETURN( input_raster2->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: input_raster2 not ready" );    
    
  /* Checking output_raster */
  
  TePDITypes::TePDIRasterPtrType output_raster;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "output_raster", 
    output_raster ),
    "Missing parameter: output_raster" );
  PDIAGN_TRUE_OR_RETURN( output_raster.isActive(),
    "Invalid parameter: output_raster inactive" );
  PDIAGN_TRUE_OR_RETURN( output_raster->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: output_raster not ready" );    
    
  /* channels1 parameter checking */

  std::vector< int > channels1;
  PDIAGN_TRUE_OR_RETURN( 
    parameters.GetParameter( "channels1", channels1 ), 
    "Missing parameter: channels1" );
  for( unsigned int channels1_index = 0 ; 
    channels1_index < channels1.size() ; 
    ++channels1_index ) {
    
    PDIAGN_TRUE_OR_RETURN( ( channels1[ channels1_index ] >= 0 ) &&
      ( channels1[ channels1_index ] < input_raster1->nBands() ),
      "Invalid parameter: channels1" );
  }
  
  /* channels2 parameter checking */

  std::vector< int > channels2;
  PDIAGN_TRUE_OR_RETURN( 
    parameters.GetParameter( "channels2", channels2 ), 
    "Missing parameter: channels2" );
  PDIAGN_TRUE_OR_RETURN( ( channels2.size() == channels1.size() ),
    "Size mismatch between channels1 and channels2" );
  for( unsigned int channels2_index = 0 ; 
    channels2_index < channels2.size() ; 
    ++channels2_index ) {
    
    PDIAGN_TRUE_OR_RETURN( ( channels2[ channels2_index ] >= 0 ) &&
      ( channels2[ channels2_index ] < input_raster2->nBands() ),
      "Invalid parameter: channels2" );
  } 
    
  /* Checking raster polygons */
  
  TePolygon raster1_polygon;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster1_polygon", 
    raster1_polygon ), "Missing parameter : raster1_polygon" );
  PDIAGN_TRUE_OR_RETURN(
    ( ! input_raster1->begin( raster1_polygon, TeBoxPixelIn, 0 ).end() ),
    "Invalid raster1_polygon" )
  PDIAGN_TRUE_OR_RETURN( ( raster1_polygon.size() == 1 ),
    "Invalid raster1_polygon" )
    
  TePolygon raster2_polygon;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_polygon", 
    raster2_polygon ), "Missing parameter : raster2_polygon" );
  PDIAGN_TRUE_OR_RETURN(
    ( ! input_raster2->begin( raster2_polygon, TeBoxPixelIn, 0 ).end() ),
    "Invalid raster2_polygon" )
  PDIAGN_TRUE_OR_RETURN( ( raster2_polygon.size() == 1 ),
    "Invalid raster2_polygon" )    
    
  /* Checking raster2 polygon offsets */
  
  double raster2_pol_offset_x = 0;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_pol_offset_x", 
    raster2_pol_offset_x ), "Missing parameter : raster2_pol_offset_x" );  

  double raster2_pol_offset_y = 0;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_pol_offset_y", 
    raster2_pol_offset_y ), "Missing parameter : raster2_pol_offset_y" );
    
  /* Checking the blending_type parameter */ 
  
  std::string blending_type;
  
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "blending_type", 
    blending_type ), "Missing parameter: blending_type" );
    
  TePDIBlendStratFactory::TeFactoryMap::iterator fac_map_it = 
    TePDIBlendStratFactory::instance().find( blending_type );
    
  PDIAGN_TRUE_OR_RETURN( 
    ( fac_map_it != TePDIBlendStratFactory::instance().end() ),
    "Invalid parameter: blending_type" );
       
  return true;
}


bool TePDIBlending::RunImplementation()
{
  /* Retriving parameters */
  
  TePDITypes::TePDIRasterPtrType input_raster1;
  params_.GetParameter( "input_raster1", input_raster1 );

  TePDITypes::TePDIRasterPtrType input_raster2;
  params_.GetParameter( "input_raster2", input_raster2 );

  TePDITypes::TePDIRasterPtrType output_raster;
  params_.GetParameter( "output_raster", output_raster );

  std::vector< int > channels1;
  params_.GetParameter( "channels1", channels1 );

  std::vector< int > channels2;
  params_.GetParameter( "channels2", channels2 );
  
  TePolygon raster1_polygon;
  params_.GetParameter( "raster1_polygon", raster1_polygon );

  TePolygon raster2_polygon;
  params_.GetParameter( "raster2_polygon", raster2_polygon );

  double raster2_pol_offset_x = 0;
  params_.GetParameter( "raster2_pol_offset_x", raster2_pol_offset_x );

  double raster2_pol_offset_y = 0;
  params_.GetParameter( "raster2_pol_offset_y", raster2_pol_offset_y );
  
  std::string blending_type;
  params_.GetParameter( "blending_type", blending_type );
  
  bool auto_equalize = false;
  if( params_.CheckParameter( "auto_equalize", 
    TePDIParameters::TePDIIntParam ) ) {
    
    auto_equalize = true;
  }
  
  /* Dumyy value definition */
 
  bool output_raster_uses_dummy = output_raster->params().useDummy_;
  double output_raster_dummy = 0;
  
  if( output_raster_uses_dummy ) {
    output_raster_dummy = output_raster->params().dummy_[ 0 ];
  }
  
  if( params_.CheckParameter( "dummy_value", 
    TePDIParameters::TePDIDoubleParam ) ) {
    
    params_.GetParameter( "dummy_value", output_raster_dummy );
    
    output_raster_uses_dummy = true;
  }

  /* Building raster 2 polygon - reference raster1 */
  
  TePolygon raster2_polygon_ref1;
  {
    TePolygon raster2_polygon_indexed_ref2;
    TePDIUtils::MapCoords2RasterIndexes( raster2_polygon,
      input_raster2, raster2_polygon_indexed_ref2 );
        
    double r2_min_x = 
      MIN( raster2_polygon_indexed_ref2.box().x1(),
      raster2_polygon_indexed_ref2.box().x2() );
    double r2_min_y = 
      MIN( raster2_polygon_indexed_ref2.box().y1(),
      raster2_polygon_indexed_ref2.box().y2() ); 
     
    for( unsigned int lr_index = 0 ; 
      lr_index < raster2_polygon_indexed_ref2.size() ; ++lr_index ) {
      for( unsigned int c2d_index = 0 ; 
        c2d_index < raster2_polygon_indexed_ref2[ lr_index ].size() ; 
        ++c2d_index ) {
          
        raster2_polygon_indexed_ref2[ lr_index ][ c2d_index ].x( 
          raster2_polygon_indexed_ref2[ lr_index ][ c2d_index ].x() +
          raster2_pol_offset_x - r2_min_x );
        raster2_polygon_indexed_ref2[ lr_index ][ c2d_index ].y( 
          raster2_polygon_indexed_ref2[ lr_index ][ c2d_index ].y() +
          raster2_pol_offset_y - r2_min_y );
      }
    }
  
    TePDIUtils::MapRasterIndexes2Coords( raster2_polygon_indexed_ref2,
      input_raster1, raster2_polygon_ref1 );
  }
    
  const short pols_relation = TeRelation( raster1_polygon, 
    raster2_polygon_ref1 );
  
  /* Reseting output raster */
  
  TeBox global_box_ref1 = raster1_polygon.box();
  updateBox( global_box_ref1, raster2_polygon_ref1.box() );
  
  TeRasterParams output_raster_params = output_raster->params();
  
  output_raster_params.nBands( channels1.size() );
  if( input_raster1->projection() != 0 ) {
    output_raster_params.projection( 
      TeProjectionFactory::make( input_raster1->projection()->params() ) );
  }
  output_raster_params.boxResolution( global_box_ref1.x1(), 
    global_box_ref1.y1(), global_box_ref1.x2(), global_box_ref1.y2(), 
    input_raster1->params().resx_, input_raster1->params().resy_ );
  if( output_raster_uses_dummy ) {
    output_raster_params.setDummy( output_raster_dummy, -1 );
  } else {
    output_raster_params.useDummy_ = false;
  }
     
  PDIAGN_TRUE_OR_RETURN( output_raster->init( output_raster_params ),
    "Output raster reset error" );    
  
  /* Auto equalizing ( if required ) */
  
  if( auto_equalize ) {
    /* Defining the overlap polygon over raster1 ( raster1 world index )*/
    
    if( pols_relation != TeDISJOINT ) {
      
      TePolygonSet ps_raster1_ref1;
      ps_raster1_ref1.add( raster1_polygon );

      TePolygonSet ps_raster2_ref1;
      ps_raster2_ref1.add( raster2_polygon_ref1 );
            
      TePolygonSet ps_intersection_ref1;
      TeOVERLAY::TeIntersection( ps_raster1_ref1, 
        ps_raster2_ref1, ps_intersection_ref1 );      
      
      PDIAGN_TRUE_OR_RETURN( ( ps_intersection_ref1.size() == 1 ),
        "Invalid polygons intersection" );  
        
      /* Creating the statistic instance to analyse the overlap
         area over raster1 ( for all bands ) */
      
      TePDIParameters stat_params;
      
      TePDITypes::TePDIRasterVectorType rasters;
      for( unsigned int channels1_index = 0 ; 
        channels1_index < channels1.size() ; ++channels1_index ) {
        
        rasters.push_back( input_raster1 );
      }
      stat_params.SetParameter( "rasters", rasters );
      stat_params.SetParameter( "bands", channels1 );
      stat_params.SetParameter( "polygonset", ps_intersection_ref1 );
      
      TePDIStatistic stat_instance;
      PDIAGN_TRUE_OR_RETURN( stat_instance.Reset( stat_params ),
        "TePDIStatistc reset error" );
        
      /* Creating a new raster2 with equalized data */
      
      TePDITypes::TePDIRasterPtrType new_input_raster2;
      TeRasterParams new_input_raster2_params = input_raster2->params();
      new_input_raster2_params.nBands( channels2.size() );
      PDIAGN_TRUE_OR_RETURN( TePDIUtils::TeAllocRAMRaster( new_input_raster2,
        input_raster2->params(), false ), 
        "Unable to allocate the new equalized raster2" );
        
      std::vector< int > new_channels2;
        
      for( unsigned int channels2_index = 0 ; 
        channels2_index < channels2.size() ; ++channels2_index ) {
        
        TePDIParameters contra_params;
        contra_params.SetParameter( "contrast_type",
          TePDIContrast::TePDIContrastStat );
        contra_params.SetParameter( "input_image", input_raster2 );
        contra_params.SetParameter( "input_band", 
          channels2[ channels2_index] );
        contra_params.SetParameter( "target_mean", 
          stat_instance.getMean( channels2_index ) );
        contra_params.SetParameter( "target_variance", 
          stat_instance.getVariance( channels2_index ) );
        contra_params.SetParameter( "output_image", new_input_raster2 );
        contra_params.SetParameter( "output_band", (int)channels2_index );
        contra_params.SetParameter( "restric_out_reset", (int)1 );
        
        TePDIContrast contra_instance;
        PDIAGN_TRUE_OR_RETURN( contra_instance.Reset( contra_params ),
          "TePDIContrast reset error" );
        PDIAGN_TRUE_OR_RETURN( contra_instance.Apply(),
          "TePDIContrast apply error" );            
        
        new_channels2.push_back( channels2_index );
      }        
      
      input_raster2 = new_input_raster2;
      channels2 = new_channels2;
    }
  }

  /* Calling the blending strategy */  
  
    
  for( unsigned int channels_index = 0 ; 
    channels_index < channels1.size() ; ++channels_index ) {
    
    if( ( pols_relation == TeDISJOINT ) || ( pols_relation == TeTOUCHES ) ) {
      /* Forcing no_blend strategy if there is no intersection between 
         polygons */
         
      TePDIParameters strat_params = params_;
      strat_params.SetParameter( "blending_type", "no_blending" );
      strat_params.SetParameter( "input_raster2", input_raster2 );
      strat_params.SetParameter( "band1", channels1[ channels_index ] );
      strat_params.SetParameter( "band2", channels2[ channels_index ] );
      strat_params.SetParameter( "out_band", (int)channels_index );
      
      TePDIBlendingStrategy::pointer blending_strategy( 
        TePDIBlendStratFactory::make( blending_type, strat_params ) );
        
      PDIAGN_TRUE_OR_RETURN( blending_strategy.isActive(), 
        "Unable to buil blend strategy" );
      
      PDIAGN_TRUE_OR_RETURN( blending_strategy->Apply( strat_params ),
        "Error applying blending strategy" );     
    } else {
      TePDIParameters strat_params = params_;
      strat_params.SetParameter( "input_raster2", input_raster2 );
      strat_params.SetParameter( "band1", channels1[ channels_index ] );
      strat_params.SetParameter( "band2", channels2[ channels_index ] );
      strat_params.SetParameter( "out_band", (int)channels_index );
      
      TePDIBlendingStrategy::pointer blending_strategy( 
        TePDIBlendStratFactory::make( blending_type, strat_params ) );
        
      PDIAGN_TRUE_OR_RETURN( blending_strategy.isActive(), 
        "Unable to buil blend strategy" );
      
      PDIAGN_TRUE_OR_RETURN( blending_strategy->Apply( strat_params ),
        "Error applying blending strategy" );    
    }
  }
  
  return true;
}

