/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIBLENDING_HPP
  #define TEPDIBLENDING_HPP

  #include "TePDIAlgorithm.hpp"
  #include "TePDIParameters.hpp"
  #include <TeSharedPtr.h>

  /**
   * @brief This is the class for digital image blending.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIAlgorithms
   *
   * @note The general required parameters:
   * @param blending_type ( std::string ) - Blending type
   * ( see TePDIBlendFactory for reference ).
   * @param input_raster1 ( TePDITypes::TePDIRasterPtrType ) - 
   *  Input raster 1.
   * @param input_raster2 ( TePDITypes::TePDIRasterPtrType ) - 
   *  Input raster 2.
   * @param output_raster ( TePDITypes::TePDIRasterPtrType ) - Output raster.
   * @param channels1 ( std::vector< int > ) - The channels to process from 
   * input_raster1.
   * @param channels2 ( std::vector< int > ) - The channels to process from 
   * input_raster2.
   * @param raster1_polygon ( TePolygon ) - Raster 1 polygon ( related to
   * input_raster1 world reference - Just 1 linear ring allowed ).
   * @param raster2_polygon ( TePolygon ) - Raster 2 polygon ( related to 
   * input_raster2 world reference - Just 1 linear ring allowed ).
   * @param raster2_pol_offset_x ( double ) - Raster 2 polygon box X offset 
   * related to raster1 polygon box in input_raster1 matricial reference
   * ( offset_x = pol2_x - pol1_x ).
   * @param raster2_pol_offset_y ( double ) - Raster 2 polygon box Y offset 
   * related to raster1 polygon box in input_raster1 matricial reference
   * ( offset_y = pol2_y - pol1_y ). 
   *
   * @note The optional parameters are:
   *
   * @param auto_equalize ( int ) - If present ( any value ), input_raster2
   * auto-equalizing will be made ( using overlap area reference ).
   * @param dummy_value (double) - A dummy pixel value for use 
   * in pixels where no data is available ( input pixels with
   * dummy values will be ignored ).    
   *
   * @note The specific parameters: See each blending strategy for reference.
   *
   * @note The output_raster parameters will follow input_raster1 parameters
   * ( box, resolution, projection ).
   */
  class TePDIBlending : public TePDIAlgorithm {
    public :
      typedef TeSharedPtr< TePDIBlending > pointer;
      typedef const TeSharedPtr< TePDIBlending > const_pointer;
      
      /**
       * Default Constructor.
       *
       */
      TePDIBlending();

      /**
       * Default Destructor
       */
      ~TePDIBlending();
      
      /**
       * Checks if the supplied parameters fits the requirements of each
       * PDI algorithm implementation.
       *
       * Error log messages must be generated. No exceptions generated.
       *
       * @param parameters The parameters to be checked.
       * @return true if the parameters are OK. false if not.
       */
      bool CheckParameters( const TePDIParameters& parameters ) const;      

    protected :
    
      /**
       * Reset the internal state to the initial state.
       *
       * @param params The new parameters referente at initial state.
       */
      void ResetState( const TePDIParameters& params );    
     
      /**
       * Runs the current algorithm implementation.
       *
       * @return true if OK. false on error.
       */
      bool RunImplementation();      
  };
  
/** @example TePDIBlending_test.cpp
 *    Shows how to use this class.
 */    

#endif
