/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDICONTRAST_HPP
  #define TEPDICONTRAST_HPP

  #include "TePDIAlgorithm.hpp"
  #include "TePDIRgbPalette.hpp"
  #include "TePDITypes.hpp"
  #include <TeSharedPtr.h>
  #include "TePDIHistogram.hpp"

  /**
   * @brief  This is the class for contrast image processing algorithms.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIAlgorithms
   *
   * @note The general input parameters are:
   *
   * @param contrast_type ( TePDIContrast::TePDIContrastType ) - The contrast 
   * to be generated.
   * @param input_image (TePDITypes::TePDIRasterPtrType) - The input
   * image.
   * @param input_band ( int ) - Band to process from input_image.
   *
   * @note The specific input parameters ( by parameter type ):
   *
   * @param min_level (double) - Minimal level for Linear Luts.
   * Required by: TePDIContrastLinear, TePDIContrastSquareRoot,
   * TePDIContrastSquare, TePDIContrastLog, TePDIContrastNegative,
   * TePDIContrastSimpleSlicer.
   *
   * @param max_level (double) - Minimal level for Linear Luts.
   * Required by: TePDIContrastLinear, TePDIContrastSquareRoot,
   * TePDIContrastSquare, TePDIContrastLog, TePDIContrastNegative,
   * TePDIContrastSimpleSlicer.
   *
   * @param rgb_palette (TePDIRgbPalette::pointer) -
   * Reference RGB palette.
   * Required by: TePDIContrastSimpleSlicer.
   *
   * @param target_mean ( double ) - The target mean value.
   * Required by: TePDIContrastStat.
   *
   * @param target_variance ( double ) - The target variance value.
   * Required by: TePDIContrastStat.
   *
   * @note The following parameters are optional and will be used if present.
   *
   * @param output_image ( TePDITypes::TePDIRasterPtrType ) -
   * The output image ( if not present, only the output LUT will be
   * generated ).   
   *
   * @param output_band ( int ) - Band to save the result on output_image
   * (required if output_image is present).
   *
   * @param restrict_out_reset (int) - If present (any value),
   * a restrictive output_image reset will be done
   * ( output_image reset will not be done if all of the following 
   * conditions are satisfied - Exact number of lines, exact number
   * of columns, sufficient number of bands, same projection, same box,
   * dummy value ), 
   * this allows output_image data keeping since
   * unused bands are not changed.
   *
   * @param outlut ( TePDITypes::TePDILutPtrType ) - The generated look-up
   * table. NOTE: If the current input_image has floating point pixel values,
   * the generated look-up table will have interpolated values.
   *
   * @param input_histogram ( TePDIHistogram::pointer ) - A pointer to an
   * active histogram instance that will be used when generating the
   * internal lut - The internal histogram generation will be disabled.
   *
   * @param histo_levels (int) - Number of histogram levels when internal
   * histogram generation is needed ( if zero, the histogram levels
   * will be automatically found ).
   *
   * @param dummy_value (double) - A dummy pixel value for use 
   * in pixels where no data is available.    
   */
  class TePDIContrast : public TePDIAlgorithm {
    public :
      typedef TeSharedPtr< TePDIContrast > pointer;
      typedef const TeSharedPtr< TePDIContrast > const_pointer;

      /**
       * Allowed contrast types.
       */
      enum TePDIContrastType{
        /** The histogram range will be changed to the supplied 
            min/max range ( linear function ). */
        TePDIContrastMinMax = 1,
        /** The histogram will be remapped using a linear function. */
        TePDIContrastLinear = 2,
        /** The histogram will be remapped using a square root function. */
        TePDIContrastSquareRoot = 3,
        /** The histogram will be remapped using a square function. */
        TePDIContrastSquare = 4,
        /** The histogram will be remapped using a logarithmic function. */
        TePDIContrastLog = 5,
        /** The histogram will be remapped using a linear inverse function. */
        TePDIContrastNegative = 6,
        /** The uniform histogram aproximation. */
        TePDIContrastHistEqualizer = 7,
        /** The histogram values will be splitted into slices and linked with
            a RGB color table. */
        TePDIContrastSimpleSlicer = 8,
        /** Image mean and variance uniformization. */
        TePDIContrastStat = 9
      };

      /**
       * Default Constructor.
       *
       */
      TePDIContrast();

      /**
       * Default Destructor
       */
      ~TePDIContrast();

      /**
       * Checks if the supplied parameters fits the requirements of each
       * PDI algorithm implementation.
       *
       * Error log messages must be generated. No exceptions generated.
       *
       * @param parameters The parameters to be checked.
       * @return true if the parameters are OK. false if not.
       */
      bool CheckParameters( const TePDIParameters& parameters ) const;

    protected :

      /**
       * Runs the current algorithm implementation.
       *
       * @return true if OK. false on error.
       */
      bool RunImplementation();

      /**
       * Reset the internal state to the initial state.
       *
       * @param params The new parameters referente at initial state.
       */
      void ResetState( const TePDIParameters& params );

      /**
       * Build a min-max look-up table.
       *
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.
       * @param hist Base histogram.
       * @return  The look-up table.
       */
      TePDITypes::TePDILutType GetMinMaxLut(
        double channel_min_level, double channel_max_level,
        TePDIHistogram::pointer hist );

      /**
       * Build a Linear look-up table.
       *
       * @param min Minimal level value.
       * @param max Maximum level value.
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.
       * @param hist Base histogram.
       * @return The look-up table.
       */
      TePDITypes::TePDILutType GetLinearLut(
        TePDIHistogram::pointer hist,
        double min, double max,
        double channel_min_level, double channel_max_level );

      /**
       * Build a Square Root look-up table.
       *
       * @param min Minimal level value.
       * @param max Maximum level value.
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.
       * @param hist Base histogram.
       * @return The look-up table.
       */
      TePDITypes::TePDILutType GetSquareRootLut(
        TePDIHistogram::pointer hist,
        double min, double max,
        double channel_min_level, double channel_max_level );

      /**
       * Build a Square look-up table.
       *
       * @param min Minimal level value.
       * @param max Maximum level value.
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.
       * @param hist Base histogram.
       * @return The look-up table.
       */
      TePDITypes::TePDILutType GetSquareLut(
        TePDIHistogram::pointer hist,
        double min, double max,
        double channel_min_level, double channel_max_level );

      /**
       * Build a Log look-up table.
       *
       * @param min Minimal level value.
       * @param max Maximum level value.
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.
       * @param hist Base histogram.
       * @return The look-up table.
       */
      TePDITypes::TePDILutType GetLogLut(
        TePDIHistogram::pointer hist,
        double min, double max,
        double channel_min_level, double channel_max_level );

      /**
       * Build a Negative look-up table.
       *
       * @param min Minimal level value.
       * @param max Maximum level value.
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.
       * @param hist Base histogram.
       * @return The look-up table.
       */
      TePDITypes::TePDILutType GetNegativeLut(
        TePDIHistogram::pointer hist,
        double min, double max,
        double channel_min_level, double channel_max_level );

      /**
       * Build a Histogram equalizer look-up table.
       *
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.       
       * @param hist A LEVEL-ORDERED Base histogram.
       * @return The look-up table.
       */
      TePDITypes::TePDILutType GetHistEqualizerLut(
        TePDIHistogram::pointer hist,
        double channel_min_level, double channel_max_level );

      /**
       * Builds a Simple slicer look-up table.
       *
       * @param hist A LEVEL-ORDERED Base histogram.
       * @param in_palette The reference RGB palette.
       * @param min Minimal level value.
       * @param max Maximum level value.
       * @param out_lut The look-up table.
       */
      void GetSimpleSlicerLut(
        TePDIHistogram::pointer hist,
        TePDIRgbPalette::pointer in_palette,
        double min,
        double max,
        TePDITypes::TePDILutType& out_lut );
        
      /**
       * Builds a mean and variance normalization lut.
       *
       * @param hist A LEVEL-ORDERED Base histogram.
       * @param target_mean The target mean value.
       * @param target_variance The target variance value.
       * @param channel_min_level The minimum channel possible level value.
       * @param channel_max_level The maximum channel possible level value.       
       * @param out_lut The look-up table.
       */
      void GetStatLut( TePDIHistogram::pointer hist, double target_mean,
        double target_variance, double channel_min_level, 
        double channel_max_level, TePDITypes::TePDILutType& out_lut );        
        
      /**
       * Do level remapping from one raster to another.
       *
       * @param inRaster The input Raster.
       * @param outRaster The Output Raster.
       * @param lut The increasing level-ordered look-up table.
       * @param in_channel in_raster channel to remap.
       * @param out_channel out_raster channel to write on.
       * @param use_dummy Force a new dummy value use ( input pixels with
       * dummy values will be ignored ).
       * @param dummy_value A dummy pixel value for use in pixels where no data 
       * is available.
       * @param fixed_step_lut A flag indication of a fixed step lut ( the
       * internal optimized fixed step algorithm will be used )
       */
      void RemapLevels(
        TePDITypes::TePDIRasterPtrType& inRaster,
        TePDITypes::TePDILutType& lut,
        int in_channel,
        int out_channel,
        TePDITypes::TePDIRasterPtrType& outRaster,
        bool use_dummy, double dummy_value,
        bool fixed_step_lut );        
  };
  
/** @example TePDIContrast_test.cpp
 *    Shows how to use this class.
 */    

#endif //TEPDICONTRAST_HPP
