 
#include "TePDIEuclidianBlendStrategy.hpp"
#include "TePDIUtils.hpp"
#include "TePDIPIManager.hpp"
#include "TePDIAgnostic.hpp"

#include <TeOverlay.h>
#include <TeUtils.h>
#include <TeGeometryAlgorithms.h>
#include <TeDefines.h>

#include <float.h>


TePDIEuclidianBlendStrategy::TePDIEuclidianBlendStrategy()
{
};      

TePDIEuclidianBlendStrategy::~TePDIEuclidianBlendStrategy()
{
};


bool TePDIEuclidianBlendStrategy::CheckParameters( 
  const TePDIParameters& parameters ) const
{
  /* Checking input_raster1 */
  
  TePDITypes::TePDIRasterPtrType input_raster1;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "input_raster1", 
    input_raster1 ),
    "Missing parameter: input_raster1" );
  PDIAGN_TRUE_OR_RETURN( input_raster1.isActive(),
    "Invalid parameter: input_raster1 inactive" );
  PDIAGN_TRUE_OR_RETURN( input_raster1->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: input_raster1 not ready" );    
    
  /* Checking input_raster2 */
  
  TePDITypes::TePDIRasterPtrType input_raster2;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "input_raster2", 
    input_raster2 ),
    "Missing parameter: input_raster2" );
  PDIAGN_TRUE_OR_RETURN( input_raster2.isActive(),
    "Invalid parameter: input_raster2 inactive" );
  PDIAGN_TRUE_OR_RETURN( input_raster2->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: input_raster2 not ready" );    
    
  /* Checking output_raster */
  
  TePDITypes::TePDIRasterPtrType output_raster;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "output_raster", 
    output_raster ),
    "Missing parameter: output_raster" );
  PDIAGN_TRUE_OR_RETURN( output_raster.isActive(),
    "Invalid parameter: output_raster inactive" );
  PDIAGN_TRUE_OR_RETURN( output_raster->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: output_raster not ready" );    
    
  /* Checking bands */    
    
  int band1 = 0;    
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "band1", band1 ),
    "Missing parameter : band1" );
  PDIAGN_TRUE_OR_RETURN( ( input_raster1->nBands() > band1 ),
    "Invalid parameter : band1" );
    
  int band2 = 0;    
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "band2", band2 ),
    "Missing parameter : band2" );
  PDIAGN_TRUE_OR_RETURN( ( input_raster2->nBands() > band2 ),
    "Invalid parameter : band2" ); 

  int out_band = 0;    
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "out_band", out_band ),
    "Missing parameter : out_band" );
  PDIAGN_TRUE_OR_RETURN( ( output_raster->nBands() > out_band ),
    "Invalid parameter : out_band" );       
    
  /* Checking raster polygons */
  
  TePolygon raster1_polygon;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster1_polygon", 
    raster1_polygon ), "Missing parameter : raster1_polygon" );
  PDIAGN_TRUE_OR_RETURN(
    ( ! input_raster1->begin( raster1_polygon, TeBoxPixelIn, 0 ).end() ),
    "Invalid raster1_polygon" )
    
  TePolygon raster2_polygon;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_polygon", 
    raster2_polygon ), "Missing parameter : raster2_polygon" );
  PDIAGN_TRUE_OR_RETURN(
    ( ! input_raster2->begin( raster2_polygon, TeBoxPixelIn, 0 ).end() ),
    "Invalid raster2_polygon" )
    
  /* Checking raster2 polygon offsets */
  
  double raster2_pol_offset_x = 0;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_pol_offset_x", 
    raster2_pol_offset_x ), "Missing parameter : raster2_pol_offset_x" );  

  double raster2_pol_offset_y = 0;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_pol_offset_y", 
    raster2_pol_offset_y ), "Missing parameter : raster2_pol_offset_y" );     
    
  return true;
}


bool TePDIEuclidianBlendStrategy::Implementation( 
  const TePDIParameters& params )
{
  /* Extracting parameters */

  TePDITypes::TePDIRasterPtrType input_raster1;
  params.GetParameter( "input_raster1", input_raster1 );
  TeRaster& input_raster1_ref = *( input_raster1.nakedPointer() );

  TePDITypes::TePDIRasterPtrType input_raster2;
  params.GetParameter( "input_raster2", input_raster2 );
  TeRaster& input_raster2_ref = *( input_raster2.nakedPointer() );

  TePDITypes::TePDIRasterPtrType output_raster;
  params.GetParameter( "output_raster", output_raster );
  TeRaster& output_raster_ref = *( output_raster.nakedPointer() );
  
  int band1 = 0;
  params.GetParameter( "band1", band1 );

  int band2 = 0;
  params.GetParameter( "band2", band2 );
  
  int out_band = 0;
  params.GetParameter( "out_band", out_band );  
  
  TePolygon raster1_polygon;
  params.GetParameter( "raster1_polygon", raster1_polygon );

  TePolygon raster2_polygon;
  params.GetParameter( "raster2_polygon", raster2_polygon );
  
  double raster2_pol_offset_x = 0;
  params.GetParameter( "raster2_pol_offset_x", raster2_pol_offset_x );

  double raster2_pol_offset_y = 0;
  params.GetParameter( "raster2_pol_offset_y", raster2_pol_offset_y );
  
  /* Dumyy value definition */
  
  const bool raster1_uses_dummy = input_raster1->params().useDummy_;
  const bool raster2_uses_dummy = input_raster2->params().useDummy_;  
 
  bool output_raster_uses_dummy = output_raster->params().useDummy_;
  double output_raster_dummy_value = 0;
  
  if( output_raster_uses_dummy ) {
    output_raster_dummy_value = output_raster->params().dummy_[ 0 ];
  }
  
  if( params.CheckParameter( "dummy_value", 
    TePDIParameters::TePDIDoubleParam ) ) {
    
    params.GetParameter( "dummy_value", output_raster_dummy_value );
    
    output_raster_uses_dummy = true;
  }  
  
  /* Raster1 positioning */
  
  double raster1_abs_index_offset_x = 0;// output raster */
  double raster1_abs_index_offset_y = 0;
  double raster1_rel_index_offset_x = 0;// input x output
  double raster1_rel_index_offset_y = 0;
  {
    TeBox pol1_box_indexed;
    TePDIUtils::MapCoords2RasterIndexes( raster1_polygon.box(),
      input_raster1, pol1_box_indexed );
          
    double r1_min_x = MIN( pol1_box_indexed.x1(),
      pol1_box_indexed.x2() );
    double r1_min_y = MIN( pol1_box_indexed.y1(),
      pol1_box_indexed.y2() );      

    TeCoord2D r1_UL_ref1_indexed( r1_min_x, r1_min_y );  
    TeCoord2D r1_UL_ref1( input_raster1->index2Coord( r1_UL_ref1_indexed ) );  
    TeCoord2D r1_UL_refout_indexed( output_raster->coord2Index( r1_UL_ref1 ) );
        
    raster1_abs_index_offset_x = r1_UL_refout_indexed.x();  
    raster1_abs_index_offset_y = r1_UL_refout_indexed.y();
      
    raster1_rel_index_offset_x = r1_UL_ref1_indexed.x() -
      raster1_abs_index_offset_x;
    raster1_rel_index_offset_y = r1_UL_ref1_indexed.y() -
      raster1_abs_index_offset_y;
  }  
  
  /* Raster2 positioning */
  
  TePolygon r2_pol_refout;
  double raster2_rel_index_offset_x = 0;
  double raster2_rel_index_offset_y = 0;  
  {
    TePolygon r2_pol_ref2_indexed;
    TePDIUtils::MapCoords2RasterIndexes( raster2_polygon,
      input_raster2, r2_pol_ref2_indexed );
      
    double r2_min_x = MIN( r2_pol_ref2_indexed.box().x1(),
      r2_pol_ref2_indexed.box().x2() );
    double r2_min_y = MIN( r2_pol_ref2_indexed.box().y1(),
      r2_pol_ref2_indexed.box().y2() );
  
    raster2_rel_index_offset_x = r2_min_x - raster1_abs_index_offset_x -
      raster2_pol_offset_x;
    raster2_rel_index_offset_y = r2_min_y - raster1_abs_index_offset_y -
      raster2_pol_offset_y;
      
    TePolygon r2_pol_refout_indexed = r2_pol_ref2_indexed;
        
    for( unsigned int lr_index = 0 ; 
      lr_index < r2_pol_refout_indexed.size() ; ++lr_index ) {
      for( unsigned int c2d_index = 0 ; 
        c2d_index < r2_pol_refout_indexed[ lr_index ].size() ; 
        ++c2d_index ) {
            
        r2_pol_refout_indexed[ lr_index ][ c2d_index ].x( 
          r2_pol_refout_indexed[ lr_index ][ c2d_index ].x() -
          r2_min_x + raster1_abs_index_offset_x + raster2_pol_offset_x );
        r2_pol_refout_indexed[ lr_index ][ c2d_index ].y( 
          r2_pol_refout_indexed[ lr_index ][ c2d_index ].y() -
          r2_min_y + raster1_abs_index_offset_y + raster2_pol_offset_y );          
      }
    }
    
    TePDIUtils::MapRasterIndexes2Coords( r2_pol_refout_indexed,
      output_raster, r2_pol_refout );
  }
  
  /* Intersection polygon - output raster reference */
  
  TePolygon int_pol_refout;
  {    
    TePolygonSet ps1;
    ps1.add( raster1_polygon );
  
    TePolygonSet ps2;
    ps2.add( r2_pol_refout );
   
    TePolygonSet ps_intersection;
    PDIAGN_TRUE_OR_RETURN(
      TeOVERLAY::TeIntersection( ps1, ps2, ps_intersection ),
      "Unable to find rasters intersection" );
      
    PDIAGN_TRUE_OR_RETURN( ( ps_intersection.size() == 1 ),
      "Invalid generated rasters intersection" );
        
    int_pol_refout = ps_intersection[ 0 ];
  }
  
  /* Building the new raster polygons with no intersection areas */
  
  TePolygon new_raster1_polygon;
  {
    TePolygonSet ps_intersection;
    ps_intersection.add( int_pol_refout );
  
    TePolygonSet ps_old_pol;
    ps_old_pol.add( raster1_polygon );
    
    TePolygonSet ps_new;
    
    if( ! TeOVERLAY::TeDifference( ps_old_pol, ps_intersection, ps_new ) ) {
      PDIAGN_TRUE_OR_RETURN( TeOVERLAY::TeDifference( ps_intersection, 
        ps_old_pol, ps_new ), 
        "Unable to find the new input_raster1 polygon" );
    }
    
    PDIAGN_TRUE_OR_RETURN( ( ps_new.size() == 1 ),
      "Invalid polygon set size" );
      
    new_raster1_polygon = ps_new[ 0 ];
  }  
  
  TePolygon new_raster2_polygon;
  {
    TePolygonSet ps_intersection;
    ps_intersection.add( int_pol_refout );
  
    TePolygonSet ps_old_pol;
    ps_old_pol.add( r2_pol_refout );
    
    TePolygonSet ps_new;
    
    if( ! TeOVERLAY::TeDifference( ps_old_pol, ps_intersection, ps_new ) ) {
      PDIAGN_TRUE_OR_RETURN( TeOVERLAY::TeDifference( ps_intersection, 
        ps_old_pol, ps_new ), 
        "Unable to find the new input_raster1 polygon" );
    }      
      
    PDIAGN_TRUE_OR_RETURN( ( ps_new.size() == 1 ),
      "Invalid polygon set size" );
      
    new_raster2_polygon = ps_new[ 0 ];
  }
  
  /* Global vars */
  
  TeRaster::iteratorPoly input_raster1_it =
    output_raster->begin( new_raster1_polygon, TeBoxPixelIn, 0 );
  PDIAGN_TRUE_OR_RETURN( ( ! input_raster1_it.end() ), 
    "Unable to create an iterator over raster1 area" ); 
    
  TeRaster::iteratorPoly input_raster2_it =
    output_raster->begin( new_raster2_polygon, TeBoxPixelIn, 0 );
  PDIAGN_TRUE_OR_RETURN( ( ! input_raster2_it.end() ), 
    "Unable to create an iterator over raster2 area" );       
    
  TeRaster::iteratorPoly intersection_refout_it =
    output_raster->begin( int_pol_refout, TeBoxPixelIn, 0 );
  PDIAGN_TRUE_OR_RETURN( ( ! intersection_refout_it.end() ), 
    "Unable to create an iterator over raster2 area" );     
  
  TePDIPIManager progress( "Rendering raster 1...", 
    input_raster1_it.nLinesInPoly(), progress_interface_enabled_ );   
    
  /* Transfering raster 1 - non-intersected area */
  {
    unsigned int curr_line = 0;
    unsigned int curr_col = 0;
    double curr_value = 0;  
    int offset_x = TeRound( raster1_rel_index_offset_x );
    int offset_y = TeRound( raster1_rel_index_offset_y );
    
    while( ! input_raster1_it.end() ) {
      curr_line = input_raster1_it.currentLine();
      curr_col = input_raster1_it.currentColumn();
        
      if( input_raster1_ref.getElement( curr_col + offset_x, 
          curr_line + offset_y, curr_value, band1 ) ) {
          
        if( ! ( output_raster_uses_dummy && 
          ( curr_value == output_raster_dummy_value ) ) ) {
          
          PDIAGN_TRUE_OR_RETURN( 
            output_raster_ref.setElement( curr_col, curr_line, curr_value, 
            out_band ),
            "Unable to write to output raster at line=" + 
            Te2String( curr_line ) + 
            " col=" + Te2String( curr_col ) );                    
        }
      } else {
        PDIAGN_TRUE_OR_RETURN( raster1_uses_dummy, 
          "Unable to read from input_raster1 at line=" + 
          Te2String( curr_line + offset_y ) + " col=" + 
          Te2String( curr_col + offset_x ) );
      }
  
      ++input_raster1_it;
       
      progress.Increment();
    } 
  }
  
  /* Transfering raster 2 - non-intersected area */
  {
    unsigned int curr_line = 0;
    unsigned int curr_col = 0;
    double curr_value = 0;  
    int offset_x = TeRound( raster2_rel_index_offset_x );
    int offset_y = TeRound( raster2_rel_index_offset_y );
    
    progress.Reset( "Rendering raster 2...", 
      input_raster2_it.nLinesInPoly() );    
      
    while( ! input_raster2_it.end() ) {
      curr_line = input_raster2_it.currentLine();
      curr_col = input_raster2_it.currentColumn();
        
      if( input_raster2_ref.getElement( curr_col + offset_x, 
          curr_line + offset_y , curr_value, band2 ) ) {
          
        if( ! ( output_raster_uses_dummy && 
          ( curr_value == output_raster_dummy_value ) ) ) {
          
          PDIAGN_TRUE_OR_RETURN( 
            output_raster_ref.setElement( curr_col, curr_line, curr_value, 
            out_band ), "Unable to write to output raster at line=" + 
            Te2String( curr_line ) + " col=" + Te2String( curr_col ) );
        }
      } else {
        PDIAGN_TRUE_OR_RETURN( raster2_uses_dummy, 
          "Unable to read from input_raster2 at line=" + 
          Te2String( curr_line + offset_y ) + " col=" + 
          Te2String( curr_col + offset_x ) );
      }
  
      ++input_raster2_it;
       
      progress.Increment();
    }
  }

  /* Blending intersection */
  {
    const short pols_relation = 
      TeRelation( raster1_polygon, r2_pol_refout );
      
    unsigned int current_line = 0;
    unsigned int current_col = 0;
    
    double value1 = 0;      
    double value2 = 0;      
    double value3 = 0;
    
    bool got_raster1_element = false;
    bool got_raster2_element = false;     
    
    int raster1_offset_x = TeRound( raster1_rel_index_offset_x );
    int raster1_offset_y = TeRound( raster1_rel_index_offset_y );    
    int raster2_offset_x = TeRound( raster2_rel_index_offset_x );
    int raster2_offset_y = TeRound( raster2_rel_index_offset_y );    
    
    if( ( pols_relation == TeCONTAINS ) || ( pols_relation == TeCOVERS ) ||
        ( pols_relation == TeWITHIN ) || ( pols_relation == TeCOVEREDBY ) ) {
        
      /* This is the case where one raster is inside onother */
      
      progress.Reset( "Blending intersection...", 
        intersection_refout_it.nLinesInPoly() );       
        
      while( ! intersection_refout_it.end() ) {
        current_line = intersection_refout_it.currentLine();
        current_col = intersection_refout_it.currentColumn();    
        
        if( input_raster1_ref.getElement( current_col + raster1_offset_x, 
            current_line + raster1_offset_y, value1, band1 ) ) {
                
          if( output_raster_uses_dummy &&
            ( value1 == output_raster_dummy_value ) ) {
          
            got_raster1_element = false;
          } else {
            got_raster1_element = true;
          }
        } else {
          got_raster1_element = false;
        }
              
        if( input_raster2_ref.getElement( current_col + raster2_offset_x, 
            current_line + raster2_offset_y, value2, band2 ) ) {
                
          if( output_raster_uses_dummy &&
            ( value2 == output_raster_dummy_value ) ) {
              
            got_raster2_element = false;
          } else {
            got_raster2_element = true;
          }
        } else {
          got_raster2_element = false;
        }          
      
        if( got_raster1_element && got_raster2_element ) {
          value3 = ( value1 + value2 ) / 2;
          
          if( ( ! output_raster_uses_dummy ) || 
            ( value3 != output_raster_dummy_value ) ) {
                
            PDIAGN_TRUE_OR_RETURN( 
              output_raster_ref.setElement( current_col, current_line, value3, 
              out_band ), "Unable to write to output raster" );
          }            
        } else if( got_raster1_element ) {
          if( ( ! output_raster_uses_dummy ) || 
            ( value1 != output_raster_dummy_value ) ) {
                
            PDIAGN_TRUE_OR_RETURN( 
              output_raster_ref.setElement( current_col, current_line, value1, 
              out_band ), "Unable to write to output raster" );
          }          
        } else {
          if( ( ! output_raster_uses_dummy ) || 
            ( value2 != output_raster_dummy_value ) ) {
              
            PDIAGN_TRUE_OR_RETURN( 
              output_raster_ref.setElement( current_col, current_line, value2, 
              out_band ), "Unable to write to output raster" );
          }          
        }
        
        progress.Increment();
        ++intersection_refout_it;
      }
    } else {
      /* Calculating the two blending lines - output raster indexed reference*/

      bool x_axis_parallels = false;
      bool y_axis_parallels = false;
      double line1_a = 0;
      double line1_b = 0;
      double line1_c = 0;
      double line1_x = 0;
      double line1_y = 0;
      double line2_a = 0;
      double line2_b = 0;
      double line2_c = 0;
      double line2_x = 0;
      double line2_y = 0;
      
      TePolygon raster1_polygon_indexed;
      TePDIUtils::MapCoords2RasterIndexes( raster1_polygon,
        output_raster, raster1_polygon_indexed );      
   
      TePolygon r2_pol_refout_indexed;
      TePDIUtils::MapCoords2RasterIndexes( r2_pol_refout,
        output_raster, r2_pol_refout_indexed );      
          
      TePolygon int_pol_refout_indexed;
      TePDIUtils::MapCoords2RasterIndexes( int_pol_refout,
        output_raster, int_pol_refout_indexed );        
          
      generateBlendingLines( raster1_polygon_indexed, r2_pol_refout_indexed, 
        int_pol_refout_indexed, 
        line1_a, line1_b, line1_c,
        line2_a, line2_b, line2_c );
        
      if( line1_a == 0.0 ) {
        x_axis_parallels = true;
        line1_y = ( -1 ) * ( line1_c / line1_b );
        line2_y = ( -1 ) * ( line2_c / line2_b );
      } else if( line1_b == 0.0 ) {
        y_axis_parallels = true;
        line1_x = ( -1 ) * ( line1_c / line1_a );
        line2_x = ( -1 ) * ( line2_c / line2_a );
      }
      
      /* Blending intersection */  
    
      double dist1 = 0;   
      double dist2 = 0;
      const double const_hip1 = sqrt( ( line1_a * line1_a ) + 
        ( line1_b * line1_b ) );
      const double const_hip2 = sqrt( ( line2_a * line2_a ) + 
        ( line2_b * line2_b ) );
        
      progress.Reset( "Blending intersection...", 
        intersection_refout_it.nLinesInPoly() );         
    
      while( ! intersection_refout_it.end() ) {
        current_line = intersection_refout_it.currentLine();
        current_col = intersection_refout_it.currentColumn();
        
        if( input_raster1_ref.getElement( current_col + raster1_offset_x, 
            current_line + raster1_offset_y, value1, band1 ) ) {
                
          if( output_raster_uses_dummy &&
            ( value1 == output_raster_dummy_value ) ) {
          
            got_raster1_element = false;
          } else {
            got_raster1_element = true;
          }
        } else {
          got_raster1_element = false;
        }
              
        if( input_raster2_ref.getElement( current_col + raster2_offset_x, 
            current_line + raster2_offset_y, value2, band2 ) ) {
                
          if( output_raster_uses_dummy &&
            ( value2 == output_raster_dummy_value ) ) {
              
            got_raster2_element = false;
          } else {
            got_raster2_element = true;
          }
        } else {
          got_raster2_element = false;
        }        
          
        if( got_raster1_element && got_raster2_element ) {
          if( x_axis_parallels ) {
            dist1 = ABS( line1_y - ( (double)current_line ) );
            dist2 = ABS( line2_y - ( (double)current_line ) );
          } else if( y_axis_parallels ) {
            dist1 = ABS( line1_x - ( (double)current_col ) );
            dist2 = ABS( line2_x - ( (double)current_col ) );
          } else {
            dist1 = ABS( ( ( line1_a * ( (double)current_line ) ) +
              ( line1_b * ( (double)current_col) ) + line1_c ) / 
              const_hip1 );
            dist2 = ABS( ( ( line2_a * ( (double)current_line ) ) +
              ( line2_b * ( (double)current_col) ) + line2_c ) / 
              const_hip2 );
          }
          
          if( dist1 == 0.0 ) {
            value3 = value1;
          } else if( dist2 == 0.0 ) {
            value3 = value2;
          } else {
            value3 = ( ( dist1 * value2 ) + ( dist2 * value1 ) ) /
              ( dist1 + dist2 );
          }
              
          PDIAGN_TRUE_OR_RETURN( 
            output_raster_ref.setElement( current_col, 
            current_line,
            value3, out_band ), "Unable to write to output raster" );                 
        } else if( got_raster1_element ) {
          PDIAGN_TRUE_OR_RETURN( 
            output_raster_ref.setElement( current_col, 
            current_line,
            value1, out_band ), "Unable to write to output raster" );
        } else if( got_raster2_element ) {
          PDIAGN_TRUE_OR_RETURN( 
            output_raster_ref.setElement( current_col, 
            current_line,
            value2, out_band ), "Unable to write to output raster" );
        }        
        
        ++intersection_refout_it;
        
        progress.Increment();      
      }       
    } 
  }
  
  return true;
}


void TePDIEuclidianBlendStrategy::generateBlendingLines( 
  const TePolygon& polygon1,  
  const TePolygon& polygon2, const TePolygon& inter_pol, 
  double& line1_a, double& line1_b, double& line1_c,
  double& line2_a, double& line2_b, double& line2_c )
{
  const TeCoord2D centroid1 = TeFindCentroid( polygon1 );
  const TeCoord2D centroid2 = TeFindCentroid( polygon2 );
  
  if( centroid1.x() == centroid2.x() ) {
    /* Centroids are lined up ( y axis ) */
    
    if( centroid1.y() > centroid2.y() ) {
      line1_a = 0.0;
      line1_b = -1.0;
      line1_c = inter_pol.box().y2() ;
      
      line2_a = 0.0;
      line2_b = -1.0;
      line2_c = inter_pol.box().y1() ;
    } else {
      line1_a = 0.0;
      line1_b = -1.0;
      line1_c = inter_pol.box().y1() ;
      
      line2_a = 0.0;
      line2_b = -1.0;
      line2_c = inter_pol.box().y2() ;
    }
  } else if( centroid1.y() == centroid2.y() ) {
    /* Centroids are lined up ( x axis ) */
      
    if( centroid1.x() > centroid2.x() ) {
      line1_a = -1.0;
      line1_b = 0.0;
      line1_c = inter_pol.box().x2() ;
      
      line2_a = -1.0;
      line2_b = 0.0;
      line2_c = inter_pol.box().x1() ;
    } else {
      line1_a = -1.0;
      line1_b = 0.0;
      line1_c = inter_pol.box().x1() ;
      
      line2_a = -1.0;
      line2_b = 0.0;
      line2_c = inter_pol.box().x2() ;
    }   
  } else {
    /* line passing throught intersection polygon centroid coeficients -
       this line is parallel with the line passing throught centroids */
    
    const TeCoord2D inter_pol_centroid = TeFindCentroid( inter_pol );
    
    const double ipc_line_m = ( centroid2.y() - centroid1.y() ) /
      ( centroid2.x() - centroid1.x() );
    const double ipc_line_q = inter_pol_centroid.y() - 
      ( ipc_line_m * inter_pol_centroid.x() );
      
    /* perpendicular line angular coeficient */
    
    const double perp_line_m = ( -1.0 ) / ipc_line_m;
    
    /* Projecting each intersection polygon point over the parallel line */
    
    std::vector< TeCoord2D > proj_points;
    
    for( unsigned int lr_index = 0 ; lr_index < inter_pol.size() ; 
      ++lr_index  ) {
      for( unsigned int coord_index = 0 ; 
        coord_index < inter_pol[ lr_index ].size() ; ++coord_index  ) {
        
        /* perpendicular line linear coef passing throught the current point */
      
        double perp_line_q = inter_pol[ lr_index ][ coord_index ].y() -
          ( perp_line_m * inter_pol[ lr_index ][ coord_index ].x() );
          
        /* Calculating the lines intersection */
        
        TeCoord2D int_point;
        int_point.x( ( perp_line_q - ipc_line_q ) / 
          ( ipc_line_m - perp_line_m  ) );
        int_point.y( ( ipc_line_m * int_point.x() ) + ipc_line_q );
        
        proj_points.push_back( int_point );
      }
    }
    
    PDIAGN_TRUE_OR_THROW( ( proj_points.size() > 1 ), 
      "Invalid projected points" );
      
    /* Finding the two pivots for the two blending lines */
    
    TeCoord2D pivot1; /* related to polygon 1 */
    TeCoord2D pivot2; /* related to polygon 2 */
    {
      /* Taking the two points  with the higher distance one to another */
      
      TeCoord2D proj_point1 = proj_points[ 0 ];
      TeCoord2D proj_point2 = proj_points[ 1 ];
      double higher_dist = TeDistance( proj_point1, proj_point2 );
      
      for( unsigned int proj_points_index = 2 ; 
        proj_points_index < proj_points.size() ; ++proj_points_index ) {
        
        double dist1 = TeDistance( proj_points[ proj_points_index ], 
          proj_point1 );
        double dist2 = TeDistance( proj_points[ proj_points_index ], 
          proj_point2 );
          
        if( dist1 > dist2 ) {
          if( dist1 > higher_dist ) {
            proj_point2 = proj_points[ proj_points_index ];
            higher_dist = dist1;
          }
        } else {
          if( dist2 > higher_dist ) {
            proj_point1 = proj_points[ proj_points_index ];
            higher_dist = TeDistance( proj_points[ proj_points_index ], 
              proj_point2 );
          }
        }
      }
      
      /* generating the equation for the perpendicular line passing 
         throught the intersection polygon centroid */
         
      double ip_cent_perp_line_a = 0;
      double ip_cent_perp_line_b = 0;
      double ip_cent_perp_line_c = 0;
      {
        double ip_centroid_perp_line_q = inter_pol_centroid.y() -
        ( perp_line_m * inter_pol_centroid.x() ); 
        
        TeCoord2D aux_point( ( inter_pol_centroid.x() + 10 ), 
          ( perp_line_m * ( inter_pol_centroid.x() + 10 ) ) + 
          ip_centroid_perp_line_q );
          
        ip_cent_perp_line_a = inter_pol_centroid.y() - aux_point.y();
        ip_cent_perp_line_b = aux_point.x() - inter_pol_centroid.x();
        ip_cent_perp_line_c = ( inter_pol_centroid.x() * aux_point.y() ) - 
        ( aux_point.x() * inter_pol_centroid.y() );          
      }
        
      /* Association of the projected points with the correct polygons */     
      
      {
        double rel_pos_proj_point1 = ( ip_cent_perp_line_a * proj_point1.x() ) 
          + ( ip_cent_perp_line_b * proj_point1.y() ) + ip_cent_perp_line_c;
        double rel_pos_centroid1 = ( ip_cent_perp_line_a * centroid1.x() ) 
          + ( ip_cent_perp_line_b * centroid1.y() ) + ip_cent_perp_line_c;
         
        if( ( rel_pos_proj_point1 > 0 ) && ( rel_pos_centroid1 > 0 ) ) {
          pivot1 = proj_point1;
          pivot2 = proj_point2;        
        } else {
          pivot2 = proj_point1;
          pivot1 = proj_point2;
        }
      }
    }
    
    /* Calculating blending line 1 equation */
    
    {
      double line1_q = pivot1.y() -
        ( perp_line_m * pivot1.x() );
        
      TeCoord2D aux_point( ( pivot1.x() + 10 ), 
        ( perp_line_m * ( pivot1.x() + 10 ) ) + line1_q );
        
      line1_a = pivot1.y() - aux_point.y();
      line1_b = aux_point.x() - pivot1.x();
      line1_c = ( pivot1.x() * aux_point.y() ) - 
        ( aux_point.x() * pivot1.y() );
    }

    /* Calculating blending line 2 equation */
    
    {
      double line2_q = pivot2.y() -
        ( perp_line_m * pivot2.x() );
        
      TeCoord2D aux_point( ( pivot2.x() + 10 ), 
        ( perp_line_m * ( pivot2.x() + 10 ) ) + line2_q );
        
      line2_a = pivot2.y() - aux_point.y();
      line2_b = aux_point.x() - pivot2.x();
      line2_c = ( pivot2.x() * aux_point.y() ) - 
        ( aux_point.x() * pivot2.y() );
    }
  }  
}


