/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIGEOMOSAICSTRATEGY_HPP
  #define TEPDIGEOMOSAICSTRATEGY_HPP

  #include "TePDIMosaicStrategy.hpp"
  
  #include <string>

  /**
   * @brief This is the class for Geographic Mosaic ( geo-referenced rasters )
   * strategy.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIStrategies
   *
   * @note The required parameters are:
   * @param mosaic_type ( std::string ) - must be "geo_mosaic".
   * @param blending_type ( std::string ) - Blending type.
   * type ( see TePDIBlendStratFactory documentation for avaliable
   * blending strategies ).   
   * @param input_raster1 ( TePDITypes::TePDIRasterPtrType ) - 
   *  Input raster 1.
   * @param input_raster2 ( TePDITypes::TePDIRasterPtrType ) - 
   *  Input raster 2.
   * @param channels1 ( std::vector< int > ) - The channels to process from 
   * input_raster1.
   * @param channels2 ( std::vector< int > ) - The channels to process from 
   * input_raster2. 
   * @param output_raster ( TePDITypes::TePDIRasterPtrType ) - Output raster.
   *
   * @note The optional parameters are:
   *
   * @param dummy_value (double) - A dummy pixel value for use 
   * in pixels where no data is available.     
   * @param keep_best_res (int) - If present (any value) the best
   * resolution raster will be used as reference.
   * @param auto_equalize ( int ) - If present ( any value ), input_raster2
   * auto-equalizing will be made ( using overlap area reference ).   
   *
   * @note The input_raster2 will be swapped with input_raster1 if 
   * the pixel resolution of input_raster2 is greater than the
   * pixel resolution of input_raster1.
   * @note The output_raster parameters will take the input_raster1 as
   * reference.
   * @note Please check each blending strategy for other specific required
   * parameters.
   */
  class TePDIGeoMosaicStrategy : public TePDIMosaicStrategy
  {
    friend class TePDIGeoMosaicSF;
  
    public :
    
      typedef TeSharedPtr< TePDIGeoMosaicStrategy > pointer;
      typedef const TeSharedPtr< TePDIGeoMosaicStrategy > const_pointer;
      
      /**
       * Default Destructor
       */
      ~TePDIGeoMosaicStrategy();
      
      /**
       * Checks if the supplied parameters fits the requirements of each
       * PDI strategy.
       *
       * Error log messages must be generated. No exceptions generated.
       *
       * @param parameters The parameters to be checked.
       * @return true if the parameters are OK. false if not.
       */
      bool CheckParameters( 
        const TePDIParameters& parameters ) const;       
      
    protected :
      
      /**
       * Default constructor
       *
       * @param factoryName Factory name.
       */
      TePDIGeoMosaicStrategy();
      
      /**
       * Runs the algorithm implementation.
       *
       * @return true if OK. false on errors.
       */
      bool Implementation( const TePDIParameters& params );
      
  };
  
#endif



