#include "TePDIHistogram.hpp"

#include "TePDIUtils.hpp"
#include "TePDIMatrix.hpp"
#include "TePDIAgnostic.hpp"
#include "TePDIPIManager.hpp"

#include <TeUtils.h>
#include <TePrecision.h>

#include <float.h>


TePDIHistogram::TePDIHistogram( unsigned int sample_step )
{
  init();
  
  sample_step_ = sample_step;
}


TePDIHistogram::TePDIHistogram( const TePDIHistogram& external )
: TePDIHMapSpec()
{
  init();
  
  operator=( external );
}


TePDIHistogram::TePDIHistogram( 
  const std::map< double, unsigned int >& external )
{
  init();
  
  operator=( external );
}


TePDIHistogram::~TePDIHistogram()
{
}


bool TePDIHistogram::SetH(
  const TePDITypes::TePDIRasterPtrType& in_raster,
  unsigned int band,
  bool zero_expansion )
{
  PDIAGN_TRUE_OR_RETURN( in_raster.isActive(),
    "Inactive raster pointer" );
  PDIAGN_TRUE_OR_RETURN( 
    in_raster->params().status_ != TeRasterParams::TeNotReady,
    "Raster not ready to read" );
  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
  PDIAGN_TRUE_OR_RETURN( ( ! TePDIUtils::IsFloatBand( in_raster, band ) ),
    "Unable to build non-interpolated histogram from float pixels type" );
    
  /* Optimization for 8 / 16 bit images */
  
  if( in_raster->params().dataType_[ band ] == TeUNSIGNEDCHAR ) {
    return Set8BitH( in_raster, band, zero_expansion );
  } else if( in_raster->params().dataType_[ band ] == TeUNSIGNEDSHORT ) {
    return Set16BitH( in_raster, band, zero_expansion );
  }
  
  TePDIHMapSpec::clear();

  unsigned int in_lines_number = in_raster->params().nlines_;
  unsigned int in_columns_number = in_raster->params().ncols_;
  const bool use_dummy = in_raster->params().useDummy_;

  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
    
  /* Building initial unordered histogram */
  
  std::map< double, unsigned int > init_hist;
  
  TePDIPIManager progress( "Analyzing raster...", 
    (unsigned int)ceil( ( (double)in_lines_number ) / 
    ( (double)( sample_step_ + 1 ) ) ), progress_int_enabled_ );
  
  {
    unsigned int rasterline ;
    unsigned int rastercolumn ;
    double pixel_value;
    std::map< double, unsigned int >::iterator it;
    TeRaster* in_raster_ptr = in_raster.nakedPointer();
    
    for( rasterline = 0 ; rasterline < in_lines_number ; 
      rasterline += ( 1 + sample_step_ ) ) {
      
      for( rastercolumn = 0 ; rastercolumn < in_columns_number ; 
        rastercolumn += ( 1 + sample_step_ ) ) {

        if( in_raster_ptr->getElement( rastercolumn, rasterline,
            pixel_value, band ) ) {
            
          it = init_hist.find( pixel_value );
            
          if( it == init_hist.end() ) {
            init_hist[ pixel_value ] = 1;
          } else {
            it->second = it->second + 1;
          }
          
        } else {
          PDIAGN_TRUE_OR_RETURN( use_dummy,
            "Unable to read from input raster at " +
            Te2String( rasterline ) + "," +
            Te2String( rastercolumn ) );
        }
      }
      
      progress.Increment();
    }
  }
  
  progress.Toggle( false );    
  
  if( init_hist.size() == 0 ) {
    return true;
  }
  
  /* Finding the lowest end highest level */

  double lowest_level = 0;
  double highest_level = 0;
  
  {
    std::map< double, unsigned int >::iterator it = init_hist.begin();
    std::map< double, unsigned int >::iterator it_end = init_hist.end();
    
    lowest_level = highest_level = it->first;
    
    while( it != it_end ) {
      if( it->first < lowest_level ) {
        lowest_level = it->first;
      }
      
      if( it->first > highest_level ) {
        highest_level = it->first;
      }
          
      ++it;
    }
  }
  
  /* Doing zero expansion */
  
  if( zero_expansion ) {
    if( 0. < lowest_level ) {
      for( double index = 0. ; index < lowest_level ; ++index ) {
        init_hist[ index ] = 0;
      }
    
      lowest_level = 0.;
    } else if( 0. > highest_level ) {
      for( double index = highest_level + 1. ; index < 1. ; ++index ) {
        init_hist[ index ] = 0;
      }
      
      std::map< double, unsigned int >::iterator it_end = init_hist.end();
      --it_end;
    
      highest_level = it_end->first;
    }
  }
  
  /* Building ordered histogram */
  
  {
    double index = 0;
    
    for( index = lowest_level ; index <= highest_level ; ++index ) {
      operator[]( index ) = 0;  
    }
  
    std::map< double, unsigned int >::iterator it = init_hist.begin();
    std::map< double, unsigned int >::iterator it_end = init_hist.end();
    
    while( it != it_end ) {
      operator[]( it->first ) = it->second;
    
      ++it;
    }
  }
  
  /* Updating global state variables */
  
  histo_min_level_ = lowest_level;
  histo_max_level_ = highest_level;
  is_discrete_histogram_ = true;
  
  update_minmax_count();

  return true;
}


bool TePDIHistogram::reset(
  const TePDITypes::TePDIRasterPtrType& in_raster,
  unsigned int band,
  unsigned int levels,
  bool zero_expansion )
{
  PDIAGN_TRUE_OR_RETURN( in_raster.isActive(),
    "Inactive raster pointer" );
  PDIAGN_TRUE_OR_RETURN( 
    in_raster->params().status_ != TeRasterParams::TeNotReady,
    "Raster not ready to read" );
  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
    
  /* Optimization - If the raster does not contains float pixels for 
     this band, then we do not need an interpolated histogram */
    
  if( ( ! TePDIUtils::IsFloatBand( in_raster, band ) ) &&
      ( levels == 0 ) ) {
      
    return SetH( in_raster, band, zero_expansion );
  }
    
  TePDIHMapSpec::clear();

  const unsigned int in_lines_number = in_raster->params().nlines_;
  const unsigned int in_columns_number = in_raster->params().ncols_;
  const bool use_dummy = in_raster->params().useDummy_;

  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
    
  /* Finding the lowest end highest level */
  
  TePDIPIManager progress( "", 0, progress_int_enabled_ );
    
  double lowest_level = 0;
  double highest_level = 0;
  
  if ( in_raster->params().decoderIdentifier_ == "DB" ) {
    
    lowest_level = in_raster->params().vmin_[ band ];
    highest_level = in_raster->params().vmax_[ band ];
  } else {
    lowest_level = DBL_MAX;
    highest_level = DBL_MIN;  
  
    progress.Reset( "Analyzing raster...", 
      (unsigned int)ceil( ( (double)in_lines_number ) / 
      ( (double)( sample_step_ + 1 ) ) ) );
  
    unsigned int rasterline = 0;
    unsigned int rastercolumn = 0;
    double current_raster_level = 0;
    TeRaster& in_raster_ref = *(in_raster.nakedPointer());
    
    for( rasterline = 0 ; rasterline < in_lines_number ; 
      rasterline += ( 1 + sample_step_ ) ) {
      
      for( rastercolumn = 0 ; rastercolumn < in_columns_number ; 
        rastercolumn += ( 1 + sample_step_ ) ) {

        if( in_raster_ref.getElement( rastercolumn, rasterline,
            current_raster_level, band ) ) {
            
          if( current_raster_level > highest_level ) {
            highest_level = current_raster_level;
          }
          if( current_raster_level < lowest_level ) {
            lowest_level = current_raster_level;
          }
        } else {
          PDIAGN_TRUE_OR_RETURN( use_dummy,
            "Unable to read from input raster at " +
            Te2String( rasterline ) + "," +
            Te2String( rastercolumn ) );
        }
      }
      
      progress.Increment();
    }
    
    progress.Toggle( false );    
  }
  
  if( zero_expansion &&
      ( ! ( ( lowest_level <= 0.0 ) && ( highest_level >= 0.0 ) ) ) ) {
  
    if( lowest_level > 0.0 ) {
      lowest_level = 0.0;
    } else if ( highest_level < 0.0 ) {
      highest_level = 0.0;
    }
  }  
  
  /* Calculating the histogram step and computed levels */

  double step = 1;
  unsigned int hist_computed_levels = 1;
  {
    if( levels == 0 ) {
      /* Auto step feature */
      PDIAGN_TRUE_OR_RETURN( ( ( highest_level - lowest_level ) > 1.0 ),
        "Histogram level range is too short for auto-step feature"
        " lowest_level=" + Te2String( lowest_level ) + " highest_level=" +
        Te2String( highest_level ) );
  
      hist_computed_levels = 1 +
        ( unsigned int ) floor( highest_level - lowest_level );
    } else {
      hist_computed_levels = levels;
    }
  
    PDIAGN_TRUE_OR_RETURN( hist_computed_levels > 0,
      "Invalid levels number" );
  
    step = ( highest_level - lowest_level ) /
          ( (double)( hist_computed_levels - 1 ) );
          
  }

  /* Ordered Histogram buffer allocation */
  
  TePDIMatrix< double > hbuffer; /* line 0 - pixel level, 
                                    line 1 - pixels count for that level */
  PDIAGN_TRUE_OR_RETURN( 
    hbuffer.Reset( 2, hist_computed_levels, 
    TePDIMatrix< double >::AutoMemPol ),
    "Unable to create histogram internal buffer" );
  {
    hbuffer( 0, 0 ) = lowest_level;
    hbuffer( 1, 0 ) = 0;
        
    for( unsigned int hbuffer_col = 1 ; 
      ( hbuffer_col < hbuffer.GetColumns() ); 
      ++hbuffer_col ) {
    
      hbuffer( 0, hbuffer_col ) = hbuffer( 0, hbuffer_col - 1 ) + step;
      hbuffer( 1, hbuffer_col ) = 0;
    }
  }
  
  /* Interpolated histogram generation */
  
  {
    unsigned int rasterline = 0;
    unsigned int rastercolumn = 0;
    double current_raster_level = 0;
    unsigned int found_index = 0;
    TeRaster& in_raster_ref = *(in_raster.nakedPointer());
    
    progress.Reset( "Generating interpolated histogram", 
      (unsigned int)ceil( ( (double)in_lines_number ) / 
      ( ( double )( 1 + sample_step_ ) ) ) );
    
    for( rasterline = 0 ; rasterline < in_lines_number ; 
      rasterline += ( 1 + sample_step_ ) ) {
      
      for( rastercolumn = 0 ; rastercolumn < in_columns_number ; 
        rastercolumn += ( 1 + sample_step_ ) ) {

        if( in_raster_ref.getElement( rastercolumn, rasterline,
            current_raster_level, band ) ) {
            
          found_index = (unsigned int) TeRound( 
            ( current_raster_level - lowest_level ) / step );          
          
          PDIAGN_DEBUG_CONDITION( ( found_index < hist_computed_levels ),
            "Invalid generated index position " + Te2String( found_index ) );
        
          hbuffer( 1, found_index ) = hbuffer( 1, found_index ) + 1;
        } else {
          PDIAGN_TRUE_OR_RETURN( use_dummy,
            "Unable to read from input raster at " +
            Te2String( rasterline ) + "," +
            Te2String( rastercolumn ) );
        }
      }
      
      progress.Increment();
    }
    
    progress.Toggle( false );     
    
  }
      
  /* Transfering values from internal buffer to internal map */  
  
  for( unsigned int hbuffer_col = 0 ; hbuffer_col < hbuffer.GetColumns() ; 
    ++hbuffer_col ) {
    
    operator[]( hbuffer( 0, hbuffer_col ) ) = 
      ( unsigned int ) hbuffer( 1, hbuffer_col );
  }
  
  /* Updating global state variables */
  
  histo_min_level_ = hbuffer( 0, 0 );
  histo_max_level_ = hbuffer( 0, hbuffer.GetColumns() - 1 );
  
  if( ( histo_min_level_ == floor( histo_min_level_ ) ) &&
      ( step == floor( step ) ) ) {
   
    is_discrete_histogram_ = true;       
  } else {
    is_discrete_histogram_ = false;
  }

  update_minmax_count();

  return true;
}


bool TePDIHistogram::hasFixedStep() const
{
  PDIAGN_TRUE_OR_THROW( size() > 0, "Trying to use an empty histogram" );
  
  const_iterator it = begin();
  const_iterator it_end = end();
  
  std::vector< double > steps;
  std::vector< double >::iterator steps_it;
  std::vector< double >::iterator steps_it_end;
  
  double curr_step = 0;
  double last_value = it->first;
  bool step_found = false;
  
  ++it;
  
  while( it != it_end ) {
    curr_step = it->first - last_value;
    
    steps_it = steps.begin();
    steps_it_end = steps.end();
    
    step_found = false;
    
    while( steps_it != steps_it_end ) {
      if( (*steps_it) == curr_step ) {
        step_found = true;
        break;
      }
      
      ++steps_it;
    }
    
    if( ! step_found ) {
      steps.push_back( curr_step );
    }
    
    last_value = it->first;
    
    ++it;
  }
  
  return ( ( steps.size() <= 1 ) ? true : false );
}


bool TePDIHistogram::IsDiscrete() const
{
  PDIAGN_TRUE_OR_THROW( size() > 0, "Trying to use an empty histogram" );
  return is_discrete_histogram_;
}


bool TePDIHistogram::Discretize()
{
  PDIAGN_TRUE_OR_RETURN( size() > 0, "Trying to use an empty histogram" );
  PDIAGN_TRUE_OR_RETURN( ( ( histo_max_level_ - histo_min_level_ ) > 1 ),
    "The current histogram range do not allow discretization" );

  if( is_discrete_histogram_ ) {
    return true;
  }
  
  TePDIPIManager progress( "Discretizing histogram...", size() * 2,
    progress_int_enabled_ );

  unsigned int progress_step = 0; 

  iterator it = begin();
  iterator it_end = end();

  TePDIHistogram temp_hist;

  while( it != it_end ) {
    progress.Update( progress_step );
    ++progress_step;  
  
    temp_hist[ TeRound( it->first ) ] = 0;

    ++it;
  }
  
  it = begin();
  
  while( it != it_end ) {
    progress.Update( progress_step );
    ++progress_step;  
  
    temp_hist[ TeRound( it->first ) ] += it->second;

    ++it;
  }

  TePDIHMapSpec::clear();
  
  operator=( temp_hist );

  progress.Toggle( false );

  /* Updating global state variables */
  
  histo_min_level_ = begin()->first;
  
  it_end = end();
  --it_end;
  histo_max_level_ = it_end->first;
  
  is_discrete_histogram_ = true;       

  update_minmax_count();

  return true;
}


double TePDIHistogram::GetMinLevel() const
{
  PDIAGN_TRUE_OR_THROW( size() > 0, "Trying to use an empty histogram" );
  return histo_min_level_;
}


double TePDIHistogram::GetMaxLevel() const
{
  PDIAGN_TRUE_OR_THROW( size() > 0, "Trying to use an empty histogram" );
  return histo_max_level_;
}


unsigned int TePDIHistogram::GetMinCount() const
{
  PDIAGN_TRUE_OR_THROW( size() > 0, "Trying to use an empty histogram" );
  return histo_min_count_;
}


unsigned int TePDIHistogram::GetMaxCount() const
{
  PDIAGN_TRUE_OR_THROW( size() > 0, "Trying to use an empty histogram" );
  return histo_max_count_;
}


void TePDIHistogram::clear()
{
  TePDIHMapSpec::clear();
  init();
}


const TePDIHistogram& TePDIHistogram::operator=( 
  const TePDIHistogram& external )
{
  clear();
  
  TePDIHMapSpec::operator=( external );
  
  histo_min_level_ = external.histo_min_level_;
  histo_max_level_ = external.histo_max_level_;
  histo_min_count_ = external.histo_min_count_;
  histo_max_count_ = external.histo_max_count_;
  is_discrete_histogram_ = external.is_discrete_histogram_;
  sample_step_ = external.sample_step_;
  
  return *this;
}


const TePDIHistogram& TePDIHistogram::operator=( 
  const std::map< double, unsigned int >& external )
{
  clear();
  
  if( external.size() > 0 ) {
    std::map< double, unsigned int >::const_iterator ext_it = external.begin();
    std::map< double, unsigned int >::const_iterator ext_it_end = 
      external.end();
    
    is_discrete_histogram_ = true;
    histo_min_level_ = DBL_MAX;
    histo_max_level_ = DBL_MIN;
    
    bool histo_min_count_not_set = true;
    
    while( ext_it != ext_it_end ) {
      if( ext_it->first != ( (double)TeRound( ext_it->first ) ) ) {
        is_discrete_histogram_ = false;
      }
      
      if( ext_it->first < histo_min_level_ ) {
        histo_min_level_ = ext_it->first;
      }
      
      if( ext_it->first > histo_max_level_ ) {
        histo_max_level_ = ext_it->first;
      }
      
      if( histo_min_count_not_set ) {
        histo_min_count_ = ext_it->second;
        histo_min_count_not_set = false;
      } else {
        if( ext_it->second < histo_min_count_ ) {
          histo_min_count_ = ext_it->second;
        }
      }
      
      if( ext_it->second > histo_max_count_ ) {
        histo_max_count_ = ext_it->second;
      }
      
      TePDIHMapSpec::operator[]( ext_it->first ) = ext_it->second;
    
      ++ext_it;
    }
  }
  
  return *this;
}


void TePDIHistogram::init()
{
  histo_min_level_ = histo_max_level_ = 0;
  histo_min_count_ = histo_max_count_ = 0;
  is_discrete_histogram_ = false;
  progress_int_enabled_ = true;
  sample_step_ = 0;
}


void TePDIHistogram::update_minmax_count()
{
  iterator it = begin();
  const iterator it_end = end();

  bool initial_min_set = false;
  bool initial_max_set = false;
  
  histo_min_count_ = 0;
  histo_max_count_ = 0;

  while( it != it_end ) {
    if( initial_min_set ) {
      if( it->second < histo_min_count_ ) {
        histo_min_count_ = it->second;
      }
    } else {
      histo_min_count_ = it->second;
      initial_min_set = true;
    }
  
    if( initial_max_set ) {
      if( it->second > histo_max_count_ ) {
        histo_max_count_ = it->second;
      }
    } else {
      histo_max_count_ = it->second;
      initial_max_set = true;
    }

    ++it;
  }
}

void TePDIHistogram::ToggleProgressInt( bool enabled )
{
  progress_int_enabled_ = enabled;
}


bool TePDIHistogram::Set8BitH(
  const TePDITypes::TePDIRasterPtrType& in_raster,
  unsigned int band,
  bool zero_expansion )
{
  PDIAGN_TRUE_OR_RETURN( in_raster.isActive(),
    "Inactive raster pointer" );
  PDIAGN_TRUE_OR_RETURN( 
    in_raster->params().status_ != TeRasterParams::TeNotReady,
    "Raster not ready to read" );
  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
  PDIAGN_TRUE_OR_RETURN( ( ! TePDIUtils::IsFloatBand( in_raster, band ) ),
    "Unable to build non-interpolated histogram from float pixels type" );
  PDIAGN_TRUE_OR_RETURN( 
    ( in_raster->params().dataType_[ band ] == TeUNSIGNEDCHAR ),
    "Invalid pixel data type" );
    
  TePDIHMapSpec::clear();

  unsigned int in_lines_number = in_raster->params().nlines_;
  unsigned int in_columns_number = in_raster->params().ncols_;
  const bool use_dummy = in_raster->params().useDummy_;

  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
    
  /* Building histogram */
  
  unsigned int init_hist[ 256 ];
  unsigned int hist_index = 0;
  
  for( hist_index = 0 ; hist_index < 256 ; ++hist_index ) {
    init_hist[ hist_index ] = 0;
  }
  
  /* Analysing raster */
  
  TePDIPIManager progress( "Analyzing raster...", 
    ( unsigned int )ceil( ( (double)in_lines_number ) / 
    ( (double)( 1 + sample_step_ ) ) ),
    progress_int_enabled_ );
  
  {
    unsigned int rasterline ;
    unsigned int rastercolumn ;
    double pixel_value;
    unsigned int ui_pixel_value = 0;
    TeRaster* in_raster_ptr = in_raster.nakedPointer();
    
    for( rasterline = 0 ; rasterline < in_lines_number ; 
      rasterline += ( 1 + sample_step_) ) {
      
      progress.Increment();
    
      for( rastercolumn = 0 ; rastercolumn < in_columns_number ; 
        rastercolumn += ( 1 + sample_step_ ) ) {

        if( in_raster_ptr->getElement( rastercolumn, rasterline,
            pixel_value, band ) ) {
            
          ui_pixel_value = ( unsigned int ) pixel_value;
            
          init_hist[ ui_pixel_value ] = init_hist[ ui_pixel_value ] + 1;
        } else {
          PDIAGN_TRUE_OR_RETURN( use_dummy, 
            "Unable to read from input raster at " +
            Te2String( rasterline ) + "," +
            Te2String( rastercolumn ) );
        }
      }
    }
  }
  
  progress.Toggle( false );    
  
  /* Finding the lowest end highest level */

  unsigned int lowest_level = 0;
  unsigned int highest_level = 255;
  
  for( hist_index = 0 ; hist_index < 256 ; ++hist_index ) {
    if( init_hist[ hist_index ] != 0 ) {
      lowest_level = hist_index;
      break;
    }
  }
  for( hist_index = 0 ; hist_index < 256 ; ++hist_index ) {  
    if( init_hist[ 255 - hist_index ] != 0 ) {
      highest_level = 255 - hist_index;
      break;
    }
  }
  
  if( zero_expansion ) {
    lowest_level = 0;
  }
  
  /* Building ordered histogram */
  
  for( hist_index = lowest_level; hist_index <= highest_level ; 
       ++hist_index ) {
       
    operator[]( ((double)hist_index) ) = init_hist[ hist_index ];
  }
  
  /* Updating global state variables */
  
  histo_min_level_ = (double)lowest_level;
  histo_max_level_ = (double)highest_level;
  is_discrete_histogram_ = true;
  
  update_minmax_count();

  return true;
}


bool TePDIHistogram::Set16BitH(
  const TePDITypes::TePDIRasterPtrType& in_raster,
  unsigned int band,
  bool zero_expansion )
{
  PDIAGN_TRUE_OR_RETURN( in_raster.isActive(),
    "Inactive raster pointer" );
  PDIAGN_TRUE_OR_RETURN( 
    in_raster->params().status_ != TeRasterParams::TeNotReady,
    "Raster not ready to read" );
  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
  PDIAGN_TRUE_OR_RETURN( ( ! TePDIUtils::IsFloatBand( in_raster, band ) ),
    "Unable to build non-interpolated histogram from float pixels type" );
  PDIAGN_TRUE_OR_RETURN( 
    ( in_raster->params().dataType_[ band ] == TeUNSIGNEDSHORT ),
    "Invalid pixel data type" );
    
  TePDIHMapSpec::clear();

  unsigned int in_lines_number = in_raster->params().nlines_;
  unsigned int in_columns_number = in_raster->params().ncols_;
  const bool use_dummy = in_raster->params().useDummy_;

  PDIAGN_TRUE_OR_RETURN( (int)band < in_raster->nBands(),
    "Invalid band" );
    
  /* Building histogram */
  
  unsigned int init_hist[ 65536 ];
  unsigned int hist_index = 0;
  
  for( hist_index = 0 ; hist_index < 65536 ; ++hist_index ) {
    init_hist[ hist_index ] = 0;
  }
  
  /* Analysing raster */
  
  TePDIPIManager progress( "Analyzing raster...", 
    (unsigned int)ceil( ( (double)in_lines_number ) / 
    ( (double)( 1 + sample_step_ ) ) ),
    progress_int_enabled_ );
  
  {
    unsigned int rasterline ;
    unsigned int rastercolumn ;
    double pixel_value;
    unsigned int ui_pixel_value = 0;
    TeRaster* in_raster_ptr = in_raster.nakedPointer();
    
    for( rasterline = 0 ; rasterline < in_lines_number ; 
      rasterline += ( 1 + sample_step_ ) ) {
      progress.Increment();
    
      for( rastercolumn = 0 ; rastercolumn < in_columns_number ; 
        rastercolumn += ( 1 + sample_step_ ) ) {

        if( in_raster_ptr->getElement( rastercolumn, rasterline,
            pixel_value, band ) ) {
            
          ui_pixel_value = ( unsigned int ) pixel_value;
            
          init_hist[ ui_pixel_value ] = init_hist[ ui_pixel_value ] + 1;
        } else {
          PDIAGN_TRUE_OR_RETURN( use_dummy,
            "Unable to read from input raster at " +
            Te2String( rasterline ) + "," +
            Te2String( rastercolumn ) );
        }
      }
    }
  }
  
  progress.Toggle( false );    
  
  /* Finding the lowest end highest level */

  unsigned int lowest_level = 0;
  unsigned int highest_level = 65535;
  
  for( hist_index = 0 ; hist_index < 65536 ; ++hist_index ) {
    if( init_hist[ hist_index ] != 0 ) {
      lowest_level = hist_index;
      break;
    }
  }
  for( hist_index = 0 ; hist_index < 65536 ; ++hist_index ) {  
    if( init_hist[ 65535 - hist_index ] != 0 ) {
      highest_level = 65535 - hist_index;
      break;
    }
  }
  
  if( zero_expansion ) {
    lowest_level = 0;
  }
  
  /* Building ordered histogram */
  
  for( hist_index = ( (unsigned int)lowest_level ); 
       hist_index <= ( (unsigned int)highest_level ); 
       ++hist_index ) {
       
    operator[]( ((double)hist_index) ) = init_hist[ hist_index ];
  }
  
  /* Updating global state variables */
  
  histo_min_level_ = (double)lowest_level;
  histo_max_level_ = (double)highest_level;
  is_discrete_histogram_ = true;
  
  update_minmax_count();

  return true;
}

