/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIHISTOGRAM_HPP
  #define TEPDIHISTOGRAM_HPP

  #include "TePDITypes.hpp"
  #include <TeSharedPtr.h>

  #include <map>

  class TePDIHMapSpec : public std::map< double, unsigned int > {};

  /**
   * @brief This class deals with histograms and related tasks.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIModule
   */
  class TePDIHistogram : public TePDIHMapSpec {
    public :
      typedef TeSharedPtr< TePDIHistogram > pointer;
      typedef const TeSharedPtr< TePDIHistogram > const_pointer;

      /**
       * Default Constructor
       *
       * @param sample_step The number of lines/columns ignored between each
       * image sample used when generating the histogram.
       *
       */
      TePDIHistogram( unsigned int sample_step = 0 );
      
      /**
       * Alternative constructor.
       *
       * @param external External histogram reference.
       *
       */
      TePDIHistogram( const TePDIHistogram& external );      
      
      /**
       * Alternative constructor from std::map.
       *
       * @param external External map reference.
       *
       */
      TePDIHistogram( const std::map< double, unsigned int >& external );

      /**
       * Default Destructor
       */
      ~TePDIHistogram();

      /**
       * Generates a level-ordered histogram ( fixed step ).
       *
       * NOTE: If the input raster has float data, the histogram will be
       * interpolated.
       *
       * @param in_raster Input raster.
       * @param band Band number.
       * @param levels The number of interpolated levels, and if set to zero,
       * the number of levels will be automaticlly found by using
       * level step with value one.
       * @param zero_expansion If true, the histogram will be expanded
       * until it contains value zero.
       * @return true if the histogram has been generated, false on error.
       */
      bool reset(
        const TePDITypes::TePDIRasterPtrType& in_raster,
        unsigned int band, unsigned int levels,
        bool zero_expansion );

      /**
       * Histogram discrete verification.
       *
       * @return true if the histogram has only discrete levels.
       */
      bool IsDiscrete() const;
      
      /**
       * Histogram fixed step verification.
       *
       * @return true if the histogram has fixed step.
       */
      bool hasFixedStep() const;      

      /**
       * Round the histogram levels to the near integer generating a
       * descrete ( fixed step ) histogram.
       *
       * @return True if the process was concluded successfully, false if not.
       */
      bool Discretize();

      /**
       * Histogram minimum level.
       *
       * @return The histogram minimum level.
       */
      double GetMinLevel() const;

      /**
       * Histogram maximum level.
       *
       * @return The histogram maximum level.
       */
      double GetMaxLevel() const;

      /**
       * Histogram minimum count value for all levels.
       *
       * @return The histogram minimum count value.
       */
      unsigned int GetMinCount() const;

      /**
       * Histogram maximum count value for all levels.
       *
       * @return The histogram maximum count value.
       */
      unsigned int GetMaxCount() const;

      /**
       * std::map::clear() overload.
       *
       */
      void clear();
      
      /**
       * operator= overload.
       *
       */
      const TePDIHistogram& operator=( const TePDIHistogram& external );      
      
      /**
       * operator= overload for simple std:maps.
       *
       */
      const TePDIHistogram& operator=( 
        const std::map< double, unsigned int >& external );
      
      /**
       * Enable / Disable the progress interface.
       *
       * @param enabled Flag to enable ( true - default ) or disable ( false ).
       */      
      void ToggleProgressInt( bool enabled );

    protected :

      /**
       * Histogram minimum value.
       */
      double histo_min_level_;

      /**
       * Histogram maximum value.
       */
      double histo_max_level_;

      /**
       * Histogram minimum count value.
       */
      unsigned int histo_min_count_;

      /**
       * Histogram maximum count value.
       */
      unsigned int histo_max_count_;

      /**
       * True if the current histogram has discrete values ( only integer ).
       */
      bool is_discrete_histogram_;
      
      /**
       * Progress interface enabled status.
       */
      bool progress_int_enabled_;
      
      /**
       * The number of lines/columns ignored between each
       * image sample used when generating the histogram.
       */      
      unsigned int sample_step_;
      
      /**
       * Reset the internal variables with default values.
       */
      void init();      

      /**
       * Updates histo_min_count_ and histo_max_count_, using the current
       * map contents.
       */
      void update_minmax_count();
      
      /**
       * Generates a level-ordered histogram ( fixed step ).
       *
       * NOTE: No trows generated.
       *
       * @param in_raster Input raster.
       * @param band Band number.
       * @param zero_expansion If true, the histogram will be expanded
       * until it contains value zero.
       * @return true if the histogram has been generated, false on error.
       */
      bool SetH(
        const TePDITypes::TePDIRasterPtrType& in_raster,
        unsigned int band, bool zero_expansion );      
      
      /**
       * Generates a level-ordered histogram ( fixed step )for an 8 bit raster.
       *
       * NOTE: No trows generated.
       *
       * @param in_raster Input raster.
       * @param band Band number.
       * @param zero_expansion If true, the histogram will be expanded
       * until it contains value zero.
       * @return true if the histogram has been generated, false on error.
       */
      bool Set8BitH(
        const TePDITypes::TePDIRasterPtrType& in_raster,
        unsigned int band,
        bool zero_expansion );
        
      /**
       * Generates a level-ordered histogram (fixed step) for an 16 bit raster.
       *
       * NOTE: No trows generated.
       *
       * @param in_raster Input raster.
       * @param band Band number.
       * @param zero_expansion If true, the histogram will be expanded
       * until it contains value zero.
       * @return true if the histogram has been generated, false on error.
       */
      bool Set16BitH(
        const TePDITypes::TePDIRasterPtrType& in_raster,
        unsigned int band,
        bool zero_expansion );          
  };
  
/** @example TePDIHistogram_test.cpp
 *    Shows how to use this class.
 */    

#endif //TEPDIHISTOGRAM_HPP
