/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIMAPPEDMEMORY_HPP
  #define TEPDIMAPPEDMEMORY_HPP
  
  #include <TeSharedPtr.h>
  #include "TePDIAgnostic.hpp"
  
  #include <string>
  
  #if defined WIN32
    #include <windows.h>
    #include <winbase.h>
  #elif defined __unix__
    #include <sys/mman.h>
  #else
    #error "Unsuported plataform"
  #endif  

  /**
    * @brief This is class deals with mapped memory resources.
    * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
    * @ingroup PDIModule
    */
  class TePDIMappedMemory{
    public :
      typedef TeSharedPtr< TePDIMappedMemory > pointer;
      typedef const TeSharedPtr< TePDIMappedMemory > const_pointer;    
      
      /**
       * Default Constructor.
       *
       * @note Throws an exception on errors.
       *
       * @param size The size in bytes of the new mapped memory.
       */
      explicit TePDIMappedMemory( unsigned long int size = 0 );      
    
      /**
       * Default Destructor.
       */
      ~TePDIMappedMemory();
      
      /**
       * Default Constructor.
       *
       * @param size The size in bytes of the new mapped memory.
       * @return true if OK. false on errors.
       */
      bool Reset( unsigned long int size = 0 );
      
      /**
       * Returns a pointer reference to the mapped memory.
       *
       * @note Throws an exception if no mapped memory is avalible.
       */
      inline void* GetPointer() const
      {
        PDIAGN_PRE_CONDITION( mapping_is_active_, 
          "Trying to access an inactive mapping" );
  
        return m_lpszFile_;      
      };
      
    protected :
    
      #if defined WIN32
      
        /**
         * Handle to file we're currently zapping
         */
        HANDLE m_hFile_;
        
        /**
         * Handle to memory-mapping of that file
         */
        HANDLE m_hMapping_;
        
        /**
         * Pointer to view of file mapped to memory
         */
        LPVOID m_lpszFile_;
        
        /**
         * 
         */
        DWORD m_dwSize_;
        
      #elif defined __unix__
  
        /**
         * Handle to file we're currently zapping
         */
        int m_hFile_;
        
        /**
         * Pointer to view of file mapped to memory
         */
        void* m_lpszFile_;
        
        /**
         * Current size.
         */
        long m_dwSize_;
        
      #else
        #error "Unsuported plataform"
      #endif
      
      /**
       * A flag indicating when this map is active.
       *
       */        
      bool mapping_is_active_;
      
      
      /**
       * The temporary disk file name.
       *
       */        
      std::string disk_file_name_;
     
      /**
       * Initiates the global initial state.
       *
       */      
      void Init();
      
    private :
      /**
       * Alternative Constructor.
       *
       * @param external External reference.
       */
      TePDIMappedMemory( const TePDIMappedMemory& ) {}; 
          
      /**
       * Operator= overload.
       *
       */
      const TePDIMappedMemory& operator=( const TePDIMappedMemory& )
      {
        return *this;
      };
  };
  
/** @example TePDIMappedMemory_test.cpp
 *    Shows how to use this class.
 */    

#endif
