 
#include "TePDINoBlendStrategy.hpp"
#include "TePDIUtils.hpp"
#include "TePDIStrategyFactory.hpp"
#include "TePDIPIManager.hpp"
#include "TePDIAgnostic.hpp"

#include <TeBox.h>
#include <TeUtils.h>

TePDINoBlendStrategy::TePDINoBlendStrategy()
{
};      

TePDINoBlendStrategy::~TePDINoBlendStrategy()
{
};


bool TePDINoBlendStrategy::CheckParameters( 
  const TePDIParameters& parameters ) const
{
  /* Checking input_raster1 */
  
  TePDITypes::TePDIRasterPtrType input_raster1;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "input_raster1", 
    input_raster1 ),
    "Missing parameter: input_raster1" );
  PDIAGN_TRUE_OR_RETURN( input_raster1.isActive(),
    "Invalid parameter: input_raster1 inactive" );
  PDIAGN_TRUE_OR_RETURN( input_raster1->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: input_raster1 not ready" );    
    
  /* Checking input_raster2 */
  
  TePDITypes::TePDIRasterPtrType input_raster2;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "input_raster2", 
    input_raster2 ),
    "Missing parameter: input_raster2" );
  PDIAGN_TRUE_OR_RETURN( input_raster2.isActive(),
    "Invalid parameter: input_raster2 inactive" );
  PDIAGN_TRUE_OR_RETURN( input_raster2->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: input_raster2 not ready" );    
    
  /* Checking output_raster */
  
  TePDITypes::TePDIRasterPtrType output_raster;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "output_raster", 
    output_raster ),
    "Missing parameter: output_raster" );
  PDIAGN_TRUE_OR_RETURN( output_raster.isActive(),
    "Invalid parameter: output_raster inactive" );
  PDIAGN_TRUE_OR_RETURN( output_raster->params().status_ != 
    TeRasterParams::TeNotReady, "Invalid parameter: output_raster not ready" );    
    
  /* Checking bands */    
    
  int band1 = 0;    
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "band1", band1 ),
    "Missing parameter : band1" );
  PDIAGN_TRUE_OR_RETURN( ( input_raster1->nBands() > band1 ),
    "Invalid parameter : band1" );
    
  int band2 = 0;    
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "band2", band2 ),
    "Missing parameter : band2" );
  PDIAGN_TRUE_OR_RETURN( ( input_raster2->nBands() > band2 ),
    "Invalid parameter : band2" ); 

  int out_band = 0;    
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "out_band", out_band ),
    "Missing parameter : out_band" );
  PDIAGN_TRUE_OR_RETURN( ( output_raster->nBands() > out_band ),
    "Invalid parameter : out_band" );       
    
  /* Checking raster polygons */
  
  TePolygon raster1_polygon;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster1_polygon", 
    raster1_polygon ), "Missing parameter : raster1_polygon" );
  PDIAGN_TRUE_OR_RETURN(
    ( ! input_raster1->begin( raster1_polygon, TeBoxPixelIn, 0 ).end() ),
    "Invalid raster1_polygon" )
    
  TePolygon raster2_polygon;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_polygon", 
    raster2_polygon ), "Missing parameter : raster2_polygon" );
  PDIAGN_TRUE_OR_RETURN(
    ( ! input_raster2->begin( raster2_polygon, TeBoxPixelIn, 0 ).end() ),
    "Invalid raster2_polygon" )
    
  /* Checking raster2 polygon offsets */
  
  double raster2_pol_offset_x = 0;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_pol_offset_x", 
    raster2_pol_offset_x ), "Missing parameter : raster2_pol_offset_x" );  

  double raster2_pol_offset_y = 0;
  PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "raster2_pol_offset_y", 
    raster2_pol_offset_y ), "Missing parameter : raster2_pol_offset_y" );    
    
  return true;
}


bool TePDINoBlendStrategy::Implementation( const TePDIParameters& params )
{
  /* Extracting parameters */

  TePDITypes::TePDIRasterPtrType input_raster1;
  params.GetParameter( "input_raster1", input_raster1 );
  TeRaster& input_raster1_ref = *( input_raster1.nakedPointer() );

  TePDITypes::TePDIRasterPtrType input_raster2;
  params.GetParameter( "input_raster2", input_raster2 );
  TeRaster& input_raster2_ref = *( input_raster2.nakedPointer() );

  TePDITypes::TePDIRasterPtrType output_raster;
  params.GetParameter( "output_raster", output_raster );
  TeRaster& output_raster_ref = *( output_raster.nakedPointer() );
  
  int band1 = 0;
  params.GetParameter( "band1", band1 );

  int band2 = 0;
  params.GetParameter( "band2", band2 );
  
  int out_band = 0;
  params.GetParameter( "out_band", out_band );  
  
  TePolygon raster1_polygon;
  params.GetParameter( "raster1_polygon", raster1_polygon );

  TePolygon raster2_polygon;
  params.GetParameter( "raster2_polygon", raster2_polygon );
  
  double raster2_pol_offset_x = 0;
  params.GetParameter( "raster2_pol_offset_x", raster2_pol_offset_x );

  double raster2_pol_offset_y = 0;
  params.GetParameter( "raster2_pol_offset_y", raster2_pol_offset_y );  
  
  /* Dummy definition */
  
  const bool raster1_uses_dummy = input_raster1->params().useDummy_;
  const bool raster2_uses_dummy = input_raster2->params().useDummy_;
  const bool output_raster_uses_dummy = output_raster->params().useDummy_;
  double output_raster_dummy_value = 0;
  if( output_raster_uses_dummy ) {
    output_raster_dummy_value = output_raster->params().dummy_[ 0 ];
  }
  
  /* Raster1 positioning */
  
  double raster1_abs_index_offset_x = 0;// output raster */
  double raster1_abs_index_offset_y = 0;
  double raster1_rel_index_offset_x = 0;// input x output
  double raster1_rel_index_offset_y = 0;
  {
    TeBox pol1_box_indexed;
    TePDIUtils::MapCoords2RasterIndexes( raster1_polygon.box(),
      input_raster1, pol1_box_indexed );
          
    double r1_min_x = MIN( pol1_box_indexed.x1(),
      pol1_box_indexed.x2() );
    double r1_min_y = MIN( pol1_box_indexed.y1(),
      pol1_box_indexed.y2() );      

    TeCoord2D r1_UL_ref1_indexed( r1_min_x, r1_min_y );  
    TeCoord2D r1_UL_ref1( input_raster1->index2Coord( r1_UL_ref1_indexed ) );  
    TeCoord2D r1_UL_refout_indexed( output_raster->coord2Index( r1_UL_ref1 ) );
        
    raster1_abs_index_offset_x = r1_UL_refout_indexed.x();  
    raster1_abs_index_offset_y = r1_UL_refout_indexed.y();
      
    raster1_rel_index_offset_x = r1_UL_ref1_indexed.x() -
      raster1_abs_index_offset_x;
    raster1_rel_index_offset_y = r1_UL_ref1_indexed.y() -
      raster1_abs_index_offset_y;
  }  
  
  /* Raster2 positioning */
  
  TePolygon r2_pol_refout;
  double raster2_rel_index_offset_x = 0;
  double raster2_rel_index_offset_y = 0;  
  {
    TePolygon r2_pol_ref2_indexed;
    TePDIUtils::MapCoords2RasterIndexes( raster2_polygon,
      input_raster2, r2_pol_ref2_indexed );
      
    double r2_min_x = MIN( r2_pol_ref2_indexed.box().x1(),
      r2_pol_ref2_indexed.box().x2() );
    double r2_min_y = MIN( r2_pol_ref2_indexed.box().y1(),
      r2_pol_ref2_indexed.box().y2() );
  
    raster2_rel_index_offset_x = r2_min_x - raster1_abs_index_offset_x -
      raster2_pol_offset_x;
    raster2_rel_index_offset_y = r2_min_y - raster1_abs_index_offset_y -
      raster2_pol_offset_y;
      
    TePolygon r2_pol_refout_indexed = r2_pol_ref2_indexed;
        
    for( unsigned int lr_index = 0 ; 
      lr_index < r2_pol_refout_indexed.size() ; ++lr_index ) {
      for( unsigned int c2d_index = 0 ; 
        c2d_index < r2_pol_refout_indexed[ lr_index ].size() ; 
        ++c2d_index ) {
            
        r2_pol_refout_indexed[ lr_index ][ c2d_index ].x( 
          r2_pol_refout_indexed[ lr_index ][ c2d_index ].x() -
          r2_min_x + raster1_abs_index_offset_x + raster2_pol_offset_x );
        r2_pol_refout_indexed[ lr_index ][ c2d_index ].y( 
          r2_pol_refout_indexed[ lr_index ][ c2d_index ].y() -
          r2_min_y + raster1_abs_index_offset_y + raster2_pol_offset_y );          
      }
    }
    
    TePDIUtils::MapRasterIndexes2Coords( r2_pol_refout_indexed,
      output_raster, r2_pol_refout );
  }  
  
  /* Global vars */
  
  TeRaster::iteratorPoly input_raster1_it =
    output_raster->begin( raster1_polygon, TeBoxPixelIn, 0 );
  PDIAGN_TRUE_OR_RETURN( ( ! input_raster1_it.end() ), 
    "Unable to create an iterator over raster1 area" ); 
    
  TeRaster::iteratorPoly input_raster2_it =
    output_raster->begin( r2_pol_refout, TeBoxPixelIn, 0 );
  PDIAGN_TRUE_OR_RETURN( ( ! input_raster2_it.end() ), 
    "Unable to create an iterator over raster2 area" );       
  
  TePDIPIManager progress( "Rendering output...", 
    input_raster1_it.nLinesInPoly() + input_raster2_it.nLinesInPoly(),
    progress_interface_enabled_ );  
    
  /* Transfering raster 1 - non-intersected area */
  {
    unsigned int curr_line = 0;
    unsigned int curr_col = 0;
    double curr_value = 0;  
    int offset_x = TeRound( raster1_rel_index_offset_x );
    int offset_y = TeRound( raster1_rel_index_offset_y );
    
    while( ! input_raster1_it.end() ) {
      curr_line = input_raster1_it.currentLine();
      curr_col = input_raster1_it.currentColumn();
        
      if( input_raster1_ref.getElement( curr_col + offset_x, 
          curr_line + offset_y, curr_value, band1 ) ) {
          
        if( ! ( output_raster_uses_dummy && 
          ( curr_value == output_raster_dummy_value ) ) ) {
          
          PDIAGN_TRUE_OR_RETURN( 
            output_raster_ref.setElement( curr_col, curr_line, curr_value, 
            out_band ),
            "Unable to write to output raster at line=" + 
            Te2String( curr_line ) + 
            " col=" + Te2String( curr_col ) );                    
        }
      } else {
        PDIAGN_TRUE_OR_RETURN( raster1_uses_dummy, 
          "Unable to read from input_raster1 at line=" + 
          Te2String( curr_line + offset_y ) + " col=" + 
          Te2String( curr_col + offset_x ) );
      }
  
      ++input_raster1_it;
       
      progress.Increment();
    } 
  }
  
  /* Transfering raster 2 - non-intersected area */
  {
    unsigned int curr_line = 0;
    unsigned int curr_col = 0;
    double curr_value = 0;  
    int offset_x = TeRound( raster2_rel_index_offset_x );
    int offset_y = TeRound( raster2_rel_index_offset_y );
      
    while( ! input_raster2_it.end() ) {
      curr_line = input_raster2_it.currentLine();
      curr_col = input_raster2_it.currentColumn();
        
      if( input_raster2_ref.getElement( curr_col + offset_x, 
          curr_line + offset_y , curr_value, band2 ) ) {
          
        if( ! ( output_raster_uses_dummy && 
          ( curr_value == output_raster_dummy_value ) ) ) {
          
          PDIAGN_TRUE_OR_RETURN( 
            output_raster_ref.setElement( curr_col, curr_line, curr_value, 
            out_band ), "Unable to write to output raster at line=" + 
            Te2String( curr_line ) + " col=" + Te2String( curr_col ) );
        }
      } else {
        PDIAGN_TRUE_OR_RETURN( raster2_uses_dummy, 
          "Unable to read from input_raster2 at line=" + 
          Te2String( curr_line + offset_y ) + " col=" + 
          Te2String( curr_col + offset_x ) );
      }
  
      ++input_raster2_it;
       
      progress.Increment();
    }
  }   
  
  return true;
}


