/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIPIMANAGER_HPP
  #define TEPDIPIMANAGER_HPP
  
  #include <TeSharedPtr.h>
  
  #include <string>

  /**
   * @brief This is the class for the active progress interface management.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIModule
   */
  class TePDIPIManager {
  
    public :

      typedef TeSharedPtr< TePDIPIManager > pointer;
      typedef const TeSharedPtr< TePDIPIManager > const_pointer;        
    
      /**
       * Default Constructor
       *
       * @param message Progress window message (label).
       * @param tsteps Progress window total steps.
       * @param enabled Flag to enable ( true ) or disable ( false ).
       */
      TePDIPIManager( const std::string& message = " ", 
        unsigned long int tsteps = 0, bool enabled = false );    
    
      /**
       * Default Destructor
       */
      ~TePDIPIManager();
      
      /**
       * Enable / Disable the current progress interface manager.
       *
       * @param enabled Flag to enable ( true ) or disable ( false ).
       */
      void Toggle( bool enabled );  
      
      /**
       * Update the current progress interface to the supplied step.
       *
       * @param step Current step.
       */
      void Update( unsigned long int step );      
      
      /**
       * Incriments the the current progress by one step.
       */
      void Increment();

      /**
       * Reset the current instance state.
       * 
       * @param message Message (label).
       * @param tsteps Total steps.
       */
      void Reset( const std::string& message = " ", 
        unsigned long int tsteps = 0 );

    private :
    
      /**
       * @brief This is the class for the active progress interface management
       * parameters.
       * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
       * @ingroup PDIModule
       */    
      class TePDIPIManagerSettings {
        public :
          /**
           * A flag indicating if the PI manager is enabled.
           */
          bool enabled;
          
          /**
           * The total steps number.
           */
          unsigned long int tsteps;
          
          /**
           * The current step.
           */
          unsigned long int currstep;
          
          /**
           * The block size (steps number) for each progress interface
           * update.
           */
          unsigned long int blksize;
          
          /**
           * The index of the next block (steps number) for each progress
           * interface update.
           */
          unsigned long int nxtblk_index;
          
          /**
           * The progress interface message (label).
           */
          std::string message;
          
          /**
           * Default constructor.
           */
          TePDIPIManagerSettings()
          : enabled( false ), tsteps( 0 ), currstep( 0 ), blksize( 0 ),
            nxtblk_index( 0 ), message( "" )
          {};
          
          /**
           * Default desstructor.
           */
          ~TePDIPIManagerSettings() {};
      };
    
      /**
       * Internal progress interface status (current instance).
       */
      mutable TePDIPIManagerSettings my_settings_;
      
      /**
       * Internal progress global interface status (last active instance).
       */
      mutable TePDIPIManagerSettings old_global_settings_;
      
      /**
       * Returns a reference to the global settings used by the last
       * active progress interface manager.
       */      
      static TePDIPIManagerSettings& GetGlobalSettings()
      {
        static TePDIPIManagerSettings global_settings;
        return global_settings;
      };
      
      /**
       * Updates the progress interface with the new settings.
       *
       * @param settings The new progress interface manager to be applied.
       */         
      void UpdateProgressInterface( 
        const TePDIPIManagerSettings& settings ) const;
  };

#endif
