/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

/**
 *   This file contains the class of structures and definitions
 *   to deal with parameters used by all PDI classes.
 */

#ifndef TEPDIPARAMETERS_HPP
  #define TEPDIPARAMETERS_HPP

  #include "TePDITypes.hpp"
  #include "TePDIRgbPalette.hpp"
  #include "TePDIFilterMask.hpp"
  #include "TePDIHistogram.hpp"
  
  #include <TeSharedPtr.h>
  #include <TeRaster.h>
  #include <TeBox.h>
  #include <TeGeometry.h>
  #include <TeLayer.h>
  
  #include <time.h>

  #include <map>
  #include <string>
  #include <vector>

/**
 * @brief This is the class to deal with parameters used by all PDI classes.
 * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
 * @ingroup PDIModule
 */
  class TePDIParameters {
    public :
      typedef TeSharedPtr< TePDIParameters > pointer;
      typedef const TeSharedPtr< TePDIParameters > const_pointer;

      /**
       * Suported parameters.
       */
      enum TePDIParamTypes{
        /** Integer parameter */
        TePDIIntParam = 1,
        /** Integers vector parameter */
        TePDIIntVectorParam = 2,
        /** Raster reference parameter */
        TePDIRasterParam = 3,
        /** Raster references vector parameter */
        TePDIRasterVectorParam = 4,
        /** Raster references vector pointer parameter */
        TePDIRastersVectorPtrParam = 5,
        /** Double parameter */
        TePDIDoubleParam = 6,
        /** Double's vector parameter */
        TePDIDoubleVectorParam = 7,
        /** String parameter */
        TePDIStringParam = 8,
        /** TeBox parameter */
        TeBoxParam = 9,
        /** PDI palette parameter */
        TePDIRgbPaletteParam = 10,
        /** PDI Filter Mask parameter */
        TePDIFilterMaskParam = 11,
        /** TePolygon parameter */
        TePolygonParam = 12,
        /** PDI Look-up table parameter */
        TePDILutParam = 13,
        /** PDI Look-up table pointer parameter */
        TePDILutPtrParam = 14,        
        /** TeLayer reference parameter */
        TePDILayerParam = 15,
        /** TePolygonSet parameter */
        TePolygonSetParam = 16,
        /** TePolygonSet pointer parameter */
        TePolygonSetPtrParam = 17,        
        /** TePDIHistogram pointer parameter */
        TePDIHistogramPtrParam = 18,
        /** TePDIHistogram pointers vector parameter */
        TePDIHistVecParam = 19,
        /** TeCoordPairVect pointer parameter */
        TeCoordPairVectPtrParam = 20,
        /** Polygon sets map pointer parameter */
        TePDIPolSetMapPtrParam = 21
      };

      /**
       * Default Constructor
       */
      TePDIParameters();

      /**
       *  Default Destructor
       */
      ~TePDIParameters();
      
      /**
       * Checks if the parameter was set.
       *
       * @param pname Parameter name.
       * @param type Parameter type.
       * @return true if OK, false if the parameter was not set.
       */
      bool CheckParameter( const std::string& pname,
                           TePDIParamTypes type ) const;  

      /**
       * Set the a name descriptor for the current parameters.
       *
       * @param dec_name A a name descriptor for the current parameters. 
       */                                                      
      void SetdecName( const std::string& dec_name );                            
                           
      /**
       * Returns the a name descriptor for the current parameters.
       *
       * @return A a name descriptor for the current parameters. 
       */                                                      
      std::string decName() const;                           
      
      /**
       * Operator == overload.
       *
       * @param ext_instance External instance reference.
       * @return true if this instance have the same internal
       * parameters that the external instance.
       */
      bool operator==( const TePDIParameters& ext_instance ) const;
      
      /**
       * Operator != overload.
       *
       * @param ext_instance External instance reference.
       * @return true if this instance don't have the same internal
       * parameters that the external instance.
       */
      bool operator!=( const TePDIParameters& ext_instance ) const;
      
      /**
       * Clear all parameter contents.
       *
       */
      void Clear();

      /**
       * Store a TeRaster parameter reference.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDIRasterPtrType& data );

      /**
       * Gets a TeRaster parameter reference.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was not found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDIRasterPtrType& data ) const;

      /**
       * Store a TeRaster vector parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDIRasterVectorType& data );

      /**
       * Gets a TeRaster vector parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was not found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDIRasterVectorType& data ) const;
                         
      /**
       * Store a TeRaster vector pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDIRasterVectorPtrType& data );

      /**
       * Gets a TeRaster vector pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was not found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDIRasterVectorPtrType& data ) const;                         
                         
      /**
       * Store a TePDIPalette parameter reference.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         TePDIRgbPalette::const_pointer& data );

      /**
       * Gets a TePDIPalette parameter reference.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was not found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDIRgbPalette::pointer& data ) const;

      /**
       * Store a Integer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         int data );

      /**
       * Gets a Integer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         int& data ) const;

      /**
       * Store a Integer vector parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         std::vector< int >& data );

      /**
       * Gets a Integer vector parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         std::vector< int >& data ) const;
                         
      /**
       * Store a double vector parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         std::vector< double >& data );

      /**
       * Gets a double vector parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         std::vector< double >& data ) const;

      /**
       * Store a String parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const std::string& data );

      /**
       * Gets a String parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         std::string& data ) const;

      /**
       * Store a Double parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         double data );

      /**
       * Gets a Double parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         double& data ) const;

      /**
       * Store a PDIBox parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TeBox& data );

      /**
       * Gets a PDIBox parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TeBox& data ) const;

      /**
       * Store a FilterMask parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDIFilterMask::pointer& data );

      /**
       * Gets a FilterMask parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDIFilterMask::pointer& data ) const;
                         
      /**
       * Store a TePolygon parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePolygon& data );

      /**
       * Gets a TePolygon parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePolygon& data ) const;
                         
      /**
       * Store a PDI Look-up table parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDILutType& data );

      /**
       * Gets a PDI Look-up table parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDILutType& data ) const;
                         
      /**
       * Store a PDI Look-up table pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDILutPtrType& data );

      /**
       * Gets a PDI Look-up table pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDILutPtrType& data ) const;                         
                         
      /**
       * Store a TeLayer reference parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDILayerPtrType& data );

      /**
       * Gets a TeLayer reference.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDILayerPtrType& data ) const;                          
                         
      /**
       * Store a TePolygonSet parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePolygonSet& data );

      /**
       * Gets a TePolygonSet parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePolygonSet& data ) const;                           

      /**
       * Store a TePolygonSet pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDIPolygonSetPtrType& data );

      /**
       * Gets a TePolygonSet pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDIPolygonSetPtrType& data ) const;
                         
      /**
       * Store a TePDIHistogram pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDIHistogram::pointer& data );

      /**
       * Gets a TePDIHistogram pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDIHistogram::pointer& data ) const;
                         
      /**
       * Store a TeCoordPairVect pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TeCoordPairVectPtrType& data );

      /**
       * Gets a TeCoordPairVect pointer parameter.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TeCoordPairVectPtrType& data ) const;
                         
      /**
       * Store a vector of histogram pointers.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const std::vector< TePDIHistogram::pointer >& data );

      /**
       * Retrive a vector of histogram pointers.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         std::vector< TePDIHistogram::pointer >& data ) const;                         
                         
      /**
       * Store a pointer of TePolygonSet map.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       */
      void SetParameter( const std::string& pname,
                         const TePDITypes::TePDIPolSetMapPtrType& data );

      /**
       * Retrive a pointer of TePolygonSet map.
       *
       * @param pname Parameter name.
       * @param data Parameter data.
       * @return true if OK, false if the parameter was no found or error.
       */
      bool GetParameter( const std::string& pname,
                         TePDITypes::TePDIPolSetMapPtrType& data ) const;                         

    protected :
    
      // Internal Type definitions of suported parameters.

      typedef std::map< std::string, std::string >
        TePDIStringMap;
      typedef std::map< std::string, int >
        TePDIIntegerMap;
      typedef std::map< std::string, std::vector< int > >
        TePDIIntegerVectorMap;
      typedef std::map< std::string, double >
        TePDIDoubleMap;
      typedef std::map< std::string, std::vector< double > >
        TePDIDoubleVectorMap;        
      typedef std::map< std::string, TePDITypes::TePDIRasterPtrType >
        TePDIRasterMap;
      typedef std::map< std::string, TePDITypes::TePDIRasterVectorType >
        TePDIRasterVecMap;
      typedef std::map< std::string, TePDITypes::TePDIRasterVectorPtrType >
        TePDIRasterVecPtrMap;        
      typedef std::map< std::string, TeBox > TeBoxMap;
      typedef std::map< std::string, TePDIRgbPalette::pointer >
        TePDIRgbPaletteMap;
      typedef std::map< std::string, TePDIFilterMask::pointer >
        TePDIFilterMaskMap;
      typedef std::map< std::string, TePolygon > TePolygonMap;
      typedef std::map< std::string, TePDITypes::TePDILutType > TePDILutMap;
      typedef std::map< std::string, TePDITypes::TePDILutPtrType > 
        TePDILutPtrMap;
      typedef std::map< std::string, TePDITypes::TePDILayerPtrType > 
        TePDILayerMap;
      typedef std::map< std::string, TePolygonSet > TePolygonSetMap;        
      typedef std::map< std::string, TePDITypes::TePDIPolygonSetPtrType > 
        TePolygonSetPtrMap;        
      typedef std::map< std::string, TePDIHistogram::pointer > 
        TePDIHistogramPtrMap;
      typedef std::map< std::string, TePDITypes::TeCoordPairVectPtrType > 
        TeCoordPairVectPtrMap;
      typedef std::map< std::string, std::vector< TePDIHistogram::pointer > > 
        HistVectMapT;
      typedef std::map< std::string, TePDITypes::TePDIPolSetMapPtrType > 
        TePDIPolSetMapPtrMapT;        
        
      /**
       * The last update time.
       */
      mutable time_t last_up_time_;

      /**
       * Stored TeRaster references Map.
       */
      mutable TePDIRasterMap RasterMap_;

      /**
       * Stored TeRaster vector Map.
       */
      mutable TePDIRasterVecMap RasterVecMap_;
      
      /**
       * Stored TeRasters vector pointers Map.
       */
      mutable TePDIRasterVecPtrMap RasterVecPtrMap_;      

      /**
       * Stored integer references Map.
       */
      mutable TePDIIntegerMap IntegerMap_;

      /**
       * Stored integer vector references Map.
       */
      mutable TePDIIntegerVectorMap IntegerVectorMap_;

      /**
       * Stored double references Map.
       */
      mutable TePDIDoubleMap DoubleMap_;

      /**
       * Stored double vector references Map.
       */
      mutable TePDIDoubleVectorMap DoubleVectorMap_;
      
      /**
       * Stored string references Map.
       */
      mutable TePDIStringMap StringMap_;

      /**
       * Stored TeBox Map.
       */
      mutable TeBoxMap PDIBoxMap_;

      /**
       * Stored PDIPalette references Map.
       */
      mutable TePDIRgbPaletteMap PDIRgbPaletteMap_;

      /**
       * Stored PDIFilterMask references Map.
       */
      mutable TePDIFilterMaskMap PDIFilterMaskMap_;
      
      /**
       * Stored TePolygon Map.
       */
      mutable TePolygonMap TePolygonMap_;      
      
      /**
       * Stored PDI Luts Map.
       */
      mutable TePDILutMap TePDILutMap_;
      
      /**
       * Stored PDI Luts pointers Map.
       */
      mutable TePDILutPtrMap TePDILutPtrMap_;       

      /**
       * Stored PDI Luts Map.
       */
      mutable TePDILayerMap TePDILayerMap_;       
      
      /**
       * Stored TePolygonSet Map.
       */
      mutable TePolygonSetMap TePolygonSetMap_;    
      
      /**
       * Stored TePolygonSet pointer Map.
       */
      mutable TePolygonSetPtrMap TePolygonSetPtrMap_;
      
      /**
       * Stored TePDIHistogram pointer Map.
       */
      mutable TePDIHistogramPtrMap TePDIHistogramPtrMap_;     
      
      /**
       * Stored TeCoordPairVectPtr pointer Map.
       */
      mutable TeCoordPairVectPtrMap TeCoordPairVectPtrMap_;        
      
      /**
       * Stored histograms pointers vector Map.
       */
      mutable HistVectMapT hist_vect_map_;
      
      /**
       * Stored TePolygonSet map pointers map.
       */      
      mutable TePDIPolSetMapPtrMapT TePDIPolSetMapPtrMapT_;
      
      /**
       * Updates the internal last update time.
       *
       * NOTE: Needed by comparison between two param instances.
       */
      void update_time();
};

/** @example TePDIParameters_test.cpp
 *    Shows how to use this class.
 */  

#endif //TEPDIPARAMETERS_HPP
