/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDIREGGROWSEGMENTATIONSTRATEGY_HPP
  #define TEPDIREGGROWSEGMENTATIONSTRATEGY_HPP
  
  #define MAXCLOSECELLS   5
  #define SQR_NC          65025
  #define TAMJAN          128
  #define STATDEAD        (char)2  

  #include <TeSharedPtr.h>
  #include "TePDIStrategy.hpp"
  #include "TePDIParameters.hpp"
  
  #include <TeRaster.h>
  #include <map>
  
  /**
  * Forward declarations
  */
  class TePDIRGSCell;
  class CloserCells;
  typedef map<long, TePDIRGSCell *> TePDIRGSCellList;  
  
  /**
    * @brief This is the class models an image segmentation process. 
    * @author Nicolas Despres <nicolasdespres@wanadoo.fr>
    * @ingroup PDIStrategies
    *
    * The segmentation
    * is performed through a region growing approach described in:  Leonardo
    * et allii, "Satellite image segmentation: a region growing approach",
    * Simposio Brasileiro de Sensoriamento Remoto, Salvador, 1996
    *
    * @note The general required parameters :
    *
    * @param input_image (TePDITypes::TePDIRasterPtrType),
    * @param euc_treshold (int) - euclidian distance treshold value.
    * @param area_min (int) - cell pixel size min value.
    *
    * @note The following parameters are optional and will be used if present.
    *
    * @param restriction_image (TePDITypes::TePDIRasterPtrType)
    * optional parameter
    * @param output_polsets ( TePDITypes::TePDIPolSetMapPtrType ) - 
    * The user supplied output polygon sets map where the generated polygons 
    * will be stored (each polygon set contains polygons related to the
    * same pixel value).   
    * @param output_image (TePDITypes::TePDIRasterPtrType),
    *       pointer to a Labelled image containing the final
    *       segmentation. Each pixel contains the cell id
    *       number to which it belongs.    
    */
  class TePDIRegGrowSegmentationStrategy : public TePDIStrategy
  {
    friend class TePDIRegGrowSegmentationSF;
  
    public :
    
      typedef TeSharedPtr< TePDIRegGrowSegmentationStrategy > pointer;
      typedef const TeSharedPtr< TePDIRegGrowSegmentationStrategy > 
        const_pointer;
      
      /**
       * Default Destructor
       */
      ~TePDIRegGrowSegmentationStrategy();
      
      /**
       * Checks if the supplied parameters fits the requirements of each
       * PDI strategy.
       *
       * Error log messages must be generated. No exceptions generated.
       *
       * @param parameters The parameters to be checked.
       * @return true if the parameters are OK. false if not.
       */
      bool CheckParameters( 
        const TePDIParameters& parameters ) const;       
      
    protected :
    
      TePDITypes::TePDIRasterPtrType  Imagein_;               //Array of pointer to images (input images)
      TePDITypes::TePDIRasterPtrType  Imagelab_;              //Pointer to the output image
      TePDITypes::TePDIRasterPtrType  Imageexc_;              //Pointer to the image containing restrictions
      TePDIRGSCellList        *ListCell_;             //Pointer to the list of cells.
      TePDIRGSCell            **WindowCell_;          //Array of cells for a subimage.
      long            OnLimbo_;       //Number of cells which have been removed from ListCell
                                                          //due to distance criteria
      unsigned long   *tuple_;                        //tuple of pixel (auxiliar);
      long            NWindow_;               //Number of subimages in the image.
      long            WindowLines_;           //Number of subimages along the lines of image
      long            WindowColumns_;         //Number of subimages along the columns of image
      long            CurrentWindow_;         //Current subimage.
      long            WindowOffset_;          //Id number of the first cell in the subimage.
      int             Areamin_;               //Minimum area allowed to the final cells.
      double          Difsim_;                        //Minimum euclidian distance between final cells.
      int             Nlin_;                  //Number of lines in Imagelab.
      int             Ncol_;                  //Number of colums in Imagelab.
      int             Nban_;                  //Numer of bands in Imagein
      int             ilin_, flin_;           //auxiliar variables
      int             icol_, fcol_;           //auxiliar variables
      int             sizelin_, sizecol_;     //auxiliar variables       
      
      /**
       * Default constructor
       *
       * @param factoryName Factory name.
       */
      TePDIRegGrowSegmentationStrategy();
      
      /**
       * Runs the algorithm implementation.
       *
       * @return true if OK. false on errors.
       */
      bool Implementation( const TePDIParameters& params );
      
      /**
      * Merges two cells
      *
      * @param c1 pointer to a cell to be merged
      * @param c2 pointer to a cell to be merged
      * @return pointer to cell resulting from merging..
      */
      TePDIRGSCell* MergeCells( TePDIRGSCell *c1, TePDIRGSCell *c2 );
  
      /**
      * Initializes a window over a image
      *
      * @return true if OK. false on error.
      */
          bool InitWindow();
  
      /**
      * Remove dead cells from list of cells.
      */
      void Adjust();
  
      /**
      * Eliminates all cells with area smaller than argument area.
      * The cells with area smaller than argument area are merged with its
      * nearest neighbour cell.
      *
      * @param area size in pixels which defines de area treshold for merging.
      * @param linmax lin max of the image to consider
      * @return true if OK. false on error.
      */
      bool MergeSmallCells( long area, int linmax);
  
      /**
      * Merges all pairs of neighbouring cells with euclidian distance
      * bellow a threshold.
      * The pair of cells say, c1 and c2 must satisfy the criteria:
      *              1 - c1 = closestneighbour( c2 ) or c2 = closestneighbour( c1 ).
      *              2 - dist( c1, c2) < threshold
      *
      * @return true if OK. false on error.
      */
      bool MergeSimilarCells();
  
      /**
      * Merges all pairs of neighbouring cells with euclidian distance
      * bellow a threshold.
      * The pair of cells say, c1 and c2 must satisfy the criteria:
      *              1 - c1 = closestneighbour( c2 ) or c2 = closestneighbour( c1 ).
      *              2 - dist( c1, c2) < threshold
      *
      * @return true if OK. false on error.
      */
      bool MergeMutuallyClosestCells();
  
      /**
      * Updates ImageLab by writing the cell id number for
      * every pixel belonging to the cell.
      *
      * @return true if OK. false on error.
      */
      bool Resort();      
  };
  
  
  /**
  * This is the basic element in the region growing process. The cell represents an area
  * in the image being segmented. It contains spectral, geometrical and contextual
  * information of this area. The behaviour of this class allows it to merge
  *
  * The behaviour of this class allows it to merge with other cells and to update its
  * information. At the initial stage in the region growing process each cell is
  * composed by one pixel of the image. It uses 4-Neighbourhood (contextual information).
  */
class TePDIRGSCell
{
public:
   /**
    * Default Constructor.
    */
        TePDIRGSCell ();

   /**
    * Alternative Constructor.
    *
    * @param tuple: pixel tuple containing spectral information.
    * @param id: Id number of the cell.
    * @param lin: line position of tuple in the image.
    * @param col: column position of tuple in the image.
    * @param nban: Number of bands.
    */
   TePDIRGSCell( unsigned long *tuple, long id, short lin, short col, short nban );

   /**
    * Default destructor.
    */
   ~TePDIRGSCell();

   /**
    * Initializes an existing TePDIRGSCell
    *
    * @param tuple: pixel tuple containing spectral information.
    * @param id: Id number of the cell.
    * @param lin: line position of tuple in the image.
    * @param col: column position of tuple in the image.
    * @param nban: Number of bands.
    */
   void Init( unsigned long *tuple, long id, short lin, short col, short nban );

   /**
    * Changes the Id  number of the cell
        *
    * @param idc: new id number
    *
    */
   void ResetId( long idc ) { Idnumber_ = idc; }

   /**
    * Returns The status of the cell
    *
    * @return If cell is dead return 1, otherwise, 0.
    */
   char Dead()  { return (Stat_ & STATDEAD); }

   /**
    * Turns cell dead
    *
        * When the cell is dead it does not represent
    * any region in the image.
    */
   void Kill() { Stat_ = STATDEAD; }

   /**
    * Return the bounding box of the cell
    *
    * @param lmin: Minimum image row wich contains a pixel belonging to the cell
    * @param cmin: Minimum image column wich contains a pixel belonging to the cell
    * @param lmax: Maximum image row wich contains a pixel belonging to the cell
    * @param cmax: Maximum image column wich contains a pixel belonging to the cell
    */
   void BoundingRectangle( int& lmin, int& cmin, int& lmax, int& cmax )
        { lmin = LinMin_; cmin = ColMin_; lmax = LinMax_; cmax = ColMax_; }

   /**
    * LinMax
    *
        * @return Maximum image line wich contains a pixel belonging to the cell
    */
   int LineMax () { return LinMax_;}

   /**
    * Merge cell c with the "this" cell.
    *
    * @return true if OK. false on error.
    */
   bool Merge( TePDIRGSCell *c );

   /**
    * Remove dead cells from the list of neighbour cells and closest cells
    */
   void AdjustNeighborhood();//{ Cc_->Adjust(); /*Neighbors_->Adjust();*/ }

   /**
    * Add neighbour to neighbour cells list and closest cells list.
    *
    * @param c being added.
    * @param dist euclidian distance to "this" cell
    * @return true if OK. false on error.
    */
   bool AddNeighbor( TePDIRGSCell* c, float dist );

   /**
    * Insert neighbour to neighbour cells list and closest cells list.
    *
    * @param c being added.
    * @param dist: euclidian distance to "this" cell
    * @return true if OK. false on error.
    */
   bool InsertNeighbor( TePDIRGSCell* c, float dist );

   /**
    * Return the cell with the minimum euclidian distance.
    *
    * @param dist: minimum distance.
    * @return Pointer to cell which has the minimum euclidian distance.
    */
   TePDIRGSCell* ClosestNeighbor( float& dist );

   /**
    * Return the euclidian distance between cell c and "this" cell.
    *
        * @param c: input cell
    * @return euclidian distance between cell c and "this" cell.
    */
   float Distance( TePDIRGSCell *c );

   /**
    * Returns the Id number of the cell.
    *
    * @return Id number of the cell.
    */
   long Id() { return Idnumber_; }

   /**
    * Return the number of pixels of the cell
    *
    * @return Number of pixels of the cell
    */
   long Area() { return Npix_; }

   /**
    * Return the number of espectral bands used
    *
    * @return Number of espectral bands used
    */
   int GetNban() { return Nban_; }

   /**
    * Prints cell state
    */
   void Print();

private:
        char            Stat_;                  //Status STATDEAD = dead, otherwise, alive.
        long            Idnumber_;              //Id number of cell
        long            Npix_;                  //Number of pixels it contains
        int             Nban_;                  //Number of spectral bands used
        float           delta_;
        int             LinMin_;                //Bounding box of cell
        int             LinMax_;                //Bounding box of cell
        int             ColMin_;                //Bounding box of cell
        int             ColMax_;                //Bounding box of cell
        float           *Media_;                //Array of spectral mean.
        float           *PreviousMedia_;        //Array containing the previous spectral mean
        TePDIRGSCellList        *Neighbors_;            //List of neighbour cells
        CloserCells     *Cc_;                   //List of closest cells

};
  
  
#endif



