#include "TePDIStatistic.hpp"

#include "TePDIAgnostic.hpp"
#include "TePDIUtils.hpp"

#include <TeUtils.h>
#include <TeBox.h>

#include <math.h>
#include <float.h>


TePDIStatistic::TePDIStatistic()
{
}


TePDIStatistic::~TePDIStatistic()
{
}


void TePDIStatistic::ResetState( const TePDIParameters& params )
{
  if( params_ != params ) {
    sum3_cache_.clear();
    sum4_cache_.clear();
    sum_pb1b2_cache_.clear();
    
    /* Building histogram cache if histograms are provided */
    
    histo_cache_.clear();
  
    if( params.CheckParameter( "histograms", 
      TePDIParameters::TePDIHistVecParam ) ) {
      
      std::vector< TePDIHistogram::pointer > histograms;
      PDIAGN_TRUE_OR_THROW( params.GetParameter( "histograms", 
        histograms ), "Missing parameter: histograms" );
        
      TePDITypes::TePDIRasterVectorType rasters;
      PDIAGN_TRUE_OR_THROW( 
        params.GetParameter( "rasters", rasters ),
        "Missing parameter: rasters" );
    
      std::vector< int > bands;
      PDIAGN_TRUE_OR_THROW( 
        params.GetParameter( "bands", bands ),
        "Missing parameter: bands" ); 
        
      for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
        std::pair< TeRaster*, int  > key;
        key.first = rasters[ index ].nakedPointer();
        key.second = bands[ index ];
      
        histo_cache_[ key ] = histograms[ index ];
      }
    }     
  }
}


unsigned int  TePDIStatistic::getRasterParamIndex( 
  const TePDITypes::TePDIRasterPtrType& raster )
{
  TePDITypes::TePDIRasterVectorType rasters;
  
  PDIAGN_TRUE_OR_THROW( params_.GetParameter( "rasters", rasters ),
    "Unable to find rasters param" );
  
  for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
    if( rasters[ index ] == raster ) {
      return index;
    }
  }
  
  PDIAGN_LOG_AND_THROW( "Unable to find the required raster pointer" );
  
  return 0;
}


std::vector< int > TePDIStatistic::getUsedBands( 
  TePDITypes::TePDIRasterPtrType& raster )
{
  std::vector< int > used_bands;
  TePDITypes::TePDIRasterVectorType rasters;
  std::vector< int > bands;
  
  
  if ( params_.GetParameter( "rasters", rasters ) &&
       params_.GetParameter( "bands", bands ) ) {
       
    for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
      if( rasters[ index ] == raster ) {
        used_bands.push_back( bands[ index ] );
      }
    }
  }
  
  return used_bands;
}


bool TePDIStatistic::RunImplementation()
{
  PDIAGN_LOG_AND_THROW( "This function cannot be used for this class" );
  return false;
}


bool TePDIStatistic::CheckParameters( 
  const TePDIParameters& parameters ) const
{
  /* Checking input rasters */
  
  TePDITypes::TePDIRasterVectorType rasters;
  PDIAGN_TRUE_OR_RETURN( 
    parameters.GetParameter( "rasters", rasters ),
    "Missing parameter: rasters" );
    
  std::vector< int > bands;
  PDIAGN_TRUE_OR_RETURN( 
    parameters.GetParameter( "bands", bands ),
    "Missing parameter: bands" );
    
  PDIAGN_TRUE_OR_RETURN( rasters.size() == bands.size(),
    "The number of rasters doesn't match the number of bands" );
    
  PDIAGN_TRUE_OR_RETURN( rasters.size() > 0, "Invalid number of rasters" );
    
  for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
    PDIAGN_TRUE_OR_RETURN( rasters[ index ].isActive(),
      "Invalid parameter: raster " + 
      Te2String( index ) + " inactive" );
      
    PDIAGN_TRUE_OR_RETURN( 
      rasters[ index ]->params().status_ != TeRasterParams::TeNotReady,
      "Invalid parameter: raster " + 
      Te2String( index ) + " not ready" );
        
    PDIAGN_TRUE_OR_RETURN( 
      rasters[ 0 ]->params().nlines_ == rasters[ index ]->params().nlines_,
      "Lines number mismatch between raster 0 and raster " +
      Te2String( index ) );
      
    PDIAGN_TRUE_OR_RETURN( 
      rasters[ 0 ]->params().ncols_ == rasters[ index ]->params().ncols_,
      "Columns number mismatch between raster 0 and raster " +
      Te2String( index ) );
      
    PDIAGN_TRUE_OR_RETURN( 
      rasters[ 0 ]->params().resx_ == rasters[ index ]->params().resx_,
      "Pixel X resolution mismatch between raster 0 and raster " +
      Te2String( index ) );      

    PDIAGN_TRUE_OR_RETURN( 
      rasters[ 0 ]->params().resy_ == rasters[ index ]->params().resy_,
      "Pixel Y resolution mismatch between raster 0 and raster " +
      Te2String( index ) );      
            
    PDIAGN_TRUE_OR_RETURN( 
      bands[ index ] >= 0, "Invalid band number (" + 
      Te2String( index ) + ")" );
    PDIAGN_TRUE_OR_RETURN( 
      bands[ index ] < rasters[ index ]->nBands(), "Invalid band number (" + 
      Te2String( index ) + ")" );
      
  }
  
  /* Checking the restriction polygon if avaliable */
  
  TePolygon polygon;
  if( parameters.CheckParameter( "polygon", 
    TePDIParameters::TePolygonParam ) ) {
    
    TePolygon polygon;
    
    PDIAGN_TRUE_OR_RETURN( parameters.GetParameter( "polygon", 
      polygon ), "Missing parameter : polygon" );
      
    TeRaster::iteratorPoly it = 
      rasters[ 0 ]->begin( polygon, TeBoxPixelIn, 0 );
      
    PDIAGN_TRUE_OR_RETURN( ( ! it.end() ), "Invalid parameter : polygon" );
    PDIAGN_TRUE_OR_RETURN( ( it.nLinesInPoly() > 0 ), 
      "Invalid parameter : polygon" );
    PDIAGN_TRUE_OR_RETURN( ( it.nColsInPoly() > 0 ), 
      "Invalid parameter : polygon" );
  }
  
  /* Checking the restriction polygon set if avaliable */
  
  if( parameters.CheckParameter( "polygonset", 
    TePDIParameters::TePolygonSetParam ) ) {
    
    TePolygonSet temp_polygonset;
    parameters.GetParameter( "polygonset", temp_polygonset );
    
    TePolygonSet::iterator it = temp_polygonset.begin();
    TePolygonSet::iterator it_next;
    TePolygonSet::iterator it_end = temp_polygonset.end();
    
    while( it != it_end ) {
      it_next = it;
      ++it_next;
      
      /* Checking the polygons relation */
      
      while( it_next != it_end ) {
        short relation = TeRelation( *it, *it_next );
        
        PDIAGN_TRUE_OR_RETURN( 
          ( ( relation == TeDISJOINT ) || ( relation == TeTOUCHES ) ),
          "All polygons must have a TeDISJOINT or TeTOUCHES relation" );
      
        ++it_next;
      }
      
      /* Checking the polygon iterator */
      
      TeRaster::iteratorPoly pol_it = 
        rasters[ 0 ]->begin( *it, TeBoxPixelIn, 0 );
        
      PDIAGN_TRUE_OR_RETURN( ( ! pol_it.end() ), 
        "Invalid parameter : polygonset" );
      PDIAGN_TRUE_OR_RETURN( ( pol_it.nLinesInPoly() > 0 ), 
        "Invalid parameter : polygonset" );
      PDIAGN_TRUE_OR_RETURN( ( pol_it.nColsInPoly() > 0 ), 
        "Invalid parameter : polygonset" );      
      
      ++it;
    }
  }
  
  /* Checking user histograms if available */
  
  if( parameters.CheckParameter( "histograms", 
    TePDIParameters::TePDIHistVecParam ) ) {
    
    std::vector< TePDIHistogram::pointer > histograms;
    
    PDIAGN_TRUE_OR_THROW( parameters.GetParameter( "histograms", 
      histograms ), "Invalid parameter: histograms" );
    PDIAGN_TRUE_OR_RETURN( ( histograms.size() == rasters.size() ),
      "Invalid parameter: histograms" );
      
    std::vector< TePDIHistogram::pointer >::iterator it =
      histograms.begin();
    std::vector< TePDIHistogram::pointer >::iterator it_end =
      histograms.end();
      
    while( it != it_end ) {
      PDIAGN_TRUE_OR_RETURN( ( (*it)->size() > 0 ),
        "Invalid parameter: histograms" );    
        
      ++it;
    }
  }  
  
  return true;
}


TePDIHistogram::pointer TePDIStatistic::getHistogram( 
  unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Building histogram, if necessary */
  
  std::pair< TeRaster*, int  > temp_pair;
  temp_pair.first = rasters[ raster_index ].nakedPointer();
  temp_pair.second = bands[ raster_index ];
  
  std::map< std::pair< TeRaster*, int  >, TePDIHistogram::pointer >::iterator
    it_cache = histo_cache_.find( temp_pair );
    
  if( it_cache == histo_cache_.end() ) {
    DoSingleRasterCCalcules( rasters[ raster_index ], bands[ raster_index ] );
    it_cache = histo_cache_.find( temp_pair );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it_cache != histo_cache_.end() ), 
    "Histogram build error" );    
  
  TePDIHistogram::pointer output_histogram( new TePDIHistogram );

  *output_histogram = *( it_cache->second );
  
  return output_histogram;
}


double TePDIStatistic::getSum( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Building histogram, if necessary */
  
  std::pair< TeRaster*, int  > temp_pair;
  temp_pair.first = rasters[ raster_index ].nakedPointer();
  temp_pair.second = bands[ raster_index ];
  
  std::map< std::pair< TeRaster*, int  >, TePDIHistogram::pointer >::iterator
    it_cache = histo_cache_.find( temp_pair );
    
  if( it_cache == histo_cache_.end() ) {
    DoSingleRasterCCalcules( rasters[ raster_index ], bands[ raster_index ] );
    it_cache = histo_cache_.find( temp_pair );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it_cache != histo_cache_.end() ), 
    "Histogram build error" );
    
  /* Calculating sum */
    
  TePDIHistogram::iterator it_hist = it_cache->second->begin();
  TePDIHistogram::iterator it_hist_end = it_cache->second->end();
  
  double sum = 0;
  
  while( it_hist != it_hist_end ) {
    sum += it_hist->first * ( (double) it_hist->second ); 
    
    ++it_hist;
  }
  
  return sum;
}


double TePDIStatistic::getSum3( const TePDITypes::TePDIRasterPtrType& raster,
  unsigned int band )
{
  PDIAGN_TRUE_OR_THROW( raster.isActive(),
    "Inactive raster pointer" );
  PDIAGN_TRUE_OR_THROW( raster->params().status_ != 
    TeRasterParams::TeNotReady,
    "Raster1 not ready to read" );
  PDIAGN_TRUE_OR_THROW( (int)band < raster->nBands(),
    "Invalid band" );
    
  std::pair< TeRaster*, int > mapkey;
  mapkey.first = raster.nakedPointer();
  mapkey.second = band;
  std::map< std::pair< TeRaster*, int >, double >::iterator it =
    sum3_cache_.find( mapkey ); 
    
  if( it == sum3_cache_.end() ) {
    DoSingleRasterCCalcules( raster, band );
    it = sum3_cache_.find( mapkey );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it != sum3_cache_.end() ),
    "Needed cache value not found" );
    
  return it->second;    
}


double TePDIStatistic::getSum4( const TePDITypes::TePDIRasterPtrType& raster,
  unsigned int band )
{
  PDIAGN_TRUE_OR_THROW( raster.isActive(),
    "Inactive raster pointer" );
  PDIAGN_TRUE_OR_THROW( raster->params().status_ != TeRasterParams::TeNotReady,
    "Raster1 not ready to read" );
  PDIAGN_TRUE_OR_THROW( (int)band < raster->nBands(),
    "Invalid band" );
    
  std::pair< TeRaster*, int > mapkey;
  mapkey.first = raster.nakedPointer();
  mapkey.second = band;
  std::map< std::pair< TeRaster*, int >, double >::iterator it =
    sum4_cache_.find( mapkey );    
  
  if( it == sum4_cache_.end() ) {
    DoSingleRasterCCalcules( raster, band );
    it = sum4_cache_.find( mapkey );
  }

  PDIAGN_DEBUG_CONDITION( ( it != sum4_cache_.end() ),
    "Needed cache value not found" );
    
  return it->second;   
}


void TePDIStatistic::DoSingleRasterCCalcules( 
  const TePDITypes::TePDIRasterPtrType& raster,
  unsigned int band )
{
  PDIAGN_TRUE_OR_THROW( raster.isActive(),
    "Inactive raster pointer" );
  PDIAGN_TRUE_OR_THROW( 
    raster->params().status_ != TeRasterParams::TeNotReady,
    "Raster1 not ready to read" );
  PDIAGN_TRUE_OR_THROW( (int)band < raster->nBands(),
    "Invalid band" );
    
  std::vector< TePolygon > polygons;
  getRestrictionPolygons( polygons );
  
  unsigned int polygons_index = 0;
  
  TeRaster::iteratorPoly itp;
  TeRaster::iteratorPoly itp_end;
  
  /* Finding the total covered lines - used by progress interface */
  
  unsigned int total_progress_steps = 0;
  
  for( polygons_index = 0 ; polygons_index < polygons.size() ;
       ++polygons_index ) {
       
    itp = raster->begin( polygons[ polygons_index ], TeBoxPixelIn, 0 );
    
    total_progress_steps += itp.nColsInPoly();       
  }
  
  /* Doing raster scan dependent calcules */
  
  double value = 0;
  double square_value = 0;
  double cubic_value = 0;
  
  double sum = 0;
  double square_sum = 0;
  double cubic_sum = 0;
  double hypercubic_sum = 0;
  
  int last_raster_line = 0;
  int current_raster_line = 0;
  
  bool dummy_used = raster->params().useDummy_;
  double dummy_value = 0;
  if( dummy_used ) {
    dummy_value = raster->params().dummy_[ band ];
  }
  
  std::map< double, unsigned int > primary_histogram;
  std::map< double, unsigned int >::iterator primary_histogram_it;
  
  StartProgInt( "Statistical Analysis : Raster " + 
    Te2String( getRasterParamIndex( raster ) ) + " Band " +
    Te2String( band ), total_progress_steps );   
  
  for( polygons_index = 0 ; polygons_index < polygons.size() ;
       ++polygons_index ) {
       
    itp = raster->begin( polygons[ polygons_index ], TeBoxPixelIn, 0 );
    itp_end = raster->end( polygons[ polygons_index ], TeBoxPixelIn, 0 );
    
    last_raster_line = 0;
    
    while( itp != itp_end ) {
      value = itp*( band );
      
      if( ( value != dummy_value ) || ( ! dummy_used ) ) {
        square_value = ( value * value );
        cubic_value = ( square_value * value );
        
        sum += value;
        square_sum += square_value;
        cubic_sum += cubic_value;
        hypercubic_sum += ( cubic_value * value );
        
        primary_histogram_it = primary_histogram.find( value );
        if( primary_histogram_it == primary_histogram.end() ) {
          primary_histogram[ value ] = 1;
        } else {
          ++( primary_histogram_it->second );
        }        
      }
      
      current_raster_line =  itp.currentLine();
      if( last_raster_line != current_raster_line ) {
        last_raster_line = current_raster_line;
        IncProgInt();
      }
    
      ++itp;
    }
       
  }
  
  std::pair< TeRaster*, int  > mapkey;
  mapkey.first = raster.nakedPointer();
  mapkey.second = band;  
  
  sum3_cache_[ mapkey ] = cubic_sum;
  sum4_cache_[ mapkey ] = hypercubic_sum;
  
  TePDIHistogram::pointer hist_ptr( new TePDIHistogram );
  ( *hist_ptr ) = primary_histogram;
  histo_cache_[ mapkey ] = hist_ptr;
}


void TePDIStatistic::DoCoupleRasterCCalcules( 
  TePDITypes::TePDIRasterPtrType& raster1, unsigned int band1, 
  TePDITypes::TePDIRasterPtrType& raster2, unsigned int band2 )
{
  PDIAGN_TRUE_OR_THROW( raster1.isActive(),
    "Inactive raster1 pointer" );
  PDIAGN_TRUE_OR_THROW( raster1->params().status_ != 
    TeRasterParams::TeNotReady,
    "raster1 not ready to read" );
  PDIAGN_TRUE_OR_THROW( raster2.isActive(),
    "Inactive raster2 pointer" );
  PDIAGN_TRUE_OR_THROW( raster2->params().status_ != 
    TeRasterParams::TeNotReady,
    "raster2 not ready to read" );

  PDIAGN_TRUE_OR_THROW( (int)band1 < raster1->nBands(),
    "Invalid band1" );
  PDIAGN_TRUE_OR_THROW( (int)band2 < raster2->nBands(),
    "Invalid band2" );
    
  std::vector< TePolygon > polygons;
  getRestrictionPolygons( polygons );
  
  unsigned int polygons_index = 0;
  
  TeRaster::iteratorPoly itp;
  TeRaster::iteratorPoly itp_end;
  
  /* Finding the total covered lines - used by progress interface */
  
  unsigned int total_progress_steps = 0;
  
  for( polygons_index = 0 ; polygons_index < polygons.size() ;
       ++polygons_index ) {
       
    itp = raster1->begin( polygons[ polygons_index ], TeBoxPixelIn, 0 );
    
    total_progress_steps += itp.nColsInPoly();       
  }
   
  /* Doing raster scan dependent calcules */  

  double value1 = 0;
  double value2 = 0;
  double sum_pb1b2 = 0;
  int it_line = 0;
  int it_col = 0;
  int last_raster_line = 0;
  bool got_value1 = false;
  bool got_value2 = false;
  double cubic_sum1 = 0;
  double hypercubic_sum1 = 0;  
  double cubic_sum2 = 0;
  double hypercubic_sum2 = 0;
  double temp_cubic = 0;
  
  bool dummy1_used = raster1->params().useDummy_;
  bool dummy2_used = raster2->params().useDummy_;
  
  std::map< double, unsigned int > primary_histogram1;
  std::map< double, unsigned int >::iterator primary_histogram1_it;

  std::map< double, unsigned int > primary_histogram2;
  std::map< double, unsigned int >::iterator primary_histogram2_it;
    
  StartProgInt( "Statistical Analysis : Raster " + 
    Te2String( getRasterParamIndex( raster1 ) ) +
    "[band " + Te2String( band1 ) + "] and Raster " +
    Te2String( getRasterParamIndex( raster2 ) ) +
    "[band " + Te2String( band1 ) + "]",
    total_progress_steps );   
    
  for( polygons_index = 0 ; polygons_index < polygons.size() ;
       ++polygons_index ) {    
       
    itp = raster1->begin( polygons[ polygons_index ], TeBoxPixelIn, 0 );
    itp_end = raster1->end( polygons[ polygons_index ], TeBoxPixelIn, 0 );       
    
    last_raster_line = 0;

    while( itp != itp_end ) {
      it_line = itp.currentLine();
      it_col = itp.currentColumn();
      
      got_value1 = raster1->getElement( it_col, it_line, value1, band1 );
      got_value2 = raster2->getElement( it_col, it_line, value2, band2 );
  
      if( got_value1 ) {
        primary_histogram1_it = primary_histogram1.find( value1 );
        if( primary_histogram1_it == primary_histogram1.end() ) {
          primary_histogram1[ value1 ] = 1;
        } else {
          ++( primary_histogram1_it->second );
        }
        
        temp_cubic = ( value1 * value1 * value1 );
        cubic_sum1 += temp_cubic;
        hypercubic_sum1 += ( temp_cubic * value1 );
        
        if( got_value2 ) {
          primary_histogram2_it = primary_histogram2.find( value2 );
          if( primary_histogram2_it == primary_histogram2.end() ) {
            primary_histogram2[ value2 ] = 1;
          } else {
            ++( primary_histogram2_it->second );
          }        
            
          sum_pb1b2 += ( value1 * value2 );
          temp_cubic = ( value2 * value2 * value2 );
          cubic_sum2 += temp_cubic;
          hypercubic_sum2 += ( temp_cubic * value2 );
        } else {
          PDIAGN_TRUE_OR_THROW( dummy2_used, "Unable to read from raster2" );
        }
      } else {
        PDIAGN_TRUE_OR_THROW( dummy1_used, "Unable to read from raster1" );      
      }

      if( last_raster_line != it_line ) {
        last_raster_line = it_line;
        IncProgInt();
      }
    
      ++itp;
    }
  }
  
  std::pair< TeRaster*, int > pair1;
  pair1.first = raster1.nakedPointer();
  pair1.second = band1;

  std::pair< TeRaster*, int > pair2;
  pair2.first = raster2.nakedPointer();
  pair2.second = band2;
  
  std::pair< std::pair< TeRaster*, int >,  std::pair< TeRaster*, int > >
    mapkey;
  mapkey.first = pair1;
  mapkey.second = pair2;    
  
  sum_pb1b2_cache_[ mapkey ] = sum_pb1b2;
  
  sum3_cache_[ pair1 ] = cubic_sum1;
  sum3_cache_[ pair2 ] = cubic_sum2;
  
  sum4_cache_[ pair1 ] = hypercubic_sum1;
  sum4_cache_[ pair2 ] = hypercubic_sum2;
  
  TePDIHistogram::pointer hist1_ptr( new TePDIHistogram );
  ( *hist1_ptr ) = primary_histogram1;
  histo_cache_[ pair1 ] = hist1_ptr;  
  TePDIHistogram::pointer hist2_ptr( new TePDIHistogram );
  ( *hist2_ptr ) = primary_histogram2;
  histo_cache_[ pair2 ] = hist2_ptr;  
}


double TePDIStatistic::getSumPB1B2(TePDITypes::TePDIRasterPtrType& raster1,
  unsigned int band1,  TePDITypes::TePDIRasterPtrType& raster2,
  unsigned int band2 )
{
  PDIAGN_TRUE_OR_THROW( raster1.isActive(),
    "Inactive raster1 pointer" );
  PDIAGN_TRUE_OR_THROW( raster1->params().status_ != TeRasterParams::TeNotReady,
    "raster1 not ready to read" );
  PDIAGN_TRUE_OR_THROW( raster2.isActive(),
    "Inactive raster2 pointer" );
  PDIAGN_TRUE_OR_THROW( raster2->params().status_ != TeRasterParams::TeNotReady,
    "raster2 not ready to read" );

  PDIAGN_TRUE_OR_THROW( (int)band1 < raster1->nBands(),
    "Invalid band1" );
  PDIAGN_TRUE_OR_THROW( (int)band2 < raster2->nBands(),
    "Invalid band2" );
    
  /* Trying to use cache */

  std::pair< TeRaster*, int > pair1;
  pair1.first = raster1.nakedPointer();
  pair1.second = band1;

  std::pair< TeRaster*, int > pair2;
  pair2.first = raster2.nakedPointer();
  pair2.second = band2;

  std::pair< std::pair< TeRaster*, int >,  std::pair< TeRaster*, int > >
    mapkey;
  mapkey.first = pair1;
  mapkey.second = pair2;
  
  std::map< std::pair< std::pair< TeRaster*, int >,
    std::pair< TeRaster*, int > >, double >::iterator it =
    sum_pb1b2_cache_.find( mapkey );
    
  if( it == sum_pb1b2_cache_.end() ) {
    DoCoupleRasterCCalcules( raster1, band1, raster2, band2 );
    it = sum_pb1b2_cache_.find( mapkey );
  }
  
  PDIAGN_DEBUG_CONDITION( 
    ( it !=  sum_pb1b2_cache_.end() ), "Needed cache value not found" );
    
  return sum_pb1b2_cache_[ mapkey ];
}


double TePDIStatistic::getMean( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Building histogram, if necessary */
  
  std::pair< TeRaster*, int  > temp_pair;
  temp_pair.first = rasters[ raster_index ].nakedPointer();
  temp_pair.second = bands[ raster_index ];
  
  std::map< std::pair< TeRaster*, int  >, TePDIHistogram::pointer >::iterator
    it_cache = histo_cache_.find( temp_pair );
    
  if( it_cache == histo_cache_.end() ) {
    DoSingleRasterCCalcules( rasters[ raster_index ], bands[ raster_index ] );
    it_cache = histo_cache_.find( temp_pair );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it_cache != histo_cache_.end() ), 
    "Histogram build error" );
    
  /* Calculating mean */
    
  TePDIHistogram::iterator it_hist = it_cache->second->begin();
  TePDIHistogram::iterator it_hist_end = it_cache->second->end();
  
  double sum = 0;
  double elements = 0;
  
  while( it_hist != it_hist_end ) {
    sum += it_hist->first * ((double)it_hist->second); 
    elements += ((double)it_hist->second);
    
    ++it_hist;
  }   
  
  return ( sum / elements ); 
}


TeMatrix TePDIStatistic::getMeanMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );  

  TeMatrix outmatrix;
  outmatrix.Init( 1, rasters.size(), 0. );

  for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
    outmatrix( 0, index ) = getMean( index );
  }

  return outmatrix;
}


double TePDIStatistic::getCovariance( unsigned int raster1_index,  
  unsigned int raster2_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster1_index < rasters.size(), "Invalid index1" );
  PDIAGN_TRUE_OR_THROW( raster2_index < rasters.size(), "Invalid index2" );
  
  /* Calculating the elements number */
  
  double elements_number = 0;
  
  TePDIHistogram::pointer histogram1 = getHistogram( raster1_index );
  
  TePDIHistogram::iterator it = histogram1->begin();
  TePDIHistogram::iterator it_end = histogram1->end();
  
  while( it != it_end ) {
    elements_number += ( (double) it->second );
    ++it;
  }
  
  /* Calculating the sum of all elements */

  double sum1 = getSum( raster1_index );
  double sum2 = getSum( raster2_index );
  
  /* Calculating the sum of the product of band1 and band2 */
  
  double sum_pb1b2 = getSumPB1B2( rasters[ raster1_index ], 
    bands[ raster1_index ], rasters[ raster2_index ], bands[ raster2_index ] );
    
  /* calculating the covariance */

  double covariance = ( sum_pb1b2 / elements_number ) -
    ( ( sum1 / elements_number ) * ( sum2 / elements_number ) );

  return covariance;                        
}


double TePDIStatistic::getVariance( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Building histogram, if necessary */
  
  std::pair< TeRaster*, int  > temp_pair;
  temp_pair.first = rasters[ raster_index ].nakedPointer();
  temp_pair.second = bands[ raster_index ];
  
  std::map< std::pair< TeRaster*, int  >, TePDIHistogram::pointer >::iterator
    it_cache = histo_cache_.find( temp_pair );
    
  if( it_cache == histo_cache_.end() ) {
    DoSingleRasterCCalcules( rasters[ raster_index ], bands[ raster_index ] );
    it_cache = histo_cache_.find( temp_pair );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it_cache != histo_cache_.end() ), 
    "Histogram build error" );
    
  /* Calculating the number of elements */

  TePDIHistogram::iterator it_hist = it_cache->second->begin();
  TePDIHistogram::iterator it_hist_end = it_cache->second->end();

  double elements_number = 0;    
  
  while( it_hist != it_hist_end ) {
    elements_number = elements_number + (double)it_hist->second; 
    
    ++it_hist;
  }  
    
  /* Calculating variance */
    
  it_hist = it_cache->second->begin();
  
  double mean = getMean( raster_index );
  double variance = 0;
  
  while( it_hist != it_hist_end ) {
    variance += ( ( (double)it_hist->second ) / elements_number ) *
      ( it_hist->first - mean ) * ( it_hist->first - mean );
    
    ++it_hist;
  }
  
  return variance;
}


double TePDIStatistic::getStdDev( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );

  return sqrt( getVariance( raster_index ) );
}


double TePDIStatistic::getEntropy( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Calculating the elements number */
  
  double elements_number = 0;
  
  TePDIHistogram::pointer histogram1 = getHistogram( raster_index );
  
  TePDIHistogram::iterator it = histogram1->begin();
  TePDIHistogram::iterator it_end = histogram1->end();
  
  while( it != it_end ) {
    elements_number += ( (double) it->second );
    ++it;
  }  
  
  /* Calculating entropy */
    
  it = histogram1->begin();
  
  double entropy = 0;
  double probability = 0;
  
  while( it != it_end ) {
    probability = ( (double)(it->second) ) / elements_number;
    
    if( probability > 0.0 ) {
      entropy += ( probability * ( log( probability ) /
        log( (double)2 ) ) );
    }

    ++it;
  }
  
  entropy = (-1.0) * entropy;

  return entropy;
}


double TePDIStatistic::getMin( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Building histogram, if necessary */
  
  std::pair< TeRaster*, int  > temp_pair;
  temp_pair.first = rasters[ raster_index ].nakedPointer();
  temp_pair.second = bands[ raster_index ];
  
  std::map< std::pair< TeRaster*, int  >, TePDIHistogram::pointer >::iterator
    it_cache = histo_cache_.find( temp_pair );
    
  if( it_cache == histo_cache_.end() ) {
    DoSingleRasterCCalcules( rasters[ raster_index ], bands[ raster_index ] );
    it_cache = histo_cache_.find( temp_pair );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it_cache != histo_cache_.end() ), 
    "Histogram build error" );
    
  /* Calculating min */
    
  TePDIHistogram::iterator it_hist = it_cache->second->begin();
  TePDIHistogram::iterator it_hist_end = it_cache->second->end();
  
  double min = DBL_MAX;
  
  while( it_hist != it_hist_end ) {
    if( it_hist->first < min ) {
      min = it_hist->first;
    }
    
    ++it_hist;
  }
  
  return min;
}


double TePDIStatistic::getMax( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Building histogram, if necessary */
  
  std::pair< TeRaster*, int  > temp_pair;
  temp_pair.first = rasters[ raster_index ].nakedPointer();
  temp_pair.second = bands[ raster_index ];
  
  std::map< std::pair< TeRaster*, int  >, TePDIHistogram::pointer >::iterator
    it_cache = histo_cache_.find( temp_pair );
    
  if( it_cache == histo_cache_.end() ) {
    DoSingleRasterCCalcules( rasters[ raster_index ], bands[ raster_index ] );
    it_cache = histo_cache_.find( temp_pair );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it_cache != histo_cache_.end() ), 
    "Histogram build error" );
    
  /* Calculating min */
    
  TePDIHistogram::iterator it_hist = it_cache->second->begin();
  TePDIHistogram::iterator it_hist_end = it_cache->second->end();
  
  double max = DBL_MIN;
  
  while( it_hist != it_hist_end ) {
    if( it_hist->first > max ) {
      max = it_hist->first;
    }
    
    ++it_hist;
  }
  
  return max;
}


double TePDIStatistic::getMode( unsigned int raster_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster_index < rasters.size(), "Invalid index" );
  
  /* Building histogram, if necessary */
  
  std::pair< TeRaster*, int  > temp_pair;
  temp_pair.first = rasters[ raster_index ].nakedPointer();
  temp_pair.second = bands[ raster_index ];
  
  std::map< std::pair< TeRaster*, int  >, TePDIHistogram::pointer >::iterator
    it_cache = histo_cache_.find( temp_pair );
    
  if( it_cache == histo_cache_.end() ) {
    DoSingleRasterCCalcules( rasters[ raster_index ], bands[ raster_index ] );
    it_cache = histo_cache_.find( temp_pair );
  }
  
  PDIAGN_DEBUG_CONDITION( ( it_cache != histo_cache_.end() ), 
    "Histogram build error" );
    
  /* Calculating min */
    
  TePDIHistogram::iterator it_hist = it_cache->second->begin();
  TePDIHistogram::iterator it_hist_end = it_cache->second->end();
  
  unsigned int frequency = 0;
  double mode = 0;
  
  while( it_hist != it_hist_end ) {
    if( it_hist->second > frequency ) {
      mode = it_hist->first;
      frequency = it_hist->second;
    }
    
    ++it_hist;
  }
  
  return mode;
}


double TePDIStatistic::getCorrelation( unsigned int raster1_index,  
  unsigned int raster2_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster1_index < rasters.size(), "Invalid index1" );
  PDIAGN_TRUE_OR_THROW( raster2_index < rasters.size(), "Invalid index2" );
  
  double covariance = getCovariance( raster1_index, raster2_index );

  double sqrt_c_1 = sqrt(
    getCovariance( raster1_index, raster1_index ) );
  double sqrt_c_2 = sqrt(
    getCovariance( raster2_index, raster2_index ) );

  double correlation;

  if( ( sqrt_c_1 * sqrt_c_2 ) == 0 ) {
    correlation = 1.;
  } else {
    correlation = covariance / ( sqrt_c_1 * sqrt_c_2 );
  }

  return correlation;  
}


double TePDIStatistic::getCCorrelation( unsigned int raster1_index,  
        unsigned int raster2_index )
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  PDIAGN_TRUE_OR_THROW( raster1_index < rasters.size(), "Invalid index1" );
  PDIAGN_TRUE_OR_THROW( raster2_index < rasters.size(), "Invalid index2" );
  
  double sum_pb1b2 = getSumPB1B2( rasters[ raster1_index ], 
    bands[ raster1_index ], rasters[ raster2_index ], bands[ raster2_index ] );
    
  /* Calculating the elements number */
  
  double elements_number = 0;
  
  TePDIHistogram::pointer histogram1 = getHistogram( raster1_index );
  
  TePDIHistogram::iterator it = histogram1->begin();
  TePDIHistogram::iterator it_end = histogram1->end();
  
  while( it != it_end ) {
    elements_number += ( (double) it->second );
    ++it;
  }    
    
  double ccorrelation = sum_pb1b2 / elements_number;

  return ccorrelation;  
}


TeMatrix TePDIStatistic::getCMMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );
  
  TeMatrix outmatrix;
  outmatrix.Init( 3, rasters.size(), 0. );

  double sum3 = 0;
  double sum4 = 0;
  double mean = 0;
  double ccorr = 0;
  double pixels_nmb = 0;

  for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
    sum3 = getSum3( rasters[ index ], bands[ index ] );
    sum4 = getSum4( rasters[ index ], bands[ index ] );
    ccorr = getCCorrelation( index, index );
    mean = getMean( index );
    
    /* Calculating the elements number */
    
    pixels_nmb = 0;
    
    TePDIHistogram::pointer histo_ptr = getHistogram( index );
    TePDIHistogram::iterator it = histo_ptr->begin();
    TePDIHistogram::iterator it_end = histo_ptr->end();
    
    while( it != it_end ) {
      pixels_nmb += ( (double) it->second );
      ++it;
    }
    
    outmatrix( 0, index ) = getCovariance( index, index );

    outmatrix( 1, index ) = sum3 / pixels_nmb - 3 * mean * ccorr +
                           2 * pow( mean, 3 );

    outmatrix( 2, index ) = sum4 / pixels_nmb - 4 * mean * sum3 / pixels_nmb +
                           6 * pow( mean, 2 ) * ccorr -
                           3 * pow( mean, 4 );
  }

  return outmatrix;
}


TeMatrix TePDIStatistic::getAssimetryMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );

  TeMatrix outmatrix;
  outmatrix.Init( 1, rasters.size(), 0. );

  double sum3 = 0;
  double mean = 0;
  double ccorr = 0;
  double cov = 0;
  double sigma = 0;
  double sigma3 = 0;
  double pixels_nmb = 0;

  for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
    cov = getCovariance( index, index );
    sigma = sqrt( cov );
    sigma3 = sigma * sigma * sigma;

    if( sigma3 != 0 ) {
      sum3 = getSum3( rasters[ index ], bands[ index ] );
      
      ccorr = getCCorrelation( index, index );
        
      mean = getMean( index );
      
      /* Calculating the elements number */
      
      pixels_nmb = 0;
      
      TePDIHistogram::pointer histo_ptr = getHistogram( index );
      TePDIHistogram::iterator it = histo_ptr->begin();
      TePDIHistogram::iterator it_end = histo_ptr->end();
      
      while( it != it_end ) {
        pixels_nmb += ( (double) it->second );
        ++it;
      }      

      outmatrix( 0, index ) = ( sum3 / pixels_nmb -  3 * mean * ccorr +
                              2 * pow( mean, 3 ) ) / sigma3;
    } else {
      outmatrix( 0, index ) = 0;
    }
  }

  return outmatrix;
}


TeMatrix TePDIStatistic::getKurtosisMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  std::vector< int > bands;
  params_.GetParameter( "bands", bands );

  TeMatrix outmatrix;
  outmatrix.Init( 1, rasters.size(), 0. );

  double pixels_nmb = 0;

  double sum3;
  double sum4;
  double mean;
  double ccorr;
  double cov;
  double sigma;
  double sigma4;

  for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
    cov = getCovariance( index, index );
    
    sigma = sqrt( cov );
    sigma4 = sigma * sigma * sigma * sigma;

    if( sigma4 != 0 ) {
      sum3 = getSum3( rasters[ index ], bands[ index ] );
      
      sum4 = getSum4( rasters[ index ], bands[ index ] );
      
      ccorr = getCCorrelation( index, index );
      
      mean = getMean( index );
      
      /* Calculating the elements number */
      
      pixels_nmb = 0;
      
      TePDIHistogram::pointer histo_ptr = getHistogram( index );
      TePDIHistogram::iterator it = histo_ptr->begin();
      TePDIHistogram::iterator it_end = histo_ptr->end();
      
      while( it != it_end ) {
        pixels_nmb += ( (double) it->second );
        ++it;
      } 

      outmatrix( 0, index ) = ( sum4 / pixels_nmb - 4 * mean * sum3 /
        pixels_nmb  + 6 * pow( mean, 2 ) * ccorr - 3 * pow( mean, 4 ) ) /
        sigma4 - 3.;
    } else {
      outmatrix( 0, index ) = 0;
    }
  }

  return outmatrix;
}


TeMatrix TePDIStatistic::getVarCoefMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  TeMatrix outmatrix;
  outmatrix.Init( 1, rasters.size(), 0. );

  double mean;

  for( unsigned int index = 0 ; index < rasters.size() ; ++index ) {
    mean = getMean( index );

    if( mean != 0 ) {
      outmatrix( 0, index ) = sqrt(
        getCovariance( index, index ) ) / mean;
    } else {
      outmatrix( 0, index ) = 0;
    }
  }

  return outmatrix;
}


TeMatrix TePDIStatistic::getCovMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  TeMatrix outmatrix;
  outmatrix.Init( rasters.size(), rasters.size(), 0. );

  for( unsigned int band1 = 0 ; band1 < rasters.size() ; ++band1 ) {
    for( unsigned int band2 = 0 ; band2 < rasters.size() ; ++band2 ) {
      outmatrix( band1, band2 ) = 
        getCovariance( band1, band2 );
    }
  }

  return outmatrix;
}


TeMatrix TePDIStatistic::getCorMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  TeMatrix outmatrix;
  outmatrix.Init( rasters.size(), rasters.size(), 0. );

  for( unsigned int band1 = 0 ; band1 < rasters.size() ; ++band1 ) {
    for( unsigned int band2 = 0 ; band2 < rasters.size() ; ++band2 ) {
      outmatrix( band1, band2 ) = 
        getCorrelation( band1, band2 );
    }
  }

  return outmatrix;
}


TeMatrix TePDIStatistic::getVarMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  TeMatrix outmatrix;
  outmatrix.Init( rasters.size(), 1, 0. );

  for( unsigned int band = 0 ; band < rasters.size() ; ++band ) {
    outmatrix( band, 0 ) = getVariance( band );
  }

  return outmatrix;
}


TeMatrix TePDIStatistic::getStdDevMatrix()
{
  TePDITypes::TePDIRasterVectorType rasters;
  params_.GetParameter( "rasters", rasters );
  
  TeMatrix outmatrix;
  outmatrix.Init( rasters.size(), 1, 0. );

  for( unsigned int band = 0 ; band < rasters.size() ; ++band ) {
    outmatrix( band, 0 ) = getStdDev( band );
  }

  return outmatrix;
}


void TePDIStatistic::getRestrictionPolygons(
  std::vector< TePolygon >& polygons )
{
  polygons.clear();
  
  if( params_.CheckParameter( "polygon", 
    TePDIParameters::TePolygonParam ) ) {
    
    TePolygon temp_polygon;
    params_.GetParameter( "polygon", temp_polygon );
    polygons.push_back( temp_polygon );
  } 
  
  if( params_.CheckParameter( "polygonset", 
    TePDIParameters::TePolygonSetParam ) ) {
    
    TePolygonSet temp_polygonset;
    params_.GetParameter( "polygonset", temp_polygonset );
    
    TePolygonSet::iterator it = temp_polygonset.begin();
    TePolygonSet::iterator it_end = temp_polygonset.end();
    
    while( it != it_end ) {
      polygons.push_back( *it );
      
      ++it;
    }
  }
  
  if( polygons.size() == 0 ) {
    TePDITypes::TePDIRasterVectorType rasters;
    params_.GetParameter( "rasters", rasters );
    
    TeBox image_bbox = rasters[ 0 ]->params().boundingBox();
  
    TePolygon temp_polygon = polygonFromBox( image_bbox );
    
    polygons.push_back( temp_polygon );
  }
}
