/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

#ifndef TEPDISTATISTIC_HPP
  #define TEPDISTATISTIC_HPP

  #include "TePDIAlgorithm.hpp"
  #include "TePDIParameters.hpp"
  #include <TeSharedPtr.h>
  #include "TePDITypes.hpp"

  #include <TeMatrix.h>
  #include <TeGeometry.h>

  #include <map>

  /**
   * @brief This is the class for statistic calcules over raster data.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup PDIAlgorithms
   *
   * @note The general required parameters are:
   *
   * @param rasters ( TePDITypes::TePDIRasterVectorType ) - A vector of all used 
   * rasters ( NOTE: When more then one band is used from the same raster,
   * this raster must appear more then one time ).
   * @param bands ( std::vector< int > ) - A vector of all used bands
   * for each raster inside the reasters vector above respectively.
   *
   * @note The following parameters are optional, and will be used if
   * present.
   *
   * @param polygon ( TePolygon ) - Restriction polygon - The area
   * where to do calcules over all rasters.
   * @param polygonset ( TePolygonSet ) - Restriction polygon set - The area
   * where to do calcules over all rasters.   
   * @param histograms ( std::vector< TePDIHistogram::pointer > ) - 
   * A vector of all raster histogram references ( internal histogram 
   * generation will be disabled );
   */
  class TePDIStatistic : public TePDIAlgorithm {
    public :
      typedef TeSharedPtr< TePDIStatistic > pointer;
      typedef const TeSharedPtr< TePDIStatistic > const_pointer;

      /**
       * Default Constructor.
       *
       */
      TePDIStatistic();

      /**
       * Default Destructor
       */
      ~TePDIStatistic();

      /**
       * Checks if the supplied parameters fits the requirements of each
       * PDI algorithm implementation.
       *
       * Error log messages must be generated. No exceptions generated.
       *
       * @param parameters The parameters to be checked.
       * @return true if the parameters are OK. false if not.
       */
      bool CheckParameters( const TePDIParameters& parameters ) const;
      
      /**
       * Generates the histogram for a given raster/band.
       *
       * @param raster_index Index for the raster ( using the
       * supplied parameters ).
       * @return The histogram pointer.
       */
      TePDIHistogram::pointer getHistogram( unsigned int raster_index );      

      /**
       * Sum of all pixels values.
       *
       * @param raster_index Index for the raster ( using the
       * supplied parameters )
       * @return The sum value.
       */
      double getSum( unsigned int raster_index );
      
      /**
       * Mean pixels value for input_image1.
       *
       * @param raster_index Index for the raster ( using the
       * supplied parameters )
       * @return The mean value.
       */
      double getMean( unsigned int raster_index );      

      /**
       * Covariance between two bands.
       *
       * @param raster1_index Index for the first used raster / band ( using the
       * supplied parameters )
       * @param raster2_index Index for the second used raster / band ( using the
       * supplied parameters )
       * @return The covariance value.
       */
      double getCovariance( unsigned int raster1_index,  
        unsigned int raster2_index );

      /**
       * Variance of one band data.
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The variance value.
       */
      double getVariance( unsigned int raster_index );
      
      /**
       * Standard deviation of one band data.
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The Standard deviation value.
       */
      double getStdDev( unsigned int raster_index );      
                
      /**
       * Entropy of one band data.
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The entropy value.
       */
      double getEntropy( unsigned int raster_index );
      
      /**
       * The minimum pixel value of one band data.
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The minimum pixel value value.
       */
      double getMin( unsigned int raster_index );      

      /**
       * The maximum pixel value of one band data.
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The maximum pixel value value.
       */
      double getMax( unsigned int raster_index );      

      /**
       * The highest frequency pixel value of one band data (mode).
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The highest frequency pixel value of one band data (mode).
       */
      double getMode( unsigned int raster_index );      
                  
      /**
       * Correlation between two bands.
       *
       * @param raster1_index Index for the first used raster / band ( using the
       * supplied parameters )
       * @param raster2_index Index for the second used raster / band ( using the
       * supplied parameters )
       * @return The correlation value.
       */
      double getCorrelation( unsigned int raster1_index,  
        unsigned int raster2_index );

      /**
       * Correlation coefficient between two bands.
       *
       * @param raster1_index Index for the first used raster / band ( using 
       * the supplied parameters ).
       * @param raster2_index Index for the second used raster / band ( using 
       * the supplied parameters ).
       * @return The cross-correlation value.
       */
      double getCCorrelation( unsigned int raster1_index,  
        unsigned int raster2_index );

      /**
       * Mean matrix calcule.
       *
       * NOTE: Lines = 1, Columns = total bands/rasters number.
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The mean matrix.
       */
      TeMatrix getMeanMatrix();

      /**
       * Central moments matrix.
       *
       * NOTE: Lines = 3, Columns = total bands/rasters number.
       *
       * @param raster_index Index for the used raster / band ( using the
       * supplied parameters )
       * @return The central moments matrix.
       */
      TeMatrix getCMMatrix();

      /**
       * Assimetry coefficients matrix.
       *
       * NOTE: Lines = 1, Columns = total bands/rasters number.
       *
       * @return The Assimetry coefficients matrix.
       */
      TeMatrix getAssimetryMatrix();

      /**
       * Kurtosis coefficients matrix.
       *
       * NOTE: Lines = 1, Columns = total bands/rasters number.
       *
       * @return The Kurtosis coefficients matrix.
       */
      TeMatrix getKurtosisMatrix();

      /**
       * Variation coefficients matrix.
       *
       * NOTE: Lines = 1, Columns = total bands/rasters number.
       *
       * @return The Variation coefficients matrix.
       */
      TeMatrix getVarCoefMatrix();
      
      /**
       * Covariance matrix.
       *
       * NOTE: Lines = total bands/rasters number, 
       *       Columns = total bands/rasters number.
       *
       * @return The Covariance matrix.
       */
      TeMatrix getCovMatrix();      
      
      /**
       * Correlation matrix.
       *
       * NOTE: Lines = total bands/rasters number, 
       *       Columns = total bands/rasters number.
       *
       * @return The Correlation matrix.
       */
      TeMatrix getCorMatrix();        
      
      /**
       * Variance matrix.
       *
       * NOTE: Lines = total bands/rasters number, 
       *       Columns = 1.
       *
       * @return The Variance matrix.
       */
      TeMatrix getVarMatrix();       
      
      /**
       * Standard deviation matrix.
       *
       * NOTE: Lines = total bands/rasters number, 
       *       Columns = 1.
       *
       * @return The Standard deviation matrix.
       */
      TeMatrix getStdDevMatrix();          

    protected :
    

      /**
       * Pixel level cubic sum cache.
       *
       * @param TeRaster Raster refenrece (key).
       * @param int Band.
       * @param double Sum.
       */
      std::map< std::pair< TeRaster*, int >, double > sum3_cache_;

      /**
       * Pixel hypercubic level cubic sum cache.
       *
       * @param TeRaster Raster refenrece (key).
       * @param int Band.
       * @param double Sum.
       */
      std::map< std::pair< TeRaster*, int >, double > sum4_cache_;

      /**
       * Sum of products between all pixels from different bands.
       *
       * @param TeRaster Raster1 reference.
       * @param unsigned int Band1 from raster1.
       * @param TeRaster Raster2 reference.
       * @param unsigned int Band2 from raster2.
       * @param double The product value.
       */
      std::map< std::pair< std::pair< TeRaster*, int >,
        std::pair< TeRaster*, int > >, double > sum_pb1b2_cache_;
        
      /**
       * A cache of level ordered histograms to avoid rebuilding
       * histograms all the times.
       *
       * @param TeRaster A raster naked pointer reference.
       * @param "unsigned int" Raster band.
       * @param TePDIHistogram::pointer A Histogram shared pointer reference.
       */
      std::map< std::pair< TeRaster*, int  >,
                TePDIHistogram::pointer > histo_cache_;
      
      /**
       * Runs the current algorithm implementation.
       *
       * @return true if OK. false on error.
       */
      bool RunImplementation();

      /**
       * Reset the internal state to the initial state.
       *
       * params The new parameters referente at initial state.
       */
      void ResetState( const TePDIParameters& params );
      
      /**
       * Locate the raster index inside current param rasters.
       *
       * @param raster The raster.
       * @return The raster index.
       */
      unsigned int getRasterParamIndex( 
        const TePDITypes::TePDIRasterPtrType& raster );      
      
      /**
       * Locate all bands of a supplied raster inside the internal parameters.
       *
       * @param raster The raster.
       * @return A vector of all used bands for the supplied raster.
       */
      std::vector< int > getUsedBands( 
        TePDITypes::TePDIRasterPtrType& raster );

      /**
       * The cubic sum of pixels values.
       *
       * @param raster Raster..
       * @param band Band number.
       * @return The pixel level sum.
       */
      double getSum3( const TePDITypes::TePDIRasterPtrType& raster,
        unsigned int band );

      /**
       * The hypercubic sum of pixels values.
       *
       * @param raster Raster..
       * @param band Band number.
       * @return The pixel level sum.
       */
      double getSum4( const TePDITypes::TePDIRasterPtrType& raster,
        unsigned int band );
        
      /**
       * Do common calcules over a single raster updating all caches.
       *
       * @param raster Raster..
       * @param band Band number.
       */
      void DoSingleRasterCCalcules( 
        const TePDITypes::TePDIRasterPtrType& raster,
        unsigned int band );
        
      /**
       * Do common calcules over two rasters updating all caches.
       *
       * @param raster1 Raster1.
       * @param band1 Band number ( from input raster1 ).
       * @param raster2 Raster2.
       * @param band2 Band number ( from input raster2 ).
       */
      void DoCoupleRasterCCalcules( TePDITypes::TePDIRasterPtrType& raster1,
        unsigned int band1, TePDITypes::TePDIRasterPtrType& raster2,
        unsigned int band2 );           
        
      /**
       * Sum of products between all pixels from different bands.
       *
       * @param raster1 Raster1.
       * @param band1 Band number ( from input raster1 ).
       * @param raster2 Raster2.
       * @param band2 Band number ( from input raster2 ).
       * @return The Sum of products.
       */
      double getSumPB1B2( TePDITypes::TePDIRasterPtrType& raster1,
        unsigned int band1,  TePDITypes::TePDIRasterPtrType& raster2,
        unsigned int band2 );
        
      /**
       * Extracts the user supplied restriction polygons ( if they don't
       * exist, one restriction polygon, build from the image
       * box, will be returned ).
       *
       * @param polygons A output vector with all restriction polygons.
       */
      void getRestrictionPolygons( std::vector< TePolygon >& polygons );
      
  };
  
/** @example TePDIStatistic_test.cpp
 *    Shows how to use this class.
 */    

#endif //TEPDISTATISTIC_HPP
