#include "TePDIUtils.hpp"

#include "TePDIAgnostic.hpp"

#include <TeDecoderMemory.h>
#include <TeDecoderMemoryMap.h>
#include <TeDecoderTIFF.h>
#include <TeRaster.h>
#include <TeRasterParams.h>
#include <TeDataTypes.h>
#include <TeUtils.h>
#include <TeDefines.h>
#include <TeVectorRemap.h>
#include <TeRasterTransform.h>
#include <TeRasterRemap.h>
#include <TeProgress.h>

#include <float.h>
#include <math.h>
#include <stdio.h>


namespace TePDIUtils{


  bool TeAllocRAMRaster(
    TePDITypes::TePDIRasterPtrType& template_raster,
    TePDITypes::TePDIRasterPtrType& RAMRaster,
    unsigned int bands, bool force_new_dt, bool force_mm,
    TeDataType pixel_type )
  {
    PDIAGN_TRUE_OR_RETURN( template_raster.isActive(),
      "Inactive Input Raster" );
    PDIAGN_TRUE_OR_RETURN( template_raster->params().status_ != TeRasterParams::TeNotReady,
      "Input not Ready" );

    TeRasterParams temp_params = template_raster->params();

    if( bands != 0 ) {
      temp_params.nBands( bands );
    }

    if( force_new_dt ) {
      temp_params.setDataType( pixel_type, -1 );
    }

    return TeAllocRAMRaster( RAMRaster, temp_params, force_mm );
  }


  bool TeAllocRAMRaster(
    TePDITypes::TePDIRasterPtrType& RAMRaster,
    unsigned int bands, unsigned int lines, unsigned int columns,
    bool force_mm, TeDataType pixel_type, TePDIRgbPalette* palette )
  {
    PDIAGN_TRUE_OR_RETURN( bands > 0, "Invalid bands number" );
    PDIAGN_TRUE_OR_RETURN( lines > 0, "Invalid lines number" );
    PDIAGN_TRUE_OR_RETURN( columns > 0, "Invalid columns number" );

    TeRasterParams temp_params;

    temp_params.nBands( bands );

    temp_params.setDataType( pixel_type, -1 );
    
    temp_params.setNLinesNColumns( lines, columns );
    
    if( palette != 0 ) {
      temp_params.setPhotometric( TeRasterParams::TePallete );

      temp_params.lutr_.clear();
      temp_params.lutg_.clear();
      temp_params.lutb_.clear();

      TePDIRgbPalette::iterator pal_it = palette->begin();
      TePDIRgbPalette::iterator pal_it_end = palette->end();

      for( unsigned int lut_index = 0 ; lut_index < palette->size() ;
           ++lut_index ) {

        temp_params.lutr_.push_back( (unsigned short)pal_it->second.red_ );
        temp_params.lutg_.push_back( (unsigned short)pal_it->second.green_ );
        temp_params.lutb_.push_back( (unsigned short)pal_it->second.blue_ );

        ++pal_it;
      }
    }    

    return TeAllocRAMRaster( RAMRaster, temp_params, force_mm );
  }
  
  bool TeAllocRAMRaster( TePDITypes::TePDIRasterPtrType& RAMRaster,
    const TeRasterParams& params, bool force_mm )
  {
    return TeAllocRAMRaster( RAMRaster, params,
      ( force_mm ? TePDIUtilsDiskMemPol : TePDIUtilsAutoMemPol ) );
  }

  
  bool TeAllocRAMRaster( TePDITypes::TePDIRasterPtrType& RAMRaster,
    const TeRasterParams& params, TePDIUtilsMemPol mempol )
  {
    TeRasterParams internal_params = params;
    
    internal_params.mode_ = 'c';
  
    /* Verifying the amount of required memory */

    unsigned long int req_mem = TeGetRasterReqMem( internal_params );
    
    PDIAGN_TRUE_OR_RETURN( TeGetTempFileName( internal_params.fileName_ ),
      "Unable to get temporary file name" );    
      
    if( mempol == TePDIUtilsRamMemPol ) {
      internal_params.decoderIdentifier_ = "MEM";
    } else if( mempol == TePDIUtilsDiskMemPol ) {
      internal_params.decoderIdentifier_ = "MEMMAP";
    } else {
      /* Auto mem pol */
      if( req_mem > TeGetFreePhysicalMemory() ) {
        internal_params.decoderIdentifier_ = "MEMMAP";
      } else {
        internal_params.decoderIdentifier_ = "MEM";
      }
    }
    
    if( RAMRaster.isActive() ) {
      RAMRaster->updateParams( internal_params );
      return RAMRaster->init( internal_params );
    } else {
      RAMRaster.reset( new TeRaster( internal_params ) );
      return RAMRaster->init();
    }
  }
    

  bool TeRaster2Geotiff(
    TePDITypes::TePDIRasterPtrType& in_raster,
    const std::string& file_name, TeDataType pixel_type )
  {
    PDIAGN_TRUE_OR_RETURN( in_raster.isActive(),
      "Inactive Input Raster" );
    PDIAGN_TRUE_OR_RETURN( in_raster->params().status_ != 
      TeRasterParams::TeNotReady, "Input not Ready" );
    PDIAGN_TRUE_OR_RETURN( file_name.size() != 0,
      "Invalid file name" );

    TeRasterParams temp_params = in_raster->params();

    temp_params.mode_ = 'c';
    temp_params.fileName_ = file_name;
    temp_params.setDataType( pixel_type );
    
    /* Creating disk output raster */
    
    TePDITypes::TePDIRasterPtrType outRaster( new TeRaster() );

    outRaster->setDecoder( new TeDecoderTIFF( temp_params ) );

    PDIAGN_TRUE_OR_RETURN( outRaster->init(),
      "Unable to init GeoTIFF Raster" );
      
    return TeCopyRasterBands( in_raster, outRaster, true, true );
  }


  bool TeCopyRasterPixels(
    TePDITypes::TePDIRasterPtrType& source_raster,
    unsigned int source_band,
    TePDITypes::TePDIRasterPtrType& target_raster,
    unsigned int target_band,
    bool discretize_levels,
    bool trunc_levels )
  {
    PDIAGN_TRUE_OR_RETURN( source_raster.isActive(),
      "Inactive Input Raster" );
    PDIAGN_TRUE_OR_RETURN( target_raster.isActive(),
      "Inactive Output Raster" );
    PDIAGN_TRUE_OR_RETURN( source_raster->params().status_ != TeRasterParams::TeNotReady,
      "Input not Ready" );
    PDIAGN_TRUE_OR_RETURN( target_raster->params().status_ == TeRasterParams::TeReadyToWrite,
      "OutPut not Ready" );

    TeRasterParams& source_params = source_raster->params();
    TeRasterParams& target_params = target_raster->params();

    PDIAGN_TRUE_OR_RETURN( (int)source_band < source_params.nBands(),
      "Invalid source band" );
    PDIAGN_TRUE_OR_RETURN( (int)target_band < target_params.nBands(),
      "Invalid target band" );
    PDIAGN_TRUE_OR_RETURN( (
      ( source_params.nlines_ == target_params.nlines_ ) &&
      ( source_params.ncols_ == target_params.ncols_ ) ),
      "Input and Output Raster's have different dimentions" );
      
    /* Dummy support */
    
    const bool source_raster_uses_dummy = source_raster->params().useDummy_;
    double source_raster_dummy = 0;
    if( source_raster_uses_dummy ) {
      source_raster_dummy = source_raster->params().dummy_[ source_band ];
    }
    
    const bool target_raster_uses_dummy = target_raster->params().useDummy_;
    double target_raster_dummy = 0;
    if( target_raster_uses_dummy ) {
      target_raster_dummy = target_raster->params().dummy_[ target_band ];
    } else if( source_raster_uses_dummy ) {
      target_raster_dummy = source_raster_dummy;
    }

    /* Defining the valid values range base on the output raster pixel type */

    double min_value = 0;
    double max_value = 0;

    PDIAGN_TRUE_OR_RETURN(
      TeGetRasterMinMaxBounds( target_raster, target_band, min_value, max_value ),
      "Unable to get raster level bounds" );

    /* Pixel by pixel copy loop */

    double pixel_value;

    for( int line = 0 ; line < source_params.nlines_ ; ++line ) {
      for( int column = 0 ; column < source_params.ncols_ ; ++column ) {
        if( source_raster->getElement( column, line, pixel_value, 
          source_band ) ) {
          
          if( discretize_levels ) {
            pixel_value = TeRound( pixel_value );
          }

          if( trunc_levels ) {
            if( pixel_value < min_value ) {
              pixel_value = min_value;
            } else if( pixel_value > max_value ) {
              pixel_value = max_value;
            }
          }

          PDIAGN_TRUE_OR_RETURN( target_raster->setElement( column, line,
            pixel_value, target_band ), "Pixel copy error" );
        } else {
          PDIAGN_TRUE_OR_RETURN( source_raster_uses_dummy,
            "Unable to read from input raster" );
            
          PDIAGN_TRUE_OR_RETURN( target_raster->setElement( column, line,
            target_raster_dummy, target_band ), "Pixel copy error" );          
        }
      }
    }

    return true;
  }


  bool TeCopyRasterBands(
    TePDITypes::TePDIRasterPtrType& source_raster,
    TePDITypes::TePDIRasterPtrType& target_raster,
    bool discretize_levels,
    bool trunc_levels )
  {
    PDIAGN_TRUE_OR_RETURN( source_raster.isActive(),
      "Inactive Input Raster" );
    PDIAGN_TRUE_OR_RETURN( target_raster.isActive(),
      "Inactive Output Raster" );
    PDIAGN_TRUE_OR_RETURN( source_raster->params().status_ != TeRasterParams::TeNotReady,
      "Input not Ready" );
    PDIAGN_TRUE_OR_RETURN( target_raster->params().status_ == TeRasterParams::TeReadyToWrite,
      "OutPut not Ready" );

    TeRasterParams& source_params = source_raster->params();
    TeRasterParams& target_params = target_raster->params();

    PDIAGN_TRUE_OR_RETURN(
      ( ( source_params.nBands() == target_params.nBands() ) &&
      ( source_params.nlines_ == target_params.nlines_ ) &&
      ( source_params.ncols_ == target_params.ncols_ ) ),
      "Input and Output Raster's have different dimentions" );

    for( int band = 0 ; band < source_params.nBands() ; ++band ) {
      PDIAGN_TRUE_OR_RETURN( TeCopyRasterPixels( source_raster, band, target_raster,
        band, discretize_levels,trunc_levels ), "Unable to copy band " +
        Te2String( band ) );
    }

    return true;
  }

  
  bool TeGetRasterMinMaxBounds( TePDITypes::TePDIRasterPtrType& raster,
    unsigned int band, double& min, double& max )
  {
    PDIAGN_TRUE_OR_RETURN( raster.isActive(),
      "Inactive Input Raster" );

    TeRasterParams& params = raster->params();

    PDIAGN_TRUE_OR_RETURN( (int)band < params.nBands(),
      "Invalid source band" );

    switch( params.dataType_[ band ] ) {
      case TeBIT :
      {
        max = 1;
        min = 0;
        break;
      }
      case TeUNSIGNEDCHAR :
      {
        max = pow( 2., (double)sizeof( unsigned char ) * 8. ) - 1;
        min = 0;
        break;
      }
      case TeUNSIGNEDSHORT :
      {
        max = pow( 2., (double)sizeof( short ) * 8. ) - 1;
        min = 0;
        break;
      }
      case TeUNSIGNEDLONG :
      {
        max = pow( 2., (double)sizeof( unsigned long ) * 8. ) - 1;
        min = 0;
        break;
      }
      case TeCHAR :
      {
        max = ( pow( 2., (double)sizeof( char ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeSHORT :
      {
        max = ( pow( 2., (double)sizeof( short ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeINTEGER :
      {
        max = ( pow( 2., (double)sizeof( int ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeLONG :
      {
        max = ( pow( 2., (double)sizeof( long ) * 8. ) ) / 2;
        min = -1 * max;
        break;
      }
      case TeFLOAT :
      {
        max = FLT_MAX;
        min = -1 * max;
        break;
      }
      case TeDOUBLE :
      {
        max = DBL_MAX;
        min = -1 * max;
        break;
      }
      default :
      {
        PDIAGN_LOG_AND_RETURN( "Invalid raster pixel type" );
        break;
      }
    }

    return true;
  }
  
  
  unsigned long int TeGetRasterReqMem( TeRasterParams& params )
  {
    unsigned long int req_mem = 0;

    for( int band = 0 ; band < params.nBands() ; ++band ) {
      req_mem += params.elementSize( band ) * 
        params.nlines_ * params.ncols_;
    }
    
    req_mem += sizeof( unsigned short ) * params.lutr_.size();
    req_mem += sizeof( unsigned short ) * params.lutg_.size();
    req_mem += sizeof( unsigned short ) * params.lutb_.size();  
    
    return req_mem;
  }

  
  bool rasterCompare( TePDITypes::TePDIRasterPtrType& raster1,
    TePDITypes::TePDIRasterPtrType& raster2 )
  {
    PDIAGN_TRUE_OR_THROW( raster1.isActive(), "Raster1 not active" );
    PDIAGN_TRUE_OR_THROW( raster2.isActive(), "Raster2 not active" );
    PDIAGN_TRUE_OR_THROW( raster1->params().status_ != TeRasterParams::TeNotReady,
      "Raster1 not ready" );
    PDIAGN_TRUE_OR_THROW( raster2->params().status_ != TeRasterParams::TeNotReady,
      "Raster2 not ready" );

    TeRasterParams par1 = raster1->params();
    TeRasterParams par2 = raster2->params();

    if( ( par1.nBands() != par2.nBands() ) ||
        ( par1.ncols_ != par2.ncols_ ) ||
        ( par1.nlines_ != par2.nlines_ ) ) {

      return false;
    }

    for( int bandt = 0 ; bandt < par1.nBands() ; ++bandt ) {
      if( ( par1.dataType_[ bandt ] != par2.dataType_[ bandt ] ) ||
          ( par1.photometric_[ bandt ] != par2.photometric_[ bandt ] ) ) {

        return false;
      }
    }

    unsigned int lines = ( unsigned int ) par1.nlines_;
    unsigned int columns = ( unsigned int ) par1.ncols_;
    double value1 = 0;
    double value2 = 0;

    for( int band = 0 ; band < par1.nBands() ; ++band ) {
      for( unsigned int line = 0 ; line < lines ; ++line ) {
        for( unsigned int column = 0 ; column < columns ; ++column ) {
          if( raster1->getElement( column, line, value1, band ) !=
              raster2->getElement( column, line, value2, band ) ) {

            return false;
          }

          if( value1 != value2 ) {
            return false;
          }
        }
      }
    }

    return true;
  }
  
  
  bool IsFloatBand( const TePDITypes::TePDIRasterPtrType& raster, 
    unsigned int band  )
  {
    PDIAGN_TRUE_OR_THROW( raster.isActive(),
      "Inactive Input Raster" );

    TeRasterParams& params = raster->params();

    PDIAGN_TRUE_OR_THROW( (int)band < params.nBands(),
      "Invalid source band" );

    switch( params.dataType_[ band ] ) {
      case TeBIT :
      {
        return false;
        break;
      }
      case TeUNSIGNEDCHAR :
      {
        return false;
        break;
      }
      case TeUNSIGNEDSHORT :
      {
        return false;
        break;
      }
      case TeUNSIGNEDLONG :
      {
        return false;
        break;
      }
      case TeCHAR :
      {
        return false;
        break;
      }
      case TeSHORT :
      {
        return false;
        break;
      }
      case TeINTEGER :
      {
        return false;
        break;
      }
      case TeLONG :
      {
        return false;
        break;
      }
      case TeFLOAT :
      {
        return true;
        break;
      }
      case TeDOUBLE :
      {
        return true;
        break;
      }
      default :
      {
        PDIAGN_LOG_AND_THROW( "Invalid raster pixel type" );
        return false;
        break;
      }
    }
  }
  
  
  void MapCoords2RasterIndexes( const TeCoord2D& g_in, 
    const TePDITypes::TePDIRasterPtrType& raster, TeCoord2D& g_out )
  {
    PDIAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
    
    g_out = raster->coord2Index( g_in );
  };
  
  
  void MapCoords2RasterIndexes( const TeBox& box_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeBox& box_out )
  {
    PDIAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
    
    TeBox temp_box( raster->coord2Index( box_in.lowerLeft() ),
      raster->coord2Index( box_in.upperRight() ) );
      
    box_out = temp_box;
  };  
  
  
  void MapRasterIndexes2Coords( const TeCoord2D& g_in, 
    const TePDITypes::TePDIRasterPtrType& raster, TeCoord2D& g_out )
  {
    PDIAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
      
    g_out = raster->index2Coord( g_in );
  };  
  
  
  void MapRasterIndexes2Coords( const TeBox& box_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeBox& box_out )
  {
    PDIAGN_TRUE_OR_THROW( raster.isActive(), "Inactive raster" );
    
    TeBox temp_box( raster->index2Coord( box_in.lowerLeft() ),
      raster->index2Coord( box_in.upperRight() ) );
      
    box_out = temp_box;
  };   
  
  
  bool IsDiscreteLut( const TePDITypes::TePDILutType& lut )
  {
    PDIAGN_TRUE_OR_THROW( ( lut.size() > 0 ), "Emptry lut" );
  
    TePDITypes::TePDILutType::const_iterator it = lut.begin();
    const TePDITypes::TePDILutType::const_iterator it_end = lut.end();
  
    while( it != it_end ) {
      if( it->first != floor( it->first ) ) {
        return false;
      }
      
      ++it;
    }
    
    return true;
  }
  

  bool FindBestDecimLevels( 
    const TePDITypes::TePDIRasterPtrType& raster1,
    const TePDITypes::TePDIRasterPtrType& raster2,
    unsigned long int& r1_level, unsigned long int& r2_level,
    unsigned long int& decim_error )
  {
    PDIAGN_TRUE_OR_RETURN( raster1.isActive(), "raster1 inactive" );
    PDIAGN_TRUE_OR_RETURN( raster2.isActive(), "raster2 inactive" );
  
    const unsigned long int r1_lines = raster1->params().nlines_;
    const unsigned long int r1_cols = raster1->params().ncols_;
    const unsigned long int r2_lines = raster2->params().nlines_;
    const unsigned long int r2_cols = raster2->params().ncols_;
  
    const unsigned long int r1_max_levels = MIN(
      TePDIMathFunctions::DecimLevels( r1_lines ),
      TePDIMathFunctions::DecimLevels( r1_cols ) );  
    const unsigned long int r2_max_levels = MIN(
      TePDIMathFunctions::DecimLevels( r2_lines ),
      TePDIMathFunctions::DecimLevels( r2_cols ) );
    
    if( ( r1_max_levels < 1 ) || ( r2_max_levels < 1 ) ) {
      r1_level = r2_level = 0;
      return false;
    } else {
      bool best_level_diff_found = false;
      
      unsigned long int curr_diff = 0;
      long int curr_r1_lines = r1_lines;
      long int curr_r1_cols = r1_cols;
      long int curr_r2_lines = 0;
      long int curr_r2_cols = 0;
      
      for( unsigned long int curr_r1_level = 1 ; 
        curr_r1_level <= r1_max_levels ; 
        ++curr_r1_level ) {
        
        curr_r1_lines = ( long int )
          ceil( ( (double)curr_r1_lines ) / 2 );
        curr_r1_cols = ( long int )
          ceil( ( (double)curr_r1_cols ) / 2 );      
          
        curr_r2_lines = r2_lines;
        curr_r2_cols = r2_cols;        
      
        for( unsigned long int curr_r2_level = 1 ; 
          curr_r2_level <= r2_max_levels ; 
          ++curr_r2_level ) {
        
          curr_r2_lines = ( long int )
            ceil( ( (double)curr_r2_lines ) / 2 );
          curr_r2_cols = ( long int )
            ceil( ( (double)curr_r2_cols ) / 2 );
            
          curr_diff = ABS( curr_r1_lines - curr_r2_lines ) +
            ABS( curr_r1_cols - curr_r2_cols );
            
          if( best_level_diff_found ) {
            if( curr_diff < decim_error ) {
              r1_level = curr_r1_level;
              r2_level = curr_r2_level;
              decim_error = curr_diff;
            }
          } else {
            r1_level = curr_r1_level;
            r2_level = curr_r2_level;
            decim_error = curr_diff;
            best_level_diff_found = true;
          }
        }
      }
      
      return true;
    }
  }
  
  bool buildDetailedBox( const TePDITypes::TePDIRasterPtrType& raster,
    TePolygon& box )
  {
    box.clear();
    
    PDIAGN_TRUE_OR_RETURN( raster.isActive(), "raster inactive" );
    
    bool raster_is_projected = true;
    if( raster->params().projection() == 0 ) {
      raster_is_projected = false;
    } else if( raster->params().projection()->name() == "NoProjection" ) {
      raster_is_projected = false;
    }
    
    TeLinearRing lr;
    double lines = (double)raster->params().nlines_;
    double cols = (double)raster->params().ncols_;
    double offset = 0;
    
    /* Adding top box border ( direction left -> right )*/
    
    for( offset = 0 ; offset < cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( TeCoord2D( offset, 0 ) ) );
      } else {
        lr.add( TeCoord2D( offset, 0 ) );
      }
    }
    
    /* Adding right box border ( direction up -> down) */
    
    for( offset = 1 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( TeCoord2D( cols - 1, offset ) ) );
      } else {
        lr.add( TeCoord2D( cols - 1, offset ) );
      }
    }
    
    /* Adding bottom box border ( direction right ->left )*/
    
    for( offset = 1 ; offset < cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( cols - 1 - offset, lines - 1 ) ) );
      } else {
        lr.add( TeCoord2D( cols - 1 - offset, lines - 1 ) );
      }
    }      
    
    /* Adding left box border ( bottom -> top )*/
    
    for( offset = 1 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( 0, lines - 1 - offset ) ) );
      } else {
        lr.add( TeCoord2D( 0, lines - 1 - offset ) );
      }
    }
        
    box.add( lr );
    
    return true;
  }
  
  
  bool buildDetailedBBox( const TePDITypes::TePDIRasterPtrType& raster,
    TePolygon& box )
  {
    box.clear();
    
    PDIAGN_TRUE_OR_RETURN( raster.isActive(), "raster inactive" );
    
    bool raster_is_projected = true;
    if( raster->params().projection() == 0 ) {
      raster_is_projected = false;
    } else if( raster->params().projection()->name() == "NoProjection" ) {
      raster_is_projected = false;
    }    
    
    TeLinearRing lr;
    const double lines = (double)raster->params().nlines_;
    const double cols = (double)raster->params().ncols_;
    double offset = 0;
    
    /* Adding top box border ( direction left -> right )*/
    
    for( offset = 0 ; offset <= cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( TeCoord2D( ( offset - 0.5 ), -0.5 ) ) );
      } else {
        lr.add( TeCoord2D( ( offset - 0.5 ), -0.5 ) );
      }
    }
    
    /* Adding right box border ( direction up -> down) */
    
    for( offset = 0 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( ( cols - 0.5 ), offset + 0.5) ) );
      } else {
        lr.add( TeCoord2D( ( cols - 0.5 ), offset + 0.5) );
      }
    }
    
    /* Adding bottom box border ( direction right ->left )*/
    
    for( offset = 0 ; offset < cols ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( ( cols - 1.5 - offset ), ( lines - 0.5 ) ) ) );
      } else {
        lr.add( TeCoord2D( ( cols - 1.5 - offset ), ( lines - 0.5 ) ) );
      }
    }      
    
    /* Adding left box border ( bottom -> top )*/
    
    for( offset = 0 ; offset < lines ; ++offset ) {
      if( raster_is_projected ) {
        lr.add( raster->index2Coord( 
          TeCoord2D( -0.5, ( lines - 1.5 - offset ) ) ) );
      } else {
        lr.add( TeCoord2D( -0.5, ( lines - 1.5 - offset ) ) );
      }
    }
        
    box.add( lr );
    
    return true;
  }  
  
  
  void getGeodeticPixelRes( TeRasterParams& params,
    double& x_res, double& y_res )
  {
    PDIAGN_TRUE_OR_THROW( ( params.projection() != 0 ),
      "Missing projection" )
    PDIAGN_TRUE_OR_THROW( ( params.projection()->name() != "NoProjection" ),
      "Invalid projection" );
      
    TeCoord2D pixel00( 0, 0 );
    TeCoord2D rightpixel( 1, 0 );
    TeCoord2D bottompixel( 0, 1 );
      
    pixel00 = params.index2Coord( pixel00 );
    rightpixel = params.index2Coord( rightpixel );
    bottompixel = params.index2Coord( bottompixel );
    
    pixel00 = params.projection()->PC2LL( pixel00 );
    rightpixel = params.projection()->PC2LL( rightpixel );
    bottompixel = params.projection()->PC2LL( bottompixel );
    
    x_res = TeDistance( rightpixel, pixel00 );
    y_res = TeDistance( bottompixel, pixel00 );  
  }
  
    
  bool reprojectRaster( const TePDITypes::TePDIRasterPtrType& input_raster,
    TeProjection& target_projection, 
    TePDITypes::TePDIRasterPtrType& output_raster )
  {
    PDIAGN_TRUE_OR_RETURN( input_raster.isActive(), 
      "input_raster not active" );
    PDIAGN_TRUE_OR_RETURN( output_raster.isActive(), "Raster2 not active" );
    PDIAGN_TRUE_OR_RETURN( 
      input_raster->params().status_ != TeRasterParams::TeNotReady,
      "input_raster not ready" );
    PDIAGN_TRUE_OR_RETURN( 
      output_raster->params().status_ != TeRasterParams::TeNotReady,
      "output_raster not ready" ); 
    PDIAGN_TRUE_OR_RETURN( ( input_raster->params().projection() != 0 ),
      "Missing input_raster projection" )
    PDIAGN_TRUE_OR_RETURN( 
      ( input_raster->params().projection()->name() != "NoProjection" ),
      "Invalid input_raster projection" );
      
    /* Building the remmaped box */
      
    TePolygon original_box_pol;
    PDIAGN_TRUE_OR_RETURN( TePDIUtils::buildDetailedBox( input_raster,
      original_box_pol ), 
      "Unable to build detailed bounding box" );
        
    TePolygon reprojected_box_pol;  
    TeVectorRemap( original_box_pol, input_raster->params().projection(),
      reprojected_box_pol, &target_projection );    
      
    /* Finding the target pixel size for the target projection */

    TeCoord2D pixel00 = TeCoord2D( 0, 0 );
    TeCoord2D pixel01 = TeCoord2D( 0, 1 );
    TeCoord2D pixel10 = TeCoord2D( 1, 0 );
    
    pixel00 = input_raster->index2Coord( pixel00 );
    pixel01 = input_raster->index2Coord( pixel01 );
    pixel10 = input_raster->index2Coord( pixel10 );
    
    pixel00 = input_raster->projection()->PC2LL( pixel00 );
    pixel01 = input_raster->projection()->PC2LL( pixel01 );
    pixel10 = input_raster->projection()->PC2LL( pixel10 );
    
    pixel00 = target_projection.LL2PC( pixel00 );
    pixel01 = target_projection.LL2PC( pixel01 );
    pixel10 = target_projection.LL2PC( pixel10 );
    
    double res_x = TeDistance( pixel00, pixel10 );
    double res_y = TeDistance( pixel00, pixel01 );
        
    /* Remmaping raster */
  
    TeRasterParams output_raster_params = output_raster->params();
    
    output_raster_params.projection( &target_projection );
    
    output_raster_params.nBands( input_raster->params().nBands() );
    
    TeBox reprojected_box = reprojected_box_pol.box();
    output_raster_params.boxResolution( reprojected_box.x1(), 
      reprojected_box.y1(), reprojected_box.x2(), reprojected_box.y2(), 
      res_x, res_y );
      
    PDIAGN_TRUE_OR_RETURN( output_raster->init( output_raster_params ),
      "Output raster init error" );
      
    TeRasterTransform transf( input_raster.nakedPointer(),
      output_raster.nakedPointer() );
    transf.setTransfFunction( TeRasterTransform::TeBand2Band );
      
    TeRasterRemap remap( input_raster.nakedPointer(), 
      output_raster.nakedPointer() ) ;
    remap.setTransformer( &transf );
    
    bool progress_interface_enabled = false;
    if( TeProgress::instance() != 0 ) {
      progress_interface_enabled = true;
    }
    
    PDIAGN_TRUE_OR_RETURN( remap.apply( progress_interface_enabled ),
      "Raster remapping error" );
      
    return true;
  }
  
  
  short TeRelation( const TeCoord2D& c, const TeTileIndexer& indexer )
  {
    TeTileIndexer::TeTileSegIndex* tindex_ptr;
    const TePolygon& pol = indexer.getPol();
    
    indexer.getTile( c.y(), &tindex_ptr );
    
    if( tindex_ptr == 0 ) {
      return TeOUTSIDE;
    }
    
    register double ty, tx;
    register bool inside_flag = false;
    register int yflag0, yflag1;
  
    TeLinearRing::iterator vtx0, vtx1;
  
    tx = c.x();
    ty = c.y();
    
    for( unsigned int i = 0 ; i < tindex_ptr->size() ; ++i ) {
      PDIAGN_DEBUG_CONDITION( ( ( *tindex_ptr )[ i ].first < pol.size() ),
        "Invalid ring index" );
      PDIAGN_DEBUG_CONDITION( ( ( ( *tindex_ptr )[ i ].second + 1 ) < 
        pol[ ( *tindex_ptr )[ i ].first ].size() ),
        "Invalid segment index" );
    
      const TeCoord2D& vtx0 = 
        pol[ ( *tindex_ptr )[ i ].first ][ ( *tindex_ptr )[ i ].second ];
      const TeCoord2D& vtx1 = 
      pol[ ( *tindex_ptr )[ i ].first ][ ( *tindex_ptr )[ i ].second + 1 ];
      
      yflag0 = (vtx0.y() >= ty);
      
      yflag1 = (vtx1.y() >= ty);
    
      if(yflag0 != yflag1)
      {
        /* TODO - Check Boundary case */
      
        if(((vtx1.y() - ty) * (vtx0.x() - vtx1.x()) >=
              (vtx1.x() - tx) * (vtx0.y() - vtx1.y())) == yflag1)
        {
          inside_flag = !inside_flag ;
        }
      }
    
    }
    
    if( inside_flag ) {
      return TeINSIDE;
    }
    
    return TeOUTSIDE;
  }  
  
} // namespace TeUtils

