/*
TerraLib - a library for developing GIS applications.
Copyright  2001, 2002, 2003 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose. The library provided hereunder is on an "as is" basis, and the
authors have no obligation to provide maintenance, support, updates,
enhancements, or modifications.
In no event shall INPE be held liable to any party
for direct, indirect, special, incidental, or consequential damages arising
out of the use of this library and its documentation.
*/

/**
 * @file TePDIUtils.hpp
 * @brief This file contains utilitary functions used by PDI algorithms. 
 * THEY SHOLD NOT BE USED BY ANYONE. The support and interfaces 
 * can be changed in future. THIS IS FOR INTERNAL USE ONLY.
 * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
 */

#ifndef TEPDIUTILS_HPP
  #define TEPDIUTILS_HPP

  #include "TePDIRgbPalette.hpp"
  #include "TePDITypes.hpp"
  #include "TePDIMathFunctions.hpp"
  
  #include "data_structs/TeTileIndexer.h"

  #include <TeRaster.h>
  #include <TeDataTypes.h>
  #include <TeBox.h>

  #include <string>
  #include <map>

 
    /**
     * @brief This namespace contains base utilitary functions used by 
     * PDI classes.
     * THEY SHOLD NOT BE USED BY ANYONE. The support and interfaces 
     * can be changed in future. THIS IS FOR INTERNAL USE ONLY.     
     * @ingroup PDIModule
     */
  namespace TePDIUtils
  {
    /**
     * PDIUtils memory policies.
     */
    enum TePDIUtilsMemPol {
      /** The memory policy will be choosed following the free RAM */
      TePDIUtilsAutoMemPol = 1,
      /** Use RAM memory */
      TePDIUtilsRamMemPol = 2,
      /** Use memory mapped disk files */
      TePDIUtilsDiskMemPol = 3
    };  

    /**
     * Allocates one RAM raster using another raster as template base.
     *
     * NOTE: If there are not enought memory, memory mapping to disk files 
     * will be used.
     *
     * @param template_raster Input raster.
     * @param RAMRaster Output raster.
     * @param bands Number of bands for the generated RAMRaster. If zero, the
     * number of bands will match the input raster.
     * @param force_new_dt If true, a new pixel data type will be used for 
     * all bands.
     * @param force_mm Force disk memory mapped raster creation.
     * @param pixel_type Pixel data type.
     * @return true if OK. false on error.
     */
    bool TeAllocRAMRaster(
      TePDITypes::TePDIRasterPtrType& template_raster,
      TePDITypes::TePDIRasterPtrType& RAMRaster,
      unsigned int bands, bool force_new_dt, bool force_mm,
      TeDataType pixel_type = TeDOUBLE );

    /**
     * Allocates a brand new RAM raster.
     *
     * NOTE: If there are not enought memory, memory mapping to disk files 
     * will be used.
     *
     * @param RAMRaster Output raster.
     * @param bands Number of bands for the generated RAMRaster.
     * @param lines Number of lines.
     * @param columns Number of Columns
     * @param force_mm Force disk memory mapped raster creation.
     * @param pixel_type Pixel data type.
     * @param palette A palette reference, for generation of a paletted
     * raster ( must be 0 if not used ).
     * @return true if OK. false on error.
     */
    bool TeAllocRAMRaster(
      TePDITypes::TePDIRasterPtrType& RAMRaster,
      unsigned int bands, unsigned int lines, unsigned int columns,
      bool force_mm, TeDataType pixel_type, TePDIRgbPalette* palette );
      
    /**
     * Allocates a brand new RAM raster using the supplied parameters.
     *
     * NOTE: If there are not enought memory, memory mapping to disk files 
     * will be used.
     *
     * @param RAMRaster Output raster.
     * @param params Raster parameters.
     * @param force_mm Force disk memory mapped raster creation.
     * @return true if OK. false on error.
     */      
    bool TeAllocRAMRaster( TePDITypes::TePDIRasterPtrType& RAMRaster,
      const TeRasterParams& params, bool force_mm );
      
    /**
     * Allocates a brand new RAM raster using the supplied parameters.
     *
     * @param RAMRaster Output raster.
     * @param params Raster parameters.
     * @param mempol Memory policy.
     */      
    bool TeAllocRAMRaster( TePDITypes::TePDIRasterPtrType& RAMRaster,
      const TeRasterParams& params, 
      TePDIUtilsMemPol mempol = TePDIUtilsAutoMemPol );      
      

    /**
     * Save the raster to a GeoTIFF file.
     *
     * @param in_raster Input raster.
     * @param file_name Output file name.
     * @param pixel_type The type of pixels.
     * @return true if OK. false on error.
     */
    bool TeRaster2Geotiff(
      TePDITypes::TePDIRasterPtrType& in_raster,
      const std::string& file_name,
      TeDataType pixel_type = TeUNSIGNEDCHAR );

    /**
     * Do pixel by pixel copy from one raster to another.
     *
     * @param source_raster Input raster.
     * @param source_band Souce band.
     * @param target_raster Output raster.
     * @param target_band Target band.
     * @param discretize_levels Levels will be rounded to the near integer.
     * @param trunc_levels Levels outside a range ( from zero
     * to the maximum Output raster allowed pixel value ) will be rounded to 
     * zero or the maximum Output raster allowed pixel value.
     * @return true if OK. false on error.
     */
    bool TeCopyRasterPixels(
      TePDITypes::TePDIRasterPtrType& source_raster,
      unsigned int source_band,
      TePDITypes::TePDIRasterPtrType& target_raster,
      unsigned int target_band,
      bool discretize_levels,
      bool trunc_levels );

    /**
     * Do pixel by pixel copy from one raster to another ( all bands ).
     *
     * @param source_raster Input raster.
     * @param target_raster Output raster.
     * @param discretize_levels Levels will be rounded to the near integer.
     * @param trunc_levels Levels outside a range ( from zero
     * to the maximum Output raster allowed pixel value ) will be rounded to 
     * zero or the maximum Output raster allowed pixel value.
     * @return true if OK. false on error.
     */
    bool TeCopyRasterBands(
      TePDITypes::TePDIRasterPtrType& source_raster,
      TePDITypes::TePDIRasterPtrType& target_raster,
      bool discretize_levels,
      bool trunc_levels );

    /**
     * Used to find the allowed maximum and minimum level values
     * that the supplied TeRaster band can have.
     *
     * @param raster Input raster.
     * @param band Band number.
     * @param min Minimum allowed value.
     * @param max Maximum allowed value..
     * @return true if OK, false on error.
     */
    bool TeGetRasterMinMaxBounds( TePDITypes::TePDIRasterPtrType& raster,
      unsigned int band, double& min, double& max );
      
    /**
     * Used to find the raster required memory.
     *
     * @param params Input raster parameters.
     * @return The amount of required memory.
     */
    unsigned long int TeGetRasterReqMem( TeRasterParams& params );      

    /**
     * Compares two rasters.
     *
     * @param raster1 Input raster 1.
     * @param raster2 Input raster 2.
     * @return true if all the pixels from both rasters have the same levels,
     * false if not.
     */
    bool rasterCompare( TePDITypes::TePDIRasterPtrType& raster1,
      TePDITypes::TePDIRasterPtrType& raster2 );
      
    /**
     * Given a raster band, verifies if the pixels are floating point
     * pixels.
     *
     * @param raster The input raster.
     * @param band Band number.
     * @return true if the the pixels are floating point pixels.
     */      
    bool IsFloatBand( const TePDITypes::TePDIRasterPtrType& raster, 
      unsigned int band );
    
    /**
     * Convert a TeCoord2D with projected coordinates to a geometry with
     * indexed coordinates ( following the raster reference of lines and
     * columns ).
     *
     * @param g_in Input geometry.
     * @param raster Reference raster.
     * @param g_out Output geometry.
     */       
    void MapCoords2RasterIndexes( const TeCoord2D& g_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeCoord2D& g_out );
      
    /**
     * Convert a TeBox with projected coordinates to a TeBox with
     * indexed coordinates ( following the raster reference of lines and
     * columns ).
     *
     * @param box_in Input box.
     * @param raster Reference raster.
     * @param box_out Output box.
     */       
    void MapCoords2RasterIndexes( const TeBox& box_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeBox& box_out );      
      
   /**
     * Convert a geometry with projected coordinates to a geometry with
     * indexed coordinates ( following the raster reference of lines and
     * columns ).
     *
     * @param g_in Input geometry.
     * @param raster Reference raster.
     * @param g_out Output geometry.
     */        
    template< typename Geometry >
    void MapCoords2RasterIndexes( const Geometry& g_in, 
      const TePDITypes::TePDIRasterPtrType& raster, Geometry& g_out )
    {
        unsigned int nItens = (unsigned int)g_in.size();
        
        for (unsigned int i=0; i < nItens ; i++ ) {
                typename Geometry::value_type e1 = g_in[i];
                typename Geometry::value_type e2;
                
                MapCoords2RasterIndexes( e1, raster, e2 );
                
                g_out.add( e2 );
        }
    }
    
    /**
     * Convert TeCoord2D with matrix indexed coordinates to a TeCoord2D with 
     * projected coordinates folowing a raster reference.
     *
     * @param g_in Input geometry.
     * @param raster Reference raster.
     * @param g_out Output geometry.
     */       
    void MapRasterIndexes2Coords( const TeCoord2D& g_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeCoord2D& g_out );     
      
    /**
     * Convert TeBox with matrix indexed coordinates to a TeBox with 
     * projected coordinates folowing a raster reference.
     *
     * @param g_in Input geometry.
     * @param raster Reference raster.
     * @param g_out Output geometry.
     */       
    void MapRasterIndexes2Coords( const TeBox& g_in, 
      const TePDITypes::TePDIRasterPtrType& raster, TeBox& g_out );       
    
   /**
     * Convert geometry with matrix indexed coordinates to a geometry with 
     * projected coordinates folowing a raster reference.
     *
     * @param g_in Input geometry.
     * @param raster Reference raster.
     * @param g_out Output geometry.
     */        
    template< typename Geometry >
    void MapRasterIndexes2Coords( const Geometry& g_in, 
      const TePDITypes::TePDIRasterPtrType& raster, Geometry& g_out )
    {
        unsigned int nItens = (unsigned int)g_in.size();
        
        for (unsigned int i=0; i < nItens ; i++ ) {
                typename Geometry::value_type e1 = g_in[i];
                typename Geometry::value_type e2;
                
                MapRasterIndexes2Coords( e1, raster, e2 );
                
                g_out.add( e2 );
        }
    }
        
    /**
     * Virifies if the lut has only discrete reference values ( integer ).
     *
     * @param lut Lut.
     * @return true if only discrete values were found, otherwise false
     * will be returned.
     */
    bool IsDiscreteLut( const TePDITypes::TePDILutType& lut );
    
    /**
     * Find the best decimation levels bringging the two rasters to
     * closer sizes.
     *
     * @param raster1 Input raster1.
     * @param raster2 Input raster2.
     * @param r1_level Raster 1 best decomposition level.
     * @param r2_level Raster 2 best decomposition level.
     * @param decim_error The error ( pixels number ) for the best level found.
     * @return true if OK. false on errors.
     */
    bool FindBestDecimLevels( 
      const TePDITypes::TePDIRasterPtrType& raster1,
      const TePDITypes::TePDIRasterPtrType& raster2,
      unsigned long int& r1_level, unsigned long int& r2_level,
      unsigned long int& decim_error );
      
    /**
     * Builds a detailed box (world indexes) using all raster border pixels.
     *
     * @note If raster do not have projection matricial indexes will be 
     * returned.
     *
     * @param raster Input raster.
     * @param box_polygon The generated box (polygon representation).
     * @return true if OK, false on errors.
     */
    bool buildDetailedBox( const TePDITypes::TePDIRasterPtrType& raster,
      TePolygon& box_polygon );
      
    /**
     * Builds a detailed bounding box (world indexes) using all raster 
     * border pixels.
     *
     * @note If raster do not have projection matricial indexes will be 
     * returned.     
     *
     * @param raster Input raster.
     * @param box_polygon The generated box (polygon representation).
     * @return true if OK, false on errors.
     */
    bool buildDetailedBBox( const TePDITypes::TePDIRasterPtrType& raster,
      TePolygon& box_polygon );
      
    /**
     * Calculate the pixel size in geodetic units.
     *
     * @param params Raster params.
     * @param x_res Pixel size ( X Axis - Geodetic units ).
     * @param y_res Pixel size ( Y Axis - Geodetic units ).
     */
    void getGeodeticPixelRes( TeRasterParams& params,
      double& x_res, double& y_res );
      
    /**
     * Reproject a raster geometry.
     *
     * @param input_raster Input raster.
     * @param target_projection Target projection.
     * @param output_raster Output raster.
     *
     * @return true if OK, false on errors.
     */
    bool reprojectRaster( const TePDITypes::TePDIRasterPtrType& input_raster,
      TeProjection& target_projection,
      TePDITypes::TePDIRasterPtrType& output_raster );
      
    /**
     * Tells the relation between a point and a polygon by using a 
     * polygon tile indexer.
     *
     * @note The boundary case is not checked.
     *
     * @param c The input point.
     * @param indexer The polygon tile indexer.
     *
     * @return TeINSIDE or TeOUTSIDE.
     */      
    short TeRelation( const TeCoord2D& c, const TeTileIndexer& indexer );
    
  };
  
/** @example TePDIUtils_test.cpp
 *    Shows how to use this namespace functions.
 */    

#endif //TEPDIUTILS_HPP
