/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2006 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeCommunicator.h
    \brief This file contains  class to deal with inter-class communication.
	\author Emiliano F. Castejon <castejon@dpi.inpe.br>
*/

#ifndef TECOMMUNICATOR_H
  #define TECOMMUNICATOR_H

  #include <list>
  #include <algorithm>
  #include <iostream>
  
  /**
   * @brief A macro for use when declaring class wrapper functions that will
   * be connected to a TeCommunicator instance.
   *
   * @param function_name The function name.
   * @param param_type The function pamrameter data type.
   * @param class_name The function's host class name.
   * @note The function body must be implemented.
   */
  #define TECOMMWRAPPER( function_name, param_type, class_name ) \
    protected : void function_name( const param_type& x ); \
    public : static void function_name( const param_type& x, void* objptr ) \
    { \
      class_name * casted_ptr = ( class_name * ) objptr; \
      casted_ptr->function_name( x ); \
    };
    
  /**
   * @brief For internal use only. DO NOT USE THIS MACRO.
   */    
  #define TECOM_TRUE_OR_THROW( value , message ) \
    if( value == 0 ) { \
      std::cerr << std::endl << "Error : " \
        << __FILE__ \
        << ":" << __LINE__ \
        << " - " << std::string( message ) \
        << std::endl; \
      throw; \
    };

  /**
   * @brief For internal use only. DO NOT USE THIS MACRO.
   */    
  #define TECOM_TRUE_OR_RETURN( value , message ) \
    if( value == 0 ) { \
      std::cout << std::endl << "Warning : " \
        << __FILE__ \
        << ":" << __LINE__ \
        << " - " << std::string( message ) \
        << std::endl; \
      return false; \
    };      

  /**
   * @brief This is the template class to deal with inter-class communication.
   * @author Emiliano F. Castejon <castejon@dpi.inpe.br>
   * @ingroup Utils
   */
  template< typename DataType >
  class TeCommunicator 
  {
    public :
    
      /**
       * @brief The objects wrapper function pointer.
       *
       * @param DataType The data type.
       * @param void A cast pointer to the active object instance.
       */        
      typedef void (*WrapperFuncPtrT)( const DataType&, void* );
            
      /**
       * @brief Default Constructor.
       *
       * @param host_obj_ptr A pointer to the active class instance.
       * @param wrapper_func_ptr A pointer to the class wrapper static
       * function.
       * @param enabled_flag The emission/reception enabling/disabling flag.
       */
      TeCommunicator( void* host_obj_ptr = 0,
        WrapperFuncPtrT wrapper_func_ptr = 0,
        bool enabled_flag = true );

      /**
       * @brief Default Destructor
       */
      ~TeCommunicator();
      
      /**
       * @brief Connects the current instance to a class wrapper function.
       *
       * @param host_obj_ptr A pointer to the active class instance.
       * @param wrapper_func_ptr A pointer to the class wrapper static
       * function.
       */      
      void setHostObj( void* host_obj_ptr = 0, 
        WrapperFuncPtrT wrapper_func_ptr = 0 );
      
      /**
       * @brief Connects the current instance to another active instance.
       *
       * @param external The external active instance reference.
       * @return true if OK, false on errors.
       */         
      bool connect( TeCommunicator< DataType>& external );
      
      /**
       * @brief Disconnects the current instance from all connected instances.
       */         
      void disconnect();
      
      /**
       * @brief Disconnects the current instance from another instance.
       *
       * @param external The external active instance reference.
       */        
      void disconnect( TeCommunicator< DataType>& external );
      
      /**
       * @brief Send data to all connected instances.
       *
       * @param data The data to be sent.
       */        
      void send( const DataType& data ) const;
      
      /**
       * @brief Enables/disables the data emission and rececption.
       *
       * @param enabled_flag The emission/reception enabling/disabling flag.
       */        
      void toggle( bool enabled_flag );      
      
    protected :
      
	  //! Communicator pointer type definition
      typedef TeCommunicator< DataType >* ComPtrT;

	  //! Connection list type definition
      typedef std::list< ComPtrT > ConListT;

      //! Connection list iterator type definition
      typedef typename ConListT::iterator ConListItT;      
    
      /**
       * @brief Used when another active instance requests to be connected to this
       * instance.
       *
       * @param my_pointer The pointer of the active instance.
       */       
      void ConnectMe( ComPtrT my_pointer );
      
      /**
       * @brief Used when another active instance requests to be disconnected from 
       * this instance.
       *
       * @param my_pointer The pointer of the active instance.
       */       
      void DisconnectMe( ComPtrT my_pointer );
      
      /**
       * @brief Used when another active instance sends data to this instance.
       *
       * @param data The sent data.
       */        
      inline void Receive( const DataType& data ) const; 
      
    private :
    
      bool emission_reception_enabled_;		//!< A flag indication for enabled/disabled.
    
      WrapperFuncPtrT wrapper_func_ptr_;	//!< A pointer to the host object wrapper function.
      
      void* host_obj_ptr_;	//!< A pointer to the host object active instance.
      
      mutable ConListT connections_;	//!< A list of pointers to the connected active instances.
      
      /**
       * @brief Copy constructor not allowed.
       */
      TeCommunicator( const TeCommunicator< DataType >& ) {};
      
      /**
       * @brief Copy operation not allowed.
       */
      const TeCommunicator< DataType >& operator=( 
        const TeCommunicator< DataType >& ) {};
      
  };
  

  template< typename DataType >  
  TeCommunicator< DataType >::TeCommunicator( void* host_obj_ptr,
    WrapperFuncPtrT wrapper_func_ptr, bool enabled_flag )
  {
    wrapper_func_ptr_ = 0;
    host_obj_ptr_ = 0;
    emission_reception_enabled_ = enabled_flag;
    
    setHostObj( host_obj_ptr, wrapper_func_ptr );
  }
  
  
  template< typename DataType >  
  TeCommunicator< DataType >::~TeCommunicator()
  {
    disconnect();
  }
  
  
  template< typename DataType >  
  void TeCommunicator< DataType >::setHostObj( void* host_obj_ptr,
    WrapperFuncPtrT wrapper_func_ptr )
  {
    TECOM_TRUE_OR_THROW( ( 
      ( ( wrapper_func_ptr == 0 ) && ( host_obj_ptr == 0 ) ) || 
      ( ( wrapper_func_ptr != 0 ) && ( host_obj_ptr != 0 ) ) ), 
      "Invalid receiver_func_ptr/host_obj_ptr parameters" );
      
    wrapper_func_ptr_ = wrapper_func_ptr;
    host_obj_ptr_ = host_obj_ptr;
  }
  
  
  template< typename DataType >
  bool TeCommunicator< DataType >::connect( 
    TeCommunicator< DataType>& external )
  {
    TECOM_TRUE_OR_RETURN( 
      ( ( wrapper_func_ptr_ != 0 ) && ( host_obj_ptr_ != 0 ) ),
      "Connecting a inactive communicator instance" );
    
    ConListItT it = find( connections_.begin(), connections_.end(),
      &external );

    if( it == connections_.end() ) {
      connections_.push_back( &external );
      external.ConnectMe( this );
    }
    
    return true;
  }
  

  template< typename DataType >
  void TeCommunicator< DataType >::disconnect()
  {
    ConListItT it = connections_.begin();
    ConListItT it_end = connections_.end();

    while( it != it_end ) {
      (*it)->DisconnectMe( this );
      
      ++it;
    }
    
    connections_.clear();
  }
  
  
  template< typename DataType >
  void TeCommunicator< DataType >::disconnect( 
    TeCommunicator< DataType>& external )
  {
    ConListItT it = find( connections_.begin(), connections_.end(),
      &external );
      
    if( it != connections_.end() ) {
      (*it)->DisconnectMe( this );
      connections_.erase( it );
    }
  }

  
  template< typename DataType >
  void TeCommunicator< DataType >::ConnectMe( 
    ComPtrT my_pointer )
  {
    ConListItT it = find( connections_.begin(), connections_.end(),
      my_pointer );
      
    if( it == connections_.end() ) {
      connections_.push_back( my_pointer );
    }
  }

  
  template< typename DataType >
  void TeCommunicator< DataType >::DisconnectMe( 
    ComPtrT my_pointer )
  {
    ConListItT it = find( connections_.begin(), connections_.end(),
      my_pointer );
      
    if( it != connections_.end() ) {
      connections_.erase( it );
    }
  }
  
  
  template< typename DataType >
  void  TeCommunicator< DataType >::send( const DataType& data ) const
  {
    if( emission_reception_enabled_ ) {
      ConListItT it = connections_.begin();
      ConListItT it_end = connections_.end();
      
      while( it != it_end ) {
        (*it)->Receive( data );
        
        ++it;  
      }
    }
  }
  
  
  template< typename DataType >
  void  TeCommunicator< DataType >::toggle( bool enabled_flag )
  {
    emission_reception_enabled_ = enabled_flag;
  }

  
  template< typename DataType >
  inline void  TeCommunicator< DataType >::Receive( 
    const DataType& data ) const
  {
    if( emission_reception_enabled_ && ( wrapper_func_ptr_ != 0 ) ) {
      wrapper_func_ptr_( data, host_obj_ptr_ );
    }
  }
  
/** @example TeCommunicator_test.cpp
 *    Shows how to use this class.
 */  
    
#endif

