/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/
/*! \file TeDecoderMemory.h
    \brief This file deals with decoding of raster structures as a mulitdimensional matrix in memory
*/
#ifndef  __TERRALIB_INTERNAL_DECODERMEMORY_H
#define  __TERRALIB_INTERNAL_DECODERMEMORY_H

#include <TeDecoder.h>

//! Implements a decoder to a raster stored as a as a mulitdimensional matrix in memory
class TeDecoderMemory : public TeDecoder
{
public:

	//! Allocates memory.
	bool allocateMemory();

	//! Reset memory with dummy values.
	bool resetMemory();

public:

	//! Empty constructor.
	TeDecoderMemory ( ) : 
		data_(0), 
		isModified_(false) 
		{ params_.decoderIdentifier_ = "MEM";};

	//! Constructor from parameters.
	TeDecoderMemory ( const TeRasterParams& par ) : 
		data_(0),
		isModified_(false)
	{	
		params_ = par; 
		params_.decoderIdentifier_ = "MEM"; 
	};

	//! Destructor.
	virtual ~TeDecoderMemory ();

	//! Sets the value of a specific raster element.
	/*!
		\param col	element column identifier
		\param lin	element line identifier
		\param val	element value being inserted
		\param band element band identifier
	*/
	bool setElement (int col,int lin, double val, int band=0);

	//! Gets an specific element (col, lin, band) of a raster data.
	/*!
		\param col	element column identifier.
		\param lin	element line identifier.
		\param val	element value being retrieved.
		\param band element band identifier.
	*/
	bool getElement (int col,int lin, double &val,int band=0);

	//! Retuns a pointer to a given band.
	void* data (int band=0);

	//! Initializes the internal structures of the decoder.
	virtual void	init	();

	//! Clears its internal structures.
	virtual bool	clear	();

	//! Sets isModified flag.
	void setModified(bool val)
	{	isModified_ = val; }

protected:
	int	nelems_;			//!< Number of elements (bands x columns x lines).
	int	size_;				//!< Memory size.
	void*	data_;			//!< Pointer to view of file mapped to memory.
	bool isModified_;		//!< A flag that indicates if a raster element was modified.
};

//! Implements a factory to build decoder to MEMORY raster.
class TeDecoderMemoryFactory : public TeDecoderFactory
{
public:

	//! Constructor.
	TeDecoderMemoryFactory(const string& name) : TeDecoderFactory(name) {}

	//! Builds a Memory decoder.
	virtual TeDecoder* build (const TeRasterParams& arg)
	{  return new TeDecoderMemory(arg); }
};
#endif

