/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

#include <TeDecoderTIFF.h>
#include <TeUtils.h>

#include <fstream>
using std::ifstream;
using std::ofstream;

// ---- Public Methods ----- 

TeDecoderTIFF::TeDecoderTIFF(const TeRasterParams& par)
{
	params_ = par;
	TImage = -1;
	TCurStrip_ = -1;
	TCurLine_ = -1;
	first = 1;
	TBuffer_ = 0;
	TBufferTile_ = 0;
	nBands_ = par.nBands();
	params_.status_= TeRasterParams::TeNotReady;
	flushStrip_ = false;

	bytesperline_ = 0;
	rowtilesize_  = 0;
	tilew_ = 0;
	tileh_ = 0;
	bytespertile_ = 0;

	tif = 0;
	gtif = 0;
	
	char* tifmode = "r";
	if (params_.mode_=='c')
	{
		tifmode = "a";
	}
	else if (params_.mode_ == 'r')	// file should exists
	{								// read tiff/geotiff parameters
		tif=XTIFFOpen(params_.fileName_.c_str(),tifmode);
		if (!tif)
		{
			return;
		}
		gtif = (GTIF*)GTIFNew(tif);
		ReadTiffDirectory();
		if (gtif)
		{
			GTIFFree(gtif);
			gtif = 0;
		}
		if (tif)
		{
			XTIFFClose(tif); 
			tif = 0;
		}
		params_.decoderIdentifier_ = "TIF";
	}
}

void
TeDecoderTIFF::init()
{
	clear();		// reset existent internal structures
	flushStrip_ = false;

	params_.status_= TeRasterParams::TeNotReady;
	char* tifmode;	// libtiff mode

	// check the initial permissions on the file
	if (params_.mode_ == 'c')		// creating a new raster file
	{
		ofstream dataFile(params_.fileName_.c_str()); // try to (re)create the file
		if (!dataFile)
		{
			params_.errorMessage_ = "Fail to (re)create the raster the file:" + params_.fileName_;
			return;
		}
		dataFile.close();
		tifmode = "a";
		params_.status_ = TeRasterParams::TeReadyToWrite;
	}
	else if ( params_.mode_ == 'r' )
	{
		ifstream dataFile(params_.fileName_.c_str()); // check if file exists
		if (!dataFile)
		{
			params_.errorMessage_ = "Fail to open the raster the file:" + params_.fileName_;
			return; 
		}
		dataFile.close();				
		tifmode = "r";
		params_.status_ = TeRasterParams::TeReadyToRead;
	}
	// TIFF image files may not be opened for both reading and writing; 
	// there is no support for altering the contents of a TIFF file. So
	// mode == 'w' never initializes this decoder.
	else if (params_.mode_ == 'w')
	{
		params_.status_= TeRasterParams::TeNotReady;
		params_.errorMessage_ = "There is no support for altering the contents of a TIFF file.";
		return;
	}
	else	// unknown mode
		return;

	// try to open the tiff file with the desired mode
	tif=XTIFFOpen(params_.fileName_.c_str(),tifmode);
	if (!tif)
	{
		params_.errorMessage_ = "Fail to instantiate a tiff file.";
		params_.status_= TeRasterParams::TeNotReady;
		return;
	}

	// try to create a geotiff instance on the tiff file 
	gtif = (GTIF*)GTIFNew(tif);
	if (!gtif)
	{
		params_.errorMessage_ = "Fail to instantiate a geotiff file.";
		params_.status_= TeRasterParams::TeNotReady;
		return;
	}

	if (params_.mode_ == 'c')	// if geotiff is being created we should set the geo flags
		SetGeoKeys();
	
	ReadTiffDirectory();			// read the tiff geokeys
	AllocateBuffer();				// allocate the internal buffers
	if (params_.mode_ == 'c' && 	// if a new image writes creates a dummy image
		!WriteDummyImage())			// if couldnt write dummy image there is an error
	{
		params_.errorMessage_ = "Fail to create the (geo)tiff file.";
		params_.status_= TeRasterParams::TeNotReady;		 
		return;
	}
}


bool 
TeDecoderTIFF::clear()
{
	params_.errorMessage_.clear();
	if (flushStrip_)	//  if we have been 
	{													//	writting save the last 
		tif->tif_curoff = 0;
		TIFFDirectory *td = &tif->tif_dir;
		td->td_stripoffset[TCurStrip_] = 0;
		TIFFWriteStrip(tif,TCurStrip_,TBuffer_,TIFFStripSize(tif));
		flushStrip_= false;
	}
	if (gtif)
		GTIFFree(gtif);
	if (tif)
		XTIFFClose(tif); 

	tif = 0;
	gtif = 0;
	first = 0;
	TImage = -1;
	TCurStrip_ = -1;
	TCurLine_ = -1;
	bytesperline_ = 0;
	rowtilesize_  = 0;
	tilew_ = 0;
	tileh_ = 0;
	bytespertile_ = 0;

	DeallocateBuffer();
	params_.status_ = TeRasterParams::TeNotReady;
	return true;
}

TeDecoderTIFF::~TeDecoderTIFF()
{
	clear();
}

bool
TeDecoderTIFF::getElement(int col,int lin, double &val,int band)
{
	if (flushStrip_)
	{
		tif->tif_curoff = 0;
		TIFFDirectory *td = &tif->tif_dir;
		td->td_stripoffset[TCurStrip_] = 0;
		TIFFWriteStrip(tif,TCurStrip_,TBuffer_,TIFFStripSize(tif));
		flushStrip_ = false;
	}
	if (TCurLine_ != (int)lin)		// if line is not the current strip in memory
	{
		if (TImage == MONOIMAGE || TImage == PALLETEIMAGE)
		{
			if (TIFFReadScanline(tif,TBuffer_,lin,0) < 0)
				return false;
		}
		else // RGB image
		{
			if (planar_ == PLANARCONFIG_CONTIG)
			{
				if (isTiled_)
				{
					if(!ReadTileImageContig(lin,TBuffer_))
						return false;
				}
				else
				{
					if (TIFFReadScanline(tif,TBuffer_,lin,0) < 0)
						return false;
				}
			}
			else	// RGB Separate
			{
				if (isTiled_)  return false; // SEPARATE and tiled => Error!
				for (int s=0; s < nBands_; s++)
					if (TIFFReadScanline(tif,&TBuffer_[s*params_.ncols_*params_.elementSize(s)],lin,s) < 0)
						return false;
			}
		}
		TCurLine_ = lin;
	}
	int pos;
	if (planar_ == PLANARCONFIG_CONTIG)
		pos = col*nBands_+band;
	else
		pos = band*params_.ncols_+col;
	if (params_.dataType_[band] == TeDOUBLE)			
		val = ((double*)TBuffer_)[pos];
	else if (params_.dataType_[band] == TeFLOAT)			
		val = ((float*)TBuffer_)[pos];
	else if (params_.dataType_[band] == TeUNSIGNEDLONG)			
		val = ((unsigned long*)TBuffer_)[pos];
	else if (params_.dataType_[band] == TeLONG)			
		val = ((long*)TBuffer_)[pos];
	else if (params_.dataType_[band] == TeINTEGER)			
		val = ((int*)TBuffer_)[pos];
	else if (params_.dataType_[band] == TeUNSIGNEDSHORT)			
		val = ((unsigned short*)TBuffer_)[pos];
	else if (params_.dataType_[band] == TeSHORT)			
		val = ((short*)TBuffer_)[pos];
	else if (params_.dataType_[band] == TeCHAR)			
		val = ((char*)TBuffer_)[pos];
	else 			
		val = TBuffer_[pos];
	return true;
}


// We allow writting only to a strip based, non compressed tiff file
bool
TeDecoderTIFF::setElement(int col, int lin, double val, int band)
{
	int strip = lin/rowsperstrip_;
	if (strip != TCurStrip_)
	{
		if (TCurStrip_ >= 0)
		{
			TIFFDirectory *td = &tif->tif_dir;
			td->td_stripoffset[TCurStrip_] = 0;
			tif->tif_curoff = 0;
			TIFFWriteStrip(tif,TCurStrip_,TBuffer_,TIFFStripSize(tif));
			flushStrip_ = false;
		}
		TCurStrip_ = strip;

		int l = strip * rowsperstrip_;
		bool res = false;
		if (TImage == MONOIMAGE || TImage == PALLETEIMAGE)
		{
			tif->tif_curstrip = 0;
			res = ReadMonoImage(l);
		}
		else if(TImage == RGBIMAGE)
		{
			tif->tif_curstrip = 0;
			res = ReadRGBImage (l);
		}
		if (!res)
		{
			TCurStrip_ = -1;
			return false;
		}
	}
	unsigned char* cp = TBuffer_ + (lin % rowsperstrip_ * bytesperline_);
	int pos = col*nBands_+band;
	if (params_.dataType_[band] == TeDOUBLE)			
		((double*)cp)[pos] = (double)val;
	else if (params_.dataType_[band] == TeFLOAT)			
		((float*)cp)[pos] = (float)val;
	else if (params_.dataType_[band] == TeUNSIGNEDLONG)			
		((unsigned long*)cp)[pos] = (unsigned long)val;
	else if (params_.dataType_[band] == TeLONG)			
		((long*)cp)[pos] = (long)val;
	else if (params_.dataType_[band] == TeINTEGER)			
		((int*)cp)[pos] = (int)val;
	else if (params_.dataType_[band] == TeUNSIGNEDSHORT)			
		((unsigned short*)cp)[pos] = (unsigned short)val;
	else if (params_.dataType_[band] == TeSHORT)			
		((short*)cp)[pos] = (short)val;
	else if (params_.dataType_[band] == TeCHAR)			
		((char*)cp)[pos] = (char)val;
	else 		
		((unsigned char*)cp)[pos] = (unsigned char)val;
	flushStrip_ = true;
	return true;
}


// --- Internal Methods ---

// --- Methods that deal with Tiff/GeoTiff Flags

void TeDecoderTIFF :: SetGeoKeys()
{

// Image Size
	TIFFSetField(gtif->gt_tif, TIFFTAG_IMAGEWIDTH, params_.ncols_);
	TIFFSetField(gtif->gt_tif, TIFFTAG_IMAGELENGTH, params_.nlines_);

// Orientation
	TIFFSetField(gtif->gt_tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);

// Number of bits
	int nb = params_.nbitsperPixel_[0];
	TIFFSetField(gtif->gt_tif, TIFFTAG_BITSPERSAMPLE, nb);
	switch (params_.dataType_[0]) 
	{
		case TeUNSIGNEDCHAR:
		case TeUNSIGNEDSHORT:
		case TeUNSIGNEDLONG:
			TIFFSetField(gtif->gt_tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_UINT);
			break;
		case TeCHAR:
		case TeSHORT:
		case TeINTEGER:
		case TeLONG:
			TIFFSetField(gtif->gt_tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_INT);
			break;
		case TeFLOAT:
		case TeDOUBLE:
			TIFFSetField(gtif->gt_tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
			break;
		default:
			TIFFSetField(gtif->gt_tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_VOID);
			break;
	}
// Number of bands
	TIFFSetField(gtif->gt_tif, TIFFTAG_SAMPLESPERPIXEL, nBands_);

// Compression
	TIFFSetField(gtif->gt_tif, TIFFTAG_COMPRESSION, COMPRESSION_NONE);

// Planar configuration
	TIFFSetField(tif,TIFFTAG_PLANARCONFIG,  PLANARCONFIG_CONTIG);

// Photometric type
	if (nBands_ == 1) // monobands
	{
		if (params_.photometric_[0] == TeRasterParams::TePallete)
		{
			TIFFSetField(gtif->gt_tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
			SaveLut();
			TImage = PALLETEIMAGE;
		}
		else
		{
			TIFFSetField(gtif->gt_tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
			TImage = MONOIMAGE;
		}
	}
	else
	{
		TIFFSetField(gtif->gt_tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
		TImage = RGBIMAGE;
	}
	TIFFSetField(gtif->gt_tif, TIFFTAG_ROWSPERSTRIP, 1);


	if (params_.dxI_ != 0. || params_.dyJ_ != 0.) // there is a rotation defined
	{
		//	It seems that most viewers arent able to deal with GTIFF_TRANSMATRIX...
		// ... so let's write a navigation file instead...
		string tfw = TeGetName(params_.fileName_.c_str())+".tfw";
		FILE* fp = fopen(tfw.c_str(),"w");
		if (fp)
		{
			fprintf(fp,"%.6f\n",params_.dxJ_);
			fprintf(fp,"%.6f\n",params_.dyI_);
			fprintf(fp,"%.6f\n",params_.dxI_);
			fprintf(fp,"%.6f\n",params_.dyJ_);
			fprintf(fp,"%.6f\n",params_.x0_);
			fprintf(fp,"%.6f\n",params_.y0_);
		}
	}
	else // tf there is no rotation use one tie point and pixel scale
	{
		double tiepoint[6];
		tiepoint[0] =0;
		tiepoint[1] =0;
		tiepoint[2] =0;
		tiepoint[3] = params_.box().x1_;
		tiepoint[4] = params_.box().y2_;
		tiepoint[5] = 0;
		TIFFSetField(gtif->gt_tif, GTIFF_TIEPOINTS, 6,&tiepoint); 

		double resolution[3];
		resolution[0] = params_.resx_;
		resolution[1] = params_.resy_;
		resolution[2] = 0;
		TIFFSetField(gtif->gt_tif, GTIFF_PIXELSCALE, 3,&resolution);
	}

	// projection parameters
	GTIFKeySet(gtif, GTRasterTypeGeoKey, TYPE_SHORT, 1, RasterPixelIsArea);
	if (!params_.projection() || params_.projection()->name() == "NoProjection")
	{
		GTIFWriteKeys(gtif);				// write so far defined geo keys 
		// at least try to write a tfw file
		string tfw = TeGetName(params_.fileName_.c_str())+".tfw";
		FILE* fp = fopen(tfw.c_str(),"w");
		if (fp)
		{
			vector<double> tfw;
			params_.getNavigationParameters(tfw);
			for (unsigned int nl=0; nl<6; ++nl)
				fprintf(fp,"%8.4f\n",tfw[nl]);
		}
		return;
	}
	const TeProjectionParams& paramas = params_.projection()->params();
	const string& projname = params_.projection()->name();
	if (projname == "LatLong")
	{
		GTIFKeySet(gtif, GTModelTypeGeoKey, TYPE_SHORT, 1, ModelTypeGeographic);
		GTIFKeySet(gtif, GeogAngularUnitsGeoKey, TYPE_SHORT, 1, Angular_Degree);
	}
	else
	{
		GTIFKeySet(gtif, GTModelTypeGeoKey, TYPE_SHORT, 1, ModelTypeProjected);
		GTIFKeySet(gtif, ProjLinearUnitsGeoKey, TYPE_SHORT, 1,Linear_Meter);
		if (projname == "UTM")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_TransverseMercator);
			GTIFKeySet(gtif, ProjNatOriginLatGeoKey, TYPE_DOUBLE,1,paramas.lat0*TeCRD);
			GTIFKeySet(gtif, ProjNatOriginLongGeoKey, TYPE_DOUBLE,1,paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjScaleAtNatOriginGeoKey, TYPE_DOUBLE,1,paramas.scale);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE,1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey, TYPE_DOUBLE,1,paramas.offy);
		}
		else if (projname == "Mercator")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_Mercator);
			GTIFKeySet(gtif, ProjNatOriginLongGeoKey, TYPE_DOUBLE,1,paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjNatOriginLatGeoKey, TYPE_DOUBLE,1,paramas.stlat1*TeCRD);
			GTIFKeySet(gtif, ProjScaleAtNatOriginGeoKey, TYPE_DOUBLE,1,paramas.scale);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE,1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey, TYPE_DOUBLE,1,paramas.offy);
		}
		else if (projname == "LambertConformal")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_LambertConfConic_2SP);
			GTIFKeySet(gtif, ProjFalseOriginLatGeoKey,TYPE_DOUBLE, 1, paramas.lat0*TeCRD);
			GTIFKeySet(gtif, ProjFalseOriginLongGeoKey, TYPE_DOUBLE, 1, paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjStdParallel1GeoKey, TYPE_DOUBLE, 1, paramas.stlat1*TeCRD);
			GTIFKeySet(gtif, ProjStdParallel2GeoKey, TYPE_DOUBLE, 1, paramas.stlat2*TeCRD);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE, 1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey,TYPE_DOUBLE, 1,paramas.offy);
		}
		else if (projname == "Polyconic")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_Polyconic);
			GTIFKeySet(gtif, ProjNatOriginLatGeoKey,TYPE_DOUBLE, 1, paramas.lat0*TeCRD);
			GTIFKeySet(gtif, ProjNatOriginLongGeoKey, TYPE_DOUBLE, 1, paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE, 1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey,TYPE_DOUBLE, 1,paramas.offy);
			GTIFKeySet(gtif, ProjScaleAtNatOriginGeoKey, TYPE_DOUBLE,1,paramas.scale);
		}
		else if (projname == "CylindricalEquidistant")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_Equirectangular);
			GTIFKeySet(gtif, ProjCenterLatGeoKey,TYPE_DOUBLE, 1, paramas.stlat1*TeCRD);
			GTIFKeySet(gtif, ProjCenterLongGeoKey, TYPE_DOUBLE, 1, paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE, 1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey,TYPE_DOUBLE, 1,paramas.offy);
		}
		else if (projname == "PolarStereographic")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_PolarStereographic);
			GTIFKeySet(gtif, ProjNatOriginLatGeoKey, TYPE_DOUBLE,1,paramas.lat0*TeCRD);
			GTIFKeySet(gtif, ProjStraightVertPoleLongGeoKey, TYPE_DOUBLE,1,paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjScaleAtNatOriginGeoKey, TYPE_DOUBLE,1,paramas.scale);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE,1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey, TYPE_DOUBLE,1,paramas.offy);
		}
		else if (projname == "Albers")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_AlbersEqualArea);
			GTIFKeySet(gtif, ProjStdParallel1GeoKey, TYPE_DOUBLE, 1, paramas.stlat1*TeCRD);
			GTIFKeySet(gtif, ProjStdParallel2GeoKey, TYPE_DOUBLE, 1, paramas.stlat2*TeCRD);
			GTIFKeySet(gtif, ProjNatOriginLatGeoKey,TYPE_DOUBLE, 1, paramas.lat0*TeCRD);
			GTIFKeySet(gtif, ProjNatOriginLongGeoKey, TYPE_DOUBLE, 1, paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE, 1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey,TYPE_DOUBLE, 1,paramas.offy);
		}
		else if (projname == "Miller")
		{
			GTIFKeySet(gtif, ProjCoordTransGeoKey, TYPE_SHORT, 1, CT_MillerCylindrical);
			GTIFKeySet(gtif, ProjCenterLatGeoKey, TYPE_DOUBLE, 1, paramas.lat0*TeCRD);
			GTIFKeySet(gtif, ProjCenterLongGeoKey, TYPE_DOUBLE, 1, paramas.lon0*TeCRD);
			GTIFKeySet(gtif, ProjFalseEastingGeoKey, TYPE_DOUBLE, 1,paramas.offx);
			GTIFKeySet(gtif, ProjFalseNorthingGeoKey,TYPE_DOUBLE, 1,paramas.offy);
		}
	}

	// datum parameters
	if (paramas.datum.name() == "CorregoAlegre")
		GTIFKeySet(gtif, GeogGeodeticDatumGeoKey, TYPE_SHORT, 1,Datum_Corrego_Alegre);
	else if (paramas.datum.name() == "WGS84")
		GTIFKeySet(gtif, GeogGeodeticDatumGeoKey, TYPE_SHORT, 1,Datum_WGS84);
	else if (paramas.datum.name() == "SAD69")
		GTIFKeySet(gtif, GeogGeodeticDatumGeoKey, TYPE_SHORT, 1,Datum_South_American_Datum_1969);
	else if (paramas.datum.name() == "Aratu")
		GTIFKeySet(gtif, GeogGeodeticDatumGeoKey, TYPE_SHORT, 1,Datum_Aratu);
	else
		GTIFKeySet(gtif, GeogGeodeticDatumGeoKey, TYPE_SHORT, 1,32767);

	GTIFKeySet(gtif, GeogCitationGeoKey, TYPE_ASCII,1,paramas.datum.name().c_str());
	GTIFKeySet(gtif, GeogSemiMajorAxisGeoKey,TYPE_DOUBLE, 1, paramas.datum.radius());
	GTIFKeySet(gtif, GeogSemiMinorAxisGeoKey,TYPE_DOUBLE, 1, paramas.datum.radius()*(1-paramas.datum.flattening()));
	GTIFWriteKeys(gtif);
}

void
TeDecoderTIFF::ReadTiffDirectory()
{
	int ncols=0, nlins = 0;
	TIFFGetField(tif,TIFFTAG_IMAGEWIDTH,&ncols);
	TIFFGetField(tif,TIFFTAG_IMAGELENGTH,&nlins);
	params_.ncols_ = ncols;
	params_.nlines_ = nlins;
		
	short samplesperPixel;
	if (!TIFFGetField(tif,TIFFTAG_SAMPLESPERPIXEL,&samplesperPixel) )
		params_.nBands(1);
	else
		params_.nBands(samplesperPixel);
	nBands_ = params_.nBands();

	nbitsperpixel_ = 0;
	if( !TIFFGetField(tif,TIFFTAG_BITSPERSAMPLE,&nbitsperpixel_) )
	{
		nbitsperpixel_ = 1;
		params_.setDataType(TeBIT);
	}
	else
	{
		short sampleformat;
		if (!TIFFGetField(tif,TIFFTAG_SAMPLEFORMAT,&sampleformat))
			sampleformat = 0;

		if (nbitsperpixel_ == 8)
		{
			if (sampleformat == SAMPLEFORMAT_INT)
				params_.setDataType(TeCHAR);
			else 
				params_.setDataType(TeUNSIGNEDCHAR);
		}
		else if (nbitsperpixel_ == 16)
		{
			if (sampleformat == SAMPLEFORMAT_INT)
				params_.setDataType(TeSHORT);
			else 
				params_.setDataType(TeUNSIGNEDSHORT);
		}
		else if (nbitsperpixel_ == 32)
		{
			if (sampleformat == SAMPLEFORMAT_INT)
				params_.setDataType(TeLONG);
			else if (sampleformat == SAMPLEFORMAT_UINT)
				params_.setDataType(TeUNSIGNEDLONG);
			else
				params_.setDataType(TeFLOAT);		
		}
		else 
			params_.setDataType(TeDOUBLE);
	}

	isTiled_ = (TIFFIsTiled(tif) != 0);			// return non-zero if is organized in tiles 
	if (isTiled_)
	{
		rowtilesize_= TIFFTileRowSize(tif);
		TIFFGetField(tif,TIFFTAG_TILEWIDTH,&tilew_);
		TIFFGetField(tif,TIFFTAG_TILELENGTH,&tileh_);
		bytespertile_ = TIFFTileSize(tif);
	}
	else
		TIFFGetField(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip_);


	bytesperline_ = TIFFScanlineSize(tif);

	TIFFGetField(tif,TIFFTAG_PLANARCONFIG,&planar_);	// storage organization
	TIFFGetField(tif,TIFFTAG_PHOTOMETRIC,&photom_);     // photometric interpretation

	if (samplesperPixel == 3 && 
	   (photom_ == PHOTOMETRIC_MINISWHITE || 
	    photom_ == PHOTOMETRIC_MINISBLACK))
		photom_ = PHOTOMETRIC_RGB;
	
	switch (photom_)
	{
	case PHOTOMETRIC_PALETTE:
		params_.setPhotometric(TeRasterParams::TePallete);
		TImage = PALLETEIMAGE;
		ReadLut();
		break;
	case PHOTOMETRIC_MINISWHITE:
	case PHOTOMETRIC_MINISBLACK:
		params_.setPhotometric(TeRasterParams::TeMultiBand);
		TImage = MONOIMAGE;
		break;
	case PHOTOMETRIC_RGB:
		params_.setPhotometric(TeRasterParams::TeRGB);
		TImage = RGBIMAGE;
		break;
	default:
		break;
	}

	TIFFGetField(tif, TIFFTAG_COMPRESSION, &compress_);
	if (compress_ != COMPRESSION_NONE)
		params_.setCompressionMode(TeRasterParams::TeTiffCompression);

	// Read georeference parameters
	GetGeoTIFF();
	if (!params_.box().isValid())
		params_.lowerLeftResolutionSize(0.5,0.5,1,1,params_.ncols_,params_.nlines_);
}

bool
TeDecoderTIFF::GetGeoTIFF()
{
	if (!gtif) 
		return false;

	params_.resx_ = params_.resy_ = 1;

	//---------------- navigation parameters 
	bool isGeoreferenced_ = false;
	if (gtif->gt_num_keys > 0)
	{
		// The three tags defined below may be used for defining the relationship between R (raster space)
		// and M (model space), and the relationship may be diagrammed as: 
		//           ModelPixelScaleTag 
		//           ModelTiepointTag         
		//    R  ------------ OR --------------> M
		//  (I,J,K)  ModelTransformationTag   (X,Y,Z)

		double *tiepoints   = 0;
		double *pixel_scale = 0;
		double *transform   = 0;

		int count=0, tiepoint_count=0, transform_count=0;

		if (!(gtif->gt_methods.get)(tif, GTIFF_TIEPOINTS, &tiepoint_count, &tiepoints ))
			tiepoint_count = 0;
		if (!(gtif->gt_methods.get)(tif, GTIFF_PIXELSCALE, &count, &pixel_scale ))
			count = 0;
		if (!(gtif->gt_methods.get)(tif, GTIFF_TRANSMATRIX, &transform_count, &transform ))
			transform_count = 0;

		if (count >= 3 && tiepoint_count >= 6) // if we have 1 tie point and pixel scale
		{
			params_.topLeftResolutionSize(tiepoints[3]-tiepoints[0]*pixel_scale[0],
			                              tiepoints[4]-tiepoints[1]*pixel_scale[1],
										  pixel_scale[0],pixel_scale[1],
										  params_.ncols_,params_.nlines_);
			isGeoreferenced_ = true;
		}
		else if (transform_count == 16) // if we have a transformation matrix
		{
			vector<double> wf;
			wf.push_back(transform[0]);
			wf.push_back(transform[1]);
			wf.push_back(transform[4]);
			wf.push_back(transform[5]);
			wf.push_back(transform[3]);
			wf.push_back(transform[7]);
			params_.setNavigationParameters(wf);
			isGeoreferenced_ = true;

		}
		// else if ( tiepoint_count > 6)
		// {
		//		we should use the set of tiepoints to interpolate... not implemented yet...
		// }
	}

	// if we couldn't get the georrefencing information from the tags
	// try to get if from an external (.tfw) world file
	if (!isGeoreferenced_) 
	{
		int dotpos = params_.fileName_.rfind('.');
		if (dotpos != std::string::npos)
		{
			vector<double> wf;
			wf.resize(6);
			string tfwfile = params_.fileName_.substr(0,dotpos+1) + "tfw";
			FILE* fp = fopen(tfwfile.c_str(),"r");
			if (!fp)
			{
				tfwfile = params_.fileName_.substr(0,dotpos+1) + "TFW";
				fp = fopen(tfwfile.c_str(),"r");
			}

			if (fp)
			{
				isGeoreferenced_ = true;
				char val[40];
				for (int ii=0; ii<6; ++ii)
				{
					if (fscanf(fp,"%s",val))
						wf[ii] = TeRoundD(atof(val)); 
					else
					{
						isGeoreferenced_ = false;
						break;
					}
				}
				fclose(fp);
			}
			if (isGeoreferenced_)
				params_.setNavigationParameters(wf);
		}
	}

	//---------------- projection parameters 
	GTIFDefn defn;        
	if (!GTIFGetDefn(gtif, &defn))
	{
		TeProjection* noproj = new TeNoProjection();
		params_.projection(noproj);
		delete noproj;
		return false;
	}
	else
	{
		// 1) Get Datum parameters
		TeDatum mDatum;
		if (defn.Datum == Datum_South_American_Datum_1969)
			mDatum = TeDatumFactory::make("SAD69");
		else if (defn.Datum == Datum_WGS84)
			mDatum = TeDatumFactory::make("WGS84");
		else if (defn.Datum == Datum_Corrego_Alegre)
			mDatum = TeDatumFactory::make("CorregoAlegre");
		else if (defn.Datum == Datum_Aratu)
			mDatum = TeDatumFactory::make("Aratu");
		else
		{
			double TISemiMajor = defn.SemiMajor, TISemiMinor = defn.SemiMinor;
			// no known datum... try to find a similar one based on SemiAxis values
			if (defn.SemiMajor == 0. && defn.SemiMinor == 0.)
				TISemiMajor =6.371000e+06;            // default parameters of a spherical ellipsoid
			if (!findDatum(TISemiMajor,(TISemiMajor - TISemiMinor)/TISemiMajor,mDatum))
				mDatum = TeDatum("UserDefined",TISemiMajor,(TISemiMajor - TISemiMinor)/TISemiMajor,0.,0.,0.);
		}

		// 2) Get projection parameters
		// if geographical model
		short mtype;
		GTIFKeyGet(gtif, GTModelTypeGeoKey,&mtype, 0, 1 );
		if (mtype == ModelTypeGeographic)	 
		{
			TeProjection* latlong = new TeLatLong(mDatum);
			params_.projection(latlong);
			delete latlong;
			return true;
		}		

		if (mtype != ModelTypeProjected)
		{
			TeProjection* noproj = new TeNoProjection();
			params_.projection(noproj);
			delete noproj;
			return false;
		}
		TeProjectionParams mProjPars;	
		mProjPars.units = "Meters"; // == not necessarily true... we should check for the units as defined in geotiff	
		mProjPars.datum = mDatum;

		// just set all parameters available from defn... some of them will not be used by some projections
		mProjPars.lat0 = defn.ProjParm[0]*defn.UOMAngleInDegrees*TeCDR;
		mProjPars.lon0 = defn.ProjParm[1]*defn.UOMAngleInDegrees*TeCDR;
		mProjPars.stlat1 = defn.ProjParm[2]*defn.UOMAngleInDegrees*TeCDR;
		mProjPars.stlat2 = defn.ProjParm[3]*defn.UOMAngleInDegrees*TeCDR;
		mProjPars.scale = defn.ProjParm[4];
		mProjPars.offx = defn.ProjParm[5];
		mProjPars.offy = defn.ProjParm[6];
		if (mProjPars.offy > 0)
			mProjPars.hemisphere = TeSOUTH_HEM;
		else
			mProjPars.hemisphere = TeNORTH_HEM;

		// now make the correspondence between geotiff projections and terralib projections
		switch (defn.CTProjection)
		{
			case CT_TransverseMercator : 
				mProjPars.name = "UTM";
			break;
			case CT_Mercator :
				mProjPars.name = "Mercator";
			break;
			case CT_LambertConfConic_2SP :
				mProjPars.name = "LambertConformal";
			break;
			case CT_Polyconic : 
				mProjPars.name = "Polyconic";
			break;
			case CT_Equirectangular :
				mProjPars.name = "CylindricalEquidistant";
			break;
			case CT_PolarStereographic :
				mProjPars.name = "PolarStereographic";
			break;
			case CT_AlbersEqualArea :
				mProjPars.name = "Albers";
			break;
			case CT_MillerCylindrical:
				mProjPars.name = "Miller";
			break;
			case CT_Sinusoidal:
				mProjPars.name = "TeSinusoidal";
			break;
			default:
				TeProjection* noproj = new TeNoProjection();
				params_.projection(noproj);
				delete noproj;
				return false;
			break;
		}
		TeProjection* proj = TeProjectionFactory::make(mProjPars);
		params_.projection(proj);
		delete proj;
		return true;
	}
}

// --- Methods that deal with the internal buffer to store the image in memory

bool 
TeDecoderTIFF::AllocateBuffer()
{
	unsigned long bufsize;
	int n;

	if (TBufferTile_)
	{
		delete []TBufferTile_;
		TBufferTile_ = 0;
	}
	if (TBuffer_)
	{
		delete []TBuffer_;
		TBuffer_ = 0;
	}

	if (isTiled_)		// we dont write tiled TIFF
	{					// let libtigg manage its raw data buffer
		// allocate a buffer to encode/decode a tile
		bufsize = bytespertile_;
		TBufferTile_ = new unsigned char[bytespertile_];
		if (!TBufferTile_)
			return false;
/*		n = TIFFReadBufferSetup(tif,TBufferTile_,bufsize);
		n = TIFFWriteBufferSetup(tif,TBufferTile_,bufsize);
*/
		// allocate a buffer to hold a line of image
		// a line may be formed by more than one tile
		TBuffer_ = new unsigned char[bytesperline_];
		if (!TBuffer_)
		{
			delete []TBufferTile_;
			TBufferTile_ = 0;
			return false;
		}
	}
	else
	{
		if (params_.mode_ == 'r')		// if we are reading let libtiff handle
		{								// the raw buffer and keeps a copy of each line
			bufsize = TIFFScanlineSize(tif) * params_.nBands();
			TBuffer_ = new unsigned char[bufsize];
			if (!TBuffer_)
				return false;
		}
		else							// if we are writting
		{								// we control the raw buffer 
			bufsize = TIFFStripSize(tif);
			TBuffer_ = new unsigned char[bufsize];
			if (!TBuffer_)
				return false;
			n = TIFFReadBufferSetup(tif,TBuffer_,bufsize);
			n = TIFFWriteBufferSetup(tif,TBuffer_,bufsize);
		}
	}
	return true;
}

void 
TeDecoderTIFF::DeallocateBuffer()
{
	if (TBuffer_)
		delete [] TBuffer_;
	TBuffer_ = 0;

	if (TBufferTile_)
		delete [] TBufferTile_;
	TBufferTile_ = 0;
}


// ---- Internal routines
bool 
TeDecoderTIFF::WriteDummyImage()
{
	long ssize = TIFFScanlineSize(gtif->gt_tif);
	unsigned char* obuf = (unsigned char*)_TIFFmalloc(ssize);
	unsigned char* pp;
	pp = obuf;
	if (TImage == RGBIMAGE)	// RGB image
	{
		for (int x = 0; x < params_.ncols_*3; x+=3) 
		{
			if (params_.dataType_[0] == TeDOUBLE)			
			{
				((double*)obuf)[x] = (double)params_.dummy_[0];
				((double*)obuf)[x+1] = (double)params_.dummy_[1];
				((double*)obuf)[x+2] = (double)params_.dummy_[2];
			}
			else if (params_.dataType_[0] == TeFLOAT)			
			{
				((float*)obuf)[x] = (float)params_.dummy_[0];
				((float*)obuf)[x+1] = (float)params_.dummy_[1];
				((float*)obuf)[x+2] = (float)params_.dummy_[2];
			}
			else if (params_.dataType_[0] == TeLONG)	
			{
				((long*)obuf)[x] = (long)params_.dummy_[0];
				((long*)obuf)[x+1] = (long)params_.dummy_[1];
				((long*)obuf)[x+2] = (long)params_.dummy_[2];
			}
			else if (params_.dataType_[0] == TeUNSIGNEDLONG)	
			{
				((unsigned long*)obuf)[x] = (unsigned long)params_.dummy_[0];
				((unsigned long*)obuf)[x+1] = (unsigned long)params_.dummy_[1];
				((unsigned long*)obuf)[x+2] = (unsigned long)params_.dummy_[2];
			}
			else if (params_.dataType_[0] == TeINTEGER)	
			{
				((int*)obuf)[x] = (int)params_.dummy_[0];
				((int*)obuf)[x+1] = (int)params_.dummy_[1];
				((int*)obuf)[x+2] = (int)params_.dummy_[2];
			}
			else if (params_.dataType_[0] == TeUNSIGNEDSHORT)	
			{
				((unsigned short*)obuf)[x] = (unsigned short)params_.dummy_[0];
				((unsigned short*)obuf)[x+1] = (unsigned short)params_.dummy_[1];
				((unsigned short*)obuf)[x+2] = (unsigned short)params_.dummy_[2];
			}
			else if (params_.dataType_[0] == TeSHORT)	
			{
				((short*)obuf)[x] = (short)params_.dummy_[0];
				((short*)obuf)[x+1] = (short)params_.dummy_[1];
				((short*)obuf)[x+2] = (short)params_.dummy_[2];
			}
			else if (params_.dataType_[0] == TeCHAR)											
			{
				obuf[x] = (char)params_.dummy_[0];
				obuf[x+1] = (char)params_.dummy_[1];
				obuf[x+2] = (char)params_.dummy_[2];
			}
			else											
			{
				obuf[x] = (unsigned char)params_.dummy_[0];
				obuf[x+1] = (unsigned char)params_.dummy_[1];
				obuf[x+2] = (unsigned char)params_.dummy_[2];
			}
		}
	}
	else		// Mono or pallete image
	{
		for (int x = 0; x < params_.ncols_; x++) 
		{
			if (params_.dataType_[0] == TeDOUBLE)			
			{
				((double*)obuf)[x] = (double)params_.dummy_[0];
			}
			else if (params_.dataType_[0] == TeFLOAT)			
			{
				((float*)obuf)[x] = (float)params_.dummy_[0];
			}
			else if (params_.dataType_[0] == TeLONG)	
			{
				((long*)obuf)[x] = (long)params_.dummy_[0];
			}
			else if (params_.dataType_[0] == TeUNSIGNEDLONG)	
			{
				((unsigned long*)obuf)[x] = (unsigned long)params_.dummy_[0];
			}
			else if (params_.dataType_[0] == TeINTEGER)	
			{
				((int*)obuf)[x] = (int)params_.dummy_[0];
			}
			else if (params_.dataType_[0] == TeUNSIGNEDSHORT)	
			{
				((unsigned short*)obuf)[x] = (unsigned short)params_.dummy_[0];
			}
			else if (params_.dataType_[0] == TeSHORT)	
			{
				((short*)obuf)[x] = (short)params_.dummy_[0];
			}
			else if (params_.dataType_[0] == TeCHAR)											
			{
				obuf[x] = (char)params_.dummy_[0];
			}
			else											
			{
				obuf[x] = (unsigned char)params_.dummy_[0];
			}
		}
	}

 	for (int row=0; row<params_.nlines_; row++)
	{
		if (!TIFFWriteScanline(gtif->gt_tif, obuf, row, 0))
			return false;
	}
	TIFFFlushData(gtif->gt_tif);
	_TIFFfree(obuf);
	return true;
}

bool 
TeDecoderTIFF::ReadMonoImage(unsigned long lin)
{	
	if (isTiled_)
	{
		if (!ReadTileImageContig(lin,TBuffer_))
			return false;
	}
	else
	{
		if (TIFFReadScanline(tif,TBuffer_,lin,0) < 0)
			return false;
	}
	return true;
}

bool
TeDecoderTIFF::ReadRGBImage (unsigned long lin)
{
	if (planar_ == PLANARCONFIG_CONTIG)
	{
		if (isTiled_)
		{
			if(!ReadTileImageContig(lin,TBuffer_))
				return false;
		}
		else
		{
			if (TIFFReadScanline(tif,TBuffer_,lin,0) < 0)
				return false;
		}
	}
	else	// RGB Separate
	{
		if (isTiled_)
			return false;
		int s;
		for (s=0; s < nBands_; s++)
			if (TIFFReadScanline(tif,&TBuffer_[s*params_.ncols_],lin,s) < 0)
				return false;
	}
	return true;
}

void TeDecoderTIFF::Print()
{
	long t = 0L;
	if(tif)
		TIFFPrintDirectory(tif,stdout,t);
}

static long Tiffpow2(short bits)
{
	unsigned long ret = 1;
	unsigned short x;
	
	for(x=0; x < bits ; x++)
		ret *= 2;
		
	return ret;
}

static unsigned short CalculateMask(unsigned short bits)
{
	unsigned long ret;
	ret = Tiffpow2(bits) -1L;
	return (unsigned short)ret;
}

static void ShiftAndLoadTo8Bits(unsigned char *ret, unsigned char *buf, short bits, unsigned long col, unsigned short mask)
{
	long shift1 = (long) (8 - bits);
	long shift = shift1;
	unsigned long y = 0;
	unsigned long x;

	for(x=0; x < col; x++)
	{
		ret[x] = (buf[y] >> shift)&mask;
		shift -= bits;
		if( shift < 0 )
		{
			shift = shift1;
			y++;
		}
	}	
}

static void CreateLut(short bits,u_char lut[])
{
	u_long 	itens = Tiffpow2(bits);
	u_long 	x;

	for (x=0; x < itens; x++)
		lut[x] = (u_char) ( (255L*x)/(itens-1));
}

void TeDecoderTIFF:: To8Bits(unsigned char* buf,unsigned char *ret,short vmax,short vmin,short nx)
{
	if (params_.nbitsperPixel_[0] == 32)
	{
		unsigned long aux;

		long j=0;
		long ls = (long)params_.ncols_*4;

		for (long i=0; i< ls; i+=4)
		{
			memcpy(&aux,&buf[i],4);
	
			ret[j] = (unsigned char)aux;
 			j++;
		}
		return;
	}

	if (params_.nbitsperPixel_[0] == 16)
	{
		short aux,aux2,i,j;
		float div = (float)((vmax-vmin)/255.);

		j=0;

		for (i=0; i<nx*2; i+=2)
		{
			memcpy(&aux,&buf[i],2);
	
			aux2 = (short) ((float)(aux-vmin) /div);
			ret[j] = (unsigned char)aux2;
 			j++;
		}
		return;
	}

	if ( params_.nbitsperPixel_[0] == 8 )
	{
		memcpy(ret,buf,params_.ncols_);
		return;
	}

	if( first )
	{
		mask = CalculateMask(params_.nbitsperPixel_[0]);
		CreateLut(params_.nbitsperPixel_[0],Lut);
		first = 0;
	}
	
	ShiftAndLoadTo8Bits(ret,buf,params_.nbitsperPixel_[0],params_.ncols_,mask);
	unsigned long x;
	for(x=0; x < (unsigned long)params_.ncols_; x++)
	{
		ret[x] = Lut[ret[x]];
	}
}

// --- Tile Functions ---/
bool 
TeDecoderTIFF::ReadTileImageContig(unsigned long lin, unsigned char *line)
{
	unsigned long offset = lin%tileh_;
	unsigned long col, i, nbytes;	
	for (col=0; col < (unsigned long)params_.ncols_; col += tilew_)
	{
		if (TIFFReadTile(tif, TBufferTile_, col, lin, 0, 0) < 0 )
			return false;
		nbytes = params_.nbitsperPixel_[0]/8;
		if (col + tilew_ > (unsigned long)params_.ncols_ )
		{
			unsigned long diff = params_.ncols_-col;
			for (i=0; i < diff*nbytes; i++)
				line[i+col] = TBufferTile_[i+offset*rowtilesize_];
		}
		else
		{
			for (i=0; i < tilew_*nbytes; i++)
				line[col+i] = TBufferTile_[i+offset*rowtilesize_];
		}
	}
	return true;
}

// --- To read and set a LUT table ---

static int checkmap(int n, unsigned short *r, unsigned short *g, unsigned short *b,long val)
{
    while (n-- >= 0)
		if (*r++ > val || *g++ > val || *b++ > val)
			return (16);
    return (8);
}

static unsigned short CVT1 (unsigned long x,long value)	//CVT is used in DEC station
{
	unsigned long den = (1L << 16) -1L;
	unsigned short num = (unsigned short)((x*value)/den); // normalize lut value
	return num;
}

bool
TeDecoderTIFF::ReadLut()
{
	unsigned short *rmap,*gmap,*bmap;
	if (!TIFFGetField(tif, TIFFTAG_COLORMAP, &rmap, &gmap, &bmap))
		return false;
	params_.lutr_.clear();
	params_.lutb_.clear();
	params_.lutg_.clear();

	unsigned int n = params_.lutr_.size();
	
	int nindexes = (1 << params_.nbitsperPixel_[0]);
	long maxval = (long)pow(2.,params_.nbitsperPixel_[0])-1;

	int x;
	if (checkmap (nindexes,rmap,gmap,bmap,maxval) == 16)
	{
		for(x=0; x<nindexes; ++x)
		{
			params_.lutr_.push_back(CVT1(rmap[x],maxval)); 
			params_.lutg_.push_back(CVT1(gmap[x],maxval));
			params_.lutb_.push_back(CVT1(bmap[x],maxval));
		}
	}
	else
	{
		for(x=0; x<nindexes; ++x)
		{
			params_.lutr_.push_back(rmap[x]);
			params_.lutg_.push_back(gmap[x]);
			params_.lutb_.push_back(bmap[x]);
		}
	}
	n = params_.lutr_.size();
	return true;
}

void
TeDecoderTIFF::SaveLut()
{
	if (!tif)
		return;

	int nentries = params_.lutr_.size();
	if (nentries <= 0) 
		return;

	unsigned short* lutr = new unsigned short[nentries];
	unsigned short* lutg = new unsigned short[nentries];
	unsigned short* lutb = new unsigned short[nentries];

	for (int i=0; i<nentries; i++)
	{
		lutr[i] = params_.lutr_[i]*255;
		lutg[i] = params_.lutg_[i]*255;
		lutb[i] = params_.lutb_[i]*255;
	}
	TIFFSetField(tif, TIFFTAG_COLORMAP,lutr,lutg,lutb);
	delete [] lutr;
	delete [] lutg;
	delete [] lutb;
}
