/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeSTElement.h
    \brief  This file contains structures and definitions to deal with instances
	        in time of a spatial element or object.  
*/

#ifndef  __TERRALIB_INTERNAL_STELEMENT_H
#define  __TERRALIB_INTERNAL_STELEMENT_H

#include "TeSTInstance.h"
#include "TeTimeInterval.h"

class TeQuerier;

#include <string>
#include <map> 
#include <vector>
using namespace std;

/*! \class TeSTElement
	\brief A class that implements a spatial element and its instances in the time. 

	This class contains a set of instances in the time of a spatial element. Each 
	spatio-temporal instance contains an attribute set and geometries that are valid in a 
	specific time.
	
	\sa 
	TeSTInstance.h
*/
class TeSTElement
{
	protected:
		string					object_id_;		//!< object identity
		TeTimeInterval			time_;			//!< valid time interval
		TeQuerier*				querier_;		//!< a querier used to load the instances from database or file

		vector<TeSTInstance>	instances_;     //!< the instance set ordered by time

		typedef vector<TeSTInstance>::iterator  STElementIterator; //!< iterator used to traverse the instance set

		//! Internal function to fill the instance set using a querier
		bool buildImpl(int slide=-1);

	public:
		//! Constructor 
		TeSTElement(const string& objId= ""); 

		//! Constructor  
		TeSTElement(TeTheme* theme, const string& objId);

		//! Copy constructor
		TeSTElement(const TeSTElement& other);  

		//! Destructor
		~TeSTElement(); 
		
		//! Copy operator
		TeSTElement& operator= (const TeSTElement& other);

		//! Returns the object identity
		string objectId () 
		{	return object_id_;	}
	
		//! Sets the object identity
		void objectId (const string& id) 
		{	object_id_ = id;	}

		//! Returns the valid time interval 
		TeTimeInterval timeInterval () 
		{	return time_;	}
	
		//! Sets the valid time interval
		void timeInterval (const TeTimeInterval& t) 
		{	time_ = t;	}

		//! Returns the querier pointer
		TeQuerier* querier() 
		{ return querier_; }

		//! Inserts a new spatio-temporal instance in the set
		bool insertSTInstance (TeSTInstance& object); 

		//! Sets the querier
		void querier(TeQuerier* qu);  
				
		//! Loads all instances of this spatial element from a database 
		bool build(bool loadGeometries=false, bool loadAllAttributes=true, vector<string> attrNames=vector<string>(), int slide=-1);

		//! Loads the instances of this spatial element from a database and groups them
		bool build(TeGroupingAttr& groupAttr, bool loadGeometries=false, int slide=-1);

		//! Returns the number of instances 
		int numSTInstance() { return instances_.size(); }

		/*! \class iterator
			\brief This class implements an iterator concept to the class TeSTElement.
			
			A structure that allows the traversal STElement 
			in a similar way as the STL iterators. 
		*/
		class iterator 
		{
			public:

				//! Constructor
				iterator(STElementIterator it) : elemIt_(it) 
				{}

				//! Moves to the next instance 
				iterator& operator++()
				{	++elemIt_; 	return (*this); }
			
				//! Returns the current instance 
				TeSTInstance& operator*()
				{	return (*elemIt_); 	}

				//! Equal operator
				bool operator==(const iterator& rhs) const
				{ return (this->elemIt_ == rhs.elemIt_ && this->elemIt_== rhs.elemIt_); }

				//! Unequal operator
				bool operator!=(const iterator& rhs) const
				{ return (this->elemIt_ != rhs.elemIt_ && this->elemIt_!= rhs.elemIt_);	}

			protected:
				STElementIterator		elemIt_; //!< a pointer to an instance set
		};
		
		//! Returns an iterator to the first instance of the element
		iterator begin();     
		
		//! Returns an iterator to the last instance of the element
		iterator end();       
};

 
#endif 

