/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeSTElementSet.h
    \brief This file contains structures to deal with a set of spatio-temporal instances 
	       of the elements or objects that belongs to a specific layer or theme.
*/

#ifndef  __TERRALIB_INTERNAL_STOBJECTSET_H
#define  __TERRALIB_INTERNAL_STOBJECTSET_H

#include "TeSTElement.h"

class TeQuerier; 
class TeSTEvent;
class TeLayer;


/*! \class TeSTElementSet
	\brief A class that represents a set of spatial temporal elements or objets.

	A set of spatial temporal elements or objects (ST elements) can be obtained 
	from a layer or theme.
		
	\sa
	TeSTInstance TeSTElement TeQuerier
*/
class TeSTElementSet  
{

protected:

	TeTimeInterval			totalTime_;		//!< valid time interval of all ST elements in the set
	
	TeQuerier*				querier_;       //!< a querier mechanism used to load the ST elements from a layer or theme
	
	map<string, TeSTElement >	elements_;	//! map of object identification to its spatial instances in the time
	
	typedef map<string, TeSTElement >::iterator  STElementSetIterator; //!< iterator used to traverse the ST element set
	typedef TeSTElement::iterator				 STElemIterator;	

	//! Internal function to fill the ST element set using a querier mechanism
	bool buildImpl(int slide=-1);

public:

	
	//! Constructor
	TeSTElementSet()
	{	querier_ = 0; }

	//! Constructor
	TeSTElementSet(TeQuerier* querier); 

	//! Constructor used when the ST elements will be obtained from a theme
	TeSTElementSet(TeTheme* theme); 
	
	//! Constructor used when the ST elements will be obtained from a layer
	TeSTElementSet(TeLayer* layer); 

	//! Copy constructor
	TeSTElementSet (const TeSTElementSet& other); 

	//! Destructor 
	~TeSTElementSet();  
	
	//! Assignment operator 
	TeSTElementSet& operator= (const TeSTElementSet& other);

	//! Equal operator
	bool operator== (const TeSTElementSet& other) const
	{ return ((totalTime_==other.totalTime_)); } 
		
	//! Returns the valid time interval for all ST elements
	TeTimeInterval totalTime() { return totalTime_;} 

	//! Sets the valid time interval for all ST elements
	void totalTime(TeTimeInterval t) { totalTime_ = t;}

	//! Returns the querier pointer
	TeQuerier* querier() { return querier_; }

	// Returns a theme pointer
	TeTheme* theme();  

	//! Returns a reference to the map of ST elements
	map<string, TeSTElement >& elements() { return elements_; }

	//! Verifies if there is a specific element or object in the set 
	bool hasElement(const string& objId) { return (elements_.find (objId) != elements_.end()); }
	
	//! Inserts a new ST element instance in the set 
	bool insertSTInstance (TeSTInstance& object); 

	// Inserts a new ST element instance from an ST event
	bool insertSTInstance(TeSTEvent& ev); 

	//! Sets a polygon set and its valid time interval to a specific object or element
	bool setGeometry(const string& object_id, const TePolygonSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Sets a line set and its valid time interval to a specific object or element
	bool setGeometry(const string& object_id, const TeLineSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Sets a point set and its valid time interval to a specific object or element
	bool setGeometry(const string& object_id, const TePointSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Sets a cell set and its valid time interval to a specific object or element
	bool setGeometry(const string& object_id, const TeCellSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Sets a geometry set and its valid time interval to a specific object or element
	bool setGeometry(const string& object_id, vector<TeGeometry*>& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Sets a multi geometry and its valid time interval to a specific object or element
	bool setGeometry(const string& object_id, const TeMultiGeometry& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Sets a text set and its valid time interval to a specific object or element
	bool setGeometry(const string& object_id, const TeTextSet& geomSet, TeTimeInterval time = TeTimeInterval()); 

	//! Gets a polygon set of an element or object and its valid time interval
	bool getGeometry(const string& object_id, TePolygonSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Gets a line set of an element or object and its valid time interval
	bool getGeometry(const string& object_id, TeLineSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Gets a point set of an element or object and its valid time interval
	bool getGeometry(const string& object_id, TePointSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Gets a cell set of an element or object and its valid time interval
	bool getGeometry(const string& object_id, TeCellSet& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Gets a geometry set of an element or object and its valid time interval
	bool getGeometry(const string& object_id, vector<TeGeometry*>& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Gets a multi geometry of an element or object and its valid time interval
	bool getGeometry(const string& object_id, TeMultiGeometry& geomSet, TeTimeInterval time = TeTimeInterval());
	//! Gets a text set of an element or object and its valid time interval
	bool getGeometry(const string& object_id, TeTextSet& geomSet, TeTimeInterval time = TeTimeInterval());
 
	//! Returns a ST instance of an object or element that is valid in a time interval  
	TeSTInstance* getSTInstance(const string& objctId, TeTimeInterval time = TeTimeInterval());

	//! Gets a property vector of an object or element that is valid in a time interval
	TePropertyVector* getPropertyVector (const string& object_id, TeTimeInterval time = TeTimeInterval()); 

	//! Gets the value of an attribute (as string) of a object that is valid in a time interval
	bool getAttributeValue (const string&  object_id, const string& attr_name, 
		                    string& val, TeTimeInterval time = TeTimeInterval()); 
	

	//! Gets the value of the i-th attribute (as string) of a object that is valid in a time interval
	bool getAttributeValue (const string& object_id, int attr_index, string& val,
							TeTimeInterval time = TeTimeInterval()); 

	//! Gets the value of the i-th attribute (as double) of a object that is valid in a time interval
	bool getAttributeValue (const string& object_id, int attr_index, double& val,
							TeTimeInterval time = TeTimeInterval()); 

	//! Returns the index of an attribute named "attrName"
	int getAttributeIndex(const string& attrName);

	//! Gets a vector of attributes or properties of a object that is valid in a time interval
	bool getPropertyVector (const string& object_id, TePropertyVector& propVec,
							TeTimeInterval time = TeTimeInterval()); 

	//! Sets the value of the property or adds it if there is no such property in the object
	bool addProperty(const string& object_id, TeProperty& prop);

	//! Verifies if there is an object in the set with more than one instance
	bool existMultiTemporalElem(const string& objId);

	//! Inserts a new ST instance set of a object or element
	void insertSTElement(TeSTElement& stElem)
	{ elements_[stElem.objectId()] = stElem;	}

	//! Retuns the number of objects or elements in the set
	int numElements() const  { return elements_.size();}

	//! Returns the number of instances of an object or element
	int numSTInstance(const string& objectId = ""); 

	//! Fills the ST element set
	bool build(bool loadGeometries=false, bool loadAllAttributes=true, vector<string> attrNames=vector<string>(), int slide=-1); 

	//! Fills the ST element set
	bool build(TeGroupingAttr& groupAttr, bool loadGeometries=false, int slide=-1); 

	//! Clears the set 
	void clear() { elements_.clear(); }
		
	/*! \class iterator
		\brief This class implements an iterator concept to the class TeSTElementSet.
		
		A structure that allows the traversal ST element set
		in a similar way as the STL iterators. This iterator passes over each 
		ST instance of all ST element of the set.
	*/
	class iterator 
	{
		public:

		//! Constructor
		iterator(TeSTElementSet* stoSet, STElementSetIterator stoSetP, STElemIterator stoVecP) : 
			curObj_(stoSetP),
			curSTO_(stoVecP),
			stoSet_(stoSet) 
		{}

		//! Moves to the next ST instance in the set
		iterator& operator++(); 
		
		//! Returns the current ST instance 
		TeSTInstance& operator*()
		{	return (*curSTO_); 	}

		//! Returns the current ST instance 
		TeSTInstance* operator->()
		{	return &(*curSTO_); }

		//! Returns the current ST instance pointer
		STElemIterator getCurSTO()
		{	return curSTO_; 	}

		//! Returns a pointer to the ST element set 
		TeSTElementSet* elemSet()
		{	return stoSet_; 	}

		//! Equal operator
		bool operator==(const iterator& rhs) const
		{ return (this->curSTO_ == rhs.curSTO_ && this->curObj_== rhs.curObj_); }

		//! Unequal operator
		bool operator!=(const iterator& rhs) const
		{ return (this->curSTO_ != rhs.curSTO_ && this->curObj_!= rhs.curObj_);	}

		protected:
			STElementSetIterator	curObj_;
			STElemIterator			curSTO_;
			TeSTElementSet*			stoSet_;
	};
		
	
	/*! \class propertyIterator
		\brief This class implements an iterator concept to the class TeSTElementSet.
		
		A structure that allows the traversal ST element set
		in a similar way as the STL iterators. This iterator passes over each 
		ST instance of the ST element set and returns a specific numeric property.
	*/
	class propertyIterator : public iterator
	{
		public:

			//! Constructor
			propertyIterator (TeSTElementSet* stoSet, 
				STElementSetIterator stoSetP, 
				STElemIterator stoVecP, 
				string attrName):
				iterator(stoSet, stoSetP, stoVecP),
				attrName_ (attrName)
				{}
		
			//! Returns the numeric property (as a double) of the current ST instance.
			double operator*();  
		
			//! Returns the numeric property of the current ST instance.
			bool getProperty(TeProperty& prop); 
				
		protected:
			string	attrName_; //<! the numeric property name

	};
	
	//! Returns an iterator to the first element of the set
	iterator begin();     
	
	//! Returns a property iterator to the first element of the set
	propertyIterator begin(const string& attrName);     
	
	//! Returns an iterator to the one past last element of the set
	iterator end();       
	
	//! Returns a property iterator to the one past last element of the set
	propertyIterator end(const string& attrName);       
};

//! Calculates the bounding box of all instances of all objects included in the set
/*! 
	\param steSet reference to element set that contains the objects
	\return the bounding box of all instances of all objects
*/
TeBox
getSTElementBox(TeSTElementSet& steSet);

/*! \example createSTElementSetFromLayer.cpp
	Shows how to create a spatio-temporal element set from a layer.
 */

/*! \example createSTElementSetFromShapeFile.cpp
	Shows how to create aspatio-temporal element set from a shapefile.
 */

/*! \example createSTElementSetFromTheme.cpp
	Shows how to create aspatio-temporal element set from a theme.
 */
#endif

