/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeSTEvent.h
    \brief This file contains structures and definitions to deal with spatial temporal events. 
	       A spatial temporal event is represented by a location (point), a time and a set of attributes. 
*/

#ifndef  __TERRALIB_INTERNAL_STEVENT_H
#define  __TERRALIB_INTERNAL_STEVENT_H

#include "TeGeometry.h"
#include "TeAttribute.h" 
#include "TeTimeInterval.h"
#include "TeKdTree.h"

class TeTheme;
class TeSTInstance;

/*! \class TeSTEvent
	\brief A class that represents spatial temporal events (STEvent).

	A spatial temporal event is represented by a location (point), 
	a time when it happened and a set of attributes.
*/
class TeSTEvent
{
protected:
	vector<string>	unique_id_;		//!< event identifications in each attribute table
	string			object_id_;		//!< event identification
	TePoint			point_;			//!< the location where the event happened  
	long			time_;			//!< the time when the event happened   

	vector<double>			attrValue_;		//!< event attribute values 
	vector<TeAttributeRep>	attrRep_;		//!< event attribute representations (name, type, size) 
	vector<int>				attrOut_;		//!< indexes of the attributes that were generated

	TeTheme*		theme_;			//!< the theme that contains this event
		
public:
	
	//! Empty constructor
	TeSTEvent()
	{ }

	//! Constructor
	TeSTEvent(const string& objId, const TePoint& point, const long& time);  

	//! Constructor
	TeSTEvent(const string& objId, const TePoint& point, const vector<double>& attrValue, 
		const vector<TeAttributeRep>& attrRep, const long& time);
	
	//! Copy constructor
	TeSTEvent(const TeSTEvent& other); 
						
	//! Assignment operator
	TeSTEvent& operator= (const TeSTEvent& other);
		
	//! Equal operator
	bool operator== (const TeSTEvent& other) const;
	
	//! Destructor
	~TeSTEvent() {} 

	//! Returns the event identification
	string objectId (void)
	{ return object_id_; }

	//! Sets the event identification
	void objectId (const string& objId)
	{ object_id_ = objId; }

	//! Returns the event identifications in all attribute tables
	vector<string>& uniqueId () 
	{	return unique_id_;	}

	//! Returns the event identification in the index_-th attribute table
	string uniqueId (int index_) 
	{	return unique_id_[index_];	}
	
	//! Sets the instance identifications
	void uniqueId (const vector<string>& ids) 
	{	unique_id_ = ids;	}
	
	//! Gets the event location (point)
	void getGeometry(TePoint& result)
	{	result = point_; }

	//! Sets the event location (point)
	void setGeometry(const TePoint& result)
	{	point_ = result; }

	//! Gets the time when the event happened
	long eventTime()
	{	return time_; }

	//! Sets the time when the event happened
	void eventTime(const long& t)
	{ time_ = t; }

	//! Sets the event location (point)
	void point(TePoint& p)
	{ point_ = p; }

	//! Gets the event location (point)
	TePoint& point(void)
	{ return point_; }
	
	//! Gets the ordinate x of the event location
	double x()
	{	return point_.location().x(); }

	//! Sets the ordinate x of the event location
	void x(const double& x)
	{	point_.location().x(x); }

	//! Gets the ordinate y of the event location
	double y()
	{	return point_.location().y(); }

	//! Sets the ordinate y of the event location
	void y(const double& y)
	{	point_.location().y(y); }

	//! Gets the time when the event happened as a TeTimeInterval structure 
	TeTimeInterval timeInterval(); 
	
	//! Gets the theme that contains this event
	TeTheme* theme();

	//! Sets the theme that contains this event
	void theme(TeTheme* t);

	//! Gets the event attribute values
	vector<double>& attrValues()
	{ return attrValue_; }

	//! Sets the event attribute values
	void attrValues (const vector<double>& attrVal)
	{ attrValue_ = attrVal; }

	//! Gets the event attribute representations (name, type, size)
	vector<TeAttributeRep>& getAttributeRep()
	{ return attrRep_; }

	//! Gets the event attribute representation of the index-th attribute
	TeAttributeRep getAttributeRep(const int& index)
	{ return  attrRep_[index]; }

	//! Sets the event attribute representations (name, type, size)
	void setAttributeRep (const vector<TeAttributeRep>& attrRep)
	{ attrRep_ = attrRep; }

	//! Gets the event attribute indexes that were generated
	vector<int>& attrOutIndexes()
	{ return attrOut_; }

	//! Sets the event attribute indexes that were generated
	void attrOutIndexes (const vector<int>& indexes)
	{ attrOut_ = indexes; }

	//! Gets the event attribute or property value of the index-th property
	double getPropertyValue(const int& index)
	{ return  attrValue_[index]; }

	//! Gets the event attribute value (as a string) of the i-th property
	bool getPropertyValue (string& val, int i= 0);	

	//! Sets the event attribute value of the indexAttr-th property
	void setPropertyValue(const int& indexAttr, const double& value);
	
	//! Gets the event attribute name of the index-th attribute
	string getPropertyName(const int& index)
	{ return  attrRep_[index].name_; }

	//! Gets the property index of the attribute called "name"
	int getPropertyIndex(const string& name); 
	
	//! Returns the value (as a double) of the index-th property
	double operator[] (const int& index)
	{ return  attrValue_[index]; 	}

	//! Adds a new property that contains the attribute representation and the attribute value
	void addProperty(const TeProperty& prop, bool attrIn = true); 

	//! Adds a new property from a attribute representation and a attribute value
	void addProperty(const double& val, TeAttributeRep rep = TeAttributeRep(), bool attrIn = true);

	//! Returns the properties or attributes (theirs representations and values)that were generated 
	TePropertyVector getOutPropertyVector(); 

	//! Returns all the properties or attributes (theirs representations and values) 
	TePropertyVector getPropertyVector(); 
};


/*! \class TeSTEventSetImpl
	\brief A class that represents spatial temporal event set.

	An abstract class that defines a base interface to different implementations 
	of spatial temporal event set.
*/
class TeSTEventSetImpl
{

protected:
	typedef TeSAM::TeAdaptativeKdTreeNode<TeCoord2D, vector<TePoint>, TePoint> kdNode;
	typedef TeSAM::TeAdaptativeKdTree<kdNode> kdTree;

	kdTree*	eventTree_;
	TeBox   box_;
	bool	kdTreeBuilt_;

public:
	
	//! Constructor
	TeSTEventSetImpl():
		eventTree_(0),
		kdTreeBuilt_(false)
	{ }

	//! Constructor
	TeSTEventSetImpl(const TeBox&  box):
		eventTree_(0),
		box_(box),
		kdTreeBuilt_(false)
	{ }

    //! Destructor
	virtual ~TeSTEventSetImpl()
	{ 
		if(eventTree_)
			delete eventTree_;
	}

	//! Sets the minimal bounding box that contains all the events
	virtual void setBox(const TeBox& b) { box_= b; }

	//! Gets the minimal bounding box that contains all the events
	TeBox box() { return box_;}

	//! Returns the number of events in the set
	virtual unsigned int size()=0;

	//! Returns the number of events in the set
	virtual int numSTInstance()=0;

	//! Returns the number of elements or objects in the set
	virtual int numElements()=0;

	//! Gets the value of the attr_index-th attribute (as double) 
	virtual bool getAttributeValue (const string& object_id, const int& attr_index, double& val, const long& time= -1)=0;

	//! Gets the attribute index of attribute called "attrName"
	virtual int getAttributeIndex(const string& attrName)=0;

	//! Searchs the coordinates inside a specific bounding box  
	virtual bool search(const TeBox& b, vector<TePoint>& result);
	
	//! Searchs the k-nearest neighbours of a specific coordinate
	virtual bool nearestNeighbourSearch(const TeCoord2D& coord, vector<TePoint>& result, vector<double>& distances, const unsigned int& k);

	/*! \class iterator
		\brief This class implements an iterator concept to the event set.

		An strucuture that allows the traversal STEvent set 
		(ST elements) in a similar way as the STL iterators. 
	*/
	class iterator 
	{
		public:

		//! Constructor
		iterator()
		{ }
		
		//! Destructor
		virtual ~iterator()
		{ }

		//! Moves to the next event
		virtual void operator++()=0;
				
		//! Returns a reference to the current event
		virtual TeSTEvent& operator*()=0;
		
		//! Equal operator
		virtual bool operator==(const iterator* rhs) const=0;
		
		//! Unequal operator
		virtual bool operator!=(const iterator* rhs) const=0; 
	};

	friend class TeSTEventSetImpl::iterator;

	//! Returns an iterator to the first event of the set
	virtual TeSTEventSetImpl::iterator* begin()=0;     
	
	//! Returns an iterator to the one past last event of the set
	virtual TeSTEventSetImpl::iterator* end()=0;       

	//! Inserts a new event in the set 
	virtual void insertSTInstance(TeSTEvent& ev)=0;

	//! Builds the kdTree that is used to search events in the set
	virtual bool buildKDTree(const int& n=-1)= 0;

	//! Clears the event set 
	virtual void clear()=0; 

	//! Returns the "index"-th event. It must be used when the events are stored in a vector.   
	virtual TeSTEvent operator[](const unsigned int& index)=0;

	//! Returns the event that has an identification equal to "id". It must be used when the events are stored in a map ordered by object id.   
	virtual TeSTEvent operator[](const string& id)=0; 
};


/*! \class TeSTEventVector
	\brief A class that represents spatial temporal event set.

	 A class that represents a spatial temporal event set stored in a vector. 
	 A kdtree can be built on event coordinates. 
*/
class TeSTEventVector : public TeSTEventSetImpl
{

protected:
	vector<TeSTEvent>	events_;		// vector of spatial temporal events

	typedef vector<TeSTEvent>::iterator			STElemVecIterator;

public:

	//! Constructor
	TeSTEventVector();

	//! Constructor
	TeSTEventVector(const TeBox& box);

	//! Returns the number of events
	unsigned int size() { return events_.size(); }

	//! Returns the number of spatial temporal events in the set
	int numSTInstance() { return size(); }

	//! Returns the number of elements or objects in the set
	int numElements() { return size(); }

	//! Gets the value of the attr_index-th attribute (as double) 
	bool getAttributeValue (const string& object_id, const int& attr_index, double& val, const long& time= -1);

	//! Gets the attribute index of attribute called "attrName"
	int getAttributeIndex(const string& attrName);

	/*! \class iterator
		\brief This class implements an iterator concept to a set of spatial temportal events.

		An strucuture that allows the traversal STEvent set 
		(ST elements) in a similar way as the STL iterators. 
	*/
	class iterator : public TeSTEventSetImpl::iterator 
	{
		public:

		//! Constructor
		iterator(STElemVecIterator curElem, TeSTEventVector* set):
			curElemVec_(curElem),
			eventVector_(set)
			{ }
			
		//! Destructor
		virtual ~iterator()
		{ }

		//! Moves to the next event
		void operator++();
				
		//! Returns a reference to the current event
		TeSTEvent& operator*() { return *curElemVec_; }
		
		//! Equal operator
		bool operator==(const TeSTEventSetImpl::iterator* rhs) const
		{ return (this->curElemVec_== ((TeSTEventVector::iterator*)rhs)->curElemVec_); }
		
		//! Unequal operator
		bool operator!=(const TeSTEventSetImpl::iterator* rhs) const
		{ return (this->curElemVec_!= ((TeSTEventVector::iterator*)rhs)->curElemVec_); }
		
		public:
			STElemVecIterator		curElemVec_;
			TeSTEventVector*		eventVector_;
	};

	friend class TeSTEventVector::iterator;
	
	//! Returns an iterator to the first event of the set
	TeSTEventSetImpl::iterator* begin();     
	
	//! Returns an iterator to the one past last event of the set
	TeSTEventSetImpl::iterator* end();       

	//! Inserts a new event in the set
	void insertSTInstance(TeSTEvent& ev);

	//! Builds the kdTree that is used to search events in the set
	bool buildKDTree(const int& n=-1);

	//! Clears the event set 
	void clear(); 

	//! Returns the "index"-th event. It must be used when the events are stored in a vector.    
	TeSTEvent operator[](const unsigned int& index) { return events_[index]; }

	//! Returns the event that has an identification equal to "id". It must be used when the events are stored in a map ordered by object id.   
	TeSTEvent operator[](const string& id); 
};


/*! \class TeSTEventMap
	\brief A class that represents spatial temporal event set.

	A class to represent a spatial temporal event set stored in a map orderd by object 
	identification. A kdtree can be built on event coordinates. 
*/
class TeSTEventMap : public TeSTEventSetImpl
{

protected:
	map<string, TeSTEvent>				events_;

	typedef map<string, TeSTEvent>::iterator	STElemMapIterator;

public:
	
	//! Constructor 
	TeSTEventMap();
	
	//! Constructor 
	TeSTEventMap(const TeBox& box);

	//! Returns the number of events
	unsigned int size() { return events_.size(); }

	//! Returns the number of spatial temporal events in the set
	int numSTInstance() { return size(); }

	//! Returns the number of elements or objects in the set
	int numElements() { return size(); }
	
	//! Gets the value of the attr_index-th attribute (as double) 
	bool getAttributeValue (const string& object_id, const int& attr_index, double& val, const long& time= -1);

	//! Gets the attribute index of attribute called "attrName"
	int getAttributeIndex(const string& attrName);

	/*! \class iterator
		\brief This class implements an iterator concept to a set of spatial temportal events.

		An strucuture that allows the traversal STEvent set 
		(ST elements) in a similar way as the STL iterators. 
	*/
	class iterator : public TeSTEventSetImpl::iterator 
	{
		public:

		//! Constructor
		iterator(STElemMapIterator curElem, TeSTEventMap* set) :
			curElemMap_(curElem),
			eventMap_(set)
			{ }

		//! Destructor
		virtual ~iterator()
		{ }
		
		//! Moves to the next event
		void operator++();
				
		//! Returns a reference to the current event
		TeSTEvent& operator*() { return (curElemMap_->second); }
		
		//! Equal operator
		bool operator==(const TeSTEventSetImpl::iterator* rhs) const
		{ return (this->curElemMap_ == ((TeSTEventMap::iterator*)rhs)->curElemMap_); }
		
		//! Unequal operator
		bool operator!=(const TeSTEventSetImpl::iterator* rhs) const
		{	return (this->curElemMap_ != ((TeSTEventMap::iterator*)rhs)->curElemMap_); }
		
		protected:
			STElemMapIterator		curElemMap_;
			TeSTEventMap*			eventMap_;
	};

	friend class TeSTEventMap::iterator;

	//! Returns an iterator to the first event of the set
	TeSTEventSetImpl::iterator* begin();     
	
	//! Returns an iterator to the one past last event of the set
	TeSTEventSetImpl::iterator* end();       

	//! Inserts a new event in the set ordered by object identification 
	void insertSTInstance(TeSTEvent& ev);

	//! Builds the kdTree that is used to search events in the set
	bool buildKDTree(const int& n=-1);

	//! Clears the event set 
	void clear(); 

	//! Returns the "index"-th event. It must be used when the events are stored in a vector.   
	TeSTEvent operator[](const unsigned int& index);

	//! Returns the event that has an identification equal to "id". It must be used when the events are stored in a map ordered by object id.   
	TeSTEvent operator[](const string& id); 
};

/*! \class TeSTEventSet
	\brief A class that represents a set of spatial temporal events.

	A set of spatial temporal events (ST events) can be obtained 
	from a layer or theme.
		
	\sa
	TeSTEvent 
*/
class TeSTEventSet
{
protected:
		
	TeSTEventSetImpl*			events_;		//!< a pointer to a class that implements the event container 

	TeChronon					chronon_;			//!< the chronon associated to the event times 
	TeTheme*					theme_;				//!< the theme that contains this event set
	bool						orderedByObjId_;	//!< if the events are ordered by object identification or not
		
public:

	//! Constructor	
	TeSTEventSet(TeTheme* theme=0, bool objIdOrdered=true); 

	//! Destructor	
	~TeSTEventSet();

	//! Gets the theme that contains this event set
	TeTheme* theme(); 

	//! Sets the theme that contains this event set
	void theme(TeTheme* t);

	//! Returns the number of events
	unsigned int size();

	//! Sets the chronon associated to the events
	void chronon(const TeChronon& chr) { chronon_ = chr; } 

	//! Gets the chronon associated to the events
	TeChronon chronon() { return chronon_;} 

	//! Sets the minimal bounding box that contains all the events
	void box(const TeBox& b);  

	//! Gets the minimal bounding box that contains all the events
	TeBox box(); 

	//! Verifies if the event set is ordered by the object identifications
	bool isOrderedByObjId() { return orderedByObjId_; }
	
	//! Gets the value of an attribute (as string) called "attr_name"
	bool getAttributeValue (const string&  object_id, const string& attr_name, 
		                    string& val, TeTimeInterval time = TeTimeInterval()); 
		
	//! Gets the value of the attr_index-th attribute (as string) 
	bool getAttributeValue (const string& object_id, const int& attr_index, string& val,
							TeTimeInterval time = TeTimeInterval());
	
	//! Gets the value of the attr_index-th attribute (as double) 
	bool getAttributeValue (const string& object_id, const int& attr_index, double& val, const long& time = -1);

	//! Gets the attribute index of the attribute called "attrName"
	int getAttributeIndex(const string& attrName);

	//! Returns the number of spatial temporal events in the set
	int numSTInstance();

	//! Returns the number of elements or objects in the set
	int numElements();  

	//! Searchs the coordinates inside a specific bounding box 
	bool search(const TeBox& b, vector<TePoint>& result);

	//! Searchs the k-nearest neighbours of a specific coordinate
	bool nearestNeighbourSearch(const TeCoord2D& coord, vector<TePoint>& result, vector<double>& distances, const unsigned int& k);

	
	/*! \class iterator
		\brief This class implements an iterator concept to the class TeSTEventSet.
		
		A strucuture that allows the traversal ST event set  
		in a similar way as the STL iterators. This iterator passes over each 
		ST event of the set.
	*/
	class iterator 
	{
		public:

		//! Constructor
		iterator(TeSTEventSet* elemSet, TeSTEventSetImpl::iterator* it):
			  elemSet_(elemSet),
			  it_(it)
			{ }

		//! Copy Constructor
		iterator(const iterator& it)
		{
			elemSet_ = it.elemSet_;	
			it_ = elemSet_->events_->begin();
			(*it_)=(*(it.it_));
		}

		//! Destructor
		~iterator()
		{ delete (it_); }

		//! Moves to the next event
		iterator& operator++();
		
		//! Copy operator
		iterator& operator=(const iterator&);
		
		//! Returns a reference to the current event
		TeSTEvent& operator*();
		
		//! Returns a pointer to the current event
		TeSTEventSet* elemSet()
		{	return elemSet_; 	}

		//! Equal operator
		bool operator==(const iterator& rhs) const;
		
		//! Unequal operator
		bool operator!=(const iterator& rhs) const; 
		
		protected:
			TeSTEventSet*					elemSet_;	
			TeSTEventSetImpl::iterator*		it_;
	};
	
	friend class TeSTEventSet::iterator;

	//! Returns an iterator to the first event of the set
	TeSTEventSet::iterator begin();     
	
	//! Returns an iterator to the one past last event of the set
	TeSTEventSet::iterator end();       

	//! Inserts a new event in the set 
	bool insertSTInstance(TeSTEvent& ev); 

	//! Inserts a new event in the set from a instance TeSTInstance
	bool insertSTInstance (TeSTInstance& object);  

	//! Builds the kdTree that is used to search events in the set
	bool buildKDTree(const int& n=-1);

	//! Clears the event set 
	void clear(); 

	//! Returns the "index"-th event. It must be used when the events are stored in a vector.   
	TeSTEvent operator[](const unsigned int& index);

	//! Returns the event that has an identification equal to "id". It must be used when the events are stored in a map ordered by object id.   
	TeSTEvent operator[](const string& id); 
};

#endif







