/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeSTInstance.h
    \brief This file contains structures and definitions to deal with an instance
	       in time of a spatial element, including their geometry and set of properties.  
*/

#ifndef  __TERRALIB_INTERNAL_STINSTANCE_H
#define  __TERRALIB_INTERNAL_STINSTANCE_H

#include "TeAttribute.h"
#include "TeTimeInterval.h"
#include "TeMultiGeometry.h"

class TeTheme;

#include <string>
#include <map> 
#include <vector>
using namespace std;

/*! \class TeSTInstance
	\brief A class that represents an instance in a time of a spatial element.

	A spatio-temporal instance is composite by an attribute set and geometries 
	of a spatial element or object that are valid in a specific time.
	
	\sa 
	TeTimeInterval TePropertyVector TeMultiGeometry
*/

class TeSTInstance
{
	
protected:
	string				object_id_;		//!< object identification
	vector<string>		unique_id_;		//!< instance identifications in each attribute table
	int					slice_;			//!< the slice or group that contains this instance  
	
	TeTimeInterval		time_;			//!< valid time interval
	TePropertyVector	properties_;	//!< set of properties
	TeMultiGeometry	    geometries_;	//!< set of geometries

	TeTheme*			theme_;			//!< the theme that contains the event

public:		

	//! Constructor
	TeSTInstance() 
	{ }

	//! Constructor 
	TeSTInstance (const string& object_id, TeProperty& prop ); 
			
	//! Copy constructor
	TeSTInstance(const TeSTInstance& other); 
						
	//! Assignment operator
	virtual TeSTInstance& operator= (const TeSTInstance& other);
		
	//! Equal operator 
	bool operator== (const TeSTInstance& other) const;
	
	//! Destructor
	virtual ~TeSTInstance() 
	{} 
	
	//! Sets the property vector to the ST instance
	void properties(TePropertyVector& p) 
	{	properties_ = p;	}

	//! Returns the property vector from the ST instance
	TePropertyVector& getPropertyVector()
	{	return properties_;	}

	//! Returns the properties of the instance that were generated by a function
	TePropertyVector& getOutPropertyVector()
	{	return properties_;	}

	//! Returns the geometries
	TeMultiGeometry& geometries() { return geometries_;}

	//! Gets the i-th property
	bool getProperty (TeProperty& prop, unsigned int i = 0);	
	
	//! Gets the property named "name"
	bool getProperty (TeProperty& prop, string name);	
	
	//! Gets the value (as a string) of the i-th property
	bool getPropertyValue (string& val, int i = 0);	

	//! Gets the value (as a string) of a property named "name"
	bool getPropertyValue (const string& name, string& val);	

	//! Verifies if the instance has polygons
	bool hasPolygons()	
	{ return (geometries_.hasPolygons()); } 
	//! Verifies if the instance has lines
	bool hasLines()		
	{ return (geometries_.hasLines()); } 
	//! Verifies if the instance has points
	bool hasPoints()		
	{ return (geometries_.hasPoints()); } 
	//! Verifies if the instance has cells
	bool hasCells()		
	{ return (geometries_.hasCells()); } 
	
	//! Gets a copy of the ST instance polygon set
	bool getGeometry(TePolygonSet& result); 
	//! Gets a copy of the ST instance line set
	bool getGeometry(TeLineSet& result);
	//! Gets a copy of the ST instance point set
	bool getGeometry(TePointSet& result);
	//! Gets a copy of the ST instance cell set
	bool getGeometry(TeCellSet& result);
	//! Gets a copy of the ST instance text set
	bool getGeometry(TeTextSet& result);
	//! Gets a copy of the ST instance geometry vector
	bool getGeometry(vector<TeGeometry*>& result);
	//! Gets a copy of the ST instance multi geometry
	bool getGeometry(TeMultiGeometry& result);

	//! Get a reference to the ST instance polygon geometry
	TePolygonSet&  getPolygons();
	//! Get a reference to the ST instance line geometry
	TeLineSet&  getLines();
	//! Get a reference to the ST instance point geometry 
	TePointSet&  getPoints();
	//! Get a reference to the ST instance cell geometry  
	TeCellSet&  getCells();
	//! Get a reference to the ST instance text geometry 
	TeTextSet&  getTexts();
	
	//! Adds a property to the ST instance
	void addProperty(TeProperty& prop) 
	{	properties_.push_back ( prop );		}

	//! Adds a property to the ST instance
	void addProperty(const double& val, TeAttributeRep rep = TeAttributeRep(), bool attrIn = true); 

	//! Sets a polygon set to the instance 
	void setGeometry(const TePolygonSet& result)
	{ geometries_.setGeometry(result); }
	//! Sets a line set to the instance 
	void setGeometry(const TeLineSet& result)
	{ geometries_.setGeometry(result); }
	//! Sets a point set to the instance 
	void setGeometry(const TePointSet& result)
	{ geometries_.setGeometry(result); }
	//! Sets a cell set to the instance 
	void setGeometry(const TeCellSet& result)
	{ geometries_.setGeometry(result); }
	//! Sets a text set to the instance 
	void setGeometry(const TeTextSet& result)
	{ geometries_.setGeometry(result); }
	//! Sets a geometry set to the instance 
	void setGeometry(vector<TeGeometry*>& result)
	{ geometries_.setGeometry(result); }
	//! Sets a multi geometry to the instance 
	void setGeometry(const TeMultiGeometry& geoms) 
	{	geometries_ = geoms; }

	//! Adds a polygon to the instance
	bool addGeometry(const TePolygon& poly);  
	//! Adds a line to the instance
	bool addGeometry(const TeLine2D& line);
	//! Adds a point to the instance
	bool addGeometry(const TePoint& point);	
	//! Adds a cell to the instance
	bool addGeometry(const TeCell& cell);		
	//! Adds a text to the instance
	bool addGeometry(const TeText& cell);		
		
	//! Sets the value (as a string) of a property named 'name' 
	bool setPropertyValue (const string& name, const string& val);	
	
	//! Returns the value (as a double) of the i-th property
	double operator[](int i);    
	
	//! Returns a centroid associated to the geometries of the instance
	bool centroid(TeCoord2D& centroid, TeGeomRep geomRep=TeGEOMETRYNONE); 
	
	//! Returns the area of the geometries of the instance
	bool area( double& a, TeGeomRep geomRep=TeGEOMETRYNONE);
	
	//! Returns the valid time interval
	TeTimeInterval timeInterval () 
	{	return time_;	}
	
	//! Sets the valid time interval
	void timeInterval (const TeTimeInterval& t) 
	{	time_ = t;	}

	//! Returns the initial time (as a string) of the valid time interval
	string getInitialDateTime(const string& mask="YYYYsMMsDDsHHsmmsSS") 
	{	return time_.getInitialDateTime(mask);	}
	
	//! Returns the final time (as a string) of the valid time interval
	string getFinalDateTime(const string& mask="YYYYsMMsDDsHHsmmsSS") 
	{	return time_.getFinalDateTime(mask);	}

	//! Returns the object identification
	string objectId () 
	{	return object_id_;	}
	
	//! Sets the object identification
	void objectId (const string& id) 
	{	object_id_ = id;	}

	//! Returns the instance identification in all attribute tables
	vector<string>& uniqueId () 
	{	return unique_id_;	}

	//! Returns the instance identification in the index_-th attribute table
	string uniqueId (int index_) 
	{	return unique_id_[index_];	}
	
	//! Sets the instance identifications
	void uniqueId (const vector<string>& ids) 
	{	unique_id_ = ids;	}

	//! Adds an instance identification
	void addUniqueId(const string& id)
	{	unique_id_.push_back(id); }

	//! Sets the group or slice that contains this instance 
	void slice (int s) 
	{	slice_ = s; }

	//! Returns the group or slice that contains this instance
	int slice()
	{	return slice_; }

	//! Returns the theme pointer that contains this instance
	TeTheme* theme() { return theme_; }

	//! Sets the theme that contains this instance
	void theme(TeTheme* theme)  { theme_ = theme; }

	//! Clears the instance, its geometries and attributes
	void clear(); 
};


#endif 
