/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/
/*! \file TeTheme.h
	\brief This file contains definitions about a theme in TerraLib
*/
#ifndef  __TERRALIB_INTERNAL_THEME_H
#define  __TERRALIB_INTERNAL_THEME_H

#include "TeLegendEntry.h"
#include "TeTable.h"
#include "TeGeometry.h"
#include "TeViewNode.h"

class TeLayer;
class TeRasterTransform;
class TeRaster;
struct TeRepresentation;

#include <string>

//! A grouping structure
/*!
	The TeGrouping is a structure that describes how the objects of a theme should
	be grouped, or separated in groups.
*/
struct TeGrouping
{
	//! Returns the attribute used to group the objects of the theme
	TeAttributeRep	groupAttribute_;
	
	//! Returns the attribute used to normalize a grouping 
	string	groupNormAttribute_;
	
	//! Returns the grouping mode 
	TeGroupingMode	groupMode_;			
	
	//! Returns the number of groups
	int	groupNumSlices_;	
	
	//! Returns the numeric precision used to group objects
	int	groupPrecision_;

	//! Returns the deviation parameter used to group objects
	double	groupStdDev_;

	//! Aggregate function
	string	groupFunction_; 

	//! Show missing data
	bool	groupNullAttr_; 

	//! Minimum value used to calculate a equal step grouping
	double groupMinVal_;

	//! Maximum value used to calculate a equal step grouping 
	double groupMaxVal_;

	//! Chronon used for grouping
	TeChronon groupChronon_;

	//! Constructor
	TeGrouping(TeAttributeRep att=TeAttributeRep(), const string& normAttr="", 
				TeGroupingMode gMode=TeNoGrouping, int numSlice=0, int gPrecision=6,  
				double	gStDev=1.0, const string& func="", TeChronon gChronon = TeNOCHRONON): 
			groupAttribute_(att),
			groupNormAttribute_(normAttr),
			groupMode_(gMode),
			groupNumSlices_(numSlice),
			groupPrecision_(gPrecision),
			groupStdDev_(gStDev),
			groupFunction_(func),
			groupNullAttr_(false),
			groupMinVal_(TeMAXFLOAT),
			groupMaxVal_(TeMINFLOAT),
			groupChronon_(gChronon)
			{}

	//! Copy constructor
	TeGrouping(const TeGrouping& other)
	{	
		groupAttribute_		= other.groupAttribute_;
		groupNormAttribute_	= other.groupNormAttribute_;
		groupMode_			= other.groupMode_;
		groupNumSlices_		= other.groupNumSlices_;
		groupPrecision_		= other.groupPrecision_;
		groupStdDev_		= other.groupStdDev_;
		groupFunction_		= other.groupFunction_;
		groupNullAttr_		= other.groupNullAttr_;
		groupMinVal_		= other.groupMinVal_;
		groupMaxVal_		= other.groupMaxVal_;
		groupChronon_			= other.groupChronon_;
	}
			
	//! Destructor
	~TeGrouping() {}
	
	//! Copy constructor
	TeGrouping& operator= (const TeGrouping& other)
	{	
		if ( this != &other )
		{		
			groupAttribute_		= other.groupAttribute_;
			groupNormAttribute_	= other.groupNormAttribute_;
			groupMode_			= other.groupMode_;
			groupNumSlices_		= other.groupNumSlices_;
			groupPrecision_		= other.groupPrecision_;
			groupStdDev_		= other.groupStdDev_;
			groupFunction_		= other.groupFunction_;
			groupNullAttr_		= other.groupNullAttr_;
			groupNullAttr_		= other.groupNullAttr_;
			groupMinVal_		= other.groupMinVal_;
			groupMaxVal_		= other.groupMaxVal_;
			groupChronon_		= other.groupChronon_;
		}
		return *this;
	}
};

using namespace std;

//! TerraLib definition of a Theme
/*!
\par 
A TeTheme represents a collection of objects selected from a TeLayer. This selection is based
on restrictions that can be non-spatial (e.g. POP > 10000), spatial (e.g. "inside box(0,0,100,100)
or temporal. The most simple selection is "all" objects of a layer.
\par 
A TeTheme contains the list of attribute tables of the layer used by a theme.
\par 
A TeTheme contains the visual presentation parameters for the objects that contains, for
all of its geometrical representations.
\par 
A TeTheme can store the parameters associated to how to separate its objects in groups. 
\par 
A TeTheme can store parameters associated to the creation of individual graphs on its objects
(e.g. pie bars that relate two or more of its attributes).
\par 
A TeTheme store the range of scales within it should be visible. 
\sa TeView TeTable
*/
class TeTheme: public TeViewNode
{
public:
	//! Constructor
    TeTheme( const string& name="", TeLayer* layer=0, TeViewNode* parent = 0, int view=0, int id=0);

	//! Copy constructor
	TeTheme (TeTheme& other);

	//! Destructor
	~TeTheme ();

	//! Assignment operator
	TeTheme& operator= (TeTheme& other); 

	/** @name Layer
		Methods related to the layer that gives origin to this theme.
	*/
	//@{	
	//! Returns the id of the source layer
	virtual int layerId() 
	{	return layerId_; }

	//! Sets the id of the source layer
	virtual void layerId(int i)
	{	layerId_ = i; }

	//! Sets the layer that is the source of objects of the theme
	/*! 
		\param layer a pointer to a TeLayer
	*/
	virtual void layer(TeLayer* layer);

	//! Returns a pointer to the layer from which the theme get its objects
	virtual TeLayer* layer() { return layer_; }
	//@}

	/** @name Scale
		Methods related to the range of scales where this theme should be visible
	*/
	//@{	
	//! Returns the minimum scale in which the theme is visible
	virtual double minScale() { return minScale_; }

	//! Sets the minimum scale in which the theme is visible
	virtual void minScale(double s) { minScale_ = s; }

	//! Returns the maximum scale in which the theme is visible
	virtual double maxScale() { return maxScale_; }

	//! Sets the maximum scale in which the theme is visible
	virtual void maxScale(double s) { maxScale_ = s; }
	//@}

	/** @name Restrictions
		Methods related to the restrictions over the theme used to generate this theme
	*/
	//@{
	//! Verifies if this theme has any kind of restrictions
	virtual bool hasRestriction() { return (!(generateAttributeRest_.empty() || 
									  generateTemporalRest_.empty()) && hasSpatialRes_); }

	//! Returns the attribute restriction (where clause) used to generate the theme  
	virtual string attributeRest() { return generateAttributeRest_; }

	//! Sets the attribute restriction (where clause) used to generate the theme
	virtual void attributeRest(const string& s) { generateAttributeRest_ = s; }

	//! Returns if there is an attribute restriction  
	virtual bool hasAttrRest () { return (!generateAttributeRest_.empty());}

	//! Returns the temporal restriction used to generate the theme
	virtual string temporalRest() { return generateTemporalRest_; }

	//! Sets the temporal restriction used to generate the theme
	virtual void temporalRest(const string& t) { generateTemporalRest_ = t; };

	//! Returns TRUE if there is a temporal restriction defined in the theme 
	bool hasTemporalRest() { return (!generateTemporalRest_.empty());}

	//! Returns the spatial restriction used to generate the theme
	virtual string spatialRest() { return generateSpatialRest_; }

	//! Sets the spatial restriction used to generate the theme
	virtual void spatialRest(const string& s) { generateSpatialRest_ = s; };
	
	//! Returns the spatial relation (e.g WITHIN, COVERED BY, etc.) associate to the spatial restriction
	virtual TeSpatialRelation	spatialRelation() { return spatialRelation_;}

	//! Sets the spatial relation (e.g WITHIN, COVERED BY, etc.) associate to the spatial restriction
	virtual void spatialRelation(TeSpatialRelation s) {spatialRelation_=s;}
	
	//! Returns TRUE if this theme has a spatial restriction  
	virtual bool hasSpatialRest() { return hasSpatialRes_;}

	//! Sets the flag that indicates that there is a spatial restriction 
	virtual void hasSpatialRest(bool a) {hasSpatialRes_ = a;}
	
	//! Returns the box associated with the spatial restriction
	virtual TeBox boxRestriction() { return boxRest_;}

	//! Sets the box associated with the spatial restriction
	virtual void boxRestriction(TeBox& b) {boxRest_ = b;}

	//! Returns a pointer to the geometry associated with the spatial restriction
	virtual TeGeometry* geomRestriction() { return geomRest_;}

	//! Sets a pointer to the geometry associated with the spatial restriction
	virtual void geomRestriction(TeGeometry* g) {geomRest_ = g;}

	//! Returns the geometry representation of the theme which will be considered in the spatial restriction  
	virtual TeGeomRep geomRepRestriction() { return geomRepRest_; }

	//! Sets the geometry representation of the theme which will be considered in the spatial restriction  
	virtual void geomRepRestriction (TeGeomRep& rep) { geomRepRest_ = rep; }

	//! Sets the spatial restriction to be a spatial relation with a box
	virtual void setSpatialRest(TeBox& box, TeGeomRep rep = TeGEOMETRYNONE, TeSpatialRelation relation = TeWITHIN); 

	//! Sets the spatial restriction to be a spatial relation with a geometry
	virtual void setSpatialRest(TeGeometry* geom, TeGeomRep rep = TeGEOMETRYNONE, TeSpatialRelation relation = TeWITHIN); 

	//! Returns the clause WHERE derived from the combination of all restricitions (spatial, attribute and temporal)
	virtual string sqlWhereRestrictions(TeRepresentation* rep=0);
	//@}

	/** @name Collection
		Methods related to the materialization in the database of the theme as a collection of objects 
	*/
	//@{
	//! Returns the name of a table used to store the ids of the objects belonging to the theme
	virtual string	collectionTable() { return collectionTable_; }

	//! Sets the name of a table used to store the ids of the objects belonging to the theme
	virtual void collectionTable(const string& s) { collectionTable_ = s; }

	//! Returns the name of the collection auxiliary table
	virtual string collectionAuxTable() { return collectionAuxTable_; }

	//! Sets the name of the collection auxiliary table
	virtual void collectionAuxTable(string name) { collectionAuxTable_ = name; }

	//! Fills the sqlGridJoin_ and sqlGridFrom_ statements according to the status of the database
	virtual void loadTablesJoin(const string& geomTable="");

	/*!
		\brief Returns a SQL JOIN statement to get all the attributes of the theme objects, 
	           the attributes of the collection table, and the attributes of the extended 
		       collection table 
	*/
	virtual string sqlGridJoin() { return sqlGridJoin_; }

	/*! \brief Returns a FROM clause of a SQL statement to get attributes of the theme objects, the attributes of the 
	           collection table, and the attributes of the extended collection table 
	*/
	virtual string sqlGridFrom(const string& geomTable=""); 

	//! Save the the theme parameters in the database
	virtual bool save();

	//! Build the theme collection: materializes the selection described in the theme in a collection table
	virtual bool buildCollection(std::string objectId = "");

	//! Create the auxiliar collection table used to represent objects with multiple versions in time  
	virtual bool createCollectionAuxTable();

	//! Populate the auxiliar collection table used to represent objects with multiple versions in time
	virtual bool populateCollectionAux(std::string objectId = "");
	//@}

	/** @name Visibility/Status
		Methods related to the visibility/status of the theme and its components: graphs,
		geometrical representations of the objects, grouping.
	*/
	//@{
	//! Sets the geometrical representation visible when this theme is drawn 
	virtual void visibleRep(int rep) { visibleRep_ = rep; }

	//! Returns geometrical representation visible when this theme is drawn 
	virtual int visibleRep() { return visibleRep_; }

	//! Returns a status of a theme
	/*! 
		- Returns 0 if the theme is not visible and not active
		- Returns 1 if the theme is visible and not active
		- Returns 2 if theme is and active and not visible
		- Returns 3 if theme is and visible and active
	 */
	virtual int visibility()
	{	return enableVisibility_; }

	//! Sets whether the theme should be visible and/or active
	/*! 
		- Returns 0 if the theme is not visible and not active
		- Returns 1 if the theme is visible and not active
		- Returns 2 if theme is and active and not visible
		- Returns 3 if theme is and visible and active
	 */
	virtual void visibility(int v)
	{	enableVisibility_ = v; }
	//@}

	/** @name Grouping
		Methods related to grouping of objects of the theme. Each group is
		represented by a specific presentation visual called legend. An slice
		represents the range of values of an attribute that characterizes a group.
	*/
	//@{
	//! Returns a grouping associated the theme 
	virtual TeGrouping& grouping() { return grouping_; }
	
	//! Sets a grouping associated to the theme
	virtual void grouping(const TeGrouping& g) 
	{ grouping_ = g; }

	//! Returns the vector of legends of the theme
	virtual TeLegendEntryVector& legend() { return legend_; }

	//! Sets the visual associated to a geometric representation in the n-th group of the theme 
	virtual bool setGroupingVisual(int n, TeVisual& visual, TeGeomRep rep);

	//! Sets the visual of the n-th group of the theme
	virtual bool setGroupingVisual(int n, TeGeomRepVisualMap& vismap);

	//! Clear the existing grouping of objects of this theme
	virtual void resetGrouping ();

	//! Clear the vector of legends
	virtual void cleanLegend(); 

	//! Save the grouping parameters in memory when there is no chronon
	virtual bool buildGrouping(const TeGrouping& g, TeSelectedObjects selectedObjects = TeAll,
		               vector<double>* dValuesVec = 0);

	//! Save the grouping parameters in memory when there is chronon
	virtual bool buildGrouping(const TeGrouping& g, TeChronon chr, vector<map<string, string> >& mapObjValVec);
					   
	//! Save grouping parameters in memory passing an arbitrary set of slices
	virtual bool buildGrouping(const TeGrouping& g, vector<TeSlice>& slices);

	//! Build the grouping and associate each object to its group in the collection table 
	virtual bool saveGrouping(TeSelectedObjects selectedObjects = TeAll);

	//! Save the theme grouping legends in the collection table  
	virtual bool saveLegendInCollection(TeSelectedObjects selectedObjects = TeAll, std::string objectId = "");

	//! Set the legend id for each object of the theme 
	virtual void setLegendsForObjects();

	//! Set the own legend id for each object of the theme 
	virtual void setOwnLegendsForObjects();

	//! Delete grouping
	virtual bool deleteGrouping(); 
	
	//! Returns the slices associated to the grouping of the theme
	virtual TeSliceVector getSlices();

	//! Generates a optimized position (x,y) in the spatial extention of each object to position label ou graphs
	virtual bool generateLabelPositions();  
	//@}

	/** @name Legends
		Legends are also used to define presentation characteristics of the objets
		of the theme that aren't grouped, according to some specific characteristics
		(such as being pointed, being queried, among others).
	*/
	//@{	
	//! Sets a default legend of the objects of the theme
	virtual void defaultLegend (TeLegendEntry& leg) 
	{ defaultLegend_ = leg; }

	//! Returns the default legend of the objects of the theme
	virtual TeLegendEntry& defaultLegend () 
	{ return defaultLegend_; }

	//! Sets the visual of the default legend for a specific geometrical representation
	virtual void setVisualDefault (TeVisual& visual, TeGeomRep rep)
	{ defaultLegend_.setVisual(visual, rep); }

	//! Sets a legend for objects of the layer that weren't selected in this theme
	/*! Useful when is necessary to see the theme in the context of the layer */
	virtual void outOfCollectionLegend (TeLegendEntry &leg) 
	{ outOfCollectionLegend_ = leg;}

	//! Returns the legend of the layer objects that arent selected in this layer.
	virtual TeLegendEntry& outOfCollectionLegend () 
	{ return outOfCollectionLegend_; } 
	
	//! Sets the visual of the non-selected objects legend for a specific geometrical representation
	virtual void setVisualOutOfCollection (TeVisual& visual, TeGeomRep rep)
	{ outOfCollectionLegend_.setVisual(visual, rep); }

	//! Sets a legend for objects that have geometries but not descriptive attributes
	/*! Useful when in intermediary situations such as editing */
	virtual void withoutDataConnectionLegend (TeLegendEntry &leg) 
	{ withoutDataConnectionLegend_ = leg; }

	//! Returns the legend of objects that have geometries but not descriptive attributes
	virtual TeLegendEntry& withoutDataConnectionLegend () 
	{ return withoutDataConnectionLegend_; }

	//! Sets the visual of the without-attributes objects legend for a specific geometrical representation
	virtual void setVisualWithoutDataConnection (TeVisual& visual, TeGeomRep rep)
	{ withoutDataConnectionLegend_.setVisual(visual, rep); }
	
	//! Sets a legend for the theme objects selected by pointing
	virtual void pointingLegend (TeLegendEntry &leg) 
	{ pointingLegend_ = leg; }

	//! Returns the legend of the theme objects selected by pointing
	virtual TeLegendEntry& pointingLegend () 
	{ return pointingLegend_; }

	//! Sets the visual of the pointed objects legend for a specific geometrical representation
	virtual void setVisualPointing (TeVisual& visual, TeGeomRep rep)
	{ pointingLegend_.setVisual(visual, rep); }

	//! Sets a legend for the theme objects selected by a query
	virtual void queryLegend (TeLegendEntry &leg) 
	{ queryLegend_ = leg; }

	//! Returns the legend of the theme objects selected by a query
	TeLegendEntry& queryLegend () 
	{ return queryLegend_; }
	
	//! Sets the visual of the queried objects legend for a specific geometrical representation
	virtual void setVisualQuery (TeVisual& visual, TeGeomRep rep)
	{ queryLegend_.setVisual(visual, rep); }

	//! Sets a legend for the theme objects selected by query and pointing
	virtual void queryAndPointingLegend (TeLegendEntry &leg) 
	{ queryAndPointingLegend_ = leg; }

	//! Returns the legend of the theme objects selected by query and pointing
	virtual TeLegendEntry& queryAndPointingLegend () 
	{ return queryAndPointingLegend_; }
	
	//! Returns the map of legend ids associated to each object
	virtual map<string, int>& getObjLegendMap() 
	{ return objLegendMap_; }

	//! Returns the map of legend ids associated to each object
	virtual map<string, int>& getObjOwnLegendMap() 
	{ return objOwnLegendMap_; }

	//! Sets the visual of the queried and pointed objects legend for a specific geometrical representation
	virtual void setVisualQueryAndPointing (TeVisual& visual, TeGeomRep rep)
	{ queryAndPointingLegend_.setVisual(visual, rep); }

	/** @name Raster Visual
	*  Methods to deal with the visual presentation of the raster representations
	*/
	//@{ 
	//! Returns the visual presentation of raster geometry
	virtual TeRasterTransform* rasterVisual() 
	{ return rasterVisual_; }

	//! Sets the visual presentation of raster geometry
	virtual void rasterVisual(TeRasterTransform* r) 
	{ rasterVisual_ = r; } 

	//! Removes the visual presentation of the raster
	virtual void removeRasterVisual();

	//! Creates an appropriate visual presentation to the raster of the theme
	virtual void createRasterVisual(TeRaster* rst=0);
	//@}
	//@}

	/** @name Attribute Tables
		A theme can use one or more attribute tables of the layer that gives its data.
		These methods are related to the manipulation of these tables.
	*/
	//@{	
	//! Add a new attribute table to a theme 
	virtual bool addThemeTable (TeTable& table);

	//! Add a new attribute table to a theme
	virtual void addThemeTable (string tableName);

	//! Verify if an attribute table is part of a theme
	virtual bool isThemeTable(int tableId);

	//! Verify if an attribute table is part of a theme
	virtual bool isThemeTable(string tableName);

	//! Returns the list of attribute tables used by this theme 
	virtual TeAttrTableVector& attrTables()
	{	return attTableVector_; }

	//! Sets the entire list of attribute tables used by this theme 
	virtual bool setAttTables(TeAttrTableVector& attrs);  

	//! Returns a vector of attribute tables, of a specific type, used by this theme
	virtual bool getAttTables(TeAttrTableVector& attrs, TeAttrTableType attType = TeAllAttrTypes); 

	//! Returns a representation of an attribute table  given name 
	virtual bool getTable(TeTable& table, const string tableName);

	//! Clears the list of attribute tables used by this theme
	virtual void clearAttTableVector() 
	{	attTableVector_.clear();	}

	//! Returns the temporal attribute table of the theme (TeEvent or TeFixedGeomDynAttr)
	/*
		\note A theme supports only one temporal attribute table
	*/
	virtual bool getTemporalTable(TeTable& table);

	//! Removes an attribute table from the list of tables of a theme
	virtual bool removeThemeTable(unsigned int index);

	//! Returns the the name of an attribute table that contains a given attribute
	virtual string getTableName(const string& attrName);

	/*!
		\brief Returns the name the index-th attribute resulting of the join of all attribute
		       tables associated to the theme tables 
	*/
	virtual string getAttribute(unsigned int index);

	//! Loads the theme tables in the database
	virtual bool loadThemeTables();
	
    //! Returns the list of attributes of theme tables 
	virtual TeAttributeList sqlAttList() { return sqlAttList_;}

	//! Clears the list of attributes associated to the theme tables 
	virtual void clearAttList() {sqlAttList_.clear();}

	//! Returns the list of numerical attributes of the theme tables 
	virtual TeAttributeList sqlNumAttList() { return sqlNumAttList_;}

	//! Clears the list of numerical attributes associated to the theme tables 
	virtual void clearNumAttList() {sqlNumAttList_.clear();}

	//! Returns a SQL JOIN statement to reach to all attribute tables used by this theme
	virtual string	sqlJoin() { return sqlJoin_;}

	//! Returns a SQL FROM CLAUSE that gives access to all attribute tables used by this theme
	virtual string	sqlFrom() { return sqlFrom_;}

	//! Returns the alias vector of the names of the theme tables
	virtual vector<string>&	aliasVector() { return aliasVector_; }

	//! fill aliasVector_
	virtual void loadAliasVector();

	//! Refresh list of attributes of all the theme tables.
	/*
		\note All attributes are stored into sqlAttList_ and numeric attributes are stored into sqlNumAttList_.
	*/
	virtual void loadAttrLists();
	//@}

	/** @name Box
	    Methods related to the bounding box of a theme
	*/
	//@{ 
	//! Returns the theme box 
	virtual TeBox& getThemeBox()
	{	return themeBox_; }

	//! Sets the theme box 
	virtual void setThemeBox(TeBox& box)
	{	themeBox_ = box; }

	//! Returns the theme box 
	virtual TeBox box()
	{	return themeBox_; }

	//! Refreshes the bounding box of a theme according to its representation
	/*!
		\note This functions is being kept for compatibility reasons with derived
		classes. IT SHOULD NOT BE USED BY APPLICATION USING ONLY THE BASIC CLASS.
	*/
	virtual void updateThemeBox() {}

	//! Updates the bounding box of a theme (in database)
	/*!
		\note This functions is being kept for compatibility reasons with derived
		classes. IT SHOULD NOT BE USED BY APPLICATION USING ONLY THE BASIC CLASS.
	*/
	virtual void updateThemeBox(const TeBox& /*box*/) {}
	//@}

	/** @name Locate geometries
	    Returns the geometry(ies) of the theme given coordinate
	*/
	//@{ 	
	virtual bool locatePolygon		(TeCoord2D &pt, TePolygon &polygon, const double& tol = 0.0);
	virtual bool locatePolygonSet   (TeCoord2D &pt, double tol, TePolygonSet &polygons);
	virtual bool locateLine		(TeCoord2D &pt, TeLine2D &line, const double& tol = 0.0);
	virtual bool locatePoint	(TeCoord2D &pt, TePoint &point, const double& tol = 0.0);
	virtual bool locateCell		(TeCoord2D &pt, TeCell &c, const double& tol = 0.0);
	//@}

	//! Returns TRUE if a theme has lower priority than another
	virtual bool operator< (const TeTheme& r) const
	{	return priority_ < r.priority_; }

	//! Verifies if there are objetcs without geometries of a specific geometry representation
	virtual bool hasObjectsWithoutGeometries(TeGeomRep geomRep);

	//! Removes the objetcs without geometries of a specific geometry representation
	virtual bool removeObjectsWithoutGeometries(TeGeomRep geomRep);

protected:
	//! Fill the sqlJoin_ and sqlFrom_ 
	virtual void loadThemeTablesJoin();
	
	//! Populate the collection table based in the theme restrictions
	virtual bool populateCollection(std::string objectId = ""); 

	//! list of attribute tables of the theme
	TeAttrTableVector	attTableVector_;

	//! List containing all the attributes of the theme tables
	TeAttributeList	sqlAttList_;
	
	//! List containing only the numeric attributes of the theme tables
	TeAttributeList	sqlNumAttList_;		

	//! Pointer to the layer that gives origin to this theme
	TeLayer*	layer_;

private:
	int			layerId_;	// layer id
	
	string		generateAttributeRest_;
	string		generateTemporalRest_;
	string		generateSpatialRest_; //future use

	TeSpatialRelation	spatialRelation_;
	bool				hasSpatialRes_;
	TeBox				boxRest_;		//box which defines the spatial restriction 
	TeGeometry*			geomRest_;		//geometry which defines the spatial restriction 
	TeGeomRep			geomRepRest_;	//geometry representation of the theme which will be 
										//considered in the spatial restriction  
	// Display scale
	double	minScale_;
	double	maxScale_;

	// collection table name
	string	collectionTable_;
	string	collectionAuxTable_;

	//representation visible in the theme
	int visibleRep_;

	//Theme status
	int	enableVisibility_;

	// ----------------- grouping information -----------------
	TeGrouping		grouping_;

	// ----------------- legend information -----------------
	TeLegendEntryVector	legend_;

	// Background Legend
	TeLegendEntry		outOfCollectionLegend_;			//group(-1) 
	TeLegendEntry		withoutDataConnectionLegend_;	//group(-2) 
	TeLegendEntry		defaultLegend_;					//group(-3) 
	TeLegendEntry		pointingLegend_;				//group(-4) 
	TeLegendEntry		queryLegend_;					//group(-5) 
	TeLegendEntry		queryAndPointingLegend_;		//group(-6)

	map<string, int>	objLegendMap_;					// object legend
	map<string, int>	objOwnLegendMap_;				// object own legend
		
	// ----------------- theme tables information -----------------

	//! clause FROM: join with the theme tables and collection table
	string	sqlFrom_;		
	
	//! clause SELECT and FROM: join with the theme tables and collection table
	string	sqlJoin_;
	
	//! clause FROM: join with the theme tables and auxiliar collection table
	string	sqlGridFrom_;
	
	//! clause SELECT and FROM: join with the theme tables and auxiliar collection table
	string	sqlGridJoin_;

	//! vector of alias to the attribute tables that are used more than once  
	vector<string>		aliasVector_;

	//! visual of raster
	TeRasterTransform*		rasterVisual_;
	TeBox	themeBox_;
};

//! A vector of pointers to theme
typedef vector<TeTheme*>	TeThemeVector;

//! A map from theme identifiers to pointers to theme
typedef map<int, TeTheme*>	TeThemeMap;

/*! \example createTheme.cpp
	Shows how to create themes in TerraLib.
 */

/*! \example themeGrouping.cpp
	Shows how to do a grouping on the objects of a TerraLib theme.
 */

/*! \example rasterSlicing.cpp
	Shows how to  create a legend over a raster data, stored in a layer TerraLib.
 */
#endif

