/************************************************************************************
TerraLib - a library for developing GIS applications.
Copyright  2001-2004 INPE and Tecgraf/PUC-Rio.

This code is part of the TerraLib library.
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

You should have received a copy of the GNU Lesser General Public
License along with this library.

The authors reassure the license terms regarding the warranties.
They specifically disclaim any warranties, including, but not limited to,
the implied warranties of merchantability and fitness for a particular purpose.
The library provided hereunder is on an "as is" basis, and the authors have no
obligation to provide maintenance, support, updates, enhancements, or modifications.
In no event shall INPE and Tecgraf / PUC-Rio be held liable to any party for direct,
indirect, special, incidental, or consequential damages arising out of the use
of this library and its documentation.
*************************************************************************************/

/*! \file TeViewNode.h
    \brief This file provides TerraLib's tree structure of views.
*/
#ifndef  __TERRALIB_INTERNAL_VIEWNODE_H
#define  __TERRALIB_INTERNAL_VIEWNODE_H

#define THEME_GROUP

#include "TeBox.h"

#include <iostream>
#include <string>
#include <vector>
using namespace std;

class TeProjection;

//! Type of node
enum TeViewNodeType
{ TeTHEME, TeTREE }; 

//! A class to deal with nodes of a view tree structure
class TeViewNode
{
public:
	//! Constructor
    TeViewNode( const string& name, TeViewNode* parent = 0, int view=0, int id=0, TeViewNodeType type=TeTHEME): 
	      myParent_(parent), 
		  id_(id), 
		  myParentId_(parent?parent->id():0),
		  name_(name), 
		  view_(view), 
		  type_(type),
		  priority_(0)
	{}

	//! Destructor
	virtual ~TeViewNode()
	{}

	//! Inserts a new child in the tree hierachy. Fails if the object is a leaf
	virtual void add (TeViewNode*) {}

    //!	Remove an existing child identified by its id from the tree hierarchy
    /*!	
		Fails if the object is a leaf
	    Should be used in conjunction with a delete 
	*/	
	virtual TeViewNode* removeID (int /* id */) { return 0; }
 
   //!	Remove an existing child identified by its name from the tree hierarchy
    /*!	
		Fails if the object is a leaf
	    Should be used in conjunction with a delete 
	*/	
	virtual TeViewNode* remove (const string& /*name*/) { return 0; }

	//! Retrieve a node identified by its identifier from the tree structure
	virtual TeViewNode* retrieve (int)  { return 0; }

	//! Returns a pointer to a parent node
	TeViewNode* parent()
	{ return myParent_; }

	//! Returns the identifier of a node parent
	int parentId()
	{
		if (myParent_)
			return myParent_->id(); 
		else
			return myParentId_;
	}

	//! Sets the identification of the parent node
	void parentId(int i) 
	{
		if (myParent_)
			myParent_->id(i);
		myParentId_ = i;
	}

	//! Sets the parent of node
	void setParent ( TeViewNode* node )
	{ 
		myParent_ = node; 
		myParentId_ = (myParent_)? myParent_->id():0;
	}

	//! Returns the identification of a node
	int		id () { return id_; }

	//! Sets the identification of a node
	virtual void	id (int i){ id_ = i; }

	//! Returns the name of a node
	string	name () { return name_; }
	//! Sets the name of a node
	void	name (const string& s) { name_ = s; }

	//! Returns the priority of a node
	int		priority () { return priority_; }
	//! Sets the priority of a node
	void	priority (int i) { priority_ = i; }

	//! Sets the view identification of a node
	void	view (int view) { view_ = view; }
	//! Returns the view identification of a node
	int		view () { return view_ ; }

	//! Returns the node type
	TeViewNodeType type() { return type_; };
	//! Sets the node type
	void type(TeViewNodeType t) { type_ = t; }

	//! Moves a node up in the tree structure
	virtual void moveUp ()
	{
		if (myParent_)
			myParent_->moveUp (this);
	}
	
	//! Moves a node down in the tree structure
	virtual void moveDown ()
	{
		if (myParent_)
			myParent_->moveDown (this);
	}

	//! Swaps nodes 
	virtual void swap ( unsigned int, unsigned int ) {}

	//! Sets the nodes visibility 
	virtual void visibility ( int ){}; 

	//! Returns the nodes visibility 
	virtual int visibility(){ return 1; };

	//! Draws a node
	virtual void draw() {}

	//! Sorts the node
	virtual void sort() {}

protected:

	//! Moves a node up in the tree structure.
	virtual void	moveUp (TeViewNode*) {}

	//! Moves a node down in the tree structure.
	virtual void	moveDown (TeViewNode*) {}

	TeViewNode* 	myParent_;		//!< Pointer to parent node.
	int				id_;			//!< Node id
	int 			myParentId_;	//!< Node parent id
	string			name_;			//!< Node name
	int				view_;			//!< view id
	TeViewNodeType  type_;			//!< Node type (like theme or tree).
	int			    priority_;		//!< Precedence when stored in a view
};


//! A class to deal with branchs in a view tree structure  
class TeViewTree: public TeViewNode
{
public:

	//! Constructor
	TeViewTree(const string& name=""): 
	  TeViewNode(name)
	{ 	  type_ = TeTREE; }

	//! Destructor
	virtual ~TeViewTree ();

	//! Sets the nodes visibility 
	virtual void visibility (int vis);

	//! Returns the nodes visibility 
	virtual int visibility ();

	//! Swap nodes
	virtual void swap (unsigned int i, unsigned int j);

	//! Moves a node up in the tree structure
	virtual void moveUp (TeViewNode* node);

	//! Moves a node down in the tree structure
	virtual void moveDown (TeViewNode* node);

	//! Adds a node to the structure
	virtual void add (TeViewNode* node);

    //!	Removes a node identified by its name
	virtual TeViewNode* remove (const string& name);

    //!	Removes a node through its identifier
	virtual TeViewNode* removeID (int id) ;

    //!	Retrieves a node through its index
	virtual TeViewNode* retrieve (unsigned int i) 
	{ return nodes_[i]; }

	//! Draw a node
	virtual void draw ();

	//! Returns the size of the structure
	unsigned int size()
    { return nodes_.size(); }

	//! Returns the vector of View nodes
	vector<TeViewNode*>& nodes()
	{ return nodes_; }

	//! Sort the vector of nodes according to their priorities
	void sort();

	//! Unlinks the nodes from the view tree
	virtual void clear();
	
	/*! \brief Move the given node to the top of it's tree.
	    \param node to be moved
	    \return the number of positions the node has been moved
	*/
	virtual int moveTop (TeViewNode* node);

	/*!	\brief Move the given node to the bottom of it's tree.
	    \param node to be moved
	    \return the number of positions the node has been moved
	*/
	int moveBottom(TeViewNode* node);

	/*! \brief Inserts the given node to at the begin of the view tree.
	    \param node node to be added
	*/
	virtual void insertFront(TeViewNode* node);

	/*! \brief Look at a View Tree for a node with the given id.
		\param id Id of the node to be found
	*/
	virtual TeViewNode* find (int id);

	/*! \brief Look at a View Tree for a node with the given name.
		\param name name of the node to be found
		\param caseS true if the given name is case sensitive (optional)
	*/
	virtual TeViewNode* find( std::string name, bool caseS = false );

	/*! \brief Returns the boundary box of the tree view themes
	    \param onlyVisible If true compute box based on visible nodes otherwise makes use of all tree nodes.
	    \param projection  The returned box coordinates will be in the given projection 
	*/
	virtual TeBox box(bool onlyVisible, TeProjection* projection );

	/*!	\brief Asserts the nodes priorities by the positions.
		\return Returns true if any priority was updated, otherwise returns false.
	*/
	virtual bool assertsPriorities();

private:
	vector<TeViewNode*> nodes_;		//!< Vector to all branchs.
};
#endif

